Package: metabom8
Type: Package
Title: A High-Performance R Package for Metabolomics Modeling and
        Analysis
Version: 1.1.0
Authors@R: person(given = "Torben",
            family = "Kimhofer",
            role = c("aut", "cre"),
            email = "tkimhofer@gmail.com",
            comment = c(ORCID = "0000-0001-7158-9930"))
Description: Tools for 1D NMR metabolomics workflows, including import and 
              preprocessing of Bruker experiments, multivariate modeling (PCA, PLS, OPLS)
              and model analytics and validation (y-permutations, cv-anova). Performance-critical 
              routines are implemented in C++ and use the Armadillo and Eigen linear algebra 
              libraries to improve runtime.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.6.0)
Imports: abind, colorRamps, ellipse, fs, ggplot2, ggrepel, graphics,
        methods, parallel, pcaMethods, plotly, pROC, progress, ptw,
        Rcpp (>= 1.0.4.6), reshape2, rlang, scales, signal, stats,
        utils
LinkingTo: Rcpp, RcppArmadillo, RcppEigen
biocViews: Metabolomics, Cheminformatics, Preprocessing, DataImport,
        Alignment, WorkflowStep
BugReports: https://github.com/tkimhofer/metabom8/issues
URL: https://tkimhofer.github.io/metabom8/
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.3
Suggests: BiocStyle, dplyr, DT, ExperimentHub, htmlTable, knitr,
        rmarkdown, testthat
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/metabom8
git_branch: devel
git_last_commit: 977c8fa
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: yes
Packaged: 2026-04-28 23:39:47 UTC; biocbuild
Author: Torben Kimhofer [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-7158-9930>)
Maintainer: Torben Kimhofer <tkimhofer@gmail.com>

# Generated by roxygen2: do not edit by hand

S3method(plot,m8_stocsy1d)
S3method(print,m8_stocsy1d)
export(add_note)
export(align_segment)
export(align_spectra)
export(balanced_boot)
export(balanced_mc)
export(binning)
export(bline)
export(calibrate)
export(cliffs_d)
export(correct_baseline)
export(correct_lw)
export(cv_anova)
export(dmodx)
export(ellipse2d)
export(es_cdelta)
export(excise)
export(fitted)
export(get.idx)
export(get_idx)
export(get_provenance)
export(hotellingsT2)
export(kfold)
export(list_preprocessing)
export(lw)
export(matspec)
export(mc)
export(minmax)
export(noise_sd)
export(norm_eretic)
export(opls)
export(pareto_scaling)
export(pca)
export(plotStocsy)
export(plot_spec)
export(pls)
export(ppick)
export(ppick2)
export(pqn)
export(prep_X)
export(print_preprocessing)
export(print_provenance)
export(read1d)
export(read1d_proc)
export(read1d_raw)
export(scRange)
export(scores)
export(spec)
export(stocsy)
export(storm)
export(stratified_kfold)
export(unscaled)
export(uv_scaling)
export(vip)
export(weights)
export(xres)
exportClasses(m8_model)
exportMethods(loadings)
exportMethods(show)
exportMethods(summary)
import(methods)
importFrom(Rcpp,evalCpp)
importFrom(abind,abind)
importFrom(colorRamps,matlab.like2)
importFrom(ellipse,ellipse)
importFrom(fs,path_common)
importFrom(fs,path_rel)
importFrom(ggplot2,aes)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,element_text)
importFrom(ggplot2,geom_hline)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_segment)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggtitle)
importFrom(ggplot2,labs)
importFrom(ggplot2,scale_colour_gradientn)
importFrom(ggplot2,scale_x_reverse)
importFrom(ggplot2,scale_y_continuous)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_bw)
importFrom(ggplot2,xlab)
importFrom(pROC,multiclass.roc)
importFrom(pROC,roc)
importFrom(parallel,detectCores)
importFrom(progress,progress_bar)
importFrom(ptw,asysm)
importFrom(reshape2,melt)
importFrom(rlang,.data)
importFrom(scales,breaks_pretty)
importFrom(signal,sgolayfilt)
importFrom(stats,aggregate)
importFrom(stats,approxfun)
importFrom(stats,ccf)
importFrom(stats,cor)
importFrom(stats,cov)
importFrom(stats,lm)
importFrom(stats,mad)
importFrom(stats,median)
importFrom(stats,pf)
importFrom(stats,pt)
importFrom(stats,qf)
importFrom(stats,quantile)
importFrom(stats,residuals)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,t.test)
importFrom(utils,capture.output)
importFrom(utils,combn)
importFrom(utils,head)
importFrom(utils,tail)
useDynLib(metabom8, .registration = TRUE)

<!-- badges: start -->
  [![Bioconductor build](https://bioconductor.org/shields/build/devel/bioc/metabom8.svg)](https://bioconductor.org/checkResults/devel/bioc-LATEST/metabom8/)
  [![R-CMD-check-bioc](https://github.com/tkimhofer/metabom8/actions/workflows/check-bioc.yml/badge.svg)](https://github.com/tkimhofer/metabom8/actions/workflows/check-bioc.yml)
  [![DOI](https://img.shields.io/badge/DOI-10.18129%2FB9.bioc.metabom8-blue)](https://doi.org/10.18129/B9.bioc.metabom8)
<!-- badges: end --><!-- badges: end -->

# metabom8 

**A High-Performance R Package for Metabolomics Modeling and Analysis**


`metabom8` (*metabo-mate*) provides:

- 1D NMR data import & preprocessing  
- Multivariate modeling (PCA, PLS, O-PLS)  
- Metabolite identification  

Core functionality is accelerated using C++ (`Rcpp`, `Armadillo`, `Eigen`) for improved performance.

---

## 🛠️ Features

- NMR data import and preprocessing, including: 
  - calibration, baseline correction, alignment, normalisation
  - provenance logging
- Multivariate statistics:  
  - Principal Components Analysis (PCA) 
  - Partial Least Squares (PLS) 
  - Orthogonal Partial Least Squares (OPLS)
- Automatic selection of the optimal number of components for supervised models
- Robust cross-validation framework, for large and small sample sizes (e.g., Monte Carlo CV)
- OPLS model diagnostics & validation (DModX, permutation testing)
- Tools supporting metabolite identification (STOCSY, STORM)
- Native C++ acceleration via `RcppArmadillo` and `RcppEigen`

---

## 📦 Installation

```r
# Install from Bioconductor
install.packages("BiocManager")
BiocManager::install("metabom8")
```

---

## 🚀 Quick Start

```r
library(metabom8)

# Load example data
data(hiit_raw, package = "metabom8")

# plot spectra interactively with plotly
plot_spec(hiit_raw)

# piped preprocessing 
hiit <- hiit_raw |> 
  calibrate(type = "tsp") |>
  excise() |>
  correct_baseline(method='asls') |>
  align_spectra() |>
  pqn()

## Provenance logging
print_provenance(hiit)


data(covid, package = "metabom8")
X <- covid$X
Y <- covid$an$type

# Modelling Context
uv = uv_scaling(center = TRUE)
mc_cv <- balanced_mc(k=15, split=2/3, type="DA")

# PCA
pca_model <- pca(covid$X, scaling = uv, ncomp=2)

# PLS
pls_model <- pls(X, Y, scaling = uv, validation_strategy=mc_cv)
show(pls_model)

# OPLS
opls_model <- opls(X, Y, scaling = uv, validation_strategy=mc_cv)

Tx <- scores(opls_model)
Px <- loadings(opls_model)
vip <- vip(opls_model)
```

---

## ⚡ Performance

Benchmarking OPLS modelling against another widely used R implementations under comparable conditions:

```r
bench::mark(
  metabom8 = metabom8::opls(X, Y, scaling=uv, validation_strategy=kfold),
  ropls    = ropls::opls(X, Y, predI = 1, orthoI = 1),
  check = FALSE
)
```

| Method     | Median Time | Iter/sec | Memory Use |
|------------|-------------|----------|------------|
| **metabom8** | 1.61 sec     | 0.617    | 247.33 MB  |
| **ropls**    | 9.12 sec     | 0.110    | 2.09 GB    |

> `metabom8` is based on C++ accelerated linear algebra and provides a >5× speed-up over `ropls`*.  

\**[`ropls`](https://bioconductor.org/packages/ropls) (uv scaling, 7-fold CV; 1 predictive and 1 orthogonal component).*

---

## 📘 Documentation

Comprehensive documentation and vignettes are available at:  
🔗 https://tkimhofer.github.io/metabom8/

---

## 📝 License & Citation

MIT License © Torben Kimhofer  
If `metabom8` or any of its components contributes to your work, please ensure appropriate citation. See [`CITATION`](https://github.com/tkimhofer/metabom8/blob/master/inst/CITATION) for citation details.

---

## 🙋 Getting Help

- Found a bug? Open an [issue](https://github.com/tkimhofer/metabom8/issues)
- Want to contribute? Fork the repo and submit a pull request

---

*Built with 💙 by [@tkimhofer](https://github.com/tkimhofer)*

#' @keywords internal
.em <- function(n, lb) {
  idx <- seq_len(n)
  out <- exp(-(idx * lb * pi) / n)
  minmax(out)
}

#' @keywords internal
.cosine <- function(n) {
  idx <- seq_len(n)
  out <- cos(pi * idx / n)
  minmax(out)
}

#' @keywords internal
.sine <- function(n) {
  idx <- seq_len(n)
  out <- sin(pi * idx / n)
  minmax(out)
}

#' @keywords internal
.sem <- function(n, lb = 1.5) {
  idx <- seq_len(n)
  out <- sin(pi * idx / n) * exp(-(idx * lb * pi) / n)
  minmax(out)
}

#' @keywords internal
.expGaus_resyG <- function(n, lb = -10, gb = 0.12, aq_t) {
  idx <- seq_len(n)
  b <- -lb / (2 * gb * aq_t)
  out <- exp((-lb * idx) - (b * idx^2))
  minmax(out)
}

#' @keywords internal
.gauss <- function(n, lb = -1.2, gb = 0.3, para) {
  sw_hz <- para$a_SW_h
  AQ <- para$a_TD / (2 * sw_hz)
  idx <- seq_len(n) - 1
  e <- (-pi * idx * lb) / sw_hz
  g <- ((lb * pi) / (2 * gb * AQ)) * (idx / sw_hz)
  out <- exp(e - g^2)
  minmax(out)
}


#' @title Apodisation function dispatcher
#' @description
#' Generates apodisation (window) functions for FID processing.
#' The function applies one of several predefined window shapes,
#' selected via \code{pars$fun}.
#' @param n Integer. Number of data points.
#' @param pars List of parameters defining the apodisation.
#' Must contain element \code{fun} specifying the window type.
#' @details
#' Supported apodisation functions and parameters:
#' \describe{
#'   \item{\code{"uniform"}}{No apodisation (all weights = 1).}
#'
#'   \item{\code{"exponential"}}{
#'     Exponential decay window.
#'     \describe{
#'       \item{\code{lb}}{Line broadening parameter.}
#'     }
#'   }
#'
#'   \item{\code{"cosine"}}{
#'     Cosine window (no additional parameters).
#'   }
#'
#'   \item{\code{"sine"}}{
#'     Sine window (no additional parameters).
#'   }
#'
#'   \item{\code{"sem"}}{
#'     Sine-modulated exponential window.
#'     \describe{
#'       \item{\code{lb}}{Exponential decay parameter.}
#'     }
#'   }
#'
#'   \item{\code{"expGaus_resyG"}}{
#'     Exponential-Gaussian hybrid window.
#'     \describe{
#'       \item{\code{lb}}{Exponential decay parameter.}
#'       \item{\code{gb}}{Gaussian broadening parameter.}
#'       \item{\code{aq_t}}{Acquisition time.}
#'     }
#'   }
#'
#'   \item{\code{"gauss"}}{
#'     Gaussian window (Bruker-style).
#'     \describe{
#'       \item{\code{lb}}{Line broadening parameter.}
#'       \item{\code{gb}}{Gaussian broadening factor.}
#'       \item{\code{para}}{List containing acquisition parameters
#'         (e.g. \code{a_SW_h}, \code{a_TD}).}
#'     }
#'   }
#' }
#' @return Numeric vector of length \code{n} with apodisation weights.
#' @examples
#' # Internal FID apidisation functions (illustrative only)
#' f_apod <- c("sine","cosine","exponential","sem")
#'
#' # create toy fid
#' n <- 200; t <- seq(0,1,len=n)
#' fid <- exp(-5*t)*cos(20*pi*t)
#'
#' # generate & apply apodisation windows
#' A  <- sapply(f_apod, \(f) metabom8:::.fidApodisationFct(n, list(fun=f, lb=-0.2)))
#' Fp <- sweep(A, 1, fid, `*`)
#'
#' # generate spectra
#' S  <- apply(Fp, 2, \(x) Mod(fft(x))[1:(n/2)])
#'
#' # graphhics
#' cols <- 1:ncol(A)
#'
#' par(mfrow=c(2,2), mar=c(3,3,2,1))
#' plot(t, fid, type="l", lwd=2, main="FID")
#' matplot(t, A, type="l", lwd=2, col=cols, main="Windows")
#' matplot(t, Fp, type="l", lwd=2, col=cols, main="Windowed FID")
#' matplot(S, type="l", lwd=2, col=cols, main="Spectrum")
#'
#' legend("topright", f_apod, col=cols, lty=1, bty="n", cex=.8)
#'
#' @keywords internal
.fidApodisationFct <- function(n, pars) {
  if (is.null(pars$fun) || !pars$fun %in% c("uniform", "exponential", "cosine", "sine", "sem", "expGaus_resyG", "gauss")) {
    stop("Invalid apodisation function specified. Choose one of: 'uniform', 'exponential', 'cosine', 'sine', 'sem', 'expGaus_resyG', or 'gauss'.", call. = FALSE)
  }

  afun <- switch(pars$fun,
                 "uniform" = rep(1, n),
                 "exponential" = {
                   if (!"lb" %in% names(pars)) stop("Exponential apodisation requires 'lb' parameter.", call. = FALSE)
                   .em(n, pars$lb)
                 },
                 "cosine" = .cosine(n),
                 "sine" = .sine(n),
                 "sem" = {
                   if (!"lb" %in% names(pars)) stop("SEM apodisation requires 'lb' parameter.", call. = FALSE)
                   .sem(n, pars$lb)
                 },
                 "expGaus_resyG" = {
                   if (!all(c("lb", "gb", "aq_t") %in% names(pars))) {
                     stop("expGaus_resyG apodisation requires 'lb', 'gb', and 'aq_t' parameters.", call. = FALSE)
                   }
                   .expGaus_resyG(n, pars$lb, pars$gb, pars$aq_t)
                 },
                 "gauss" = {
                   if (!all(c("lb", "gb", "para") %in% names(pars))) {
                     stop("Gaussian apodisation requires 'lb', 'gb', and 'para' parameters.", call. = FALSE)
                   }
                   .gauss(n, pars$lb, pars$gb, pars$para)
                 }
  )

  if (!is.null(pars$plot) && isTRUE(pars$plot)) {
    plot(afun, type = "l", main = paste("Apodisation function:", pars$fun))
  }

  return(afun)
}


#' @title Filter Bruker NMR Experiments
#' @description Filters Bruker 1D experiments based on acquisition metadata and
#'   optionally limits the number of returned entries. Character parameters are
#'   matched by value, while numeric parameters support equality, set inclusion,
#'   range filtering (list(range = c(min, max))), and generic operators
#'   (list(op = ">", value = x)).
#' @param pars Data frame. Parsed acquisition and processing parameters.
#' @param exp_type Named list. Filtering conditions for parameters. Elements may
#'   be character vectors, numeric values/vectors, or operator/range lists for
#'   numeric fields.
#' @param f_list List. File paths of spectra and associated metadata (e.g., f_fid, f_1r).
#' @param n_max Integer. Maximum number of experiments to retain.
#' @return A list containing filtered and ordered `f_list` and `pars`.
#' @keywords internal
#' @seealso \code{\link{.extract_acq_pars1d}}, \code{\link{.check1d_files_fid}}
#' @importFrom stats aggregate
.filterExp_files <- function(pars, exp_type, f_list, n_max) {

  if (!is.data.frame(pars)) {
    if (is.null(names(pars))) stop("`pars` must be named when not a data.frame.", call. = FALSE)
    pars <- as.data.frame(as.list(pars), stringsAsFactors = FALSE)
  }

  pars <- as.data.frame(pars, stringsAsFactors = FALSE)

  idx <- match(toupper(names(exp_type)), toupper(gsub("[ap]_", "", colnames(pars))))

  if (length(idx) == 0) {
    stop("No parameter(s) found that match the specification. Check 'exp_type' and parameter choices in 'acqus' and 'procs'.", call. = FALSE)
  }

  idx_na <- which(is.na(idx))
  if (length(idx_na) > 0) {
    if (length(idx_na) == length(idx)) {
      stop("No matching parameter names found. Check input argument 'exp_type'.", call. = FALSE)
    } else {
      message(sprintf("Experiment filter %s not in NMR acquisition list. Using remaining arguments to filter: %s",
                      paste(names(exp_type)[idx_na], collapse = ", "),
                      paste(names(exp_type)[-idx_na], collapse = ", ")))
      idx <- idx[-idx_na]
    }
  }

  fmat <- vapply(seq_along(idx), function(i) {
    col <- pars[, idx[i]]
    cond <- exp_type[[i]]

    if (is.numeric(col)) {
      if (is.list(cond)) {
        if (!is.null(cond$range)) {
          lo <- cond$range[1]; hi <- cond$range[2]
          return(!is.na(col) & col >= lo & col <= hi)
        }
        if (!is.null(cond$op) && !is.null(cond$value)) {
          op <- cond$op
          val <- cond$value
          return(!is.na(col) & do.call(op, list(col, val)))
        }
        stop(sprintf("Unsupported numeric filter for '%s'.", names(exp_type)[i]), call. = FALSE)
      }
      if (is.numeric(cond)) {
        if (length(cond) == 1) return(!is.na(col) & col == cond)
        return(!is.na(col) & col %in% cond)
      }
      if (is.character(cond) && all(grepl("^[0-9.]+$", cond))) {
        condn <- as.numeric(cond)
        if (length(condn) == 1) return(!is.na(col) & col == condn)
        return(!is.na(col) & col %in% condn)
      }
      stop(sprintf("Numeric column '%s' needs numeric filter values.", names(exp_type)[i]), call. = FALSE)
    }

    vars <- gsub("^<|>$", "", as.character(col))
    vars %in% cond
  }, logical(nrow(pars)))

  if (is.matrix(fmat) && nrow(fmat) > 0) {
    idx_filt <- apply(fmat == 1, 1, all)
  } else {
    idx_filt <- (fmat == 1)
  }
  if (all(!idx_filt)) {
    stop("No files found that match the specified parameter specification levels.", call. = FALSE)
  }

  f_list <- lapply(f_list, function(x) x[idx_filt])
  idx_filt <- which(idx_filt)

  if (length(idx_filt) > n_max) {
    idx_filt <- idx_filt[seq_len(n_max)]
    f_list <- lapply(f_list, function(x) x[seq_len(n_max)])
  }

  pars <- pars[idx_filt, ]

  fnam <- strsplit(
    if (is.null(f_list$f_1r)) f_list$f_fid else ifelse(!is.na(f_list$f_1r) & f_list$f_1r != 0, f_list$f_1r, f_list$f_fid),
    .Platform$file.sep
  )

  fmat_comb <- do.call(rbind, fnam)
  uniq_cols <- vapply(seq_len(ncol(fmat_comb)), function(i) length(unique(fmat_comb[, i])) == nrow(fmat_comb), logical(1))
  idx_keep <- which(uniq_cols)

  if (length(idx_keep) > 1) {
    rack_ <- t(vapply(seq_along(fnam), function(i) {
      c(fnam[[i]][idx_keep[length(idx_keep) - 1]], pars$a_DATE[i])
    }, FUN.VALUE = c("", "")))

    colnames(rack_) <- c("a", "b")

    df <- as.data.frame(rack_)

    rack_order_ <- aggregate(as.POSIXct(b) ~ a, data = df, FUN = mean)
    colnames(rack_order_) <- c("a", "mean_b")
    rord_fac <- order(as.numeric(rack_order_$mean_b)) * 1e5

    fnam1 <- vapply(fnam, function(x) x[idx_keep[length(idx_keep) - 1]], FUN.VALUE = "")
    rord_fac <- rord_fac[match(fnam1, rack_order_$a)]

    exp_ <- vapply(fnam, function(x) x[idx_keep[length(idx_keep)]], FUN.VALUE = "")

    if (!.is_numeric_trycatch(exp_)) {
      exp_ <- factor(exp_)
    }
    rr <- order(as.numeric(exp_) + rord_fac)
  } else {
    exp_ <- vapply(fnam, function(x) x[idx_keep], FUN.VALUE = "")
    if (any(is.na(as.numeric(exp_)))) {
      exp_ <- factor(exp_)
    }
    rr <- order(as.numeric(exp_))
  }

  pars <- pars[rr, ]
  f_list <- lapply(f_list, function(x) x[rr])
  pars$a_DATE <- as.POSIXct(pars$a_DATE)

  return(list(f_list, pars))
}



#' @title Extract Bruker NMR Acquisition Parameters
#' @description Parses `acqus` files to extract acquisition parameters for each experiment.
#' @param f_list List. Contains paths to NMR experiment folders (e.g., as returned by \code{.check1d_files_fid()}).
#' @return A data frame containing extracted acquisition parameters for each experiment.
#' @keywords internal
#' @seealso \code{\link{.filterExp_files}}, \code{\link{.check1d_files_fid}}
.extract_acq_pars1d <- function(f_list) {
  out <- lapply(f_list[[1]], function(fil) {
    f_acqu_path <- file.path(fil, "acqus")
    fhand <- file(f_acqu_path, open = "r")
    f_acqu <- readLines(fhand, warn = FALSE)
    close(fhand)

    idx <- grep("..", f_acqu, fixed = TRUE)
    f_acqu[idx] <- vapply(idx, function(i) {
      gsub(" .*", f_acqu[i + 1], f_acqu[i])
    }, FUN.VALUE = "")

    parsed <- strsplit(gsub("^##\\$", "", grep("^##\\$", f_acqu, value = TRUE)), "=")
    d_acqu_val <- gsub("^ ", "", vapply(parsed, `[[`, 2, FUN.VALUE = ""))
    names(d_acqu_val) <- paste0("a_", vapply(parsed, `[[`, 1, FUN.VALUE = ""))

    idx_date <- grep("date", names(d_acqu_val), ignore.case = TRUE)
    d_acqu_val[idx_date] <- as.character(
      as.POSIXct("1970-01-01 00:00:00", format = "%Y-%m-%d %H:%M:%S") +
        as.numeric(d_acqu_val[idx_date])
    )

    return(d_acqu_val)
  })

  if (is.list(out)) {
    le <- unique(vapply(out, length, FUN.VALUE = integer(1)))
    if (length(le) == 1) {
      out <- do.call(rbind, out)
    } else {
      unam <- unique(names(unlist(out)))
      out <- do.call(rbind, lapply(out, `[`, unam))
      colnames(out) <- unam
    }
  }

  if (nrow(out) != length(f_list[[1]])) {
    out <- t(out)
  }

  out <- as.data.frame(out, stringsAsFactors = FALSE)
  dtype_num <- vapply(out, function(x) .is_numeric_trycatch(x), logical(1))
  out[, dtype_num] <- lapply(out[, dtype_num, drop = FALSE], as.numeric)
  rownames(out) <- f_list[[2]]

  return(out)
}


#' @title Check for Intact Bruker NMR File Structures
#' @description Verifies that each NMR experiment has a matching `acqus` and `fid` file.
#' @param datapath Character. Directory containing NMR spectra.
#' @param n_max Integer. Maximum number of experiments to process.
#' @param filter Logical. Whether to filter out incomplete experiments.
#' @param recursive Logical. Whether to search directories recursively.
#' @param verbose Integer. Controls verbosity level (0 = silent, 1 = basic, 2 = detailed).
#' @return A list with paths to intact experiments and matching metadata.
#' @keywords internal
#' @seealso \code{\link{.extract_acq_pars1d}}, \code{\link{.filterExp_files}}
.check1d_files_fid <- function(datapath, n_max = 10, filter = TRUE, recursive, verbose) {
  datapath <- gsub(paste0(.Platform$file.sep, "$"), "", datapath)

  f_acqus <- list.files(datapath, pattern = "^acqus$", full.names = TRUE, recursive = recursive, ignore.case = TRUE)
  f_fid <- list.files(datapath, pattern = "^fid$", full.names = TRUE, recursive = recursive, ignore.case = TRUE)

  id_a <- gsub(paste("^", datapath, "/|/acqus$", sep = ""), "", f_acqus)
  id_fid <- gsub(paste("^", datapath, "/|/fid$", sep = ""), "", f_fid)

  idx_a <- which(id_a %in% id_fid)
  idx_fid <- which(id_fid %in% id_a)

  if (length(idx_a) != length(id_a) || length(idx_fid) != length(id_fid)) {
    if (filter) {
      if (verbose > 1) {
        message("Incomplete files detected - reading only experiments with intact folder structure.")
      }
      f_acqus <- f_acqus[idx_a]
      id_a <- id_a[idx_a]
      f_fid <- f_fid[idx_fid]
      id_fid <- id_fid[idx_fid]
    } else {
      message("File system appears to be corrupt for some experiments. Consider setting `filter = TRUE`.")
      return(NULL)
    }
  }

  idm_fid <- match(id_a, id_fid)
  if (any(is.na(idm_fid)) || any(diff(idm_fid) > 1)) {
    stop("Check matching of experiment folders.", call. = FALSE)
  }

  if (length(unique(c(length(f_acqus), length(f_fid)))) != 1) {
    stop("Mismatch in number of acquisitions and FID files.", call. = FALSE)
  }

  if (n_max < length(f_fid)) {
    idx_nmax <- seq_len(n_max)
    f_acqus <- f_acqus[idx_nmax]
    f_fid <- f_fid[idx_nmax]
    id_fid <- id_fid[idx_nmax]
    message("Reached n_max - not all spectra read-in.")
  }

  p_intact <- gsub("/acqus$", "", f_acqus)
  exp_no <- id_fid

  return(list(path = p_intact, exp_no = exp_no, f_acqus = f_acqus, f_fid = f_fid))
}

#' Convert named list to compact string representation
#'
#' Internal utility that converts a named list into a single
#' character string of the form \code{name=value}. Nested
#' sublists are represented recursively using braces.
#'
#' @param lst A named list.
#'
#' @return A character string.
#'
#' @details
#' This function is primarily used for compact parameter logging
#' in provenance metadata and should not be called directly by users.
#'
#' @keywords internal
.list_to_string <- function(lst) {
  paste(
    vapply(names(lst), function(nm) {
      val <- lst[[nm]]
      if (is.list(val)) {
        sprintf("%s={%s}", nm, .list_to_string(val))
      } else {
        sprintf("%s=%s", nm, val)
      }
    }, character(1)),
    collapse = ", "
  )
}

#' @title Select Indices for a Chemical Shift Region
#'
#' @description
#' Returns the indices of the chemical shift vector (\code{ppm}) that fall
#' within the specified range.
#'
#' @param range Numeric vector of length 2 specifying lower and upper bounds
#' (order does not matter).
#' @param ppm Numeric vector. The full chemical shift axis (in ppm).
#'
#' @return Integer vector of indices corresponding to \code{ppm} values within the given range.
#'
#' @export
#'
#' @examples
#' data(covid_raw)
#' X <- covid_raw$X
#' ppm <- covid_raw$ppm
#' idx_tsp <- get_idx(c(-0.1, 0.1), ppm)
#' ppm[range(idx_tsp)]
#' plot(ppm[idx_tsp], X[1, idx_tsp], type = 'l')
#'
#' @family NMR
get_idx <- function(range = c(1, 5), ppm) {
  if (anyNA(ppm) || any(!is.finite(ppm)))
    stop("`ppm` contains NA or non-finite values.", call. = FALSE)

  if (length(range) != 2L)
    stop("`range` must be a numeric vector of length 2.", call. = FALSE)

  if (!is.numeric(ppm))
    stop("`ppm` must be numeric.", call. = FALSE)

  lo <- min(range)
  hi <- max(range)

  which(ppm >= lo & ppm <= hi)
}

#' @rdname get_idx
#' @export
get.idx <- function(range = c(1, 5), ppm) {
  warning("`get.idx` is deprecated and will be removed in future versions. Use `get_idx` instead.", call. = FALSE)
  get_idx(range, ppm)
}

#' @title Subset Optimisation by Reference Matching (STORM)
#'
#' @description
#' Selects an optimal subset of spectra that best match a specified target signal
#' region, improving downstream correlation-based structural analysis such as STOCSY.
#'
#' @param X Numeric matrix (or data.frame) of NMR spectra with samples in rows
#'   and spectral variables in columns.
#' @param ppm Numeric vector of chemical shift values corresponding to the columns of \code{X}.
#' @param b Integer. Half-window size expressed as number of spectral variables
#'   (data points). The effective window width therefore depends on the ppm spacing.
#' @param q Numeric. P-value threshold for including spectral variables when
#'   updating the reference region.
#' @param idx.refSpec Integer. Row index of \code{X} defining the initial reference spectrum.
#' @param shift Numeric vector of length 2 giving the ppm range (minimum, maximum)
#'   defining the initial target signal region.
#'
#' @details
#' STORM iteratively refines a subset of spectra exhibiting consistent signal
#' position and multiplicity within a specified ppm region.
#'
#' Starting from an initial reference spectrum and ppm window:
#' \enumerate{
#'   \item Spectra positively correlated with the current reference signal are retained.
#'   \item A driver peak (maximum intensity within the reference window) is identified.
#'   \item Correlation and covariance are evaluated within a local window of size \code{b}
#'         around the driver peak.
#'   \item The reference region is updated using variables that satisfy the p-value
#'         threshold (\code{q}) and show positive correlation.
#' }
#'
#' The procedure continues until the selected subset stabilises. The resulting
#' row indices define spectra that most consistently represent the structural
#' pattern of the target signal.
#'
#' STORM does not perform metabolite identification directly. Instead, it refines
#' the dataset to enhance structural coherence prior to correlation-based
#' interpretation methods such as STOCSY.
#'
#' @return Integer vector of row indices of \code{X} defining the selected spectral subset.
#'
#' @references
#' Posma, J. M., et al. (2012). Subset optimisation by reference matching (STORM):
#' an optimised statistical approach for recovery of metabolic biomarker structural
#' information from \eqn{^1H} NMR spectra of biofluids.
#' \emph{Analytical Chemistry}, 84(24), 10694–10701.
#'
#' @family structural_annotation
#' @importFrom stats cor cov pt
#' @export
#'
#' @examples
#' ## Simulated example with three Gaussian signals
#' set.seed(123)
#'
#' n <- 100          # number of spectra
#' S <- 1000         # number of spectral variables
#' ppm <- seq(10, 0, length.out = S)
#'
#' gauss <- function(x, centre, width, height) {
#'   height * exp(-((x - centre)^2) / (2 * width^2))
#' }
#'
#' ## Generate base signals
#' sig1 <- gauss(ppm, 7,   0.05, 10)
#' sig2 <- gauss(ppm, 3.5, 0.10,  8)
#' sig3 <- gauss(ppm, 1,   0.07,  5)
#'
#' spectra <- matrix(0, n, S)
#'
#' for (i in seq_len(n)) {
#'   spectra[i, ] <- sig1 + sig2 + rnorm(S, 0, 0.1)
#'   if (i <= 25) {
#'     spectra[i, ] <- spectra[i, ] + sig3
#'   }
#' }
#'
#' ## Apply STORM to refine spectra containing the 1 ppm signal
#' idx <- storm(
#'   X = spectra,
#'   ppm = ppm,
#'   b = 30,
#'   q = 0.05,
#'   idx.refSpec = 1,
#'   shift = c(0.75, 1.25)
#' )
#'
#' length(idx)   # number of spectra retained
storm <- function(X, ppm, b=30, q=0.05, idx.refSpec, shift){

  X_sc <- .scaleMatRcpp(X, seq.int(0, nrow(X) - 1), center=TRUE, scale_type=0)
  X <- X_sc[[1]]
  ref.idx <- get_idx(range=shift, ppm)
  ref <- X[idx.refSpec, ref.idx]

  subset <- 0
  subset1 <- seq_len(nrow(X))
  i <- 1

  subset  <- integer(0)
  subset1 <- seq_len(nrow(X))

  while (!setequal(subset, subset1)) {

    if (length(subset1) <= 1L) {
      message("STORM reduced to a single spectrum; stopping loop.")
      break
    }

    subset <- subset1

    Xr <- X[subset, ref.idx, drop = FALSE]
    r  <- stats::cor(t(Xr), ref)
    df1 <- length(ref) - 2L
    a  <- abs(r) * sqrt(df1 / pmax(1 - r^2, .Machine$double.eps))
    pval <- 2 * stats::pt(-a, df = df1)

    keep <- which(r > 0)
    subset1 <- subset[keep]
    pval <- pval[keep]

    ord <- order(pval)
    subset1 <- subset1[ord]

    idx_driver <- which.max(ref)
    driver <- ref.idx[idx_driver]

    ppm_win <- seq.int(max(1L, driver - (b + 1L)), min(ncol(X), driver + (b + 1L)))

    r  <- stats::cor(X[subset1, ppm_win, drop = FALSE], X[subset1, driver])
    co <- stats::cov(X[subset1, ppm_win, drop = FALSE], X[subset1, driver])
    co <- as.numeric(co)

    df2 <- length(subset1) - 2L
    a <- abs(r) * sqrt(df2 / pmax(1 - r^2, .Machine$double.eps))
    pval <- 2 * stats::pt(-a, df = df2)

    iid <- which(pval < q & r > 0)
    if (length(iid) == 0L) {
      message("No variables passed p/q criteria; stopping.")
      break
    }

    idx_max <- iid[which.max(co[iid])]

    ref_center <- ppm_win[idx_max]
    ref.idx <- seq.int(max(1L, ref_center - (b + 1L)), min(ncol(X), ref_center + (b + 1L)))

    ref <- stats::cov(X[subset1, ref.idx, drop = FALSE], X[subset1, driver])[, 1]
  }

  return(subset1)
}

#' @title Cross-validated ANOVA for O-PLS models
#' @description
#' Performs a cross-validated ANOVA (CV-ANOVA) test for OPLS models. The function compares residuals from a null model and a model using cross-validated predictive scores to assess the significance of the OPLS model.
#'
#' @param smod An object of class \code{m8_model}, generated by function \pkg{opls} in the \pkg{metabom8} package.
#'
#' @details
#' \strong{Interpretation of the CV-ANOVA table}
#'
#' The CV-ANOVA compares two nested linear models:
#' \itemize{
#'   \item \emph{Null model}: \eqn{Y = \beta_0 + \epsilon}
#'   \item \emph{Full model}: \eqn{Y = \beta_0 + \beta_1 t_{\text{pred,cv}} + \epsilon}
#' }
#'
#' where \eqn{t_{\text{pred,cv}}} represents the cross-validated predictive component
#' score(s) obtained from the OPLS model.
#'
#' The ANOVA table contains:
#' \itemize{
#'   \item \strong{SS} – Sum of Squares
#'   \item \strong{DF} – Degrees of Freedom
#'   \item \strong{MS} – Mean Squares (SS / DF)
#'   \item \strong{F_value} – F statistic comparing model vs. null
#'   \item \strong{p_value} – P-value from the F-test
#' }
#'
#' The \emph{Regression} row quantifies the reduction in residual variance
#' achieved by including the cross-validated predictive score(s).
#' The \emph{Residual} row represents the unexplained variance of the full model.
#'
#' The F-statistic is computed as:
#' \deqn{
#' F = \frac{(RSS_0 - RSS_1) / df_{reg}}{RSS_1 / df_{res}}
#' }
#' where \eqn{RSS_0} and \eqn{RSS_1} are the residual sums of squares of the
#' null and full models, respectively.
#'
#' \strong{Predictive interpretation}
#'
#' A small p-value (typically < 0.05) indicates that the cross-validated
#' predictive score(s) significantly reduce residual variance compared to an
#' intercept-only model. This suggests that the OPLS model captures
#' statistically meaningful predictive structure in \eqn{Y}.
#'
#' A large p-value indicates that the predictive component does not explain
#' \eqn{Y} significantly better than chance, implying weak or unstable
#' predictive performance.
#'
#' Importantly, CV-ANOVA evaluates the \emph{linear explanatory power of the
#' cross-validated predictive scores}, not the descriptive separation of the
#' latent space. A model may show visual class separation or moderate R\eqn{^2}
#' yet fail CV-ANOVA if predictive performance is weak.
#'
#' Sample size strongly affects statistical power. With small \eqn{n}, even
#' models with moderate predictive strength may not reach statistical
#' significance.
#'
#' CV-ANOVA is intended for continuous response (regression) models.
#'
#' @return A \code{data.frame} containing:
#' \itemize{
#'   \item \code{SS} - Sum of Squares
#'   \item \code{DF} - Degrees of Freedom
#'   \item \code{MS} - Mean Squares
#'   \item \code{F_value} - F statistic
#'   \item \code{p_value} - P-value from the F-test
#' }
#'
#' @references
#' Eriksson, L., et al. (2008). CV-ANOVA for significance testing of PLS and OPLS models. \emph{Journal of Chemometrics}, 22(11-12), 594–600.
#'
#'
#' @examples
#' data("covid")
#'
#' X <- covid$X
#' Y <- as.numeric(factor(covid$an$type)) - 1
#'
#' scaling <- uv_scaling(center=TRUE)
#' cv <- balanced_mc(10, split=2/3, type='R', probs = c(0, 0.5, 1))
#' mod <- opls(X, Y, scaling, cv)
#'
#' cv_anova(mod)
#'
#' @family model_validation
#' @export
#' @importFrom stats lm pf residuals
cv_anova <- function(smod) {
  if (!(inherits(smod, "m8_model") && smod@engine == "opls" && smod@ctrl$type == "R")) {
    stop("Requires an m8_model with engine == 'opls' and type regression", call. = FALSE)
  }

  # pick response: numeric Y preferred; otherwise first dummy column
  y <- NULL
  if (!is.null(smod@fit$Y) && is.numeric(smod@fit$Y)) {
    y <- as.numeric(smod@fit$Y)
  } else {
    stop("Could not find a usable response in smod@fit$Y.", call. = FALSE)
  }

  tcv <- scores(smod, cv=TRUE)
  if (length(y) != length(tcv)) stop("Length mismatch between Y and t_pred_cv.", call. = FALSE)

  n <- length(y)

  m0 <- stats::lm(y ~ 1)
  m1 <- stats::lm(y ~ 1 + tcv)

  rss0 <- sum(residuals(m0)^2)
  rss1 <- sum(residuals(m1)^2)

  ss_reg <- rss0 - rss1
  ss_res <- rss1
  ss_tot <- rss0

  df_tot <- n - 1
  df_res <- n - 2
  df_reg <- df_tot - df_res  # = 1

  ms_reg <- ss_reg / df_reg
  ms_res <- ss_res / df_res

  F_val <- ms_reg / ms_res
  p_val <- stats::pf(F_val, df_reg, df_res, lower.tail = FALSE)

  out <- data.frame(
    SS = c(ss_tot, ss_reg, ss_res, NA_real_),
    DF = c(df_tot, df_reg, df_res, NA_integer_),
    MS = c(ss_tot/df_tot, ms_reg, ms_res, NA_real_),
    F_value = c(NA_real_, NA_real_, NA_real_, F_val),
    p_value = c(NA_real_, NA_real_, NA_real_, p_val),
    row.names = c("Total corrected", "Regression", "Residual", "RESULT")
  )

  out
}

#' @title Check Y for missing values (PLS context)
#' @description Checks input Y for NA/NaN/Inf values, determines analysis type
#' (Regression or Discriminant Analysis), and returns cleaned Y matrix,
#' associated levels, and type.
#' @param Y A vector, matrix, or data.frame. Target variable.
#' @return A list: cleaned Y matrix, levels (if applicable), and type ('R' or 'DA')
#' @keywords internal
.checkYclassNas <- function(Y) {
  if (!inherits(Y, c("matrix", "data.frame"))) {
    Y <- matrix(Y, nrow = length(Y), ncol = 1)
  } else {
    Y <- as.matrix(Y)
  }

  if (anyNA(Y) || any(is.nan(Y)) || any(is.infinite(Y))) {
    stop("Input Y contains NA, NaN, or Inf values.", call. = FALSE)
  }

  if (is.numeric(Y)) {
    type <- "R"
    Y_out <- .prepareY(Y)
    Y <- Y_out[[1]]
    levs <- data.frame()
  } else {
    type <- "DA"
    Y_out <- .prepareY(Y)
    Y <- Y_out[[1]]

    levs <- unique(apply(Y, 2, function(x) length(unique(x))))
    if (length(levs) == 1 && levs[1] == 1) {
      stop("Input Y has only a single level.", call. = FALSE)
    }
    message("Performing discriminant analysis.")
  }

  if (ncol(Y) > 1) {
    type <- paste0(type, "-mY")
  }

  return(c(Y_out, type))
}


#' @title Check X matrix (PLS context)
#' @description Validates that input \code{X} is a numeric matrix and does not contain missing, NaN, or infinite values.
#' @param X Input data (univariate or multivariate), formatted as a matrix.
#' @return \code{NULL} if all checks pass; otherwise throws an informative error.
#' @keywords internal
#' @rdname checkXclassNas
.checkXclassNas <- function(X) {
  if (!is.matrix(X) || !is.numeric(X)) {
    stop(sprintf("Input X must be a numeric matrix. Use matrix() or as.matrix() to convert."), call. = FALSE)
  }

  if (any(is.na(X) | is.nan(X) | is.infinite(X))) {
    stop("Input X contains missing (NA), NaN, or infinite values.", call. = FALSE)
  }

  return(invisible(NULL))
}

#' @title Check Dimensions of X and Y Matrices (PLS Context)
#' @description
#' Checks that the input matrices \code{X} and \code{Y} have compatible dimensions
#' for Partial Least Squares (PLS) analysis.
#' Specifically, the number of observations (rows) must match, and
#' the number of variables (columns) in \code{X} must be greater than in \code{Y}.
#'
#' @param X Numeric matrix. Predictor data matrix (observations x variables).
#' @param Y Numeric matrix. Response data matrix (observations x variables).
#'
#' @return
#' Invisibly returns \code{NULL} if checks pass.
#' Throws an error if dimensions are incompatible.
#' @keywords internal
.checkDimXY <- function(X, Y) {
  if (nrow(X) != nrow(Y)) {
    stop("Dimensions of input X and Y do not match.", call. = FALSE)
  }
  if (ncol(X) <= ncol(Y)) {
    stop("Number of variables (columns) in X should be higher than in Y.", call. = FALSE)
  }
  invisible(NULL)
}

#' Evaluate model fit progression based on cross-validated performance
#'
#' Determines whether an additional component should be fitted based on
#' cross-validated generalization performance. For regression models
#' (`type = "R"`), the decision is based on the \eqn{Q^2} statistic.
#' For discriminant analysis models (`type = "DA"`), the decision is
#' based on the cross-validated AUROC (area under the ROC curve).
#'
#' AUROC is computed on cross-validation test folds to ensure that it
#' reflects out-of-sample classification performance, analogous to the
#' use of \eqn{Q^2} in regression.
#'
#' @param type Character. Model type: `"R"` for regression or `"DA"` for
#'   discriminant analysis. The value may optionally include a suffix
#'   such as `"-mY"` to indicate multi-response models.
#' @param q2s Numeric vector of \eqn{Q^2} values for the fitted
#'   components (used for regression models).
#' @param cv_auc Numeric vector of cross-validated AUROC values for
#'   the fitted components (used for classification models).
#' @param pc_max Integer. Maximum number of components allowed.
#' @param min_delta Numeric. Minimum improvement in the cross-validated
#'   performance metric required to justify fitting an additional
#'   component.
#' @param sat_q2 Numeric. Saturation threshold for \eqn{Q^2} in
#'   regression models.
#' @param sat_auc Numeric. Saturation threshold for cross-validated
#'   AUROC in discriminant analysis models.
#' @param min_q2 Numeric. Minimum acceptable \eqn{Q^2} value for the
#'   first component.
#' @param min_auc Numeric. Minimum acceptable cross-validated AUROC for
#'   the first component.
#'
#' @return A list containing:
#' \describe{
#'   \item{stop}{Logical indicating whether model fitting should stop.}
#'   \item{reason}{Character string describing the stopping condition
#'     (e.g., `"pc_max"`, `"cv_decreased"`, `"cv_improvement_negligible"`).}
#'   \item{metric}{The current cross-validated performance value.}
#'   \item{delta}{Improvement relative to the previous component.}
#' }
#'
#' @keywords internal
.evalFit <- function(type, q2s, cv_auc,
                     pc_max = 5,
                     min_delta = 0.05,
                     sat_q2  = 0.98,
                     sat_auc = 0.97,
                     min_q2    = 0.05,
                     min_auc   = 0.60
) {

  type <- strsplit(type, '-')[[1]][1]

  cv_est <- if (type == "R") q2s else cv_auc

  if (any(is.na(cv_est))) {
    stop("Something went wrong: CV metric contains NA.", call. = FALSE)
  }

  nc <- length(cv_est)

  if (type == "R" && nc ==1 && cv_est[nc] <= min_q2) {
    return(list(
      stop   = TRUE,
      reason = sprintf("q2_below_%s", min_q2),
      metric = cv_est[nc],
      delta  = if (nc > 1) cv_est[nc] - cv_est[nc - 1] else NA_real_
    ))
  }

  if (type == "DA" && nc ==1 && cv_est[nc] <= min_auc) {
    return(list(
      stop   = TRUE,
      reason = sprintf("auc_below_%s", min_auc),
      metric = cv_est[nc],
      delta  = if (nc > 1) cv_est[nc] - cv_est[nc - 1] else NA_real_
    ))
  }

  if (nc < 2) {
    return(list(
      stop   = FALSE,
      reason = NULL,
      metric = cv_est[nc],
      delta  = NA_real_
    ))
  }

  if (nc == pc_max) {
    return(list(
      stop   = TRUE,
      reason = "pc_max",
      metric = cv_est[nc],
      delta  = if (nc > 1) cv_est[nc] - cv_est[nc - 1] else NA_real_
    ))
  }

  delta <- cv_est[nc] - cv_est[nc - 1]

  thresh <- if (type == "R") sat_q2 else sat_auc

  if (delta < 0) {
    return(list(
      stop   = TRUE,
      reason = "cv_decreased",
      metric = cv_est[nc],
      delta  = delta
    ))
  }

  if (delta < min_delta) {
    return(list(
      stop   = TRUE,
      reason = "cv_improvement_negligible",
      metric = cv_est[nc],
      delta  = delta
    ))
  }

  if (cv_est[nc] > thresh) {
    return(list(
      stop   = TRUE,
      reason = "saturation",
      metric = cv_est[nc],
      delta  = delta
    ))
  }

  return(list(
    stop   = FALSE,
    reason = NULL,
    metric = cv_est[nc],
    delta  = delta
  ))
}


#' @title Prepare Response Vector for OPLS/OPLS-DA
#' @description
#' Converts a response vector \code{Y} into a numeric matrix suitable for regression or classification models.
#' - Numeric input is returned as a column matrix.
#' - Factor or character input:
#'   - If binary (2 levels): returns a single numeric column (values 1 and 2).
#'   - If multiclass (>2 levels): returns a dummy matrix with 1 for presence and -1 for absence.
#' Also returns a mapping between original labels and numeric values (only for categorical input).
#'
#' @param Y A numeric, factor, or character vector.
#'
#' @return A list with:
#' \describe{
#'   \item{\code{[[1]]}}{Numeric matrix of processed response.}
#'   \item{\code{[[2]]}}{Data frame mapping class labels to encoding (empty for numeric input).}
#' }
#' @keywords internal
.prepareY <- function(Y) {
  if (!is.numeric(Y)) {
    Y_factor <- factor(Y)
    Y_levels <- levels(Y_factor)
    n_levels <- length(Y_levels)

    if (n_levels == 2) {
      Y_num <- as.numeric(Y_factor)
      mapping <- data.frame(Level = Y_levels,
                            Code = sort(unique(Y_num)),
                            stringsAsFactors = FALSE)
      return(list(matrix(Y_num, ncol = 1), mapping))

    } else {
      Y_mat <- vapply(
        Y_levels,
        function(lvl) ifelse(Y_factor == lvl, 1, -1),
        numeric(length(Y_factor))
      )
      colnames(Y_mat) <- Y_levels

      mapping <- data.frame(Level = Y_levels,
                            Column = seq_along(Y_levels),
                            stringsAsFactors = FALSE)

      return(list(Y_mat, mapping))
    }

  } else {
    if (is.matrix(Y) && ncol(Y) > 1) {
      stop("Only single-column numeric input is supported for Y at this time.", call. = FALSE)
    }
    return(list(matrix(Y, ncol = 1), data.frame()))
  }
}




#' @title Check if input is numeric-like using tryCatch
#' @description
#' Tests whether a vector \code{x} can be safely coerced to numeric without
#' producing \code{NA}s (excluding original \code{NA}s) or warnings.
#'
#' This function attempts to coerce the input to numeric and returns \code{TRUE}
#' if all non-missing elements convert without coercion failure; otherwise \code{FALSE}.
#' It catches warnings and errors during coercion and treats those as non-numeric.
#' @param x A vector to test for numeric coercion.
#' @return Logical \code{TRUE} if \code{x} is numeric-like, \code{FALSE} otherwise.
#' @keywords internal
.is_numeric_trycatch <- function(x) {
  tryCatch({
    num_x <- as.numeric(x)
    if (any(is.na(num_x) & !is.na(x))) {
      FALSE
    } else {
      TRUE
    }
  }, warning = function(w) {
    FALSE
  }, error = function(e) {
    FALSE
  })
}



#' @title Check Validity of Selected Principal or Orthogonal Component
#' @description
#' Internal consistency check for principal (or orthogonal) component selection used in plotting or extraction functions.
#'
#' @param pc Character or numeric. Selected component, e.g., \code{1} or \code{"o1"}.
#' @param mod An object of class \code{PCA_metabom8} or \code{OPLS_metabom8}.
#' @param le Integer. Expected maximum length of \code{pc}; usually 1.
#' @param type Character. Type of component to check: \code{"p"} for projection or \code{"t"} for scores. Currently only used for logic branching.
#'
#' @return Nothing. Function is used for its side effect (i.e., error checking).
#'
#' @keywords internal
.check_pc_model <- function(pc, mod, le = 1, type = 'p') {

  if (is.na(pc) || is.infinite(pc) || length(pc) > le) {
    stop("Check 'pc' argument: It must be finite and of expected length.", call. = FALSE)
  }

  mod_class <- class(mod)[1]

  if (mod_class == "PCA_metabom8") {
    if (!is.numeric(pc)) stop("PC value must be numeric for PCA.", call. = FALSE)
    if (max(pc) > ncol(mod@p)) stop("PC value exceeds number of PCA components.", call. = FALSE)
  }

  if (mod_class == "OPLS_metabom8") {
    idx_orth <- grepl("o", pc)
    if (any(idx_orth)) {
      pc1 <- as.numeric(gsub("o", "", pc))
      if (any(is.na(pc1)) || any(is.infinite(pc1))) {
        stop("Invalid orthogonal component identifier in 'pc'.", call. = FALSE)
      }
      if (any(pc1 > nrow(mod@p_orth))) {
        stop("Requested orthogonal component exceeds available components.", call. = FALSE)
      }
    } else {
      if (!is.numeric(as.numeric(pc))) {
        stop("Predictive component index should be numeric or coercible to numeric.", call. = FALSE)
      }
      if (length(mod@p_pred) == 0) {
        stop("Predictive component appears to be missing in model object.", call. = FALSE)
      }
    }
  }
}

#' @title \eqn{R^2} and \eqn{Q^2} Calculation for OPLS Models
#'
#' @description
#' Computes the coefficient of determination (\eqn{R^2} or \eqn{Q^2}) for OPLS regression models using the formula:
#' \deqn{R^2 = 1 - \frac{PRESS}{TSS}}, where PRESS is the prediction error sum of squares,
#' and TSS is the total sum of squares. If `ytss` is not provided, it is computed directly from `Y`.
#'
#' This function supports matrix inputs (e.g., for multi-class outcomes) and averages across columns
#' to return a single summary \eqn{R^2}/\eqn{Q^2} value.
#'
#' @param Y Numeric matrix. True response values. For classification, this should be a dummy matrix.
#' @param Yhat Numeric matrix. Predicted response values from the model.
#' @param ytss Numeric (optional). Total sum of squares of `Y`. If not provided, it will be computed internally.
#'
#' @return A single numeric value representing \eqn{R^2} or \eqn{Q^2}.
#'
#' @keywords internal
.r2 <- function(Y, Yhat, ytss = NULL) {
  if (!is.numeric(Y)) stop("Y must be numeric")
  if (!is.numeric(Yhat)) stop("Yhat must be numeric")

  if (is.null(dim(Y))) {
    if (length(Y) != length(Yhat)) stop("Lengths of Y and Yhat must match")
    ncol_Y <- 1
  } else {
    if (!all(dim(Y) == dim(Yhat))) stop("Dimensions of Y and Yhat must match")
    ncol_Y <- ncol(Y)
  }

  if (!is.null(ytss) && !is.numeric(ytss)) stop("ytss must be numeric or NULL")

  press <- sum((Y - Yhat)^2, na.rm = TRUE) / ncol_Y

  if (is.null(ytss)) {
    ytss <- sum((Y - mean(Y))^2, na.rm = TRUE) / ncol_Y
  }

  return(1 - (press / ytss))
}


#' @title Check Consistency of Spectral Data and ppm Vector
#'
#' @description
#' Validates that the NMR data matrix and the chemical shift reference vector
#' (\code{ppm}) are consistent.
#' Specifically, it checks that
#' \itemize{
#'   \item \code{ppm} contains no NA or infinite values.
#'   \item The number of columns in \code{X} matches the length of \code{ppm}.
#' }
#'
#' This function is typically used as a preliminary data validation step before downstream spectral analysis.
#'
#' @param X Numeric matrix. NMR spectral matrix with samples in rows and variables in columns (e.g., intensities).
#' @param ppm Numeric vector. Chemical shift reference values corresponding to columns of `X`.
#'
#' @return Logical. Returns \code{TRUE} if inputs are valid, \code{FALSE} otherwise.
#'
#' @keywords internal
.check_X_ppm <- function(X, ppm) {

  if (is.null(ppm) || !is.numeric(ppm)) return(FALSE)

  if (anyNA(ppm) || any(!is.finite(ppm))) return(FALSE)

  if (is.null(dim(X)) || ncol(X) != length(ppm)) return(FALSE)

  if (anyDuplicated(ppm)) return(FALSE)
  TRUE
}


#' @title Ensure Input is a Row Matrix
#'
#' @description
#' Ensures that the input object `X` is in row-matrix form. If `X` is a numeric
#' vector  (i.e., has no dimensions), it is converted to a matrix with 1 row
#' and \code{length(X)} columns. This is helpful for standardizing inputs before
#' applying matrix operations.
#'
#' @param X Numeric vector, matrix, or array. Typically NMR data where rows represent spectra.
#'
#' @return A numeric matrix with one row if input was a vector; otherwise, returns `X` unchanged.
#'
#' @keywords internal
.dimX <- function(X) {

  if (!is.numeric(X))
    stop("X must be numeric.", call. = FALSE)

  if (is.atomic(X) && is.null(dim(X))) {
    return(matrix(X, nrow = 1L))
  }

  if (is.matrix(X) || is.data.frame(X)) {
    return(as.matrix(X))
  }

  stop("X must be a numeric vector, matrix, or data.frame.", call. = FALSE)
}

### .checkXclassNas

test_that(".checkXclassNas passes for valid numeric matrix", {
  X <- matrix(rnorm(10), nrow = 5)
  expect_silent(.checkXclassNas(X))
})

test_that(".checkXclassNas fails for non-matrix input", {
  x_df <- data.frame(a = 1:5, b = 6:10)
  expect_error(.checkXclassNas(x_df), "Input X must be a numeric matrix")
})

test_that(".checkXclassNas fails with NA/NaN/Inf values", {
  X <- matrix(rnorm(9), nrow = 3)
  X[1, 1] <- NA
  expect_error(.checkXclassNas(X), "contains missing")

  X[1, 1] <- NaN
  expect_error(.checkXclassNas(X), "contains missing")

  X[1, 1] <- Inf
  expect_error(.checkXclassNas(X), "contains missing")
})

test_that(".checkXclassNas fails for non-numeric matrix", {

  X <- matrix(letters[1:6], nrow = 3)

  expect_error(
    .checkXclassNas(X),
    "Input X must be a numeric matrix"
  )

})

### .checkDimXY

test_that(".checkDimXY passes for valid X and Y", {
  X <- matrix(rnorm(30), nrow = 10)
  Y <- matrix(rnorm(20), nrow = 10)
  expect_silent(.checkDimXY(X, Y))
})

test_that(".checkDimXY fails if row counts differ", {
  X <- matrix(rnorm(20), nrow = 5)
  Y <- matrix(rnorm(24), nrow = 6)
  expect_error(.checkDimXY(X, Y), "Dimensions of input X and Y do not match")
})

test_that(".checkDimXY fails if X has fewer columns than Y", {
  X <- matrix(rnorm(20), nrow = 10, ncol = 2)
  Y <- matrix(rnorm(30), nrow = 10, ncol = 3)
  expect_error(.checkDimXY(X, Y), "Number of variables .* should be higher than")
})


### .checkYclassNas

test_that(".checkYclassNas works for numeric Y (regression)", {
  Y <- rnorm(10)
  result <- .checkYclassNas(Y)
  expect_true(is.matrix(result[[1]]))
  expect_equal(ncol(result[[1]]), 1)
  expect_equal(result[[3]], "R")
})

test_that(".checkYclassNas works for binary factor Y (classification)", {
  Y <- factor(rep(c("A", "B"), each = 5))
  result <- .checkYclassNas(Y)
  expect_true(is.matrix(result[[1]]))
  expect_equal(ncol(result[[1]]), 1)
  expect_equal(result[[3]], "DA")
})

test_that(".checkYclassNas works for multi-class factor Y", {
  Y <- factor(c("A", "B", "C", "A", "B", "C"))
  result <- .checkYclassNas(Y)
  expect_true(is.matrix(result[[1]]))
  expect_gt(ncol(result[[1]]), 1)
  expect_equal(result[[3]], "DA-mY")
})

test_that(".checkYclassNas errors for constant class", {
  Y <- rep("A", 5)
  expect_error(.checkYclassNas(Y), "only a single level")
})

test_that(".checkYclassNas errors on NA/NaN/Inf", {
  Y <- c(1, 2, NA, 4)
  expect_error(.checkYclassNas(Y), "contains NA")

  Y <- c(1, 2, NaN, 4)
  expect_error(.checkYclassNas(Y), "contains NA")

  Y <- c(1, 2, Inf, 4)
  expect_error(.checkYclassNas(Y), "contains NA")
})

test_that(".checkYclassNas preserves row order", {

  Y <- factor(c("A","B","A","B"))
  result <- .checkYclassNas(Y)
  expect_equal(nrow(result[[1]]), length(Y))

})

### .prepareInputs Tests

test_that("Regression task: numeric Y vector and matrix", {
  set.seed(1)
  X <- matrix(rnorm(50), nrow = 10)
  Y <- rnorm(10)

  result <- .prepareInputs(X, Y, center = TRUE, scale = "None")
  expect_true(is.matrix(result$X))
  expect_equal(result$type, "R")
  expect_true(is.numeric(result$Y))
  expect_true(is.matrix(result$Y))
  expect_true(ncol(result$Y) == 1)

  result2 <- .prepareInputs(X, matrix(Y, ncol = 1), center = TRUE, scale = "UV")
  expect_equal(result2$type, "R")
  expect_true(is.matrix(result$Y))
  expect_equal(ncol(result$Y),1)
})

test_that("Binary classification: factor Y vector and matrix", {
  set.seed(2)
  X <- matrix(rnorm(50), nrow = 10)
  Y <- factor(sample(c("A", "B"), 10, replace = TRUE))

  result <- .prepareInputs(X, Y, center = FALSE, scale = "Pareto")
  expect_equal(result$type, "DA")
  expect_true(is.numeric(result$Y))
  expect_equal(ncol(result$Ydummy), 2)

  # Y as single-column matrix
  Y_mat <- matrix(as.character(Y), ncol = 1)
  result2 <- .prepareInputs(X, Y_mat, center = FALSE, scale = "None")
  expect_equal(result2$type, "DA")
  expect_equal(ncol(result2$Ydummy), 2)
})

test_that("Multiclass classification: 3-level factor", {
  set.seed(3)
  X <- matrix(rnorm(60), nrow = 10)
  Y <- factor(sample(c("D1", "B1", "D2"), 10, replace = TRUE))

  result <- .prepareInputs(X, Y, center = TRUE, scale = "UV")
  expect_equal(result$type, "DA-mY")
  expect_equal(ncol(result$Ydummy), 2)
  expect_equal(nrow(result$Ydummy), 3)
})

test_that("Invalid center or scale throws errors", {
  X <- matrix(rnorm(20), nrow = 5)
  Y <- rnorm(5)

  expect_error(.prepareInputs(X, Y, center = "yes", scale = "UV"), "Check center")
  expect_error(.prepareInputs(X, Y, center = TRUE, scale = "BadScale"), "Check scale")
})

test_that(".prepareY keeps class order deterministic", {
  y <- c("b","a","c","a")
  result <- .prepareY(y)
  expect_equal(colnames(result[[1]]), sort(unique(y)))
})


### .prepareY

test_that(".prepareY handles numeric vector (regression case)", {
  y <- c(1.2, 3.4, 5.6)
  result <- .prepareY(y)

  expect_type(result, "list")
  expect_true(is.matrix(result[[1]]))
  expect_equal(dim(result[[1]]), c(3, 1))
  expect_equal(result[[1]][, 1], y)
  expect_equal(ncol(result[[1]]), 1)
  expect_equal(result[[2]], data.frame())  # mapping should be empty
})

test_that(".prepareY handles binary classification (factor)", {
  y <- factor(c("A", "B", "A", "B"))
  result <- .prepareY(y)

  expect_type(result, "list")
  expect_true(is.matrix(result[[1]]))
  expect_equal(dim(result[[1]]), c(4, 1))
  expect_equal(unique(as.vector(result[[1]])), c(1, 2))
  expect_true(all(result[[2]]$Level %in% c("A", "B")))
  expect_equal(sort(result[[2]]$Code), c(1, 2))
})

test_that(".prepareY handles binary classification (character)", {
  y <- c("yes", "no", "yes", "no")
  result <- .prepareY(y)

  expect_type(result, "list")
  expect_true(is.matrix(result[[1]]))
  expect_equal(dim(result[[1]]), c(4, 1))
  expect_equal(sort(unique(as.vector(result[[1]]))), c(1, 2))
  expect_equal(nrow(result[[2]]), 2)
})

test_that(".prepareY handles multiclass classification", {
  y <- factor(c("cat", "dog", "cat", "mouse", "dog"))
  result <- .prepareY(y)

  y_mat <- result[[1]]
  mapping <- result[[2]]

  expect_type(result, "list")
  expect_true(is.matrix(y_mat))
  expect_equal(dim(y_mat), c(5, 3))
  expect_true(all(y_mat %in% c(1, -1)))
  expect_equal(mapping$Level, c("cat", "dog", "mouse"))
  expect_equal(mapping$Column, 1:3)
})

test_that(".prepareY returns correct colnames for multiclass", {
  y <- c("low", "medium", "high", "medium", "high")
  result <- .prepareY(y)
  expect_equal(colnames(result[[1]]), c("high", "low", "medium"))
})


test_that(".prepareY errors on multi-column numeric input", {
  Y <- matrix(1:6, nrow = 3, ncol = 2)
  expect_error(.prepareY(Y), "Only single-column numeric input is supported for Y at this time")
})


test_that("prepareInputs produces consistent modeling input", {

  set.seed(1)

  X <- matrix(rnorm(50), nrow = 10)
  Y <- factor(rep(c("A","B"),5))

  res <- .prepareInputs(X, Y, center = TRUE, scale = "UV")

  expect_true(is.matrix(res$X))
  expect_true(is.matrix(res$Y))
  expect_equal(nrow(res$X), nrow(res$Y))

})

#' Fit an Orthogonal Partial Least Squares (O-PLS) model
#'
#' Fits a supervised Orthogonal Partial Least Squares (O-PLS) model
#' using a NIPALS-based algorithm with optional cross-validation and
#' automatic component selection.
#'
#' O-PLS decomposes the predictor matrix into:
#' \itemize{
#'   \item One predictive component capturing variation correlated with \code{Y}
#'   \item Orthogonal components capturing structured variation in \code{X}
#'         unrelated to \code{Y}
#' }
#'
#' @param X Numeric matrix of predictors (rows = samples, columns = variables).
#' @param Y Numeric matrix or factor vector of responses.
#' @param scaling A scaling strategy object (e.g., \code{uv_scaling(center = TRUE)}),
#'   specifying model-internal centering and/or scaling applied during fitting.
#'   This does not modify the original spectral matrix.
#' @param validation_strategy A cross-validation strategy object defining how
#'   resampling is performed (e.g., k-fold, Monte Carlo).
#'
#' @details
#' Predictive and orthogonal components are estimated sequentially.
#' Cross-validated performance metrics (e.g., Q², R², classification AUC)
#' are computed for each model configuration according to the supplied
#' \code{validation_strategy}.
#'
#' The model extracts a single predictive component and iteratively
#' adds orthogonal components until the \code{stopRule} indicates
#' overfitting or \code{maxPCo} is reached.
#'
#' Scaling specified via \code{scaling} is applied internally during model
#' fitting and does not alter the input matrix \code{X}. Spectral preprocessing
#' (e.g., alignment or baseline correction) should be performed prior to
#' model fitting.
#'
#' The returned model object stores:
#' \itemize{
#'   \item Predictive and orthogonal component models
#'   \item Cross-validation results
#'   \item Performance metrics (R², Q², AUC)
#'   \item Model control parameters
#'   \item Input data provenance metadata
#'   \item Session information for reproducibility
#' }
#'
#' @return
#' An object of class \code{m8_model} containing the fitted O-PLS model,
#' cross-validation results, and performance statistics.
#'
#' @seealso \code{\link{pls}}, \code{\link{uv_scaling}}
#'
#' @family modelling
#' @examples
#' data(covid)
#'
#' cv <- balanced_mc(k=5, split=2/3)
#' scaling <- uv_scaling(center=TRUE)
#' model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
#'
#' show(model)
#' summary(model)
#'
#' # scores
#' Tp <- scores(model)
#' To <- scores(model, orth=TRUE)
#'
#' t2 <- hotellingsT2(cbind(Tp, To))
#' ell <-ellipse2d(t2)
#'
#' plot(Tp, To, asp = 1,
#'   col = as.factor(covid$an$type),
#'   xlim = range(c(Tp, ell$x)),
#'   ylim = range(c(To, ell$y))
#'  )
#' lines(ell$x, ell$y, col = "grey", lty=2)
#'
#' # loadings & vip's
#' Pp <- loadings(model)
#' Po <- loadings(model, orth=TRUE)
#' vips <- vip(model)
#'
#' x=covid$ppm
#' y = Pp * apply(covid$X, 2, sd)
#'
#' palette <-  colorRampPalette(c("blue", "cyan", "yellow", "red"))(100)
#' idx <- cut(vips, breaks = 100, labels = FALSE)
#' plot(x, y, type = "n", xlim = rev(range(x)), xlab='ppm', ylab='t_pred_sc')
#'
#' for (i in seq_len(length(x) - 1)) {
#'   segments(x[i], y[i], x[i+1], y[i+1], col = palette[idx[i]], lwd = 2)
#' }
#'
#' @export
opls <- function(X, Y, scaling, validation_strategy) {

  cl <- match.call()

  engine <- .oplsEngine(X, Y, scaling, validation_strategy, maxPCo=5)

  obj <- methods::new("m8_model",

      engine = "opls",
      fit = engine$full,
      prep = scaling,
      cv = engine$cv,
      ctrl = engine$ctrl,
      dims = engine$dims,

      provenance = attributes(X),
      call = cl,
      session = list(
        engine_version = "nipals_v1",
        R.version = getRversion(),
        pkg = list(
          m8 = as.character(utils::packageVersion("metabom8")),
          Rcpp = as.character(utils::packageVersion("Rcpp"))
        )
      )
  )

  methods::validObject(obj)
  obj

}

.oplsEngine <- function(X, Y, scaling, validation_strategy, maxPCo, cv=NULL) {

  #### checking scaling
  .arg_check_scaling(scaling)

  inputs <- .prepareInputs(X, Y, scaling$center, scaling$scale)
  type <- inputs$type
  is_multi_Y <- grepl('mY', type)
  if (is_multi_Y) {stop("Multi-response Y is currently not supported", call. = FALSE)}
  n  <- nrow(inputs$Y)
  p  <- ncol(inputs$X)
  mo <- maxPCo

  cv <- .arg_check_cv(cv_pars=validation_strategy, model_type=type, n=n, Y_prepped=inputs$Y)

  preppedX <- prep_X(scaling, inputs$X)
  XcsTot <- preppedX$X
  YcsTot <- .scaleMatRcpp(inputs$Y, 0:(nrow(inputs$Y) - 1), center = scaling$center, scale_type = inputs$scale_code)[[1]]

  tssx <- .tssRcpp(XcsTot)
  tssy <- .tssRcpp(YcsTot) / ncol(YcsTot)

  Ycs_fold <- lapply(cv$train, function(idc)
    .scaleMatRcpp(inputs$Y, idc - 1, scaling$center, inputs$scale_code)[[1]]
  )

  tt         <- NULL
  opls_filt  <- NULL

  t_orth <- matrix(0, n, mo)
  p_orth <- matrix(0, mo, p)
  w_orth <- matrix(0, mo, p)

  q <- ncol(inputs$Y)
  k <- length(cv$train)

  acc <- list(
    sum_test  = matrix(0, n, q),
    n_test    = integer(n),
    sum_train = matrix(0, n, q),
    n_train   = integer(n)
  )


  r2_comp <- r2x_comp <- q2_comp <- aucs_tr <- aucs_te <- array()
  nc <- 1
  overfitted <- FALSE

  while(!overfitted){

    res <- if (nc == 1)
      .oplsComponentCv(
        X        = inputs$X,
        cv.set   = cv$train,
        Ycs_fold = Ycs_fold,
        nc       = nc,
        mod.cv   = NULL,
        acc      = acc
      )
    else
      .oplsComponentCv(
        X        = NULL,
        cv.set   = cv$train,
        Ycs_fold = Ycs_fold,
        nc       = nc,
        mod.cv   = tt,
        acc      = acc
      )

    tt  <- res$mod.cv
    acc <- res$acc

    preds_test <- acc$sum_test
    preds_test[acc$n_test > 0,] <- preds_test[acc$n_test > 0,] / acc$n_test[acc$n_test > 0]

    preds_train <- acc$sum_train
    preds_train[acc$n_train > 0,] <- preds_train[acc$n_train > 0,] / acc$n_train[acc$n_train > 0]

    perf <- .evalComponentPerformance(
      cv_inst      = cv,
      type         = type,
      is_multi_Y   = is_multi_Y,
      Y            = Y,
      preds_train  = preds_train,
      preds_test   = preds_test,
      class_memb   = if (is.data.frame(inputs$Ydummy)) inputs$Ydummy$Level[inputs$Ydummy$Column] else NULL,
      YcsTot       = YcsTot,
      tssy         = tssy
    )

    q2_comp[nc] <- perf$q2
    r2_comp[nc] <- perf$r2
    aucs_te[nc] <- perf$aucs_te
    aucs_tr[nc] <- perf$aucs_tr

    fit_eval <- .evalFit(type, q2_comp, aucs_te, maxPCo)
    overfitted <- fit_eval$stop


    if (nc >= 2) {
      if (nc == 2) {
        opls_filt <- .nipOplsRcpp(XcsTot, YcsTot)
      } else {
        opls_filt <- .nipOplsRcpp(opls_filt$X_res, YcsTot)
      }
      t_orth[,nc-1] <- opls_filt$t_o
      p_orth[nc-1,] <- opls_filt$p_o
      w_orth[nc-1,] <- t(opls_filt$w_o)

      pred_comp <- .nipPlsCompRcpp(opls_filt$X_res, YcsTot, it_max = 800, eps = 1e-8)

      Tx <- .ensure_matrix(pred_comp$t_x, ncol = 1L)
      Px <- .ensure_matrix(pred_comp$p_x, ncol=length(pred_comp$p_x))

      X_str <- Tx %*% Px

      r2x_comp[nc - 1] <- .r2(opls_filt$X_res, X_str, NULL)
    }

    if (overfitted) {
      if(nc >1){
        message(sprintf("An O-PLS-%s model with 1 predictive and %d orthogonal components was fitted.", type, nc - 1))
        pred_comp$t_o_cv <- matrix(
          .extMeanCvFeat(tt, "t_xo")[, seq_len(nc-1), drop = FALSE],
          ncol = nc - 1
        )

        pred_comp$t_cv <- matrix(
          .extMeanCvFeat(tt, "t_xp")[, nc-1, drop = FALSE],
          ncol = 1
        )
        pred_comp$t_o <- t_orth
      } else{
        pred_comp <- "not fitted as model does not generalise"
        message(sprintf("The O-PLS-%s model does not gerenalise: %s -> no stable predictive structure detected.", type, fit_eval$reason))
      }

      break

    }else{
      nc <- nc + 1
    }
  }

  nc_orth <- max(0L, nc - 1L)

  t_orth_out <- if (nc_orth > 0L) t_orth[, seq_len(nc_orth), drop = FALSE] else
    matrix(numeric(0), nrow = n, ncol = 0)

  p_orth_out <- if (nc_orth > 0L) p_orth[seq_len(nc_orth), , drop = FALSE] else
    matrix(numeric(0), nrow = 0, ncol = p)

  w_orth_out <- if (nc_orth > 0L) w_orth[seq_len(nc_orth), , drop = FALSE] else
    matrix(numeric(0), nrow = 0, ncol = p)

  list(

    full = list(
      pred_comp = pred_comp,
      t_orth = t_orth_out,
      p_orth = p_orth_out,
      w_orth = w_orth_out,
      X_mean = preppedX$prep$X_mean,
      X_sd = preppedX$prep$X_sd,
      X_prepped = preppedX$X,
      Y = inputs$Y
    ),
    cv = cv,
    ctrl = list(
      nc_pred = if(nc > 1) 1 else 0,
      nc_orth = nc_orth,
      r2x_comp = r2x_comp,

      q2 = q2_comp,
      r2 = r2_comp,

      aucs_tr = aucs_tr,
      aucs_te = aucs_te,

      tssx = tssx,
      tssy = tssy,

      type = type,
      is_multy_Y = is_multi_Y,
      maxPCo = maxPCo,

      ncomp_selected = (if(nc > 1) 1 else 0) + nc_orth,  # orth comps
      ncomp_tested   = nc + 1,  # total orth stages evaluated
      stop_reason = fit_eval$reason,
      stop_metric = fit_eval$metric,
      stop_delta  = fit_eval$delta

    ),
    dims = list(n=n, p=p)
  )
}

.evalComponentPerformance <- function(
    cv_inst,
    type,
    is_multi_Y,
    Y,
    preds_train,
    preds_test,
    class_memb,
    YcsTot,
    tssy
){

  if (grepl('DA', type)) {

    if (is_multi_Y) {

      colnames(preds_test)  <- class_memb
      colnames(preds_train) <- class_memb

      auc_te <- multiclass.roc(
        response  = factor(Y),
        predictor = preds_test,
        quiet     = TRUE
      )$auc

      auc_tr <- multiclass.roc(
        response  = factor(Y),
        predictor = preds_train,
        quiet     = TRUE
      )$auc

    } else {

      auc_te <- roc(
        response  = as.vector(Y),
        predictor = as.vector(preds_test),
        quiet     = TRUE
      )$auc

      auc_tr <- roc(
        response  = as.vector(Y),
        predictor = as.vector(preds_train),
        quiet     = TRUE
      )$auc
    }

    return(list(
      q2      = NA,
      r2      = NA,
      aucs_te = auc_te,
      aucs_tr = auc_tr
    ))
  }

  list(
    q2      = .r2(YcsTot, preds_test, tssy),
    r2      = .r2(YcsTot, preds_train, NULL),
    aucs_te = NA,
    aucs_tr = NA
  )
}

.prepareInputs <- function(X, Y, center, scale) {
  if (is.data.frame(X)) X <- as.matrix(X)
  if (!is.logical(center)) stop("Check center parameter argument!", call. = FALSE)

  sc_num <- switch(scale,
                   "None" = 0,
                   "UV" = 1,
                   "Pareto" = 2,
                   stop("Check scale parameter argument!", call. = FALSE))

  x_check <- .checkXclassNas(X)
  y_check <- .checkYclassNas(Y)
  .checkDimXY(X, y_check[[1]])

  list(X = X, Y = y_check[[1]], Ydummy = y_check[[2]],
       type = y_check[[3]], scale_code = sc_num)
}

set.seed(15)
### pca

test_that("pca() runs and returns valid PCA_metabom8 object", {

  X <- matrix(rnorm(100 * 10), nrow = 100, ncol = 10)
  colnames(X) <- paste0("V", 1:10)
  rownames(X) <- paste0("S", 1:100)

  sc <- uv_scaling(center = TRUE)
  model <- pca(X, ncomp = 2, scaling = sc, method = "nipals")

  Tx <- scores(model)
  Px <- loadings(model)

  expect_s4_class(model, "m8_model")

  expect_true(is.matrix(Tx))
  expect_true(is.matrix(Px))
  expect_equal(nrow(Tx), nrow(X))
  expect_equal(ncol(Tx), 2)
  expect_equal(ncol(Px), ncol(X))

  expect_true(all(c("method","type","ncomp_selected","ncomp_tested") %in%
                    names(model@ctrl)))

  R <- cor(Tx)
  expect_true(all(abs(R[lower.tri(R)]) < 1e-3))

  expect_true(all(model@ctrl$r2x_comp >= 0))
  expect_true(all(model@ctrl$r2x_comp <= 1))
})


test_that("pca() handles too many components gracefully", {
  X <- matrix(rnorm(30), nrow = 5)

  sc <- uv_scaling(center = TRUE)
  expect_warning(
    model <- pca(X, ncomp = 10, scaling = sc, method = "nipals"),
    regexp = "Too many"
    )

  expect_s4_class(model, "m8_model")
  expect_lte(model@ctrl$ncomp_selected, min(nrow(X)-1, ncol(X)))
})


# pls
test_that("PLS-DA model runs correctly with numeric input", {
  data(iris)

  idx = iris$Species %in% c('setosa', 'virginica')

  X <- as.matrix(iris[idx,1:4])
  Y <- as.character(iris$Species[idx])

  sc <- uv_scaling(center = TRUE)
  kf <- kfold(k=3)
  model <- pls(X, Y, scaling = sc, validation_strategy = kf)

  Tx <- scores(model)
  Px <- loadings(model)
  Wx <- weights(model)

  expect_s4_class(model, "m8_model")

  expect_true(is.matrix(Tx))
  expect_true(is.matrix(Px))
  expect_true(is.matrix(Wx))

  expect_equal(nrow(Tx), nrow(X))
  expect_equal(ncol(Px), ncol(X))
  expect_equal(ncol(Wx), ncol(X))

  expect_equal(ncol(Tx), model@ctrl$ncomp_selected)
  expect_equal(nrow(Px), model@ctrl$ncomp_selected)
  expect_equal(nrow(Wx), model@ctrl$ncomp_selected)

  expect_true(all(c("type","ncomp_selected","ncomp_tested", "stop_reason") %in%
                    names(model@ctrl)))

  expect_true(model@ctrl$type %in% c("DA","DA-mY"))

})


test_that("PLS regression model runs correctly", {

  n <- 80
  p <- 10

  X <- matrix(rnorm(n*p), n, p)

  beta <- c(1.2, -0.8, 0.5, rep(0, p-3))
  Y <- X %*% beta + rnorm(n, sd = 0.2)

  sc <- uv_scaling(center = TRUE)
  kf <- kfold(k=3)

  model <- pls(X, Y, scaling = sc, validation_strategy = kf)

  Tx <- scores(model)
  Px <- loadings(model)
  Wx <- weights(model)

  expect_s4_class(model, "m8_model")

  expect_true(is.matrix(Tx))
  expect_true(is.matrix(Px))
  expect_true(is.matrix(Wx))

  expect_equal(nrow(Tx), nrow(X))
  expect_equal(ncol(Px), ncol(X))
  expect_equal(ncol(Wx), ncol(X))

  expect_equal(ncol(Tx), model@ctrl$ncomp_selected)
  expect_true(model@ctrl$q2[1] > 0)
})


test_that("OPLS-DA model runs correctly", {

  data(iris)

  idx = iris$Species %in% c('setosa', 'virginica')

  X <- as.matrix(iris[idx,1:4])
  Y <- as.character(iris$Species[idx])

  sc <- uv_scaling(center = TRUE)
  kf <- kfold(k=3)

  model <- opls(X, Y, scaling = sc, validation_strategy = kf)

  Tp <- scores(model)
  To <- scores(model, orth = TRUE)

  Pp  <- loadings(model)
  Po  <- loadings(model, orth = TRUE)
  Wx  <- weights(model)

  expect_s4_class(model, "m8_model")

  expect_true(is.matrix(Tp))
  expect_true(is.matrix(Pp))
  expect_true(is.matrix(Wx))

  expect_equal(nrow(Tp), nrow(X))
  expect_equal(ncol(Pp), ncol(X))
  expect_equal(ncol(Wx), ncol(X))

  expect_equal(ncol(Tp), model@ctrl$ncomp_selected-1)

  expect_true(all(c("type","ncomp_selected","ncomp_tested","stop_reason") %in%
                    names(model@ctrl)))

})


test_that("OPLS pred component correlates with Y, orthogonal comp is independent of Y", {

  idx = iris$Species %in% c('setosa', 'virginica')

  X <- as.matrix(iris[idx,1:4])
  Y <- as.character(iris$Species[idx])

  sc <- uv_scaling(center = TRUE)
  kf <- kfold(k=3)

  model <- opls(X, Y, scaling = sc, validation_strategy = kf)
  Tp <- scores(model)

  Ynum <- as.numeric(as.factor(Y))
  expect_true(abs(cor(Tp[,1], Ynum)) > 0.2)

  Txo <- scores(model, orth = TRUE)

  if (!is.null(Txo)) {
    r <- cor(Txo[,1], Ynum)
    expect_true(abs(r) < 0.3)
  }

})

#' @title Align NMR Spectra via Cross-Correlation
#'
#' @description
#' Aligns rows of an NMR spectral matrix to a reference spectrum by maximizing
#' cross-correlation. Optionally uses the row-wise median spectrum as reference.
#' Useful for minor spectral misalignments.
#'
#' @param seg Numeric matrix. Each row is a 1D NMR spectrum segment.
#' @param idx_ref Integer. Row index to use as reference spectrum. Ignored if
#'   \code{med = TRUE}.
#' @param clim Numeric. Minimum cross-correlation threshold. Segments with lower
#'   similarity are not shifted.
#' @param med Logical. If \code{TRUE}, the row-wise median spectrum is used as
#'   the reference.
#' @param norm Logical. If \code{TRUE}, rows are z-scaled prior to alignment.
#'
#' @return A numeric matrix of aligned spectra with the same dimensions as
#'   \code{seg}. If \code{med = TRUE}, the output excludes the median row.
#'
#' @details
#' Each spectrum is aligned to the reference by computing the cross-correlation
#' and applying a lag-based linear interpolated shift. The shift is applied only
#' if the cross-correlation exceeds \code{clim}. This function is intended for
#' fine-tuning alignment across short ppm windows (i.e., segments).
#'
#' @importFrom stats ccf
#' @keywords internal
.alignSegment <- function(seg, idx_ref = 1, clim = 0.7,
                         med = TRUE, norm = FALSE,
                         lag.max = 20) {

  if (!is.matrix(seg) || nrow(seg) < 2)
    stop("Input must be a matrix with at least two rows.", call. = FALSE)

  seg <- .dimX(seg)

  ref <- if (med) apply(seg, 2, median) else seg[idx_ref, ]

  if (norm) {
    seg <- t(apply(seg, 1, scale))
    ref <- as.numeric(scale(ref))
  }

  out <- t(vapply(seq_len(nrow(seg)), function(i) {

    cc <- ccf(ref, seg[i, ],
              type = "correlation",
              lag.max = lag.max,
              plot = FALSE)

    acf <- as.vector(cc$acf)
    lags <- as.vector(cc$lag)
    imax <- which.max(acf)

    cval <- acf[imax]
    k <- lags[imax]

    if (is.finite(cval) && cval >= clim)
      .shift_interp(seg[i, ], k)
    else
      seg[i, ]

  }, FUN.VALUE = numeric(ncol(seg))))

  out
}

#' @title Linear Interpolated Shift (Internal)
#'
#' @description
#' Applies a linear shift to a numeric vector using interpolation.
#' Values outside the original domain are extended
#' using the boundary values (no zero padding).
#'
#' @param x Numeric vector.
#' @param lag Numeric scalar. Shift in index units. Positive values shift the
#'   signal to the right, negative values shift to the left.
#'
#' @return Numeric vector of the same length as \code{x}.
#'
#' @keywords internal
.shift_interp <- function(x, lag, ppm = NULL) {

  n <- length(x)

  if (is.null(ppm)) {
    idx <- seq_len(n)
    f <- stats::approxfun(idx, x, rule = 2)
    f(idx - lag)
  } else {
    f <- stats::approxfun(ppm, x, rule = 2)
    f(ppm - lag)
  }
}


#' @title Align NMR Spectra in a Selected Shift Region
#'
#' @description
#' Aligns spectra within a specified ppm window using cross-correlation.
#' Only the selected region is aligned; the remainder of the spectra
#' is unchanged.
#'
#' @param dat Named list with elements:
#'   \describe{
#'     \item{X}{Numeric matrix (spectra in rows)}
#'     \item{ppm}{Numeric vector of chemical shift values}
#'     \item{meta}{Optional metadata}
#'   }
#' @param shift Numeric vector of length 2 specifying ppm region to align.
#' @param idx_ref Integer. Row index to use as reference spectrum.
#' @param med Logical. Use row-wise median spectrum as reference?
#' @param clim Numeric. Minimum correlation threshold.
#' @param norm Logical. Z-scale before alignment.
#' @param lag.max Integer. Maximum lag allowed.
#'
#' @return Updated \code{dat} list with aligned spectra in selected region.
#'
#' @family preprocessing
#' @export
#' @examples
#' data(hiit_raw)
#' plot_spec(hiit_raw, shift=c(1.3,1.4))
#' hiit_aligned <- align_segment(hiit_raw, c(1.3, 1.35))
#' plot_spec(hiit_aligned, shift=c(1.3,1.4)) # aligned segment
#' plot_spec(hiit_aligned, shift=c(3, 3.1)) # segment not aligned
align_segment <- function(dat,
                         shift,
                         idx_ref = 1,
                         med = TRUE,
                         clim = 0.7,
                         norm = FALSE,
                         lag.max = 20) {

  if (!is.list(dat) || !all(c("X", "ppm") %in% names(dat)))
    stop("dat must be a list with elements 'X' and 'ppm'.", call. = FALSE)

  X   <- dat$X
  ppm <- dat$ppm

  if (!is.matrix(X))
    stop("dat$X must be a matrix.", call. = FALSE)

  if (ncol(X) != length(ppm))
    stop("ppm length must match ncol(X).", call. = FALSE)

  if (ppm[1] < ppm[length(ppm)]) {
    ord <- order(ppm, decreasing = TRUE)
    ppm <- ppm[ord]
    X   <- X[, ord, drop = FALSE]
  }

  idx <- get_idx(shift, ppm)

  if (length(idx) < 3)
    stop("Selected shift window too small.", call. = FALSE)

  # extract segment
  seg <- X[, idx, drop = FALSE]

  aligned_seg <- .alignSegment(
    seg      = seg,
    idx_ref  = idx_ref,
    clim     = clim,
    med      = med,
    norm     = norm,
    lag.max  = lag.max
  )

  X[, idx] <- aligned_seg

  dat$X   <- X
  dat$ppm <- ppm

  dat
}


#' @title Dynamic Interval Segmentation (RSPA-style)
#'
#' @description
#' Defines peak-system intervals from the median spectrum for
#' recursive segment-wise alignment (RSPA).
#'
#' @param X Numeric matrix (spectra in rows).
#' @param ppm Numeric ppm vector.
#' @param half_win_ppm Numeric. Half window size around detected peaks.
#'
#' @return List of index vectors defining alignment intervals.
#'
#' @family NMR
#' @keywords internal
.dynamicIntervalsMedian <- function(X, ppm,
                                    half_win_ppm = 0.02,
                                    min_snr = 8) {

  ref <- matrix(apply(X, 2, median), nrow = 1)

  pk <- ppick2(ref,
               ppm = ppm,
               type = "max",
               min_snr = min_snr,
               min_distance_ppm = 0.005)[[1]]

  if (is.null(pk) || nrow(pk) < 1L)
    return(list(seq_along(ppm)))

  intervals <- lapply(pk$idc, function(i0) {
    get_idx(c(ppm[i0] - half_win_ppm,
              ppm[i0] + half_win_ppm), ppm)
  })

  intervals <- intervals[order(vapply(intervals, `[`, 1L, FUN.VALUE = 1L))]

  merged <- list(intervals[[1]])

  for (i in 2:length(intervals)) {
    cur <- intervals[[i]]
    last <- merged[[length(merged)]]

    if (min(cur) <= max(last)) {
      merged[[length(merged)]] <- sort(unique(c(last, cur)))
    } else {
      merged[[length(merged) + 1L]] <- cur
    }
  }

  merged
}


#' @title Cohort-Guided Interval Alignment for 1D NMR Spectra
#'
#' @description
#' Aligns 1D NMR spectra using signal-adaptive ppm intervals derived from the
#' cohort median spectrum. Intervals are constructed around median peak systems
#' and aligned locally via cross-correlation.
#' The function accepts either a metabom8-style `dat` list or plain matrix/vector with `ppm`.
#'
#' @param x A numeric matrix/vector of spectra or a named list with elements
#'   `X`, `ppm`, and optionally `meta`.
#' @param ppm Numeric vector of chemical shift values (ppm). Ignored if `x` is
#'   a `dat` list.
#' @param half_win_ppm Numeric scalar controlling alignment interval width.
#' @param lag.max Integer. Maximum allowed lag for cross-correlation alignment.
#'
#' @details
#' Alignment intervals are derived from peaks in the cohort median spectrum.
#'
#' The `half_win_ppm` parameter controls interval granularity:
#'
#' \itemize{
#'   \item Smaller values (e.g. 0.005–0.007 ppm) generate more, narrower
#'   intervals (RSPA-like behaviour).
#'   \item Larger values (e.g. 0.008–0.015 ppm) merge nearby multiplets into
#'   broader regions (icoshift-like behaviour).
#' }
#'
#' Typical 600–800 MHz \eqn{^1}H NMR data:
#' \itemize{
#'   \item 0.006–0.008 ppm: stable multiplet-level alignment
#'   \item <0.005 ppm: may split J-coupled systems
#'   \item >0.015 ppm: may merge unrelated resonances
#' }
#'
#' Alignment parameters and interval definitions are recorded in
#' \code{attr(X, "m8_prep")}.
#'
#' @return
#' If input is `dat`, returns updated `dat`. Otherwise returns aligned matrix.
#'
#' @seealso \code{\link{align_segment}}
#' @family preprocessing
#' @examples
#' data(hiit_raw)
#' plot_spec(hiit_raw, shift=c(1.3,1.4))
#' hiit_aligns <- align_spectra(hiit_raw)
#' plot_spec(hiit_aligns, shift=c(1.3,1.4)) # aligned segment
#' plot_spec(hiit_aligns, shift=c(3, 3.1)) # aligned segment
#' @export
align_spectra <- function(x,
                                ppm = NULL,
                                half_win_ppm = 0.007,
                                lag.max = 200) {

  inp <- x
  u   <- .m8_unpack_dat(x, ppm = ppm)

  X   <- .dimX(u$X)
  ppm <- u$ppm
  X0  <- X

  if (!.check_X_ppm(X, ppm))
    stop("Dimensions of X and ppm must match.", call. = FALSE)

  ints <- .dynamicIntervalsMedian(X, ppm,
                                  half_win_ppm = half_win_ppm)

  bounds <- lapply(ints, function(i)
    c(ppm[i[1]], ppm[i[length(i)]]))

  for (int in ints) {

    seg <- X[, int, drop = FALSE]
    seg <- .alignSegment(seg,
                        med = TRUE,
                        lag.max = lag.max)

    X[, int] <- seg
  }

  colnames(X) <- formatC(ppm, format = "fg", digits = 10)
  rownames(X) <- rownames(X0)

  X <- .m8_copy_attrs(X0, X)
  attr(X, "m8_axis") <- list(ppm = ppm)

  X <- .m8_stamp(
    X,
    step = "align spectra",
    params = list(
      half_win_ppm = half_win_ppm,
      n_intervals  = length(ints),
      lag.max      = lag.max,
      intervals_ppm = bounds
    ),
    notes = "Median-derived chemical shift intervals aligned independently using cross-correlation."
  )

  if (is.list(inp) && !is.null(inp$X)) {
    inp$X <- X
    inp$ppm <- ppm
    return(inp)
  }

  X
}

expect_valid_spectra <- function(x) {

  expect_true(is.list(x))

  expect_named(x, c("X","ppm","meta"), ignore.order = TRUE)

  expect_true(is.matrix(x$X))
  expect_type(x$ppm, "double")
  expect_s3_class(x$meta, "data.frame")

  expect_equal(ncol(x$X), length(x$ppm))

  expect_true(all(diff(x$ppm) != 0))

  expect_false(anyNA(x$X))
}

data("hiit_raw")
hiit_raw = list(X = hiit_raw$X[2:3,], ppm = hiit_raw$ppm, meta = hiit_raw$meta[2:3,])

test_that("normalize_pqn preserves spectral structure", {

  res <- pqn(hiit_raw)

  expect_valid_spectra(res)

})



test_that("preprocessing pipeline preserves invariants", {

  hiit <-
    hiit_raw |>
    calibrate(type = "tsp") |>
    excise() |>
    correct_baseline() |>
    align_spectra() |>
    pqn()

  expect_valid_spectra(hiit)

})

preproc_funs <- list(
  calibrate = function(x) calibrate(x),
  correct_lw = function(x) correct_lw(x),
  excise = function(x) excise(x),
  correct_baseline_lin = function(x) correct_baseline(x, method='linear'),
  correct_baseline_als = function(x) correct_baseline(x, method='asls'),
  binning = function(x) binning(x, width=0.01),
  norm_eretic     = function(x) norm_eretic(x),
  norm_pqn     = function(x) pqn(x),
  align_segment = function(x) align_segment(x, shift = c(3, 3.1)),
  align_spectra = function(x) align_spectra(x)
)

for (nm in names(preproc_funs)) {

  test_that(paste(nm, "preserves spectral invariants"), {
    res <- preproc_funs[[nm]](hiit_raw)
    expect_valid_spectra(res)
  })

}

#' Linewidth correction by scaling spectra to a reference linewidth
#'
#' Applies a multiplicative correction to each spectrum to compensate for
#' linewidth-induced peak height variation, using a reference peak region
#' (e.g. TSP). The correction assumes an empirical power-law relationship
#' between peak height and linewidth:
#'
#' \deqn{I \propto \mathrm{FWHM}^{-\beta}}
#'
#' Spectra are scaled such that all samples are adjusted to a common reference
#' linewidth \code{lw_ref}:
#'
#' \deqn{X_{\mathrm{adj}} = X \cdot \left(\frac{\mathrm{FWHM}}{lw_{\mathrm{ref}}}\right)^{\beta}}
#'
#' where FWHM is estimated per spectrum within the specified \code{shift}
#' region (typically containing an internal reference signal).
#'
#' If \code{estimate_beta = TRUE}, the exponent \code{beta} is estimated from a
#' log-log regression between peak height (max within \code{shift}) and FWHM
#' across samples. Otherwise, the user-supplied \code{beta} is used.
#'
#' The correction is only applied if it reduces the coefficient of variation (CV)
#' of the reference peak height across samples when \code{only_if_improves = TRUE}.
#'
#' Input may be either:
#' \itemize{
#'   \item a numeric matrix \code{X} (samples x variables), with ppm axis supplied
#'         via \code{attr(X, "m8_axis")$ppm} or numeric \code{colnames(X)}, or
#'   \item a named list with elements \code{X}, \code{ppm}, and optionally \code{meta}.
#' }
#'
#' Attributes \code{"m8_prep"}, \code{"m8_axis"}, and \code{"m8_meta"} are
#' preserved and the applied correction is appended to the preprocessing log
#' via \code{.m8_stamp()}.
#'
#' @param x A numeric matrix (samples x variables) or a named list containing
#'   \code{X}, \code{ppm}, and optionally \code{meta}.
#' @param lw_ref Numeric reference linewidth (FWHM) to which spectra are
#'   normalized. If \code{NULL}, the median linewidth across samples is used.
#' @param beta Numeric exponent governing the linewidth-to-height relationship.
#'   Ignored if \code{estimate_beta = TRUE} and sufficient samples are available.
#' @param shift Numeric vector of length 2 specifying the ppm region used for
#'   linewidth estimation and peak height measurement (e.g. \code{c(-0.1, 0.1)}
#'   for TSP).
#' @param ppm Optional numeric ppm axis. If \code{NULL}, inferred from
#'   \code{attr(X, "m8_axis")$ppm} or numeric \code{colnames(X)}.
#' @param sf Optional spectrometer frequency (MHz). If \code{NULL}, obtained from
#'   \code{meta$a_SFO1} in list input or \code{attr(X, "m8_meta")$a_SFO1}.
#' @param estimate_beta Logical; if \code{TRUE}, estimate \code{beta} from a
#'   log-log regression of reference peak height versus FWHM.
#' @param beta_min_n Minimum number of valid samples required to estimate
#'   \code{beta}.
#' @param only_if_improves Logical; if \code{TRUE}, apply correction only if the
#'   CV of the reference peak height decreases after correction.
#' @param notes Optional character string appended to the preprocessing log.
#'
#' @return An object of the same type as input:
#'   \itemize{
#'     \item If input is a matrix, returns a corrected matrix with attributes
#'           preserved and updated.
#'     \item If input is a list, returns the same list with corrected \code{$X}.
#'   }
#'
#' @details
#' This correction removes systematic peak height variation due to spectral
#' broadening (e.g. shimming differences) under the assumption that the internal
#' reference signal has constant concentration across samples. The exponent
#' \code{beta} reflects the empirical sensitivity of peak height to linewidth
#' in the current acquisition and processing regime.
#'
#' This adjustment improves comparability of signal intensities across spectra
#' by reducing linewidth-induced amplitude bias, thereby enhancing the
#' detectability of small effects in downstream multivariate analyses
#' (e.g. PLS-type models).
#'
#' @seealso \code{\link{lw}}
#'
#' @examples
#' data(covid_raw)
#' # matrix input with ppm in m8_axis
#' Xcor <- correct_lw(covid_raw, shift = c(-0.1, 0.1))
#'
#' @family preprocessing
#' @export

correct_lw <- function(x,
                       lw_ref = 0.8,
                       beta = 0.6,
                       shift = c(-0.1, 0.1),     # region used for FWHM and intensity (e.g. TSP)
                       ppm = NULL,
                       sf = NULL,
                       estimate_beta = TRUE,
                       beta_min_n = 6,
                       only_if_improves = TRUE,
                       notes = NULL) {

  cv <- function(v) stats::sd(v, na.rm = TRUE) / mean(v, na.rm = TRUE)

  is_list_input <- is.list(x) && !is.null(x$X)
  if (is_list_input) {
    dat <- x
    X <- dat$X
    if (is.null(ppm)) ppm <- dat$ppm
    meta <- dat$meta
  } else {
    X <- x
    meta <- NULL
  }

  X <- .dimX(X)

  if (is.null(ppm)) {
    ax <- attr(X, "m8_axis", exact = TRUE)
    if (is.list(ax) && !is.null(ax$ppm)) {
      ppm <- ax$ppm
    } else if (!is.null(colnames(X))) {
      ppm <- as.numeric(colnames(X))
      if (anyNA(ppm)) ppm <- NULL
    }
  }
  if (is.null(ppm)) stop("correct_lw(): `ppm` not provided and not found in attr(X,'m8_axis') or numeric colnames(X).", call. = FALSE)
  ppm <- as.numeric(ppm)
  if (!.check_X_ppm(X, ppm)) stop("correct_lw(): X and ppm mismatch or ppm contains NA/Inf.", call. = FALSE)

  if (is.null(sf)) {
    if (is_list_input && !is.null(meta) && "a_SFO1" %in% names(meta)) {
      sf <- meta$a_SFO1
    } else {
      m <- attr(X, "m8_meta", exact = TRUE)
      if (is.data.frame(m) && "a_SFO1" %in% names(m)) sf <- m$a_SFO1
    }
  }

  fwhm <- lw(X, ppm = ppm, shift = shift, sf = sf)

  idx <- get_idx(shift, ppm)
  if (length(idx) < 3L) stop("correct_lw(): selected `shift` region too small.", call. = FALSE)

  I_ref <- apply(X[, idx, drop = FALSE], 1, max, na.rm = TRUE)

  beta_used <- beta
  beta_source <- "user"

  if (isTRUE(estimate_beta)) {
    ok <- is.finite(I_ref) & I_ref > 0 & is.finite(fwhm) & fwhm > 0
    if (sum(ok) >= beta_min_n) {
      fit <- stats::lm(log(I_ref[ok]) ~ log(fwhm[ok]))
      bhat <- -stats::coef(fit)[2]
      if (is.finite(bhat) && bhat > 0) {
        beta_used <- as.numeric(bhat)
        beta_source <- "estimated"
      }
    }
  }

  lw_adj <- (fwhm / lw_ref)^beta_used
  X_new <- sweep(X, 1, lw_adj, `*`)

  I_ref_new <- apply(X_new[, idx, drop = FALSE], 1, max, na.rm = TRUE)

  improved <- is.finite(cv(I_ref)) && is.finite(cv(I_ref_new)) && (cv(I_ref_new) < cv(I_ref))

  if (only_if_improves && !improved) {
    return(x)
  }

  X_new <- .m8_copy_attrs(from = X, to = X_new, keys = c("m8_prep", "m8_axis", "m8_meta"))

  step_notes <- if (!is.null(notes)) notes else {
    "Scaled spectra by (FWHM/lw_ref)^beta using FWHM estimated in the reference shift window."
  }

  X_new <- .m8_stamp(
    X_new,
    step = "correct_lw",
    params = list(
      shift = shift,
      lw_ref = lw_ref,
      beta = beta_used,
      beta_source = beta_source,
      cv_before = cv(I_ref),
      cv_after = cv(I_ref_new)
    ),
    notes = step_notes
  )

  if (is_list_input) {
    out <- x
    out$X <- X_new
    if (is.null(out$ppm)) out$ppm <- ppm
    return(out)
  }

  X_new
}

#' Baseline Correction for Spectral Data
#'
#' Applies baseline correction to each spectrum (row) of a spectral matrix.
#' Multiple correction algorithms are available and selected via the
#' `method` argument.
#'
#' @param X Numeric matrix or metabom8 `dat` object containing spectra
#'   in rows and spectral variables in columns.
#' @param method Character specifying the baseline correction algorithm.
#'   One of:
#'   \describe{
#'     \item{`"asls"`}{Asymmetric least squares baseline estimation.}
#'     \item{`"linear"`}{Linear baseline estimated from edge regions.}
#'   }
#' @param ... Additional parameters passed to the selected method.
#'
#' **Arguments for `method = "asls"`**
#' \describe{
#'   \item{lambda}{Numeric smoothing parameter controlling baseline stiffness.
#'   Larger values produce smoother baselines. Default `1e7`.}
#'   \item{iter_max}{Maximum number of iterations used in the baseline
#'   estimation procedure. Default `30`.}
#' }
#'
#' **Arguments for `method = "linear"`**
#' \describe{
#'   \item{ppm}{Numeric vector describing the spectral axis (e.g. chemical shift).
#'   Must have length equal to `ncol(X)`.}
#'   \item{edge_frac}{Fraction of points at each spectrum edge used to
#'   estimate the baseline. Default `0.1`.}
#' }
#'
#' @return Numeric matrix containing baseline-corrected spectra with the
#'   same dimensions as `X`. If `X` is a metabom8 `dat` object, the corrected
#'   matrix replaces the `X` component while preserving associated metadata.
#'
#' @details
#' Baseline correction is performed independently for each spectrum.
#'
#' The `"asls"` method estimates a smooth baseline using asymmetric least
#' squares smoothing, which is well suited for spectra with positive peaks
#' such as NMR metabolomics data.
#'
#' The `"linear"` method estimates a straight baseline from the edges of
#' each spectrum and subtracts the fitted trend.
#'
#' The returned object includes a `.stamp` attribute recording the
#' baseline correction method and parameters used.
#'
#' @references
#' Eilers PHC, Boelens HFM (2005).
#' Baseline correction with asymmetric least squares smoothing.
#'
#' @family preprocessing
#'
#' @examples
#' data(hiit_raw)
#'
#' plot_spec(hiit_raw$X[1,], hiit_raw$ppm, shift=c(3.1,4), backend='base')
#' hiit_proc <-
#'   hiit_raw |>
#'   excise() |>
#'   correct_baseline()
#'
#' plot_spec(hiit_proc$X[1,], hiit_proc$ppm, shift=c(3.1,4), backend='base', add=TRUE, col='red')
#'
#' @export
correct_baseline <- function(X, method = c("asls", "linear"), ...) {

  method <- match.arg(method)

  inp <- X
  u   <- .m8_unpack_dat(X)

  X0   <- u$X
  Xmat <- .dimX(X0)
  ppm  <- u$ppm
  meta <- u$meta

  args <- list(...)

  Xbc <- switch(
    method,
    linear = .linear_baseline(Xmat, ppm, ...),
    asls   = .baseline_asls(Xmat, ...)
  )

  Xbc <- .m8_copy_attrs(X0, Xbc)

  if (!is.null(ppm))
    attr(Xbc, "m8_axis") <- list(ppm = ppm)

  Xbc <- .m8_stamp(
    Xbc,
    step   = "baseline",
    params = list(method = method, args = args),
    notes  = "Baseline correction applied."
  )

  if (is.list(inp) && !is.null(inp$X)) {
    inp$X    <- Xbc
    inp$ppm  <- ppm
    inp$meta <- meta
    return(inp)
  }

  Xbc
}

#'NL Baseline Correction for 1D NMR Spectra
#' The implementation relies on \code{\link[ptw]{asysm}} from the \pkg{ptw} package.
#' @noRd
.baseline_asls <- function(X, lambda = 1e7, iter_max = 60) {
  if (any(is.na(X))) {
    message("X contains missing values, replacing with zeros.")
    X[is.na(X)] <- 0
  }

  if (!is.numeric(lambda) || lambda <= 0)
    stop("`lambda` must be positive numeric.", call. = FALSE)

  X0 <- X
  X <- .dimX(X)
  X.bl <- t(vapply(seq_len(nrow(X)), function(i) {
    x <- X[i, ]
    x - ptw::asysm(x, maxit = iter_max, lambda = lambda)
  }, numeric(ncol(X))))

  X.bl <- .m8_copy_attrs(X0, X.bl)
  X.bl <- .m8_stamp(
    X.bl,
    step = "baseline_asysm",
    params = list(lambda = lambda, iter_max = iter_max),
    notes = "Missing values replaced by 0 prior to baseline correction (if present)."
  )

  X.bl
}

#' @noRd
.linear_baseline <- function(X, ppm, edge_frac = 0.1) {

  n <- length(ppm)
  stopifnot(n == ncol(X))
  stopifnot(is.numeric(edge_frac), edge_frac > 0, edge_frac < 0.5)

  k <- floor(n * edge_frac)
  idx <- c(seq_len(k), seq(n - k + 1, n))

  A     <- cbind(1, ppm[idx])
  Afull <- cbind(1, ppm)

  beta <- qr.solve(A, t(X[, idx, drop = FALSE]))
  baseline <- t(Afull %*% beta)
  X - baseline
}

#' @rdname correct_baseline
#' @export
bline <- function(X, ...) {
  warning("`bline()` will be removed in a future version; please use `correct_baseline(method = \"asls\")` instead.",
          call. = FALSE)
  correct_baseline(X, method = "asls", ...)
}

#' Principal Component Analysis (PCA)
#'
#' @description
#' Fits an unsupervised Principal Component Analysis (PCA) model
#' to a spectral data matrix. Model-internal centering and scaling
#' are controlled via a \code{ScalingStrategy} object and do not
#' modify the input matrix.
#'
#' The default backend uses an internal NIPALS implementation.
#' Alternatively, PCA algorithms from the \pkg{pcaMethods} package
#' (e.g. \code{"ppca"}, \code{"svd"}, \code{"robustPca"}) can be used.
#'
#' @param X Numeric matrix (rows = samples, columns = variables).
#' @param scaling A \code{ScalingStrategy} object defining model-internal
#'   centering and scaling (e.g. \code{uv_scaling(center = TRUE)}).
#' @param ncomp Integer. Number of principal components to compute.
#' @param method Character. PCA backend. Use \code{"nipals"} for the
#'   internal implementation or any method supported by
#'   \code{pcaMethods::pca()}.
#'
#' @details
#' PCA decomposes \eqn{X} into orthogonal score and loading matrices:
#' \deqn{X \approx T P}
#' where:
#' \itemize{
#'   \item \eqn{T} contains the principal component scores
#'   \item \eqn{P} contains the loadings
#' }
#'
#' The number of components is fixed by \code{ncomp}.
#' Unlike supervised models, PCA does not use cross-validation
#' or stopping rules.
#'
#' Scaling and centering are applied internally during model fitting.
#' The original input matrix is not modified.
#'
#' @return
#' An object of class \code{m8_model} with \code{engine = "pca"}.
#' The object contains:
#' \itemize{
#'   \item \code{fit$t}: Score matrix (samples × components)
#'   \item \code{fit$p}: Loading matrix (components × variables)
#'   \item \code{ctrl}: Model control information (variance explained, scaling settings)
#'   \item \code{provenance}: Attributes inherited from the input matrix
#' }
#'
#' @seealso \code{\link{uv_scaling}}
#'
#' @family modelling
#' @examples
#' data(covid)
#'
#' uv <- uv_scaling(center=TRUE)
#' model <- pca(X=covid$X, scaling=uv, ncomp=2)
#'
#' show(model)
#' summary(model)
#'
#' Tx <- scores(model)
#' Px <- loadings(model)
#'
#' t2 <- hotellingsT2(Tx)
#' ell <-ellipse2d(t2)
#'
#' # scores plot
#' plot(Tx, asp = 1,
#'   col = as.factor(covid$an$type),
#'   xlim = range(c(Tx[1,], ell$x)),
#'   ylim = range(c(Tx[2,], ell$y))
#'  )
#' lines(ell$x, ell$y, col = "grey", lty=2)
#' @export
pca <- function(X, scaling, ncomp, method = "nipals") {

  cl <- match.call()

  engine <- .pcaEngine(X, scaling, ncomp, method=method)

  mod <- methods::new("m8_model",

      engine = "pca",
      fit = engine$full,
      prep = scaling,
      cv = NULL,
      ctrl = engine$ctrl,
      dims = engine$dims,

      provenance = attributes(X),
      # data_ref = "list",
      call = cl,
      session = list(
        engine_version = "ppca lib",
        R.version = getRversion(),
        pkg = list(
          m8 = as.character(utils::packageVersion("metabom8")),
          Rcpp = as.character(utils::packageVersion("Rcpp"))
        )
      )
  )

  methods::validObject(mod)
  mod
}

.pcaEngine <- function(X, scaling, ncomp = 2L, method) {

  if (is.data.frame(X)) X <- as.matrix(X)
  if (!is.matrix(X)) stop("`X` must be a numeric matrix.", call. = FALSE)
  storage.mode(X) <- "double"

  if (!is.numeric(ncomp) || length(ncomp) != 1L || is.na(ncomp) || ncomp < 1)
    stop("`ncomp` must be a positive integer.", call. = FALSE)
  ncomp <- as.integer(ncomp)

  sc_num <- switch(as.character(scaling$scale),
                   "None"   = 0L,
                   "UV"     = 1L,
                   "Pareto" = 2L,
                   stop("Unsupported scaling@scale. Use 'None', 'UV', or 'Pareto'.", call. = FALSE)
  )

  pc_max <- min(nrow(X)-1, ncol(X))
  if (ncomp > pc_max) {
    warning("Too many components requested; resetting to maximum allowed (", pc_max, ").")
    ncomp <- pc_max
  }

  msd_x  <- .sdRcpp(X)
  XcsTot <- .scaleMatRcpp(
    X,
    0:(nrow(X) - 1),
    center     = scaling$center,
    scale_type = sc_num
  )[[1]]

  tssx <- .tssRcpp(XcsTot)

  n <- nrow(XcsTot)
  p <- ncol(XcsTot)


  method_lc <- tolower(method)

  if (identical(method_lc, "nipals")) {

    comps <- vector("list", ncomp)
    for (i in seq_len(ncomp)) {
      comps[[i]] <- if (i == 1L) .nipPcaCompRcpp(XcsTot) else .nipPcaCompRcpp(X = comps[[i - 1L]][[1]])
    }

    Tpc <- vapply(comps, function(z) as.numeric(z[[2]]), FUN.VALUE = numeric(n))
    Tpc <- matrix(Tpc, nrow = n, ncol = ncomp)

    Ppc <- vapply(comps, function(z) as.numeric(z[[3]]), FUN.VALUE = numeric(p))
    Ppc <- t(matrix(Ppc, nrow = p, ncol = ncomp))  # A x p

    r2x_comp <- vapply(seq_len(ncomp), function(i) {
      Xi <- Tpc[, i] %o% Ppc[i, ]
      .tssRcpp(Xi) / tssx
    }, numeric(1))

    full <- list(
      t = Tpc,
      p = Ppc,
      X_mean = msd_x[[2]],
      X_sd   = msd_x[[1]]
    )

    ctrl <- list(
      method = "nipals",
      type = 'unsupervised',
      ncomp_selected  = ncomp,
      ncomp_tested  = ncomp,
      r2x_comp = r2x_comp,
      tssx = tssx,
      center = scaling$center,
      scale  = as.character(scaling$scale)
    )

  } else {

    mod <- pcaMethods::pca(
      XcsTot,
      nPcs   = ncomp,
      scale  = "none",
      center = FALSE,
      method = method
    )

    Tpc <- as.matrix(mod@scores)

    Ppc <- t(as.matrix(mod@loadings))

    r2x_comp <- NA_real_
    if (!is.null(mod@R2cum) && length(mod@R2cum) >= 2L) {
      r2 <- mod@R2cum
      r2[-1L] <- diff(r2)
      r2x_comp <- r2[seq_len(min(length(r2), ncomp))]
    }

    full <- list(
      t = Tpc,
      p = Ppc,
      X_mean = msd_x[[2]],
      X_sd   = msd_x[[1]]
    )

    ctrl <- list(
      method = method,
      type = 'unsupervised',
      ncomp_selected  = ncomp,
      ncomp_tested  = ncomp,
      r2x_comp = r2x_comp,
      tssx = tssx,
      center = scaling$center,
      scale  = as.character(scaling$scale),
      type = "unsupervised"
    )
  }

  list(full = full, cv = NULL, ctrl = ctrl,  dims = list(n=n, p=p))
}

#' Compute Variable Importance in Projection (VIP)
#'
#' Calculates VIP scores for a PLS/OPLS model using predictive
#' components only.
#'
#' @param Tx Numeric matrix (n x A). Score matrix.
#' @param W Numeric matrix (A x p). Weight matrix.
#'          Rows correspond to components, columns to variables.
#' @param C Numeric vector or matrix of length A. Y-loadings.
#'
#' @details
#' VIP is computed as:
#'
#' \deqn{
#' VIP_j = \sqrt{ p \frac{\sum_a SSY_a w_{aj}^2}{\sum_a SSY_a} }
#' }
#'
#' where
#'   SSY_a = (t_a^T t_a) * c_a^2
#'
#' By construction, mean(VIP^2) = 1.
#'
#' @return Numeric vector of length p containing VIP scores.
#'
#' @keywords internal
.vip <- function(Tx, W, C) {

  if (!is.matrix(Tx))
    stop("T must be a numeric matrix (n x A).", call. = FALSE)

  if (!is.matrix(W))
    stop("W must be a numeric matrix (A x p).", call. = FALSE)

  if (!is.numeric(Tx) || !is.numeric(W))
    stop("T and W must be numeric.", call. = FALSE)

  A <- nrow(W)
  p <- ncol(W)

  if (is.matrix(C)) {

    if (!all(dim(C) %in% c(1L, A)) || length(C) != A)
      stop("C must be length A, or a 1 x A / A x 1 matrix.", call. = FALSE)
    C <- as.numeric(C)
  } else if (is.numeric(C)) {
    if (length(C) != A)
      stop("Length of C must equal number of components (A).", call. = FALSE)
  } else {
    stop("C must be numeric.", call. = FALSE)
  }

  if (ncol(Tx) != A)
    stop("Number of components in T must match rows of W.", call. = FALSE)

  C <- as.numeric(C)

  if (length(C) != A)
    stop("Length of C must equal number of components.", call. = FALSE)

  SSY <- colSums(Tx^2) * (C^2)
  if (all(SSY == 0))
    stop("All SSY values are zero; cannot compute VIP.", call. = FALSE)
  Wnorm <- W / sqrt(rowSums(W^2))
  vip_sq <- p * colSums((Wnorm^2) * SSY) / sum(SSY)

  vip <- sqrt(vip_sq)

  vip
}

#' @title Chemical Shift Calibration
#'
#' @description
#' Aligns 1D \eqn{^1}H NMR spectra to a reference signal on the existing ppm grid.
#' Supports singlet (e.g. TSP or custom range) and predefined doublet references
#' (glucose, alanine).
#'
#' @param X Numeric matrix/vector of spectra or a metabom8 data list.
#' @param ppm Numeric chemical shift vector. If \code{X} is a metabom8 list,
#'   this is taken from \code{X$ppm}.
#' @param type Character (\code{"tsp"}, \code{"glucose"}, \code{"alanine"}),
#' or list.
#'
#' @details
#' In addition to predefined references, custom calibration targets can be supplied
#' as a list with elements:
#' \itemize{
#'   \item \code{mode}: \code{"singlet"} or \code{"doublet"}
#'   \item \code{window}: numeric vector of length 2 defining the ppm search region
#'   \item \code{centre}: target ppm position (optional; defaults to mean(window))
#'   \item \code{j}: numeric vector of length 2 specifying expected J-coupling
#'         (required for doublet calibration)
#' }
#' Custom doublet calibration requires the expected J-coupling range (`j`)
#' in ppm to distinguish the two peaks of the multiplet.
#'
#' For example:
#' \preformatted{
#' calibrate(
#'   X, ppm,
#'   list(
#'     mode   = "doublet",
#'     window = c(1.2, 1.35),
#'     j      = c(0.007, 0.009)
#'   )
#' )
#' }
#' @return Calibrated spectra in the same structure as input.
#'
#' @examples
#' data("covid_raw")
#' X=covid_raw$X
#' ppm=covid_raw$ppm
#' X_tsp <- calibrate(X, ppm, type = "tsp")
#' X_glu <- calibrate(X, ppm, type = "glucose")
#' X_custom <- calibrate(X, ppm, type = c(1.9, 2.1))
#' @family preprocessing
#' @export
calibrate <- function(X, ppm = NULL, type = "tsp") {

  inp <- X
  u   <- .m8_unpack_dat(X, ppm = ppm)

  X0   <- u$X
  Xmat <- .dimX(X0)
  ppm  <- u$ppm
  meta <- u$meta

  if (is.null(ppm))
    stop("ppm must be supplied or inferable.", call. = FALSE)

  ppm <- as.numeric(ppm)

  target <- .resolve_calib_target(type)

  Xc <- switch(
    target$mode,
    singlet = .calibrate_singlet(
      Xmat,
      ppm,
      target$window,
      target$centre
    ),
    doublet = .calibrate_doublet(
      Xmat,
      ppm,
      window   = target$window,
      centre   = target$centre,
      j_const  = target$j
    )
  )
  Xc <- .m8_copy_attrs(X0, Xc)
  attr(Xc, "m8_axis") <- list(ppm = ppm)

  Xc <- .m8_stamp(
    Xc,
    step   = "calibrate",
    params = list(target = target),
    notes  = "Chemical shift calibration applied."
  )

  if (is.list(inp) && !is.null(inp$X)) {
    inp$X    <- Xc
    inp$ppm  <- ppm
    inp$meta <- meta
    return(inp)
  }

  Xc
}

#' Resolve calibration specification into a numeric target definition.
#' @keywords internal
#' @noRd
.resolve_calib_target <- function(type) {

  # tsp
  if (identical(type, "tsp")) {
    return(list(
      mode   = "singlet",
      window = c(-0.2, 0.2),
      centre = 0
    ))
  }

  # glucose
  if (identical(type, "glucose")) {
    return(list(
      mode   = "doublet",
      window = c(5.15, 5.30),
      centre = 5.233,
      j      = c(0.006, 0.007)
    ))
  }

  # alanine
  if (identical(type, "alanine")) {
    return(list(
      mode   = "doublet",
      window = c(1.40, 1.56),
      centre = 1.48,
      j      = c(0.0115, 0.0135)
    ))
  }

  # custom singlet
  if (is.numeric(type) && length(type) == 2L) {
    return(list(
      mode   = "singlet",
      window = type,
      centre = mean(type)
    ))
  }

  # custom full specification
  if (is.list(type) ) {

    if (!all(c("mode","window") %in% names(type)))
      stop("Custom calibration list must include 'mode' and 'window'.", call. = FALSE)

    if (!type$mode %in% c("singlet","doublet"))
      stop("`mode` must be 'singlet' or 'doublet'.", call. = FALSE)

    if (type$mode == "doublet" && is.null(type$j))
      stop("Custom doublet requires 'j'.", call. = FALSE)

    if (is.null(type$centre))
      type$centre <- mean(type$window)


    out <- list(
      mode   = type$mode,
      window = type$window,
      centre = type$centre
    )

    if (type$mode == "doublet")
      out$j <- type$j

    return(out)
  }

  stop("Unsupported calibration target.", call. = FALSE)
}

#' Align spectra to the maximum peak within a ppm interval.
#' @keywords internal
#' @noRd
.calibrate_singlet <- function(X, ppm, range, centre = mean(range)) {

  idx <- get_idx(range, ppm)
  if (length(idx) == 0L){
    warning("No ppm values found in calibration window; returning X unchanged.", call.=FALSE)
    return(X)
    }

  target <- centre
  zeroAbs <- idx[which.min(abs(ppm[idx] - target))]

  Xc <- matrix(0, nrow(X), ncol(X))

  for (i in seq_len(nrow(X))) {

    peakAbs <- idx[which.max(X[i, idx])]
    shift <- zeroAbs - peakAbs

    if (shift == 0L) {
      Xc[i, ] <- X[i, ]
    } else if (shift > 0L) {
      Xc[i, seq.int(shift + 1L, ncol(X))] <- X[i, seq_len(ncol(X) - shift)]
    } else {
      sh <- abs(shift)
      Xc[i, seq_len(ncol(X) - sh)] <- X[i, seq.int(sh + 1L, ncol(X))]
    }
  }

  Xc
}


#' Calibrate spectra to a known doublet using J-coupling constraints.
#' @importFrom signal sgolayfilt
#' @importFrom utils combn
#' @importFrom ptw asysm
#' @keywords internal
#' @noRd
.calibrate_doublet <- function(X,
                               ppm,
                               window,
                               centre,
                               j_const,
                               sg_length = 13) {

  idx <- get_idx(window, ppm)

  if (length(idx) < 3L) {
    warning("Calibration region contains too few points; returning X unchanged.", call. = FALSE)
    return(X)
  }

  picks_list <- lapply(seq_len(nrow(X)), function(i) {
    x <- X[i, idx]
    smoothed <- signal::sgolayfilt(x - ptw::asysm(x, lambda = 100, maxit = 1000),
                                   p = 3, n = sg_length)
    picks <- ppick(smoothed, ppm[idx], type = "max")
    if (is.null(picks) || length(picks) == 0L || is.null(picks[[1]])) return(NULL)
    picks[[1]]
  })

  if (all(vapply(picks_list, is.null, logical(1)))) {
    warning("Could not calibrate: no signals detected in calibration window; returning X unchanged.",
            call. = FALSE)
    return(X)
  }

  dbl <- lapply(seq_along(picks_list), function(i) {
    peaks <- picks_list[[i]]
    if (is.null(peaks) || nrow(peaks) < 2L) return(NULL)

    ptab <- peaks[peaks$Etype > 0, , drop = FALSE]
    if (nrow(ptab) < 2L) return(NULL)

    combs <- utils::combn(nrow(ptab), 2)
    ppm_diffs <- apply(combs, 2, function(id) abs(diff(ptab$ppm[id])))
    valid <- which(ppm_diffs > j_const[1] & ppm_diffs < j_const[2])
    if (length(valid) == 0L) return(NULL)

    if (length(valid) == 1L) {
      out <- ptab[combs[, valid], , drop = FALSE]
    } else {
      avg_int <- vapply(valid, function(v) {
        id <- combs[, v]
        mean(ptab$Int[id])
      }, numeric(1))
      out <- ptab[combs[, valid[which.max(avg_int)]], , drop = FALSE]
    }

    out$mm <- mean(out$ppm)
    out
  })

  Xc <- matrix(0, nrow = nrow(X), ncol = ncol(X))
  for (i in seq_len(nrow(X))) {
    if (is.null(dbl[[i]]) || nrow(dbl[[i]]) < 2L) {
      Xc[i, ] <- X[i, ]
      next
    }

    delta <- (dbl[[i]]$mm - centre)

    shift_fun <- stats::approxfun(x = ppm - delta, y = X[i, ],
                                  rule = 2)  # rule=2 avoids NA at edges
    xi <- shift_fun(ppm)

    if (anyNA(xi)) xi[is.na(xi)] <- 0
    Xc[i, ] <- xi
  }

  rownames(Xc) <- rownames(X)
  colnames(Xc) <- colnames(X)

  Xc
}

#' @title Hotelling T^2 Statistic
#'
#' @description Computes the Hotelling T^2 statistic defining a multivariate
#' confidence region for a set of observations. The region corresponds to an
#' ellipsoid in p-dimensional space and is commonly visualised as an ellipse
#' in two-dimensional (OPLS) score plots.
#' @param X Numeric matrix with observations in rows and variables (dimensions)
#' in columns.
#' @param alpha Numeric scalar. Confidence level for the T^2 region. Default is 0.95.
#' @return A named \code{list} with elements:
#' \describe{
#'   \item{center}{Numeric vector of column means.}
#'   \item{cov}{Sample covariance matrix.}
#'   \item{c2}{Squared radius of the Hotelling T^2 region.}
#' }
#' @details
#' The Hotelling T^2 region is defined as
#' \deqn{(x - \mu)^T S^{-1} (x - \mu) \le c^2}
#' where \eqn{c^2 = (p (n - 1) / (n - p)) F_{p, n-p}(\alpha)}.
#' @examples
#' set.seed(1)
#' X <- matrix(rnorm(200), ncol = 2)
#' t2 <- hotellingsT2(X)
#' ell <- ellipse2d(t2)
#' plot(X, asp = 1)
#' lines(ell$x, ell$y, col = "red", lwd = 2)
#' @importFrom stats cov qf
#' @export
hotellingsT2 <- function(X, alpha = 0.95) {
  X <- as.matrix(X)
  n <- nrow(X)
  p <- ncol(X)

  center <- colMeans(X)
  S <- cov(X)

  c2 <- (p * (n - 1) / (n - p)) * qf(alpha, p, n - p)

  list(center = center, cov = S, c2 = c2)
}


#' @title Calculate 2D Hotelling T^2 Ellipse
#' @description Generates coordinates of a two-dimensional ellipse corresponding
#' to a Hotelling T^2 region projected onto selected dimensions.
#' @param obj A \code{list} as returned by \code{hotellingsT2()}.
#' @param dims Integer vector of length 2 specifying which dimensions to project
#' onto. Default is \code{c(1, 2)}.
#' @param npoints Integer. Number of points used to approximate the ellipse.
#' Default is 100.
#' @return A \code{data.frame} with columns \code{x} and \code{y} containing
#' ellipse coordinates.
#' @details
#' The ellipse is obtained by projecting the multivariate T^2 ellipsoid onto
#' the specified dimensions. The scaling is derived from the Hotelling T^2
#' statistic and accounts for sample size and dimensionality.
#' @examples
#' set.seed(1)
#' X <- cbind(rnorm(100), rnorm(100) + 0.5)
#'
#' t2 <- hotellingsT2(X)
#' ell <- ellipse2d(t2)
#'
#' plot(X[,1], X[,2], asp = 1, pch = 16,
#'      xlab = "Score 1", ylab = "Score 2")
#' lines(ell$x, ell$y, col = "red", lwd = 2)
#' @importFrom ellipse ellipse
#' @export
ellipse2d <- function(obj, dims = c(1,2), npoints = 100) {
  S <- obj$cov[dims, dims]
  center <- obj$center[dims]

  ell <- ellipse::ellipse(S, centre = center, t = sqrt(obj$c2), npoints = npoints)
  as.data.frame(ell)
}

# internal helper: append preprocessing step to a matrix attribute

.m8_stamp <- function(X, step, params = list(), notes = NULL) {
  rec <- list(
    step = step,
    params = params,
    notes = notes,
    time = as.character(Sys.time()),
    pkg = "metabom8",
    pkg_version = as.character(utils::packageVersion("metabom8"))
  )

  log <- attr(X, "m8_prep", exact = TRUE)
  if (is.null(log)) log <- list()
  log[[length(log) + 1L]] <- rec
  attr(X, "m8_prep") <- log
  X
}

# internal helper: safely carry attributes from old X to new X
.m8_copy_attrs <- function(from, to, keys = c("m8_prep", "m8_axis")) {
  for (k in keys) {
    val <- attr(from, k, exact = TRUE)
    if (!is.null(val)) attr(to, k) <- val
  }
  to
}

.format_params <- function(x, indent = 8, detail = FALSE, max_items = 8) {

  if (is.null(x) || length(x) == 0)
    return(character())

  space <- strrep(" ", indent)

  if (!is.list(x)) {
    return(paste0(space, .format_value(x)))
  }

  keys <- names(x)

  if (is.null(keys)) {
    return(vapply(seq_along(x), function(i) {
      paste0(space, "- ", .format_value(x[[i]]))
    }, character(1)))
  }

  max_width <- max(nchar(keys))

  out <- character()

  for (k in keys) {

    value <- x[[k]]

    if (!detail && .should_collapse(value, max_items = max_items)) {
      preview <- .preview_list(value, n = 3)
      out <- c(out, paste0(
        space,
        format(k, width = max_width, justify = "left"),
        " : ",
        sprintf("<%d items>", length(value)),
        if (nzchar(preview)) paste0(" [", preview, "]") else ""
      ))
      next
    }

    if (is.list(value)) {
      out <- c(out, paste0(
        space,
        format(k, width = max_width, justify = "left"),
        " :"
      ))
      out <- c(out, .format_params(value,
                                   indent = indent + 4,
                                   detail = detail,
                                   max_items = max_items
      ))
    } else {
      out <- c(out, paste0(
        space,
        format(k, width = max_width, justify = "left"),
        " : ",
        .format_value(value)
      ))
    }
  }

  out
}

.format_value <- function(v) {

  if (is.numeric(v) && length(v) > 1) {
    return(sprintf("[%s]",
                   paste(signif(v, 6), collapse = ", ")))
  }

  if (length(v) > 1) {
    return(sprintf("[%s]", paste(v, collapse = ", ")))
  }

  as.character(v)
}

#' @importFrom utils capture.output
.should_collapse <- function(x,
                             max_items = 8,
                             max_chars = 1200) {

  if (!is.list(x))
    return(FALSE)

  if (length(x) > max_items)
    return(TRUE)

  est_size <- nchar(paste(utils::capture.output(utils::str(x, max.level = 2)), collapse = ""))
  if (est_size > max_chars)
    return(TRUE)

  FALSE
}

.preview_list <- function(x, n = 3) {

  if (!is.list(x) || length(x) == 0)
    return("")

  x_sub <- x[seq_len(min(n, length(x)))]
  items <- vapply(x_sub, .format_value, character(1))
  preview <- paste(items, collapse = ", ")

  if (length(x) > n)
    preview <- paste0(preview, ", ...")

  preview
}


get_attr <- function(X, field) {
  attributes(X)[[field]]
}


#' Retrieve metabom8 provenance metadata
#'
#' Extracts preprocessing provenance stored in the \code{"m8_prep"} attribute.
#' Allows access to the full processing log, a specific step, or a specific
#' parameter within a step.
#'
#' @param x A metabom8 object (named list with element \code{X}) or a numeric
#'   matrix containing metabom8 provenance metadata.
#' @param step Optional. Either:
#'   \itemize{
#'     \item Numeric index of the preprocessing step
#'     \item Character string matching the recorded step name
#'   }
#'   If \code{NULL}, the full provenance log is returned.
#' @param param Optional character string specifying a parameter name within
#'   the selected step. If provided, only this parameter value is returned.
#'
#' @return
#' Depending on the arguments:
#' \itemize{
#'   \item Full provenance list (if \code{step = NULL})
#'   \item A single preprocessing step (if \code{step} specified)
#'   \item A single parameter value (if \code{param} specified)
#' }
#'
#' @details
#' Provenance metadata are recorded automatically by metabom8 preprocessing
#' functions and stored as structured attributes on the spectral matrix.
#' This function provides programmatic access to these records.
#'
#' @examples
#' data(hiit_raw)
#'
#' hiit_proc <- hiit_raw |>
#'   calibrate(type = "tsp") |>
#'   excise()
#'
#' # Retrieve full log
#' log <- get_provenance(hiit_proc)
#'
#' # Retrieve specific step
#' get_provenance(hiit_proc, step = 2)
#'
#' # Retrieve parameter from a named step
#' get_provenance(hiit_proc, step = "calibrate", param = "target")
#'
#' @family provenance
#' @export
get_provenance <- function(x, step = NULL, param = NULL) {

  obj  <- x$X %||% x
  prep <- attr(obj, "m8_prep")

  if (is.null(prep))
    stop("No metabom8 provenance found.", call. = FALSE)

  if (is.null(step))
    return(prep)

  if (is.numeric(step)) {
    if (step > length(prep))
      stop("Step index out of range.", call. = FALSE)

    sel <- prep[[step]]
  }

  else {
    idx <- which(vapply(prep, `[[`, character(1), "step") == step)

    if (length(idx) == 0)
      stop("Step not found.", call. = FALSE)

    sel <- prep[[idx]]
  }

  if (is.null(param))
    return(sel)

  if (!param %in% names(sel$params))
    stop("Parameter not found in selected step.", call. = FALSE)

  sel$params[[param]]
}



`%||%` <- function(a, b) if (!is.null(a)) a else b



.format_provenance <- function(x, detail = FALSE, max_items = 8) {
  prep <- attr(x$X %||% x, "m8_prep", exact = TRUE)

  if (is.null(prep) || length(prep) == 0) {
    return("No metabom8 preprocessing metadata found.")
  }

  out <- c(
    "metabom8 processing pipeline:",
    "=============================="
  )

  for (i in seq_along(prep)) {
    step <- prep[[i]]
    out <- c(out, sprintf("[%d] %s", i, step$step))

    if (!is.null(step$params)) {
      out <- c(out, "     params:")
      out <- c(out, .format_params(step$params, indent = 8, detail = detail, max_items = max_items))
    }

    if (!is.null(step$notes)) {
      out <- c(out, paste0("     notes: ", step$notes))
    }

    out <- c(out, "")
  }

  out
}


#' Print metabom8 preprocessing pipeline
#'
#' Displays the recorded preprocessing history attached to a metabom8
#' spectral matrix. The function reads the \code{"m8_prep"} attribute
#' and prints each processing step in chronological order, including
#' parameters and notes.
#'
#' The input can be either:
#' \itemize{
#'   \item A metabom8-style list containing \code{$X}, or
#'   \item A matrix with metabom8 provenance attributes attached.
#' }
#'
#' If no preprocessing metadata is found, a message is printed.
#'
#' @param x A metabom8 object or spectral matrix with attached
#'   \code{"m8_prep"} provenance metadata.
#' @param detail Prints full information trail, incl. timestamp / versioning
#' @param max_items Limits individual list entries (e.g., parameter) to specified
#' number
#'
#' @return Invisibly returns \code{NULL}. This function is called for
#'   its side effect of printing pipeline information.
#'
#' @details
#' metabom8 records preprocessing steps as an ordered list of
#' transformation descriptors stored in the \code{"m8_prep"} attribute.
#' Each step typically contains:
#' \itemize{
#'   \item \code{step}: Name of the preprocessing operation
#'   \item \code{params}: Parameters used
#'   \item \code{notes}: Optional description
#'   \item \code{time}: Timestamp
#'   \item \code{pkg}: Package name and version
#' }
#'
#' This function provides a compact audit trail of the processing
#' workflow, facilitating reproducibility and provenance inspection.
#'
#' @examples
#' params <- list(
#'   runtime  = "docker",
#'   image    = Sys.getenv("IMAGE", "docker-image-dummy"),
#'   workflow = Sys.getenv("M8_WORKFLOW", "std_prof-urine"),
#'   agent    = paste0("snakemake/", Sys.getenv("SNAKEMAKE_VERSION", "v?")),
#'   run_id   = Sys.getenv("M8_RUN_ID", "m8-2605-001")
#' )
#'
#' data(hiit_raw)
#' print_provenance(hiit_raw)
#'
#' hiit_proc <- hiit_raw |>
#'   calibrate(type = "tsp") |>
#'   excise() |>
#'   add_note('dilution-adaptive acquisition mode -> verify snr after normalisation',
#'     params)
#'
#' print_provenance(hiit_proc)
#' @family provenance
#' @export
print_provenance <- function(x, detail = FALSE, max_items = 8) {
  txt <- .format_provenance(x, detail = detail, max_items = max_items)
  writeLines(txt)
  invisible(txt)
}

#' Add user note to metabom8 provenance
#' Appends a user annotation to the \code{"m8_prep"} attribute.
#' The step title is formatted as \code{"note {username}"}.
#' The timestamp is stored in \code{params}, and the user message
#' is stored in \code{notes}.
#' @param x A metabom8 object or matrix with \code{"m8_prep"} metadata.
#' @param note Character string describing the annotation.
#' @param params Named list providing paramter key-value pairs.
#' @return The input object with updated provenance metadata.
#' @family provenance
#' @examples
#' params <- list(
#'   runtime  = "docker",
#'   image    = Sys.getenv("IMAGE", "docker-image-dummy"),
#'   workflow = Sys.getenv("M8_WORKFLOW", "std_prof-urine"),
#'   agent    = paste0("snakemake/", Sys.getenv("SNAKEMAKE_VERSION", "v?")),
#'   run_id   = Sys.getenv("M8_RUN_ID", "m8-2605-001")
#' )
#'
#' data(hiit_raw)
#' print_provenance(hiit_raw)
#'
#' hiit_proc <- hiit_raw |>
#'   calibrate(type = "tsp") |>
#'   excise() |>
#'   add_note('dilution-adaptive acquisition mode -> verify snr after normalisation',
#'     params)
#'
#' print_provenance(hiit_proc)
#'
#' @export
add_note <- function(x, note, params=NULL) {

  stopifnot(
    is.character(note), length(note) == 1
  )

  target <- if (is.list(x) && !is.null(x$X)) x$X else x

  target <- .m8_stamp(
    target,
    step   = "user note",
    params = params,
    notes  = note
  )

  if (is.list(x) && !is.null(x$X)) {
    x$X <- target
    return(x)
  }

  target
}


#' List available preprocessing steps
#'
#' Returns a named character vector describing the preprocessing
#' operations implemented in \pkg{metabom8}. For more information
#' on individual functionalities please refer to the function help
#' pages.
#'
#' @return A named character vector with preprocessing function
#'   identifiers as names and short descriptions as values.
#' @examples
#' list_preprocessing()
#' @export
list_preprocessing <- function() {
  c(
    calibrate        = "Chemical shift calibration to reference signal",
    excise           = "Remove spectral regions (e.g., residual water, urea)",
    baseline_correct = "Baseline correction",
    norm_erectic     = "Normalise spectra based on ERETIC signal",
    pqn              = "Apply Probabilistic Quantile Normalisation (PQN)",
    binning          = "Bin spectra",
    correct_lw       = "Apply line width correction (experimental)",
    align_segment    = "Spectral alignment of a single chemical shift segment",
    align_spectra    = "Automated segment-wise full spectrum alignment"
  )
}

#' List available preprocessing functions
#' Returns the preprocessing utilities provided by \pkg{metabom8}.
#' @return A named character vector describing preprocessing functions.
#' @examples
#' list_preprocessing()
#' @family preprocessing
#' @export
print_preprocessing <- function() {
  fns <- list_preprocessing()
  out <- c("Available preprocessing functions:", "")
  out <- c(out, paste0(" - ", format(names(fns), width = 18), fns))
  writeLines(out)
  invisible(fns)
}

#' @title Spectral data binning
#' @description
#' Equidistant binning of spectra by summarising intensities within ppm bins.
#'
#' @param X Numeric matrix or data frame with spectra in rows, or a named list as
#'   returned by \code{\link{read1d}}/\code{read1d_proc} containing \code{X} and \code{ppm}.
#' @param ppm Numeric vector of chemical shift positions (length must match \code{ncol(X)}).
#'   If \code{NULL}, \code{ppm} is inferred in the following order:
#'   \enumerate{
#'     \item \code{X$ppm} if \code{X} is a list input,
#'     \item \code{attr(X, "m8_axis")$ppm} (if present),
#'     \item numeric \code{colnames(X)} (if present).
#'   }
#' @param width Numeric. Bin size in ppm, or \code{NULL} if \code{npoints} is specified.
#' @param npoints Integer. Desired number of bins per spectrum, or \code{NULL} if \code{width} is specified.
#'   If both are provided, \code{npoints} is used.
#' @param fun Function. Summary function applied to intensities within each bin.
#'   Must return a single numeric value (e.g. \code{sum}, \code{mean}, \code{max}).
#'
#' @details
#' If present, preprocessing provenance is appended to \code{attr(X, "m8_prep")} using
#' \code{.m8_stamp()}. The ppm axis is updated in \code{attr(X, "m8_axis")$ppm} and
#' column names are set to the bin centres.
#'
#' When \code{width} is specified, spectra are interpolated onto a regular ppm grid and
#' then aggregated within bins (\code{interp = TRUE} in provenance). When \code{npoints}
#' is specified, aggregation is performed by index bins on the original grid
#' (\code{interp = FALSE} in provenance).
#'
#' @return Numeric matrix with spectra in rows and binned ppm variables in columns.
#'
#' @examples
#' set.seed(1)
#' X <- matrix(rnorm(2 * 100), nrow = 2)
#' ppm <- round(seq(10, 0.5, length.out = 100), 3)
#' colnames(X) <- ppm
#'
#' Xb <- binning(X, ppm, width = 0.5)
#' Xb_mean <- binning(X, ppm, width = 0.5, fun = mean)
#' dim(Xb)
#'
#' @family preprocessing
#' @export
binning <- function(X, ppm = NULL, width = NULL, npoints = NULL, fun = sum) {

  .return_list <- is.list(X) && !is.null(X$X)
  meta <- NULL

  if (.return_list) {
    u <- .m8_unpack_dat(X, ppm = ppm)
    X   <- u$X
    ppm <- u$ppm
    meta <- u$meta
  }

  X <- .dimX(X)

  if (!is.function(fun))
    stop("`fun` must be a function.", call. = FALSE)
  .test <- fun(c(1, 2))
  if (!is.numeric(.test) || length(.test) != 1L)
    stop("`fun` must return a single numeric value.", call. = FALSE)

  X0 <- X

  if (is.null(ppm)) {
    ax <- attr(X, "m8_axis", exact = TRUE)
    if (is.list(ax) && !is.null(ax$ppm)) {
      ppm <- ax$ppm
    } else if (!is.null(colnames(X))) {
      ppm <- as.numeric(colnames(X))
      if (anyNA(ppm)) ppm <- NULL
    }
  }
  if (is.null(ppm))
    stop("`ppm` not provided and not found in list input, attr(X,'m8_axis'), or numeric colnames(X).", call. = FALSE)

  ppm <- as.numeric(ppm)
  if (!.check_X_ppm(X, ppm))
    stop("Non-matching dimensions X matrix and ppm vector or missing values in ppm.", call. = FALSE)

  if (!is.null(width) && !is.null(npoints)) width <- NULL
  if (is.null(width) && is.null(npoints))
    stop("Define bin width in ppm or desired number of bins.", call. = FALSE)

  ord <- order(ppm, decreasing = TRUE)
  ppm <- ppm[ord]
  X <- as.matrix(X)[, ord, drop = FALSE]

  ppm_in <- ppm
  axis_in <- list(n = length(ppm_in), range = range(ppm_in, finite = TRUE))

  .restore_names <- function(out, X_in) {
    rownames(out) <- rownames(X_in)
    out
  }

  if (!is.null(width)) {

    dppm <- abs(diff(ppm))
    res <- stats::median(dppm, na.rm = TRUE)

    if (!is.finite(res) || res <= 0) stop("Cannot determine ppm grid spacing.", call. = FALSE)
    if (width <= res) stop("Bin width is smaller or equal to the ppm grid spacing.", call. = FALSE)

    step <- max(1L, as.integer(round(width / res)))
    new_res <- width / step

    ppm_new <- seq(from = max(ppm), to = min(ppm), by = -new_res)

    nbins <- floor(length(ppm_new) / step)
    if (nbins < 1L) stop("Binning width too large for the ppm range.", call. = FALSE)

    ybin <- rep(seq_len(nbins), each = step)[seq_len(nbins * step)]
    ppm_new <- ppm_new[seq_along(ybin)]

    Xb <- t(apply(X, 1, function(x) {
      sInter <- stats::approxfun(ppm, x, rule = 2)
      s <- sInter(ppm_new)
      vapply(seq_len(nbins), function(i) fun(s[ybin == i]), numeric(1))
    }))

    ppm_bin <- vapply(seq_len(nbins), function(i) mean(ppm_new[ybin == i]), numeric(1))
    colnames(Xb) <- formatC(ppm_bin, format = "fg", digits = 10)
    Xb <- .restore_names(Xb, X)

    axis_out <- list(n = length(ppm_bin), range = range(ppm_bin, finite = TRUE))

    Xb <- .m8_copy_attrs(X0, Xb)
    attr(Xb, "m8_axis") <- list(ppm = ppm_bin)
    Xb <- .m8_stamp(
      Xb,
      step = "binning",
      params = list(
        width = width,
        npoints = NULL,
        fun = deparse(substitute(fun)),
        interp = TRUE,
        grid_res_est = res,
        step_points = step,
        axis_in = axis_in,
        axis_out = axis_out
      ),
      notes = "Binning via interpolation to a regular grid followed by aggregation within bins."
    )

    if (.return_list) {
      return(list(X = Xb, ppm = ppm_bin, meta = meta))
    } else {
      return(Xb)
    }
  }

  if (!is.null(npoints)) {

    npoints <- as.integer(npoints)
    if (npoints >= length(ppm)) stop("npoints cannot be larger or equal than length(ppm).", call. = FALSE)
    if (npoints < 2L) stop("npoints must be >= 2.", call. = FALSE)

    breaks <- floor(seq(0, length(ppm), length.out = npoints + 1L))
    breaks[1] <- 0L
    nbins <- npoints

    Xb <- t(apply(X, 1, function(s) {
      vapply(seq_len(nbins), function(i) {
        idx <- (breaks[i] + 1L):breaks[i + 1L]
        fun(s[idx])
      }, numeric(1))
    }))

    ppm_bin <- vapply(seq_len(nbins), function(i) {
      idx <- (breaks[i] + 1L):breaks[i + 1L]
      mean(ppm[idx])
    }, numeric(1))

    colnames(Xb) <- formatC(ppm_bin, format = "fg", digits = 10)
    Xb <- .restore_names(Xb, X)

    axis_out <- list(n = length(ppm_bin), range = range(ppm_bin, finite = TRUE))

    Xb <- .m8_copy_attrs(X0, Xb)
    attr(Xb, "m8_axis") <- list(ppm = ppm_bin)
    Xb <- .m8_stamp(
      Xb,
      step = "binning",
      params = list(
        width = NULL,
        npoints = npoints,
        fun = deparse(substitute(fun)),
        interp = FALSE,
        axis_in = axis_in,
        axis_out = axis_out
      ),
      notes = "Binning by index aggregation on the original grid."
    )

    if (.return_list) {
      return(list(X = Xb, ppm = ppm_bin, meta = meta))
    } else {
      return(Xb)
    }
  }

  NULL
}

# collection of validators

.arg_check_scaling <- function(x){
  stopifnot(exprs = {
    is.list(x)
    !is.null(names(x))
    length(x) == 2
    all(names(x) %in% c('center', 'scale'))
    is.logical(x$center) && length(x$center) == 1
    is.character(x$scale) && length(x$scale) == 1
    tolower(x$scale) %in% c('none', 'uv', 'pareto')
  })
}

.check_preprocess <- function(x){
  stopifnot(exprs = {
    is.list(x)
    !is.null(names(x))
    length(x) == 4
    all(names(x) %in% c('center', 'scale', 'X_mean', 'X_sd'))
    length(X_mean) == length(X_sd)
    all(!is.infinite(x$X_mean) && !is.null(x$X_mean))
    all(!is.infinite(x$X_sd) && !is.null(x$X_sd))
  })
}

#' @importFrom stats runif
.check_probs <- function(cv_pars) {
  if (!"probs" %in% names(cv_pars)) {
    stop("quantile probs argument missing", call. = FALSE)
  }

  probs <- cv_pars$probs

  if (!is.numeric(probs)) {
    stop("probs must be numeric", call. = FALSE)
  }

  if (length(probs) < 2L) {
    stop("at least two quantile probs must be provided", call. = FALSE)
  }

  if (any(probs < 0 | probs > 1)) {
    stop("quantile probs need to be in [0,1]", call. = FALSE)
  }

  if (anyDuplicated(probs)) {
    stop("probs must be strictly increasing (no duplicates)", call. = FALSE)
  }

  if (!isTRUE(all.equal(sort(probs), probs))) {
    stop("probs must be sorted in increasing order", call. = FALSE)
  }

  if (probs[1] != 0 || tail(probs, 1) != 1) {
    stop("probs must start at 0 and end at 1", call. = FALSE)
  }
}

.check_split <- function(cv_pars) {
  if (!"split" %in% names(cv_pars)) {
    stop("split argument missing", call. = FALSE)
  }

  split <- cv_pars$split

  if (!is.numeric(split) || length(split) != 1L || is.na(split)) {
    stop("split must be a single numeric value", call. = FALSE)
  }

  if (!(split > 0 && split < 1)) {
    stop("split must be strictly between 0 and 1", call. = FALSE)
  }

  if (split < 0.05 || split > 0.95) {
    warning("Very small or large split may lead to unstable estimates")
  }
}

.arg_check_cv <- function(cv_pars, model_type=c('DA', 'R'), n, Y_prepped){

  if (!exists(".Random.seed", envir = .GlobalEnv))
    runif(1)

  seed <- .Random.seed

  type <- match.arg(model_type, c("DA", "R"))

  k <- cv_pars$k

  stopifnot(
    is.list(cv_pars),
    !is.null(names(cv_pars)),
    all(c('method', 'k') %in% names(cv_pars)),
    length(k) == 1L && !is.na(k) && is.numeric(k) && k %% 1 == 0 && k >= 2L
  )

  out <- switch(
    cv_pars$method,

    "kfold" = {
      stopifnot("k must be <= number of samples" = k <= n)

      idc <- .cvSetsMethod(Y=Y_prepped, type, method = "k-fold", k = k)
      list(
        train    = idc,
        strategy = "KFold",
        n        = n,
        seed     = seed
      )
    },

    "StratifiedKFold" = {
      if (type == "R") {
        .check_probs(cv_pars)
        probs <- cv_pars$probs
      } else{ probs <- NULL}

      idc <- .cvSetsMethod(Y=Y_prepped, type, method = "k-fold_stratified", k = k, probs=probs)
      list(
        train    = idc,
        strategy = "StratifiedKFold",
        n        = n,
        seed     = seed
      )
    },

    "MonteCarlo" = {
      .check_split(cv_pars)
      idc <- .cvSetsMethod(Y_prepped, type, method = "MC", k = k, split = cv_pars$split)
      list(
        train    = idc,
        strategy = "MonteCarlo",
        n        = n,
        seed     = seed
      )
    },

    "BalancedMonteCarlo" = {
      if (type == "R") {
        .check_probs(cv_pars)
        probs <- cv_pars$probs
      } else{ probs <- NULL}
      .check_split(cv_pars)

      idc <- .cvSetsMethod(Y=Y_prepped, type=type, method = "MC_balanced", k = k, split = cv_pars$split, probs=probs)
      list(
        train    = idc,
        strategy = "BalancedMonteCarlo",
        n        = n,
        seed     = seed
      )
    },

    "BalancedBoot" = {
      if (type == "R") {
        .check_probs(cv_pars)
        probs <- cv_pars$probs
      } else{ probs <- NULL}
      .check_split(cv_pars)

      idc <- .cvSetsMethod(Y=Y_prepped, type=type, method = "Boot_balanced", k = k, split = cv_pars$split, probs=probs)
      list(
        train    = idc,
        strategy = "Boot_balanced",
        n        = n,
        seed     = seed
      )
    }

  )

  out
}

#' @title Read raw FIDs and process to spectra
#'
#' @description
#' Reads Bruker 1D NMR FIDs, corrects the digital filter (group delay),
#' applies apodisation (windowing), optional zero-filling, FFT, phasing and
#' ppm calibration.
#'
#' Returns either absorption-, dispersion-, or magnitude-mode
#' spectra.
#'
#' If \code{to_global = TRUE}, objects are assigned to the global environment.
#'
#' @param path Character. Path to the root directory containing Bruker experiment folders.
#' @param exp_type Named list. Acquisition-parameter filter to select experiments
#'   (e.g., \code{list(PULPROG = "noesygppr1d")}).
#' @param apodisation Named list. Apodisation function and parameters. \code{fun}
#'   must be one of \code{"exponential"}, \code{"cosine"}, \code{"sine"}, \code{"sem"}.
#' @param zerofill Integer. Zero-filling exponent (\code{1} doubles points, \code{2} quadruples, …).
#' @param mode Character. Spectrum type to return: \code{"absorption"}, \code{"dispersion"}, or \code{"magnitude"}.
#' @param verbose Integer. Verbosity level: \code{0} = silent, \code{1} = summary (default),
#'   \code{2} = detailed, \code{3} = debug.
#' @param recursive Logical. Recursively search subdirectories for FIDs.
#' @param n_max Integer. Maximum number of experiments to process.
#' @param filter Logical. Remove experiments with incomplete file structures.
#' @param to_global Logical. If TRUE, objects are also assigned to the global
#' environment; otherwise, only an invisible list is returned.
#' @details
#' FIDs are read from Bruker acquisition folders and processed by the following pipeline:
#' \enumerate{
#'   \item Digital-filter (group-delay) correction: the initial \emph{n} complex points
#'         are invalid due to the causal DSP decimation filter and are discarded; \emph{n}
#'         equals \code{GRPDLY} when present, or is looked up from Bruker tables indexed by
#'         \code{DECIM} and \code{DSPFVS} on older systems.
#'   \item Apodisation (windowing).
#'   \item Zero-filling (optional).
#'   \item FFT to the frequency domain.
#'   \item Phase correction.
#'   \item PPM calibration (e.g., to TSP).
#' }
#' A common ppm scale is then interpolated across spectra.
#'
#' \strong{Digital filter note:} On newer systems \code{GRPDLY} is written in \code{acqus}/\code{acqu2s}
#' and should be used directly. For older data sets (\code{GRPDLY < 0} or missing), the group delay
#' is derived from \code{DECIM} and \code{DSPFVS} via an internal look-up table.
#'
#' \strong{Apodisation functions}:
#' \itemize{
#'   \item \code{"uniform"}
#'   \item \code{"cosine"},
#'   \item \code{"sine"},
#'   \item \code{"exponential"} (parameter: \code{lb})
#'   \item \code{"sem"} (sine * exponential, parameter: \code{lb})
#'   \item \code{"gauss"} (parameter: \code{lb}, \code{'gb'}, and \code{'para'})
#'   \item \code{"expGaus_resyG"} (parameter: \code{lb}, \code{'gb'}, and \code{'aq_t'})
#' }
#'
#' @return
#' A named list with three elements:
#' \describe{
#'   \item{X}{A numeric matrix of spectra (rows = samples, columns = ppm values).}
#'   \item{ppm}{A numeric vector of chemical-shift axis (ppm).}
#'   \item{meta}{A data frame of acquisition metadata, row-aligned with \code{X}.}
#' }
#'
#' If \code{to_global = TRUE}, these objects are also assigned to the global environment.
#' In that case, any existing objects with the same names will be overwritten.
#'
#' @seealso \code{\link{read1d_proc}} for importing TopSpin-processed spectra
#'
#' @examples
#' path <- system.file("extdata", package = "metabom8")
#' read1d_raw(
#'   path,
#'   exp_type    = list(PULPROG = "noesygppr1d"),
#'   apodisation = list(fun = "exponential", lb = 0.2),
#'   zerofill    = 1,
#'   n_max       = 3
#' )
#'
#' @importFrom stats approxfun
#' @export
read1d_raw <- function(path,
                       exp_type = list(exp = "PROF_PLASMA_CPMG128_3mm", pulprog = "noesygppr1d"),
                       apodisation = list(fun = "exponential", lb = 0.2),
                       zerofill = 1L,
                       mode = c("absorption", "dispersion", "magnitude"),
                       verbose = 1,
                       recursive = TRUE,
                       n_max = 1000,
                       filter = TRUE,
                       to_global = FALSE) {

  path <- path.expand(path)
  mode <- match.arg(mode, c("absorption", "dispersion", "magnitude"))

  if (verbose > 1) message("Searching for spectral data...")
  f_list <- .check1d_files_fid(path, n_max = 1e6, filter = filter, recursive = recursive, verbose = verbose)

  if (verbose > 1) message("Found ", length(f_list[[1]]), " experiment(s).")

  if (verbose > 1) message("Extracting spectrometer metadata.")
  pars <- .extract_acq_pars1d(f_list)

  if (verbose > 1) message("Filtering experiments using exp_type conditions.")
  exp_filt <- .filterExp_files(pars, exp_type, f_list, n_max)
  f_list <- exp_filt[[1]]
  pars <- exp_filt[[2]]

  if (verbose > 0) {
    if (length(f_list[[1]])==1) message("Processing 1 experiment.") else{
      message("Processing ", length(f_list[[1]]), " experiments.")
    }
  }

  if (length(unique(pars$a_TD)) > 2 || length(unique(pars$a_GRPDLY)) > 1) {
    stop("Unequal TD or GRPDLY across experiments. Cannot proceed.", call. = FALSE)
  }

  ppm_ref <- .defineChemShiftPpm(
    pars$a_SFO1[1], pars$a_SW_h[1], pars$a_TD[1],
    dref = 4.79, ref = TRUE
  )[, 1]
  ppm_ref <- ppm_ref - ppm_ref[which.min(abs(ppm_ref - 0))]

  if (verbose >= 2) message("Defining apodisation function.")

  if (!'a_GRPDLY' %in% colnames(pars)){
    if (verbose >= 1) message('Group delay (GRPDLY) value not in acqus - using look-up table.')
    pars$a_GRPDLY <- vapply(seq_len(nrow(pars)), function(i){
      get_grpdly(pars$a_DECIM[i], pars$a_DSPFVS[i])
    }, 1.)
  }

  apoFct <- .fidApodisationFct(n = (as.numeric(pars$a_TD[1]) - (floor(as.numeric(pars$a_GRPDLY[1])) * 2)), apodisation)

  if (verbose > 1) message("Reading FIDs and processing spectra.")

  out <- vapply(seq_along(f_list[[1]]), function(s) {
    if (verbose > 1) message(f_list[[1]][s])

    byteorda <- c(little = 0, big = 1)
    dtypa <- c(int = 0, double = 2)

    fid <- readBin(
      file.path(f_list[[1]][s], "fid"),
      what = names(dtypa)[match(pars$a_DTYPA[s], dtypa)],
      n = pars$a_TD[s], size = 4L,
      endian = names(byteorda)[match(pars$a_BYTORDA[s], byteorda)]
    )
    fid <- fid * 2^(-pars$a_NC[s])

    fid_corF <- fid[-(seq_len(floor(pars$a_GRPDLY[s]) * 2))]

    if (length(apoFct) != length(fid_corF)) warning("Apodisation length mismatch.")
    spec_lb <- fid_corF * apoFct

    if (zerofill != floor(zerofill)) stop("zerofill must be integer (e.g., 1, 2).", call. = FALSE)
    spec_zf <- .zerofill(fid = spec_lb, zf = zerofill, le_ori = length(fid))
    sp <- .cplxFft(spec_zf)[, 1]

    sp_re <- Re(sp)
    sp_im <- Im(sp)
    sp_mag <- Mod(sp)

    ppm <- .defineChemShiftPpm(pars$a_SFO1[s], pars$a_SW_h[s], length(sp_re),
                               dref = 4.79, ref = FALSE)

    if (abs(min(sp_re[seq_len(length(sp_re) %/% 3)])) > max(sp_re[seq_len(length(sp_re) %/% 3)])) {
      sp_re <- -sp_re
    }
    sp_re <- .phaseTsp(sp_re, sp_im, ppm, seq(0, pi, by = 0.01), 0,
                       idx_tsp = get_idx(c(-0.15, 0.15), ppm) - 1)[, 1]
    if (abs(min(sp_re[seq_len(length(sp_re) %/% 3)])) > max(sp_re[seq_len(length(sp_re) %/% 3)])) {
      sp_re <- -sp_re
    }

    ppm <- .calibTsp(sp_re, ppm)

    sp_out <- switch(mode,
                     absorption = sp_re,
                     dispersion = sp_im,
                     magnitude = sp_mag)

    fspec <- approxfun(ppm, sp_out, rule = 2)
    spec_out <- fspec(ppm_ref)
    return(spec_out)
  }, FUN.VALUE = ppm_ref)

  out <- t(out)
  colnames(out) <- ppm_ref

  pars$exp_path <- f_list[[1]]

  fnam <- strsplit(f_list[[1]], .Platform$file.sep)
  idx_keep <- which(apply(do.call(rbind, fnam), 2, function(x) length(unique(x))) > 1)
  fnam <- vapply(fnam, function(x) paste(x[idx_keep], collapse = .Platform$file.sep), FUN.VALUE = "")
  rownames(out) <- fnam
  rownames(pars) <- fnam

  steps <- c("read_fid","digital_filter_correction","apodisation")
  if (zerofill > 0) steps <- c(steps, "zero_filling")
  steps <- c(steps, "fft","phase_correction","ppm_calibration","interpolation")

  prov <- list(
    source = list(
      vendor = "bruker",
      input = "fid",
      processed_upstream = FALSE
    ),
    path = list(
      root = path,
      experiments = pars$exp_path,
      recursive = recursive,
      filter_incomplete = filter,
      n_max = n_max
    ),
    time = list(
      import_time_utc = format(Sys.time(), tz = "UTC", usetz = TRUE)
    ),
    pipeline = list(
      fct = "metabom8::read1d_raw",
      mode = mode,
      steps = steps,
      params = list(
        exp_type = exp_type,
        apodisation = apodisation,
        zerofill = zerofill
      )
    ),
    axis = list(
      ppm_source = "computed",
      common_grid = TRUE,
      interpolation = list(
        applied = TRUE,
        method = "linear",
        rule = 2,
        from = "ppm_native",
        to = "ppm_ref",
        notes = "Each spectrum interpolated onto common ppm_ref grid."
      )
    ),
    session = list(
      metabom8 = as.character(utils::packageVersion("metabom8")),
      R = paste0(R.version$major, ".", R.version$minor)
    )
  )

  pars$import_time <- prov$time$import_time
  pars$import_root <- path
  pars$source <- "bruker_raw_fid"

  attr(out, "m8_axis") <- list(ppm = ppm_ref)
  attr(out, "m8_provenance") <- prov

  out <- .m8_stamp(
    out,
    step = "import_raw",
    params = list(source = "bruker_raw_fid", mode = mode, zerofill = zerofill, apodisation = apodisation),
    notes = "Raw FIDs processed to spectra (digital filter correction, apodisation, ZF, FFT, phasing, ppm calibration, interpolation)."
  )

  collect_out <- list(X = out, ppm = ppm_ref, meta = pars)

  if (to_global) {
    if (verbose > 0) message("Assigning objects X, ppm, meta to global environment.")
    list2env(collect_out, envir = .GlobalEnv)
  }

  invisible(collect_out)
}


# Bruker DSP (DECIM × DSPFVS ... group delay in complex points)
# filter data obtained from nmrglue
# https://github.com/jjhelmus/nmrglue/blob/master/nmrglue/fileio/bruker.py
# (link last accessed 11/10/25)
bruker_dsp_table <- list(
  `10` = c(
    `2`=44.75, `3`=33.5, `4`=66.625, `6`=59.083333333333333, `8`=68.5625,
    `12`=60.375, `16`=69.53125, `24`=61.020833333333333, `32`=70.015625,
    `48`=61.34375, `64`=70.2578125, `96`=61.505208333333333, `128`=70.37890625,
    `192`=61.5859375, `256`=70.439453125, `384`=61.626302083333333,
    `512`=70.4697265625, `768`=61.646484375, `1024`=70.48486328125,
    `1536`=61.656575520833333, `2048`=70.492431640625
  ),
  `11` = c(
    `2`=46.0, `3`=36.5, `4`=48.0, `6`=50.166666666666667, `8`=53.25,
    `12`=69.5, `16`=72.25, `24`=70.166666666666667, `32`=72.75, `48`=70.5,
    `64`=73.0, `96`=70.666666666666667, `128`=72.5, `192`=71.333333333333333,
    `256`=72.25, `384`=71.666666666666667, `512`=72.125,
    `768`=71.833333333333333, `1024`=72.0625, `1536`=71.916666666666667,
    `2048`=72.03125
  ),
  `12` = c(
    `2`=46.0, `3`=36.5, `4`=48.0, `6`=50.166666666666667, `8`=53.25,
    `12`=69.5, `16`=71.625, `24`=70.166666666666667, `32`=72.125, `48`=70.5,
    `64`=72.375, `96`=70.666666666666667, `128`=72.5, `192`=71.333333333333333,
    `256`=72.25, `384`=71.666666666666667, `512`=72.125,
    `768`=71.833333333333333, `1024`=72.0625, `1536`=71.916666666666667,
    `2048`=72.03125
  ),
  `13` = c(
    `2`=2.75, `3`=2.8333333333333333, `4`=2.875, `6`=2.9166666666666667,
    `8`=2.9375, `12`=2.9583333333333333, `16`=2.96875,
    `24`=2.9791666666666667, `32`=2.984375, `48`=2.9895833333333333,
    `64`=2.9921875, `96`=2.9947916666666667
  )
)

# Helper to get group delay value; returns NA if not found
get_grpdly <- function(decim, dspfvs, table = bruker_dsp_table, default = NA_real_) {
  vec <- table[[as.character(dspfvs)]]
  if (is.null(vec)) return(default)
  val <- vec[[as.character(decim)]]
  if (is.null(val)) default else unname(val)
}

#' @title OPLS Model Validation via Y-Permutation
#'
#' @description
#' Performs Y-permutation tests to assess the robustness of OPLS models by comparing model performance statistics on real vs. permuted response labels.
#'
#' @param smod An OPLS model object of class \code{OPLS_metabom8}.
#' @param n Integer. Number of permutations to perform.
#' @param plot Logical. If \code{TRUE}, generates a visual summary of permutation statistics.
#' @param mc Logical. If \code{TRUE}, enables multicore processing (currently not implemented).
#'
#' @return A \code{data.frame} with model metrics (e.g., R2, Q2, AUROC) from both permuted and original models.
#'
#' @details
#' Each permutation shuffles the response labels and fits a new OPLS model. The function captures model statistics (R2, Q2, AUC) to compare against the non-permuted model. This helps determine whether the original model performance is better than expected by chance.
#'
#' @references
#' Wiklund, S. et al. (2008). A Tool for Improved Validation of OPLS Models. *Journal of Chemometrics*, 22(11–12), 594–600.
#'
#' @family model_validation
#' @importFrom ggplot2 ggplot theme_bw labs geom_point element_rect
#' @importFrom reshape2 melt
#' @importFrom stats cor median
#' @importFrom parallel detectCores
#' @importFrom progress progress_bar
#' @importFrom rlang .data
# examples
# data(covid)
# X <- covid$X
# an <- covid$an
#
# model <- opls(X, Y = an$type)
# perm_results <- opls_perm(model, n = 10)
opls_perm <- function(smod, n = 10, plot = TRUE, mc = FALSE) {
  if (!inherits(smod, "m8_model") || smod@engine != "opls") {
    stop("Input must be an m8_model instance with opls engine.", call. = FALSE)
  }

  nc_o <- smod@ctrl$ncomp_selected - 1
  type <- smod@ctrl$type

  # full X/Y
  Y <- smod@fit$Y
  XcsTot <- smod@fit$X_prepped

  cv <- smod@cv

  pb <- progress::progress_bar$new(
    format = "  Permutations [:bar] :percent ETA: :eta",
    total = n,
    clear = FALSE,
    width = 60
  )

  perms <- lapply(seq_len(n), function(i) {
    pb$tick()

    Y_perm <- Y[sample(seq_len(nrow(Y))), ]
    stats  <- .permYmod(XcsTot, Y_perm, cv, type, nc_o)

    list(
      ind = stats,
      r   = mean(abs(diag(cor(Y_perm, Y))))
    )
  })

  out_df <- as.data.frame(do.call(rbind, lapply(perms, function(p){
    ind_vec <- unlist(p$ind)
    r_val <- unlist(p$r)
    c(ind_vec, r = p$r)
  } )))
  colnames(out_df) <- c("q2_comp", "r2_comp", "aucs_te", "aucs_tr", "r")
  out_df$model <- paste0("perm_", seq_len(nrow(out_df)))

  idx <- smod@ctrl$nc_orth

  add <- as.data.frame(list(q2_comp = smod@ctrl$q2[idx], r2_comp = smod@ctrl$r2[idx], aucs_te = smod@ctrl$aucs_te[idx], aucs_tr = smod@ctrl$aucs_tr[idx], r=1, model='non-permuted'))
  out_df <- rbind(out_df, add)

  idx_rm <- vapply(out_df, function(x) all(is.na(x)), logical(1))
  out_df <- out_df[, !idx_rm, drop = FALSE]
  out_df$r_abs <- abs(as.numeric(out_df$r))

  if (plot) {

    signed_log1p <- function(x) sign(x) * log1p(abs(x))
    out_df$q2_plot <- signed_log1p(out_df$q2_comp)

    if (grepl('DA', type)){
      dd <- reshape2::melt(out_df[,c('aucs_tr', 'aucs_te', 'r_abs', 'model')], id.vars = c("model", "r_abs"))
    }else{
      dd <- reshape2::melt(out_df[, c('r2_comp', 'q2_plot', 'r_abs', 'model')], id.vars = c("model", "r_abs"))
    }

    dd$r_abs <- as.numeric(dd$r_abs)
    dd$value <- as.numeric(dd$value)
    dd$variable <- as.character(dd$variable)

    map_parameter <- c(aucs_tr = "AUROC_train", aucs_te = "AUROC_test", r2_comp = "R2", q2_comp = "Q2", q2_plot = "Q2_s_log1p")
    dd$variable <- map_parameter[match(dd$variable, names(map_parameter))]

    y_lwer <- min(dd$value, na.rm = TRUE)

    g <- ggplot(dd, aes(x = .data$r_abs, y = .data$value, colour = .data$variable)) +
      geom_point() +
      theme_bw() +
      labs(
        x = "|cor(Y_perm, Y)|",
        y = NULL,
        colour = "Metric",
        caption = paste0("O-PLS-", type, ": ", n, " Y-permutations.")
      ) +
      scale_y_continuous(limits = c(y_lwer, 1)) +
      theme(
        legend.position = c(0.95, 0.05),  # x, y in npc coordinates
        legend.justification = c("right", "bottom"),
        legend.background = element_rect(fill = "white", colour = "black")
      )

    plot(g)
  }

  .perm_test_from_table(out_df)
}


#' Permutation-test summary from an opls_perm out_df table
#'
#' @description
#' Takes the `out_df` you showed (perm rows + one "non-permuted" row) and returns
#' observed values, permutation distributions, and permutation p-values.
#'
#' @param out_df data.frame with columns like q2_comp, r2_comp, aucs_te, aucs_tr,
#'   model (with one row "non-permuted"), and optionally r/r_abs.
#' @param observed_label character. Label used in `model` for the observed row.
#' @param alternative character. "greater" (default) tests obs > perm; "less" tests obs < perm.
#' @param add_one logical. If TRUE, uses (1 + count)/(B + 1) p-value correction.
#' @param na_rm logical. Drop NA values before computing p-values.
#'
#' @return A list with:
#' \itemize{
#'   \item \code{observed}: named numeric vector of observed metrics
#'   \item \code{perm}: list of numeric vectors for each metric
#'   \item \code{p_value}: named numeric vector of permutation p-values
#'   \item \code{B}: number of permutations used
#' }
.perm_test_from_table <- function(out_df,
                                 observed_label = "non-permuted",
                                 alternative = c("greater", "less"),
                                 add_one = TRUE,
                                 na_rm = TRUE) {
  alternative <- match.arg(alternative)

  if (!is.data.frame(out_df)) stop("out_df must be a data.frame.", call. = FALSE)
  if (!("model" %in% names(out_df))) stop("out_df must have a 'model' column.", call. = FALSE)

  obs_idx <- which(out_df$model == observed_label)
  if (length(obs_idx) != 1L) {
    stop("Expected exactly one observed row with model == '", observed_label, "'. Found: ", length(obs_idx), call. = FALSE)
  }

  perm_df <- out_df[out_df$model != observed_label, , drop = FALSE]
  B <- nrow(perm_df)
  if (B < 1L) stop("No permutation rows found (model != '", observed_label, "').", call. = FALSE)

  drop_cols <- c("model", "r_abs")  # keep r_abs out unless you want it as a metric
  metric_cols <- setdiff(names(out_df), drop_cols)
  metric_cols <- metric_cols[vapply(out_df[metric_cols], is.numeric, logical(1))]

  if (length(metric_cols) == 0L) stop("No numeric metric columns found in out_df.")

  obs <- as.list(out_df[obs_idx, metric_cols, drop = FALSE])
  obs <- unlist(obs, use.names = TRUE)

  perm <- lapply(metric_cols, function(col) perm_df[[col]])
  names(perm) <- metric_cols

  perm_p <- function(perm_vals, obs_val) {
    if (na_rm) perm_vals <- perm_vals[!is.na(perm_vals)]
    if (is.na(obs_val)) return(NA_real_)
    B_eff <- length(perm_vals)
    if (B_eff == 0L) return(NA_real_)

    if (alternative == "greater") {
      k <- sum(perm_vals >= obs_val, na.rm = TRUE)
    } else {
      k <- sum(perm_vals <= obs_val, na.rm = TRUE)
    }

    if (add_one) (k + 1) / (B_eff + 1) else k / B_eff
  }

  pvals <- vapply(metric_cols, function(col) perm_p(perm[[col]], obs[[col]]), numeric(1))
  names(pvals) <- metric_cols

  list(
    observed = obs,
    perm = perm,
    p_value = pvals,
    B = B
  )
}


#' @title OPLS Y-permutation Modeling
#' @description
#' Performs OPLS modeling on permuted Y to estimate cross-validated model performance under the null hypothesis.
#' Extracts predictive performance (R2, Q2, AUROC) from cross-validation for either regression or classification.
#'
#' @param XcsTot Numeric matrix. Input data matrix (samples x features).
#' @param Y Numeric matrix. Response variable (numeric or dummy-coded).
#' @param cv List. Cross-validation parameters, including \code{method} and \code{cv_sets}.
#' @param type Character. Model type: \code{"DA"}, \code{"R"}, possibly with \code{"-mY"} suffix for multi-column Y.
#' @param nc_o Integer. Number of orthogonal components to be fitted.
#'
#' @return List with elements:
#' \itemize{
#'   \item \code{r2_comp}: Numeric, R2 statistic for test data.
#'   \item \code{q2_comp}: Numeric, Q2 statistic from cross-validation.
#'   \item \code{aucs_tr}: Numeric, AUROC on training data (for classification only).
#'   \item \code{aucs_te}: Numeric, AUROC on test data (for classification only).
#' }
#'
#' @details
#' When \code{nc_o > 1}, the function fits additional orthogonal components sequentially, updating the model object.
#' Classification performance is computed using AUROC (via \code{pROC}). Regression performance uses R2 and Q2.
#'
#' @importFrom pROC roc multiclass.roc
#' @keywords internal
.permYmod <- function(XcsTot, Y, cv, type, nc_o) {

  if(is.vector(Y) && is.null(dim(Y))){
    Y <- matrix(Y, ncol=1)
  }

  Ycs_fold <- lapply(cv@train, function(idc)
    .scaleMatRcpp(Y, idc - 1, 0, 0)[[1]]
  )

  n  <- nrow(XcsTot)
  p  <- ncol(XcsTot)
  q <- ncol(Y)

  acc <- list(
    sum_test  = matrix(0, n, q),
    n_test    = integer(n),
    sum_train = matrix(0, n, q),
    n_train   = integer(n)
  )

  for (i in seq_len(nc_o)) {
    res <- if (nc_o == 1)
      .oplsComponentCv(
        X        = XcsTot,
        cv.set   = cv@train,
        Ycs_fold = Ycs_fold,
        nc       = nc_o,
        mod.cv   = NULL,
        acc      = acc
      )
    else
      .oplsComponentCv(
        X        = NULL,
        cv.set   = cv@train,
        Ycs_fold = Ycs_fold,
        nc       = nc_o,
        mod.cv   = tt,
        acc      = acc
      )
    tt  <- res$mod.cv
  }

  tt  <- res$mod.cv
  acc <- res$acc

  preds_test <- acc$sum_test
  preds_test[acc$n_test > 0,] <- preds_test[acc$n_test > 0,] / acc$n_test[acc$n_test > 0]

  preds_train <- acc$sum_train
  preds_train[acc$n_train > 0,] <- preds_train[acc$n_train > 0,] / acc$n_train[acc$n_train > 0]

  tssy <- .tssRcpp(Y) / ncol(Y)

  perf <- .evalComponentPerformance(
    cv_inst      = cv,
    type         = type,
    is_multi_Y   = FALSE,
    Y            = Y,
    preds_train  = preds_train,
    preds_test   = preds_test,
    class_memb   = NULL,
    YcsTot       = Y,
    tssy         = tssy
  )

  list(q2 = perf$q2, r2 = perf$r2, aucs_te = perf$aucs_te, aucs_tr = perf$aucs_tr)

}

#' @title Excise Chemical Shift Regions from 1D NMR Spectra
#'
#' @description
#' Removes specified chemical shift regions from 1D \eqn{^1}H NMR spectra.
#' By default, commonly excluded metabolomics regions are removed
#' (upfield/downfield noise, water, urea).
#'
#' @param x Numeric matrix or vector. Spectra in rows and variables in columns.
#' @param ppm Numeric vector of chemical shift positions (ppm).
#'   If omitted, \code{ppm} is inferred from \code{colnames(X)}.
#' @param regions Named list of numeric vectors (length 2), specifying ppm
#'   regions to remove. Each element must define lower and upper bounds.
#'
#' @details
#' Default regions removed (ppm):
#' \itemize{
#'   \item Upfield noise: \code{[min(ppm), 0.25]}
#'   \item Residual water: \code{[4.5, 5.2]}
#'   \item Urea region: \code{[5.5, 6.0]}
#'   \item Downfield noise: \code{[9.7, max(ppm)]}
#' }
#'
#' Removed regions are recorded in \code{attr(X, "m8_prep")} if present.
#' Updated ppm values are stored in column names and in
#' \code{attr(X, "m8_axis")$ppm}.
#'
#' @return Numeric matrix with specified chemical shift regions removed.
#' @examples
#' set.seed(1)
#' ppm <- seq(0, 10, length.out = 1000)
#' X <- matrix(rnorm(100 * length(ppm)), nrow = 100)
#'
#' Xe <- excise(X, ppm)
#'
#' dim(Xe)
#' names(attributes(Xe))
#' @export
excise <- function(x, ppm = NULL, regions = NULL) {

  inp <- x
  u   <- .m8_unpack_dat(x, ppm = ppm)
  X0  <- u$X
  X   <- .dimX(X0)
  ppm <- u$ppm

  if (!.check_X_ppm(X, ppm))
    stop("Non-matching dimensions X matrix and ppm vector or invalid ppm values.", call. = FALSE)

  ppm <- as.numeric(ppm)

  if (is.null(regions)) {
    regions <- list(
      upfield_noise  = c(min(ppm), 0.25),
      water          = c(4.5, 5.2),
      urea           = c(5.5, 6.0),
      downfield_noise = c(9.7, max(ppm))
    )
  }

  if (!is.list(regions))
    stop("`regions` must be a list of numeric vectors (length 2).", call. = FALSE)

  for (nm in names(regions)) {
    r <- regions[[nm]]
    if (!is.numeric(r) || length(r) != 2L)
      stop("Each element of `regions` must be a numeric vector of length 2.", call. = FALSE)
  }

  idx_rm <- unique(unlist(
    lapply(regions, function(r) get_idx(range(r), ppm))
  ))

  keep_idx <- setdiff(seq_along(ppm), idx_rm)

  Xc  <- X[, keep_idx, drop = FALSE]
  ppc <- ppm[keep_idx]

  colnames(Xc) <- formatC(ppc, format = "fg", digits = 10)
  rownames(Xc) <- rownames(X)

  Xc <- .m8_copy_attrs(X0, Xc)
  attr(Xc, "m8_axis") <- list(ppm = ppc)

  Xc <- .m8_stamp(
    Xc,
    step = "excise1d",
    params = list(regions = regions),
    notes = "Specified chemical shift regions removed using get_idx()."
  )

  if (is.list(inp) && !is.null(inp$X)) {
    inp$X   <- Xc
    inp$ppm <- ppc
    return(inp)
  }

  Xc
}

#' Fit a Partial Least Squares (PLS) model
#'
#' Fits a supervised Partial Least Squares (PLS) model using a NIPALS-based
#' algorithm with optional cross-validation and automatic component selection.
#'
#' @param X Numeric matrix of predictors (rows = samples, columns = variables).
#' @param Y Numeric matrix or factor vector of responses.
#' @param scaling A scaling strategy object (e.g., \code{uv_scaling(center = TRUE)}),
#'   specifying model-internal centering and/or scaling applied during fitting.
#'   This does not modify the original spectral matrix.
#' @param validation_strategy A cross-validation strategy object defining how
#'   resampling is performed (e.g., k-fold, Monte Carlo).
#' @param maxPCo Integer. Maximum number of predictive components to evaluate.
#'
#' @details
#' Model components are estimated sequentially using a NIPALS-based algorithm.
#' For each component, cross-validated performance metrics (e.g., Q², R²,
#' classification AUC) are computed according to the supplied
#' \code{validation_strategy}. Component extraction stops when the
#' \code{stopRule} indicates overfitting or when \code{maxPCo} is reached.
#'
#' Scaling specified via \code{scaling} is applied internally during model
#' fitting and does not alter the input matrix \code{X}. Spectral preprocessing
#' steps (e.g., alignment, baseline correction) should be performed prior to
#' model fitting.
#'
#' The returned model object stores:
#' \itemize{
#'   \item Fitted component models
#'   \item Cross-validation results
#'   \item Performance metrics (R², Q², AUC)
#'   \item Model control parameters
#'   \item Input data provenance metadata
#'   \item Session information for reproducibility
#' }
#'
#' @return
#' An object of class \code{m8_model} containing the fitted PLS model,
#' cross-validation results, and performance statistics.
#'
#' @seealso \code{\link{opls}}, \code{\link{uv_scaling}}
#'
#' @family modelling
#' @examples
#' data(covid)
#'
#' cv <- balanced_mc(k=5, split=2/3)
#' scaling <- uv_scaling(center=TRUE)
#' model <-pls(X=covid$X, Y=covid$an$type, scaling, cv)
#'
#' show(model)
#' summary(model)
#'
#' Tp <- scores(model)
#' Pp <- loadings(model)
#' @export
pls <- function(X, Y, scaling, validation_strategy, maxPCo=5) {

  cl <- match.call()

  engine <- .plsEngine(X, Y, scaling, validation_strategy,  maxPCo=maxPCo)

  mod <- methods::new("m8_model",

      engine = "pls",
      fit = engine$full,
      prep = scaling,
      cv = engine$cv,
      ctrl = engine$ctrl,
      dims = engine$dims,

      provenance = attributes(X),
      # data_ref = "list",
      call = cl,
      session = list(
        engine_version = "nipals_v1",
        R.version = getRversion(),
        pkg = list(
          m8 = as.character(utils::packageVersion("metabom8")),
          Rcpp = as.character(utils::packageVersion("Rcpp"))
          )
        )
      )

  methods::validObject(mod)
  mod
}


.plsEngine <- function(X, Y, scaling, validation_strategy, maxPCo){

  #### checking scaling
  .arg_check_scaling(scaling)

  inputs <- .prepareInputs(X, Y, scaling$center, scaling$scale)
  type <- inputs$type
  is_multi_Y <- grepl('mY', type)
  n  <- nrow(inputs$Y)
  p  <- ncol(inputs$X)
  q <- ncol(inputs$Y)
  mo <- maxPCo

  cv <- .arg_check_cv(cv_pars=validation_strategy, model_type=type, n=n, Y_prepped=inputs$Y)
  k <- length(cv$train)

  preppedX <- prep_X(scaling, inputs$X)
  XcsTot <- preppedX$X
  YcsTot <- .scaleMatRcpp(inputs$Y, 0:(nrow(inputs$Y) - 1), center = scaling$center, scale_type = inputs$scale_code)[[1]]

  tssx <- .tssRcpp(XcsTot)
  tssy <- .tssRcpp(YcsTot) / ncol(YcsTot)

  Ycs_fold <- lapply(cv$train, function(idc)
    .scaleMatRcpp(inputs$Y, idc - 1, scaling$center, inputs$scale_code)[[1]]
  )

  tt <- NULL
  full_mod <- list()

  # n  <- nrow(XcsTot)
  # q <- ncol(inputs$Y)
  # k <- length(cv@train)

  acc <- list(
    sum_test  = matrix(0, n, q),
    n_test    = integer(n),
    sum_train = matrix(0, n, q),
    n_train   = integer(n)
  )


  r2_comp <- r2x_comp <- q2_comp <- aucs_tr <- aucs_te <- array()
  nc <- 1
  overfitted <- FALSE

  while (!overfitted) {

    res <- if (nc == 1) {
      .plsComponentCv(inputs$X, cv.set=cv$train, Ycs_fold = Ycs_fold,  nc=nc, mod.cv = NULL, acc=acc) # , it_max=800, eps=1e-6
    } else {
      .plsComponentCv(NULL,cv.set=cv$train, Ycs_fold = Ycs_fold,  nc=nc, mod.cv = mod_cv, acc=acc)
    }

    mod_cv <- res$mod.cv
    acc <- res$acc

    preds_test <- acc$sum_test
    preds_test[acc$n_test > 0,] <- preds_test[acc$n_test > 0,] / acc$n_test[acc$n_test > 0]

    preds_train <- acc$sum_train
    preds_train[acc$n_train > 0,] <- preds_train[acc$n_train > 0,] / acc$n_train[acc$n_train > 0]

    perf <- .evalComponentPerformance(
      cv_inst      = cv,
      type         = type,
      is_multi_Y   = is_multi_Y,
      Y            = Y,
      preds_train  = preds_train,
      preds_test   = preds_test,
      class_memb   = if (is.data.frame(inputs$Ydummy)) inputs$Ydummy$Level[inputs$Ydummy$Column] else NULL,
      YcsTot       = YcsTot,
      tssy         = tssy
    )

    q2_comp[nc] <- perf$q2
    r2_comp[nc] <- perf$r2
    aucs_te[nc] <- perf$aucs_te
    aucs_tr[nc] <- perf$aucs_tr


    if (nc == 1) {
      full_mod[[1]] <- .nipPlsCompRcpp(XcsTot, YcsTot, it_max=800, eps=1e-6)
    } else {
      Xinp <- full_mod[[nc - 1]]$x_res
      full_mod[[nc]] <- .nipPlsCompRcpp(Xinp, YcsTot, it_max=800, eps=1e-6)
    }

    full_mod[[nc]]$t_x

    full_mod[[nc]]$t_x <- .ensure_matrix(full_mod[[nc]]$t_x)
    full_mod[[nc]]$p_x <- .ensure_matrix(full_mod[[nc]]$p_x, ncol=ncol(XcsTot))

    r2x_comp[nc] <- .r2(XcsTot, full_mod[[nc]]$t_x %*% full_mod[[nc]]$p_x, NULL)

    fit_eval <- .evalFit(type, q2_comp, aucs_te, maxPCo)
    overfitted <- fit_eval$stop

    if (!overfitted) {
      nc <- nc + 1
      }
    else{


      if (nc == 1){
        message(sprintf("The PLS-%s model does not gerenalise: %s -> no stable predictive structure detected.", type, fit_eval$reason))
        full <-  list("not fitted as model does not generalise")
      }else{
        if (nc == 2){
          message(sprintf("A PLS-%s model with 1 component was fitted.", type))
        } else if (nc > 2){
          message(sprintf("A PLS-%s model with %d components was fitted.", type, nc-1))
        }

        full <- list(comp=full_mod[seq_len(nc-1)])

        full$X_mean <- preppedX$prep$X_mean
        full$X_sd <- preppedX$prep$X_sd
        full$X_prepped <- preppedX$X
        full$Y <- inputs$Y
      }

      break
    }
  }

  list(
    full = full,
    cv = cv,
    ctrl = list(
      ncomp_selected  = nc - 1,
      ncomp_tested  = nc,
      stop_reason = fit_eval$reason,
      stop_metric = fit_eval$metric,
      stop_delta  = fit_eval$delta,

      r2x_comp = r2x_comp,

      q2 = q2_comp,
      r2 = r2_comp,

      aucs_tr = aucs_tr,
      aucs_te = aucs_te,

      tssx = tssx,
      tssy = tssy,

      type = type,
      is_multy_Y = is_multi_Y,
      maxPCo = maxPCo
    ),
    dims = list(n=n, p=ncol(X))
  )

}

.plsComponentCv <- function(X, cv.set, Ycs_fold, nc, mod.cv, acc) {

  n <- nrow(Ycs_fold[[1]])

  if (nc == 1) {
    q <- ncol(Ycs_fold[[1]])
    p <- ncol(X)
    mod.cv <- vector("list", length(cv.set))

    for (i in seq_along(cv.set)) {
      mod.cv[[i]] <- list(
        t_xp = matrix(0, n, 1),
        y_pred_train = matrix(0, n, q),
        y_pred_test = matrix(0, n, q),
        x_res = matrix(0, n, p)
      )
    }

  } else {

    for (i in seq_along(mod.cv)) {
      mod.cv[[i]]$t_xp <- cbind(mod.cv[[i]]$t_xp, 0)
    }

  }

  for (k in seq_along(cv.set)) {

    idc <- cv.set[[k]]
    not_idc <- setdiff(seq_len(n), unique(idc))

    Xcs <- if (nc == 1)
      .scaleMatRcpp(X, idc - 1, center = TRUE, scale_type = 1)[[1]]
    else
      mod.cv[[k]]$x_res

    Ycs <- Ycs_fold[[k]]

    comp <- .nipPlsCompRcpp(
      X = Xcs[idc, , drop = FALSE],
      Y = Ycs[idc, , drop = FALSE],
      it_max = 800,
      eps = 1e-8
    )

    p <- length(comp$w_x)
    comp$w_x <- .ensure_matrix(comp$w_x, ncol=p)
    comp$p_x <- .ensure_matrix(comp$p_x, ncol=p)
    comp$p_y <- .ensure_matrix(comp$p_y)
    comp$b   <- as.numeric(comp$b)

    Xte <- as.matrix(Xcs[not_idc, , drop = FALSE])
    storage.mode(Xte) <- "double"

    t_pred <- Xte %*% t(comp$w_x)
    if (nrow(comp$p_y) == 1)
      comp$p_y <- t(comp$p_y)
    y_pred <- comp$b * t_pred %*% t(comp$p_y)

    mod.cv[[k]]$t_xp[not_idc, nc] <- t_pred
    mod.cv[[k]]$y_pred_train[idc, ] <- comp$y_pred
    mod.cv[[k]]$y_pred_test[not_idc, ] <- y_pred

    mod.cv[[k]]$x_res[idc, ] <- comp$x_res
    mod.cv[[k]]$x_res[not_idc, ] <- Xte - t_pred %*% comp$p_x

    acc$sum_test[not_idc, ] <- acc$sum_test[not_idc, ] + y_pred
    acc$n_test[not_idc] <- acc$n_test[not_idc] + 1

    uid <- unique(idc)
    agg_pred <- rowsum(comp$y_pred, idc)
    idx <- as.integer(rownames(agg_pred))

    acc$sum_train[idx, ] <- acc$sum_train[idx, ] + agg_pred
    acc$n_train[idx] <- acc$n_train[idx] + tabulate(idc, nbins = nrow(acc$sum_train))[idx]
  }

  return(list(mod.cv = mod.cv, acc = acc))
}

set.seed(1)

test_that(".plsComponentCv returns valid structure for first component", {

  n <- 30
  p <- 5
  q <- 1

  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n * q), n, q)

  cv.set <- list(1:15, 16:30)
  Ycs_fold <- list(Y, Y)

  acc <- list(
    sum_test = matrix(0, n, q),
    n_test   = rep(0, n),
    sum_train = matrix(0, n, q),
    n_train   = rep(0, n)
  )

  res <- .plsComponentCv(
    X = X,
    cv.set = cv.set,
    Ycs_fold = Ycs_fold,
    nc = 1,
    mod.cv = NULL,
    acc = acc
  )

  expect_true(is.list(res))
  expect_true(all(c("mod.cv","acc") %in% names(res)))

  expect_length(res$mod.cv, length(cv.set))

  expect_true(all(vapply(res$mod.cv, function(m) {
    all(c("t_xp","y_pred_train","y_pred_test","x_res") %in% names(m))
  }, logical(1))))

})


test_that(".plsComponentCv accumulates predictions", {

  n <- 20
  p <- 4

  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n), n, 1)

  cv.set <- list(1:10, 11:20)
  Ycs_fold <- list(Y, Y)

  acc <- list(
    sum_test = matrix(0, n, 1),
    n_test = rep(0, n),
    sum_train = matrix(0, n, 1),
    n_train = rep(0, n)
  )

  res <- .plsComponentCv(
    X,
    cv.set,
    Ycs_fold,
    nc = 1,
    mod.cv = NULL,
    acc = acc
  )

  expect_true(any(res$acc$n_test > 0))
  expect_true(any(res$acc$n_train > 0))

})


test_that(".plsComponentCv appends components when nc > 1", {

  n <- 20
  p <- 4

  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n), n, 1)

  cv.set <- list(1:10, 11:20)
  Ycs_fold <- list(Y, Y)

  acc <- list(
    sum_test = matrix(0, n, 1),
    n_test = rep(0, n),
    sum_train = matrix(0, n, 1),
    n_train = rep(0, n)
  )

  res1 <- .plsComponentCv(X, cv.set, Ycs_fold, 1, NULL, acc)

  res2 <- .plsComponentCv(
    X,
    cv.set,
    Ycs_fold,
    nc = 2,
    mod.cv = res1$mod.cv,
    acc = res1$acc
  )

  expect_true(ncol(res2$mod.cv[[1]]$t_xp) == 2)

})

test_that(".plsComponentCv produces matrices with correct dimensions", {

  n <- 30
  p <- 6

  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n), n, 1)

  cv.set <- list(1:15, 16:30)
  Ycs_fold <- list(Y, Y)

  acc <- list(
    sum_test = matrix(0, n, 1),
    n_test = rep(0, n),
    sum_train = matrix(0, n, 1),
    n_train = rep(0, n)
  )

  res <- .plsComponentCv(X, cv.set, Ycs_fold, 1, NULL, acc)

  m <- res$mod.cv[[1]]

  expect_equal(nrow(m$t_xp), n)
  expect_equal(nrow(m$x_res), n)
  expect_equal(ncol(m$x_res), p)

})


test_that(".oplsComponentCv returns valid structure for first component", {

  n <- 30
  p <- 6

  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(sample(c(0,1), n, replace = TRUE), n, 1)

  cv.set <- list(1:15, 16:30)
  Ycs_fold <- list(Y, Y)

  acc <- list(
    sum_test  = matrix(0, n, 1),
    n_test    = rep(0, n),
    sum_train = matrix(0, n, 1),
    n_train   = rep(0, n)
  )

  res <- .oplsComponentCv(
    X = X,
    Ycs_fold = Ycs_fold,
    cv.set = cv.set,
    nc = 1,
    mod.cv = NULL,
    acc = acc
  )

  expect_true(is.list(res))
  expect_true(all(c("mod.cv","acc") %in% names(res)))

  expect_length(res$mod.cv, length(cv.set))

  expect_true(all(vapply(res$mod.cv, function(m) {
    all(c("t_xo","t_xp","y_pred_train","y_pred_test","x_res") %in% names(m))
  }, logical(1))))

})


test_that(".oplsComponentCv produces matrices with correct dimensions", {

  n <- 40
  p <- 8

  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n), n, 1)

  cv.set <- list(1:20, 21:40)
  Ycs_fold <- list(Y, Y)

  acc <- list(
    sum_test  = matrix(0, n, 1),
    n_test    = rep(0, n),
    sum_train = matrix(0, n, 1),
    n_train   = rep(0, n)
  )

  res <- .oplsComponentCv(X, Ycs_fold, cv.set, 1, NULL, acc)

  m <- res$mod.cv[[1]]

  expect_equal(nrow(m$t_xo), n)
  expect_equal(nrow(m$t_xp), n)
  expect_equal(nrow(m$x_res), n)

  expect_equal(ncol(m$x_res), p)

})


test_that(".oplsComponentCv updates acc counters", {

  n <- 20
  p <- 4

  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n), n, 1)

  cv.set <- list(1:10, 11:20)
  Ycs_fold <- list(Y, Y)

  acc <- list(
    sum_test  = matrix(0, n, 1),
    n_test    = rep(0, n),
    sum_train = matrix(0, n, 1),
    n_train   = rep(0, n)
  )

  res <- .oplsComponentCv(X, Ycs_fold, cv.set, 1, NULL, acc)

  expect_true(any(res$acc$n_test > 0))
  expect_true(any(res$acc$n_train > 0))

})


test_that(".oplsComponentCv appends components when nc > 1", {

  n <- 30
  p <- 5

  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n), n, 1)

  cv.set <- list(1:15, 16:30)
  Ycs_fold <- list(Y, Y)

  acc <- list(
    sum_test  = matrix(0, n, 1),
    n_test    = rep(0, n),
    sum_train = matrix(0, n, 1),
    n_train   = rep(0, n)
  )

  res1 <- .oplsComponentCv(X, Ycs_fold, cv.set, 1, NULL, acc)

  res2 <- .oplsComponentCv(
    X,
    Ycs_fold,
    cv.set,
    nc = 2,
    mod.cv = res1$mod.cv,
    acc = res1$acc
  )

  expect_equal(ncol(res2$mod.cv[[1]]$t_xo), 2)
  expect_equal(ncol(res2$mod.cv[[1]]$t_xp), 2)

})


test_that(".oplsComponentCv stores predictions in correct rows", {

  n <- 20
  p <- 4

  X <- matrix(rnorm(n * p), n, p)
  Y <- matrix(rnorm(n), n, 1)

  cv.set <- list(1:10, 11:20)
  Ycs_fold <- list(Y, Y)

  acc <- list(
    sum_test  = matrix(0, n, 1),
    n_test    = rep(0, n),
    sum_train = matrix(0, n, 1),
    n_train   = rep(0, n)
  )

  res <- .oplsComponentCv(X, Ycs_fold, cv.set, 1, NULL, acc)

  m <- res$mod.cv[[1]]

  expect_true(any(m$y_pred_test[-cv.set[[1]], ] != 0))
  expect_true(any(m$y_pred_train[cv.set[[1]], ] != 0))

})

#########################
### MODEL VALIDATION ###
#########################

setValidity("m8_model", function(object) {
  msg <- character()

  if (length(object@engine) != 1L || is.na(object@engine) || object@engine == "") {
    msg <- c(msg, "slot 'engine' must be a single non-empty character string.")
  } else if (!object@engine %in% c("pca", "pls", "opls")) {
    msg <- c(msg, "slot 'engine' must be one of: 'pca', 'pls', 'opls'.")
  }

  if (!is.list(object@dims) || is.null(object@dims$n) || is.null(object@dims$p)) {
    msg <- c(msg, "slot 'dims' must be a list containing elements 'n' and 'p'.")
  } else {
    n <- object@dims$n
    p <- object@dims$p
    if (!is.numeric(n) || length(n) != 1L || is.na(n) || n <= 0) {
      msg <- c(msg, "dims$n must be a single positive numeric value.")
    }
    if (!is.numeric(p) || length(p) != 1L || is.na(p) || p <= 0) {
      msg <- c(msg, "dims$p must be a single positive numeric value.")
    }
  }

  if (!is.list(object@ctrl)) {
    msg <- c(msg, "slot 'ctrl' must be a list.")
  } else {
    ctrl <- object@ctrl
    if (!is.null(ctrl$type)) {
      if (!is.character(ctrl$type) || length(ctrl$type) != 1L || is.na(ctrl$type)) {
        msg <- c(msg, "ctrl$type must be a single character string.")
      } else if (!ctrl$type %in% c("R", "DA", "DA-mY", "unsupervised")) {
        msg <- c(msg, "ctrl$type must be one of: 'R', 'DA', 'unsupervised'.")
      }
    } else {
      msg <- c(msg, "ctrl$type is missing for engine.")
    }

    n_sel <- ctrl$ncomp_selected
    n_tst <- ctrl$ncomp_tested

    if (!is.null(n_sel) && (!is.numeric(n_sel) || length(n_sel) != 1L || is.na(n_sel) || n_sel < 0)) {
      msg <- c(msg, "ctrl$ncomp_selected must be a single non-negative numeric value.")
    }
    if (!is.null(n_tst) && (!is.numeric(n_tst) || length(n_tst) != 1L || is.na(n_tst) || n_tst < 0)) {
      msg <- c(msg, "ctrl$ncomp_tested must be a single non-negative numeric value.")
    }

    if (!is.null(n_sel) && !is.null(n_tst) &&
        is.numeric(n_sel) && is.numeric(n_tst) &&
        !is.na(n_sel) && !is.na(n_tst)) {
      if (n_sel > n_tst) {
        msg <- c(msg, "ctrl$ncomp_selected cannot exceed ctrl$ncomp_tested.")
      }
      if (object@engine %in% c("pls", "opls") && n_tst < 1) {
        msg <- c(msg, "For PLS/OPLS models, ctrl$ncomp_tested should be >= 1.")
      }
    }

    check_len <- function(x, name, n_req, offset=0) {
      if (!is.null(x) && !is.null(n_req) && is.numeric(n_req) && !is.na(n_req) && n_req > 0) {
        if (!is.numeric(x) && !is.logical(x))
          return(paste0(name, " must be numeric (or logical NA)."))
        expected <- n_req - offset
        if (length(x) < expected)
          return(paste0(name, " has length ", length(x),
                        " but expected at least ", expected,
                        " based on ncomp_tested = ", n_req, "."))
      }
      NULL
    }

    if (!is.null(n_tst) && is.numeric(n_tst) && !is.na(n_tst) && n_tst > 0) {
      offset <- if (object@engine == "opls") 1 else 0
      m <- c(
        check_len(ctrl$q2, "ctrl$q2", n_tst, offset),
        check_len(ctrl$r2, "ctrl$r2", n_tst, offset),
        check_len(ctrl$aucs_te, "ctrl$aucs_te", n_tst, offset),
        check_len(ctrl$aucs_tr, "ctrl$aucs_tr", n_tst, offset)
      )
      m <- m[!vapply(m, is.null, logical(1))]
      if (length(m)) msg <- c(msg, m)
    }

    if (identical(object@engine, "opls")) {
      if (!is.null(ctrl$nc_pred) && (!is.numeric(ctrl$nc_pred) || length(ctrl$nc_pred) != 1L || is.na(ctrl$nc_pred) || ctrl$nc_pred > 1)) {
        msg <- c(msg, "For OPLS, ctrl$nc_pred must be a single numeric value >= 1.")
      }
      if (!is.null(ctrl$nc_orth) && (!is.numeric(ctrl$nc_orth) || length(ctrl$nc_orth) != 1L || is.na(ctrl$nc_orth) || ctrl$nc_orth < 0)) {
        msg <- c(msg, "For OPLS, ctrl$nc_orth must be a single non-negative numeric value.")
      }
      if (!is.null(ctrl$nc_pred) && !is.null(ctrl$nc_orth) && !is.null(ctrl$ncomp_selected) &&
          is.numeric(ctrl$nc_pred) && is.numeric(ctrl$nc_orth) && is.numeric(ctrl$ncomp_selected) &&
          !is.na(ctrl$nc_pred) && !is.na(ctrl$nc_orth) && !is.na(ctrl$ncomp_selected)) {
        if ((ctrl$nc_pred + ctrl$nc_orth) != ctrl$ncomp_selected) {
          msg <- c(msg, "For OPLS, ctrl$ncomp_selected must equal ctrl$nc_pred + ctrl$nc_orth.")
        }
      }
    }
  }

  if (!is.list(object@fit)) msg <- c(msg, "slot 'fit' must be a list.")
  if (!is.list(object@provenance)) msg <- c(msg, "slot 'provenance' must be a list.")
  if (!is.list(object@session)) msg <- c(msg, "slot 'session' must be a list.")

  if (!is.list(object@cv) && !is.null(object@cv)) {
    msg <- c(msg, "slot 'cv' must be a list'.")
  }
  if (length(msg)) msg else TRUE
})

# --- Model -----------------------------------------------------------
#' @import methods
NULL

setClassUnion("listOrNULL", c("list", "NULL"))

#' @title m8_model class
#' Model object returned by \code{pca()}, \code{pls()}, and \code{opls()}.
#' @slot engine Character. Model engine ("pca", "pls", "opls").
#' @slot ctrl List. Engine-specific control and performance information.
#' @slot fit List. Fitted data (engine specific).
#' @slot cv Resampling instance (may be NULL if not used).
#' @slot prep Scaling and centering information
#' @slot provenance Preprocessing attributes of spectral matrix X
#' @slot session R-session information
#' @slot call Function call
#' @slot dims List with \code{n} and \code{p}.
#' @return
#' An object of class `m8_model`.
#' @name m8_model-class
#' @examples
#' data(covid)
#' cv <- balanced_mc(k=5, split=2/3)
#' scaling <- uv_scaling(center=TRUE)
#' model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
#' class(model)
#' show(model)
#' @aliases m8_model
#' @exportClass m8_model
setClass("m8_model",
         slots = c(
           engine = "character",
           ctrl = "list",
           fit = "list",
           cv = "listOrNULL",
           prep = "list",# "m8_preprocess",
           provenance = "list",
           session = "list",
           call = "call",
           dims = "list"
         )
)

#' @title Distance to the Model in X-Space (DModX)
#'
#' @description
#' Calculates the orthogonal distance of each observation to an OPLS model in X-space.
#' DModX can be used for identifying outliers.
#'
#' @param mod An OPLS model object of class \code{m8_model} (engine = "opls").
#' @param plot Logical. If TRUE, a plot of DModX values with an approximate cutoff is shown.
#'
#' @return A data frame with columns \code{ID}, \code{DmodX}, and \code{passed}.
#'
#' @details
#' DModX is computed from the X-residual matrix as a scaled RMSE of residuals.
#' The empirical cutoff (dashed line in plot) uses a t-based confidence interval
#' and assumes approximate normality of DModX values. This assumption may not be
#' satisfied in all datasets, so the resulting threshold should be regarded
#' as a pragmatic heuristic for outlier detection.
#'
#' @family model_validation
#' @importFrom stats t.test sd
#' @importFrom ggplot2 ggplot aes geom_point geom_segment geom_hline
#'   scale_y_continuous xlab theme_bw theme element_blank element_text
#'   scale_colour_gradientn labs
#' @importFrom colorRamps matlab.like2
#' @importFrom rlang .data
#' @examples
#' data(covid)
#' cv <- balanced_mc(k=5, split=2/3)
#' scaling <- uv_scaling(center=TRUE)
#' model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
#' dX <- dmodx(model)
#' print(dX[[1]])
#' df <- dX[[2]]; head(df)
#'
#' @export
dmodx <- function(mod, plot = TRUE) {

  if (!inherits(mod, "m8_model") || !identical(mod@engine, "opls")) {
    stop("Please provide an m8_model object with engine = 'opls'.", call. = FALSE)
  }

  E <- xres(mod)

  if (is.null(E)) {
    stop("Could not find X residual matrix.", call. = FALSE)
  }

  N <- nrow(E)
  K <- ncol(E)
  A <- 1L

  A0 <- if (isTRUE(mod@prep$center)) 1L else 0L

  if ((K - A) <= 0) stop("Invalid dimensions for DModX: need K - A > 0.", call. = FALSE)
  if ((N - A - A0) <= 0) stop("Invalid dimensions for DModX: need N - A - A0 > 0.", call. = FALSE)

  ss_res <- rowSums(E^2)
  denom_global <- sqrt(sum(ss_res) / ((N - A - A0) * (K - A)))
  dmodX <- sqrt(ss_res / (K - A)) / denom_global

  tt <- t.test(dmodX, alternative = "less")
  ci95 <- tt$conf.int[2] + 2 * sd(dmodX)

  df <- data.frame(
    ID     = seq_len(N),
    DmodX  = dmodX,
    passed = dmodX < ci95
  )

  if (isTRUE(plot)) {

    df$outlier <- df$DmodX > ci95
    df$label <- as.character(df$ID)

    grp <- NULL
    if (!is.null(mod@fit$Y)) {
      grp <- mod@fit$Y
      if (mod@ctrl$type=='DA'){
        grp <- as.factor(grp)
      }
      }
    if (!is.null(grp) && length(grp) == nrow(df)) {
      df$Group <- grp
    } else {
      df$Group <- factor("all")
    }


    g <- ggplot2::ggplot(df, ggplot2::aes(x = .data$ID, y = .data$DmodX)) +
      ggplot2::geom_segment(
        ggplot2::aes(xend = .data$ID, yend = 0),
        colour = "grey85", linewidth = 0.2
      ) +
      ggplot2::geom_hline(yintercept = ci95, linetype = 2, colour = "black") +
      ggplot2::scale_y_continuous(
        limits = c(0, max(df$DmodX, ci95) * 1.05),
        expand = c(0, 0),
        name = "DModX"
      ) +
      ggplot2::xlab("Sample index") +
      ggplot2::theme_bw() +
      ggplot2::theme(
        panel.grid.minor.x = ggplot2::element_blank(),
        panel.grid.major.x = ggplot2::element_blank(),
        panel.grid.minor.y = ggplot2::element_blank(),
        axis.text = ggplot2::element_text(colour = "black")
      ) +
      ggplot2::scale_x_continuous(
        breaks = scales::breaks_pretty(n = 10)
      ) +
      ggplot2::labs(caption = "Dashed line indicates one-sided empirical cutoff (t-based, 95% limit).")

    g <- g +
      ggplot2::geom_point(
        data = df[!df$outlier, , drop = FALSE],
        ggplot2::aes(colour = .data$Group),
        size = 1.7, alpha = 0.9
      )

    g <- g +
      ggplot2::geom_point(
        data = df[df$outlier, , drop = FALSE],
        colour = "red3",
        size = 3.0, alpha = 1
      )

    if (any(df$outlier)) {
      g <- g +
        ggrepel::geom_text_repel(
          data = df[df$outlier, , drop = FALSE],
          ggplot2::aes(label = .data$label),
          colour = "red3",
          size = 3,
          min.segment.length = 0,
          box.padding = 0.25,
          point.padding = 0.2,
          max.overlaps = Inf
        )
    }

    if (!is.null(mod@ctrl$type) && identical(mod@ctrl$type, "R") && !is.factor(df$Group)) {
      g <- g + scale_colour_gradientn(colours = matlab.like2(10), name = "Y")
    }

    plot(g)
  }

  df
}

#' @title Import 1D NMR spectra (TopSpin processed)
#'
#' @description
#' Imports TopSpin-processed 1D NMR spectra together with spectrometer acquisition
#' and TopSpin processing parameters (`acqus` and `procs`, respectively).
#'
#' @param path Character. Directory path containing Bruker NMR experiments.
#'
#' @param exp_type Named list. Optional filtering specification based on
#' acquisition or processing metadata. Each list element must correspond to a
#' metadata field (e.g. \code{pulprog}, \code{ns}, \code{rg}). Filtering supports:
#' \itemize{
#'   \item Exact match: \code{list(pulprog = "noesygppr1d")}
#'   \item Membership: \code{list(pulprog = c("zg30", "noesygppr1d"))}
#'   \item Numeric range: \code{list(ns = list(range = c(16, 128)))}
#'   \item Generic comparison: \code{list(ns = list(op = ">=", value = 32))}
#' }
#' Multiple fields are combined using logical AND.
#'
#' @param n_max Integer. Maximum number of spectra to import. Default: 1000.
#' @param filter Logical. Filter out experiments with incomplete file systems.
#' @param recursive Logical. Search \code{path} recursively. Default: TRUE.
#' @param verbose Logical or numeric. Verbosity level.
#' @param to_global Logical. If \code{TRUE}, the returned objects are additionally
#' assigned to the global environment.
#'
#' @return
#' A named list with three elements:
#' \describe{
#'   \item{X}{A numeric matrix of spectra (rows = samples, columns = ppm values).}
#'   \item{ppm}{A numeric vector of chemical shift values (ppm).}
#'   \item{meta}{A data frame of acquisition and processing metadata,
#'   row-aligned with \code{X}.}
#' }
#'
#' If \code{to_global = TRUE}, objects with the same names in the global
#' environment will be overwritten.
#'
#' @examples
#' path <- system.file("extdata", package = "metabom8")
#'
#' read1d_proc(path, exp_type = list(pulprog = "noesygppr1d"), n_max = 2)
#'
#' @importFrom fs path_common path_rel
#' @importFrom stats approxfun
#' @export
read1d <- function(path,
                   exp_type = list(pulprog = "noesygppr1d"),
                   n_max = 1000,
                   filter = TRUE,
                   recursive = TRUE,
                   verbose = 1,
                   to_global = FALSE) {
  path <- path.expand(path)

  if (as.character(match.call()[[1]]) == "read1d") {
    warning("`read1d` will be deprecated; please use `read1d_proc()`.", call. = FALSE)
  }

  f_list <- .detect1d_procs(path, n_max = n_max, filter = filter, recursive = recursive, verbose = verbose)
  if (is.null(f_list)) stop("No valid experiments found.", call. = FALSE)

  pars <- .extract_pars1d(f_list)
  exp_filt <- .filterExp_files(pars, exp_type, f_list, n_max = n_max)
  f_list <- exp_filt[[1]]
  pars <- exp_filt[[2]]

  ppm_ref <- .chemShift(swidth = pars$a_SW[1], offset = pars$p_OFFSET[1], si = pars$p_SI[1])

  out <- vapply(seq_along(f_list[[1]]), function(s) {
    cs_ppm <- .chemShift(swidth = pars$a_SW[s], offset = pars$p_OFFSET[s], si = pars$p_SI[s])
    byteord <- c(little = 0, big = 1)
    spec <- readBin(
      f_list$f_1r[s], what = "int", n = pars$p_FTSIZE[s], size = 4,
      signed = TRUE, endian = names(byteord)[match(pars$p_BYTORDP[s], byteord)]
    )
    spec <- spec * (2 ^ pars$p_NC_proc[s])
    f_spec <- approxfun(x = cs_ppm, y = spec, rule = 2)
    f_spec(ppm_ref)
  }, FUN.VALUE = numeric(length(ppm_ref)))

  out <- t(out)
  colnames(out) <- ppm_ref


  pars$exp_dir <- dirname(f_list$f_1r)

  if (length(f_list$f_1r) ==1){
    x <- strsplit(f_list$f_1r, .Platform$file.sep)[[1]]
    xr <- x[(length(x) - 3):(length(x)-1)]
    fnam <- paste(xr, collapse = .Platform$file.sep)
  } else{
    cpath <- path_common(f_list$f_1r)
    fnam <- path_rel(pars$exp_dir, cpath)
  }

  rownames(out) <- fnam
  rownames(pars) <- fnam
  out[is.na(out)] <- 0


  if (verbose) {
    if (nrow(out)==1) message("Imported ", nrow(out), " spectrum.") else{
      message("Imported ", nrow(out), " spectra.")
    }}

  prov <- list(
    source = list(
      vendor = "bruker",
      input = "topspin_proc",
      processed_upstream = TRUE
    ),
    path = list(
      root = path,
      experiments = f_list$path,
      recursive = recursive,
      filter_incomplete = filter,
      n_max = n_max
    ),
    time = list(
      import_time_utc = format(Sys.time(), tz = "UTC", usetz = TRUE)
    ),
    pipeline = list(
      fct = "metabom8::read1d_proc",
      mode = "absorption",
      steps = c("read_1r", "ppm_from_acqus_procs", "interpolation"),
      params = list( exp_type = exp_type)
    ),
    axis = list(
      ppm_source = "topspin_procs",
      common_grid = TRUE,
      interpolation = list(
        applied = TRUE,
        method = "linear",
        rule = 2,
        from = "cs_ppm",
        to = "ppm_ref",
        notes = "Each spectrum interpolated onto common ppm_ref grid."
      )
    ),
    session = list(
      metabom8 = as.character(utils::packageVersion("metabom8")),
      R = paste0(R.version$major, ".", R.version$minor)
    )
  )

  attr(out, "m8_axis") <- list(ppm = ppm_ref)
  attr(out, "m8_provenance") <- prov

  out <- .m8_stamp(
    out,
    step = "import ft, processed spectra (1r)",
    params = list(source_dir = path, exp_filter = .list_to_string(exp_type), n_spectra = nrow(out)),
    notes = "TopSpin-processed spectra imported and interpolated onto a common ppm grid."
  )

  collect_out <- list(X = out, ppm = ppm_ref, meta = pars)

  if (to_global) {
    if (verbose > 0) message("Assigning objects X, ppm, meta to global environment.")
    list2env(collect_out, envir = .GlobalEnv)
    # assign("X", out, envir = .GlobalEnv)
    # assign("ppm", ppm_ref, envir = .GlobalEnv)
    # assign("meta", pars, envir = .GlobalEnv)
  }

  invisible(collect_out)
}

#' @rdname read1d
#' @export
read1d_proc <- read1d


#' @title Read Bruker NMR Parameter Files
#'
#' @description
#' Helper function used by `read1d()` to extract acquisition (`acqus`) and processing (`procs`) parameters
#' from Bruker-formatted 1D NMR experiments.
#'
#' @param f_list List of file paths. Output from `.detect1d_procs()`.
#'
#' @return A data frame of extracted acquisition and processing metadata. Row names correspond to spectrum filenames.
#'
#' @importFrom stats setNames
#' @keywords internal
.extract_pars1d <- function(f_list) {
  out <- lapply(seq_along(f_list[[1]]), function(i) {
    # Read procs file
    f_procs_lines <- readLines(f_list$f_procs[i], warn = FALSE)
    idx_procs <- grep("..", f_procs_lines, fixed = TRUE)
    f_procs_lines[idx_procs] <- vapply(idx_procs, function(j) {
      gsub(" .*", f_procs_lines[j + 1], f_procs_lines[j])
    }, FUN.VALUE = "")
    procs_vals <- strsplit(gsub("^##\\$", "", grep("^##\\$", f_procs_lines, value = TRUE), fixed = FALSE), "=")
    d_procs_val <- gsub("^ ", "", vapply(procs_vals, "[[", 2, FUN.VALUE = ""))
    names(d_procs_val) <- paste0("p_", vapply(procs_vals, "[[", 1, FUN.VALUE = ""))

    # Read acqus file
    f_acqus_lines <- readLines(f_list$f_acqus[i], warn = FALSE)
    idx_acqus <- grep("..", f_acqus_lines, fixed = TRUE)
    f_acqus_lines[idx_acqus] <- vapply(idx_acqus, function(j) {
      gsub(" .*", f_acqus_lines[j + 1], f_acqus_lines[j])
    }, FUN.VALUE = "")
    acqus_vals <- strsplit(gsub("^##\\$", "", grep("^##\\$", f_acqus_lines, value = TRUE), fixed = FALSE), "=")
    d_acqu_val <- gsub("^ ", "", vapply(acqus_vals, "[[", 2, FUN.VALUE = ""))
    names(d_acqu_val) <- paste0("a_", vapply(acqus_vals, "[[", 1, FUN.VALUE = ""))

    # Convert datetime
    idx_date <- grep("date", names(d_acqu_val), ignore.case = TRUE)
    d_acqu_val[idx_date] <- as.character(
      as.POSIXct("1970-01-01 00:00:00", tz = "UTC") + as.numeric(d_acqu_val[idx_date])
    )

    c(d_acqu_val, d_procs_val)
  })

  # Handle unequal metadata entries
  out_len <- vapply(out, length, FUN.VALUE = 1L)
  if (length(unique(out_len)) > 1) {
    all_names <- unique(unlist(lapply(out, names)))
    out <- lapply(out, function(x) {
      out_row <- setNames(rep(NA, length(all_names)), all_names)
      out_row[names(x)] <- x
      out_row
    })
  }

  out_df <- as.data.frame(do.call(rbind, out), stringsAsFactors = FALSE)

  is_num <- vapply(out_df, .is_numeric_trycatch, logical(1))
  out_df[is_num] <- lapply(out_df[is_num], as.numeric)

  rownames(out_df) <- f_list[[2]]
  return(out_df)
}


#' @title Check for Intact File Systems - Helper Function for read1d
#' @description
#' Scans the specified directory recursively to find intact sets of Bruker NMR files:
#' 'procs', 'acqus', and '1r' files. Optionally filters incomplete experiments.
#'
#' @param datapath character. Path to directory containing spectra.
#' @param n_max integer. Maximum number of spectra to read (default 10).
#' @param filter logical. Whether to filter out incomplete file sets (default TRUE).
#' @param recursive logical. Whether to search directories recursively.
#' @param verbose integer. Verbosity level for messaging.
#'
#' @return
#' A list with elements:
#' \describe{
#'   \item{path}{Character vector of experiment folder paths (intact).}
#'   \item{exp_no}{Character vector of experiment folder IDs.}
#'   \item{f_procs}{Character vector of paths to 'procs' files.}
#'   \item{f_acqus}{Character vector of paths to 'acqus' files.}
#'   \item{f_1r}{Character vector of paths to '1r' files.}
#' }
#'
#' @keywords internal
.detect1d_procs <- function(datapath, n_max = 10, filter = TRUE, recursive = TRUE, verbose = 1) {

  datapath <- gsub(paste0(.Platform$file.sep, "$"), "", path.expand(datapath))

  f_procs <- list.files(datapath, pattern="^procs$", full.names=TRUE, recursive=recursive, ignore.case=TRUE)
  f_acqus <- list.files(datapath, pattern="^acqus$", full.names=TRUE, recursive=recursive, ignore.case=TRUE)
  f_1r    <- list.files(datapath, pattern="^1r$",    full.names=TRUE, recursive=recursive, ignore.case=TRUE)

  if (length(f_acqus) == 0 || length(f_procs) == 0 || length(f_1r) == 0) return(NULL)

  # experiment directory = folder containing acqus
  expdir_a <- dirname(f_acqus)

  # for processed files, experiment directory = strip "/pdata/..." tail
  expdir_p <- sub(paste0(.Platform$file.sep, "pdata", .Platform$file.sep, ".*$"), "", f_procs)
  expdir_1 <- sub(paste0(.Platform$file.sep, "pdata", .Platform$file.sep, ".*$"), "", f_1r)

  # keep only experiments that have all three
  common <- Reduce(intersect, list(expdir_a, expdir_p, expdir_1))

  if (length(common) == 0) {
    if (filter) return(NULL)
    stop("No experiments with matching acqus, procs and 1r found.", call. = FALSE)
  }

  # map to common exp dirs
  ia <- match(common, expdir_a)
  ip <- match(common, expdir_p)
  i1 <- match(common, expdir_1)

  f_acqus <- f_acqus[ia]
  f_procs <- f_procs[ip]
  f_1r    <- f_1r[i1]

  # limit
  if (length(f_1r) > n_max) {
    keep <- seq_len(n_max)
    f_acqus <- f_acqus[keep]; f_procs <- f_procs[keep]; f_1r <- f_1r[keep]
    common <- common[keep]
    if (verbose) message("Reached n_max - not all spectra read-in.")
  }

  return(list(
    path   = common,
    exp_no = basename(common),   # "162" for your example
    f_procs = f_procs,
    f_acqus = f_acqus,
    f_1r = f_1r
  ))
}


#' @title Calculate Chemical Shift Axis
#'
#' @description
#' Computes a 1D NMR chemical shift axis based on sweep width, offset, and size.
#'
#' @param swidth Numeric. Sweep width (Hz).
#' @param offset Numeric. Frequency offset (ppm).
#' @param si Integer. Number of data points.
#'
#' @return Numeric vector of chemical shift values (ppm).
#'
#' @keywords internal
.chemShift <- function(swidth, offset, si) {
  dppm <- swidth / (si - 1)
  seq(offset, offset - swidth, by = -dppm)
}

#### ppm

test_that("chemical shift axis is computed correctly", {
  ppm <- .chemShift(swidth = 1000, offset = 10, si = 5)
  expected <- seq(10, 10 - 1000, by = -1000 / (5 - 1))
  expect_equal(ppm, expected, tolerance = 1e-12)
  expect_length(ppm, 5)
})


#### parsing metadata
test_that(".extract_pars1d correctly parses metadata", {
  temp_dir <- tempfile("10_")
  dir.create(temp_dir)
  on.exit(unlink(temp_dir, recursive = TRUE), add = TRUE)

  acqus_path <- file.path(temp_dir, "acqus")
  procs_path <- file.path(temp_dir, "pdata", "1", "procs")
  dir.create(dirname(procs_path), recursive = TRUE, showWarnings = FALSE)

  f_1r_path <- file.path(temp_dir, "pdata", "1", "1r")
  dir.create(dirname(f_1r_path), recursive = TRUE, showWarnings = FALSE)
  writeLines("dummy data", f_1r_path)

  writeLines(c("##$DATE= 1640995200", "##$EXP= test"), acqus_path)
  writeLines(c("##$SI= 1024", "##$SW= 5000", "##$OFFSET= 4.7"), procs_path)

  expect_true(dir.exists(dirname(procs_path)))
  expect_true(dir.exists(dirname(f_1r_path)))
  expect_true(file.exists(acqus_path))
  expect_true(file.exists(procs_path))
  expect_true(file.exists(f_1r_path))

  f_list <- list(
    f_acqus = acqus_path,
    f_procs = procs_path,
    f_1r = f_1r_path,
    path = temp_dir,
    exp_no = "10"
  )

  out <- .extract_pars1d(f_list)

  expect_s3_class(out, "data.frame")
  expect_true("a_DATE" %in% names(out))
  expect_true("p_SI" %in% names(out))
  expect_equal(out$p_SI, 1024)
})


### read1d_procs
test_that("read1d_proc returns named list with X, ppm meta", {
  skip_if_not_installed("testthat")
  skip_if_not(file.exists(system.file("extdata", package = "metabom8")),
              "Example data not installed.")

  path <- system.file("extdata", package = "metabom8")

  dat <- read1d_proc(
    path,
    exp_type = list(pulprog = 'noesygppr1d'),
    n_max = 1,
    verbose = 1
    )

  expect_true(is.list(dat))
  expect_named(dat, c("X","ppm","meta"), ignore.order = TRUE)
  expect_s3_class(dat$meta, "data.frame")
  expect_type(dat$ppm, "double")
  expect_true(is.matrix(dat$X))
  expect_equal(ncol(dat$X), length(dat$ppm))
})



test_that("read1d_proc returns expected global objects", {
  skip_if_not(dir.exists(system.file("extdata", package = "metabom8")),
              "extdata not found / installed.")
  on.exit(rm(list=c("X","ppm","meta"), envir=.GlobalEnv), add=TRUE)

  path <- system.file("extdata", package = "metabom8")

  dat <- read1d_proc(
    path,
    exp_type = list(pulprog = 'noesygppr1d'),
    n_max = 1,
    verbose = 1,
    to_global = TRUE
    )

  expect_true(exists("X", envir = .GlobalEnv))
  expect_true(exists("ppm", envir = .GlobalEnv))
  expect_true(exists("meta", envir = .GlobalEnv))
  expect_s3_class(get("meta", envir = .GlobalEnv), "data.frame")
  expect_type(get("ppm", envir = .GlobalEnv), "double")
  expect_type(get("X", envir = .GlobalEnv), "double")

})


### read1d_raw



test_that("read1d_raw returns named list with X, ppm meta", {
  skip_if_not(dir.exists(system.file("extdata", package = "metabom8")),
              "extdata not found / installed.")

  path <- system.file("extdata", package = "metabom8")

  dat <- read1d_raw(
    path = path,
    exp_type = list(exp = "PROF_PLASMA_NOESY"),
    apodisation = list(fun = "exponential", lb = 0.2),
    zerofill = 1L,
    mode = "absorption",
    verbose = 0,
    n_max = 2,
    recursive = TRUE,
    filter = TRUE
  )

  expect_true(is.list(dat))
  expect_named(dat, c("X","ppm","meta"), ignore.order = TRUE)
  expect_s3_class(dat$meta, "data.frame")
  expect_type(dat$ppm, "double")
  expect_true(is.matrix(dat$X))

  expect_equal(ncol(dat$X), length(dat$ppm))
  expect_equal(nrow(dat$X), nrow(dat$meta))
})



test_that("read1d_raw returns expected global objects", {
  skip_if_not(dir.exists(system.file("extdata", package = "metabom8")),
              "extdata not found / installed.")
  on.exit(rm(list=c("X","ppm","meta"), envir=.GlobalEnv), add=TRUE)

  path <- system.file("extdata", package = "metabom8")

  read1d_raw(
    path = path,
    exp_type = list(exp = "PROF_PLASMA_NOESY"),
    apodisation = list(fun = "exponential", lb = 0.2),
    zerofill = 1L,
    mode = "absorption",
    verbose = 0,
    n_max = 2,
    recursive = TRUE,
    filter = TRUE,
    to_global = TRUE
  )


  expect_true(exists("X", envir = .GlobalEnv))
  expect_true(exists("ppm", envir = .GlobalEnv))
  expect_true(exists("meta", envir = .GlobalEnv))

  expect_s3_class(get("meta", envir = .GlobalEnv), "data.frame")
  expect_type(get("ppm", envir = .GlobalEnv), "double")
  expect_true(is.matrix(get("X", envir = .GlobalEnv)))

  expect_equal(ncol(get("X", envir = .GlobalEnv)),
               length(get("ppm", envir = .GlobalEnv)))
  expect_equal(nrow(get("X", envir = .GlobalEnv)),
               nrow(get("meta", envir = .GlobalEnv)))

})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  error = FALSE,
  collapse = TRUE,
  comment = "#>"
)

## ----get-data-----------------------------------------------------------------
library(metabom8)

## ----import, fig.width=7.2, fig.height=4--------------------------------------
## Example: import 1D spectra
exp_dir <- system.file("extdata", package = "metabom8")
exp_type <- list(pulprog="noesygppr1d")

nmr_data <- read1d_proc(exp_dir, exp_type, n_max = 100)

names(nmr_data)

## ----pars_table, echo=FALSE---------------------------------------------------
knitr::kable(
  data.frame(
    Prefix = c("a_", "p_"),
    Meaning = c("Acquisition parameters",
                "Processing parameters"),
    Examples = c("a_PULPROG, a_SFO1, a_RG, a_SW_h, a_TE",
                 "p_WDW, p_LB, p_SI, p_PHC*, p_BC*, p_NC_proc")
  ),
  align = "lll"
)

## ----read-in-raw, fig.show='hold'---------------------------------------------
# import 1D NMR data
nmr_raw <- read1d_raw(
  exp_dir,
  exp_type,
  apodisation = list(fun = "exponential", lb = 0.2),
  zerofil = 1L,
  mode = "absorption"
  )

names(nmr_raw)

## ----apod, eval=TRUE, fig.width=8---------------------------------------------
# selected FID apodisation functions
f_apod <- c("sine","cosine","exponential","sem")

# create toy fid
n <- 200; t <- seq(0,1,len=n)
fid <- exp(-5*t)*cos(20*pi*t)

# apply apodisation
A  <- sapply(f_apod, \(f) metabom8:::.fidApodisationFct(n, list(fun=f, lb=-0.2)))
Fp <- sweep(A, 1, fid, `*`)
S  <- apply(Fp, 2, \(x) Mod(fft(x))[1:(n/2)])

# compare graphically
cols <- 1:ncol(A)

par(mfrow=c(2,2), mar=c(3,3,2,1))
plot(t, fid, type="l", lwd=2, main="FID")
matplot(t, A, type="l", lwd=2, col=cols, main="Apodisation windows")
matplot(t, Fp, type="l", lwd=2, col=cols, main="Windowed FID")
matplot(S, type="l", lwd=2, col=cols, main="Spectra")

legend("topright", f_apod, col=cols, lty=1, bty="n", cex=.8)

## ----viz_covid, fig.width=7.2, fig.height=4-----------------------------------
data("covid", package = "metabom8")

## plot directly from the dataset object
plot_spec(covid, shift = c(0.5, 2))

## alternatively extract data components: X, ppm
X <- covid$X
ppm <- covid$ppm

dim(X)
head(ppm)

plot_spec(X[1:3, ], ppm, shift = c(0.5, 2)) # backend='plotly' by default
plot_spec(X[1:3, ], ppm, shift = c(0.5, 2), backend='base')
plot_spec(X[1:3, ], ppm, shift = c(0.5, 2), backend='ggplot2')

## ----preproc_pipe-------------------------------------------------------------
data("hiit_raw", package = "metabom8") # urine NMR (HIIT exercise experiment)

names(hiit_raw) # X, ppm, meta

# piped preprocessing 
hiit_proc <- hiit_raw |> 
  calibrate(type = "tsp") |>
  excise() |>
  correct_baseline(method='asls') |>
  align_spectra() |>
  pqn()

dim(hiit_proc$X)

## ----viz-raw-hiit, fig.width=7.2, fig.height=4--------------------------------
plot_spec(hiit_raw, shift = c(4,4.1))

## ----viz-proc-hiit, fig.width=7.2, fig.height=4-------------------------------
plot_spec(hiit_proc, shift = c(4,4.1))

## ----preproc_stepwise, fig.width=7.2, fig.height=4----------------------------
X <- hiit_raw$X
ppm <- hiit_raw$ppm

# perform TSP calibration
X_cal <- calibrate(X, ppm, type='tsp')

# excise chemical shift regions
regions <- list(
      upfield_noise   = c(min(ppm), 0.25),
      residual_water  = c(4.5, 5.2),
      downfield_noise = c(9.7, max(ppm))
    )
X_exi <- excise(X_cal, ppm, regions)
ppm_exi <- attr(X_exi, 'm8_axis')$ppm

# baseline correction
X_bli <- correct_baseline(X_exi)

# plot spectrum
plot_spec(X_bli[1,], ppm_exi, shift = c(0,10))

## ----preproc_provenance-------------------------------------------------------

## Inspect preprocessing steps
print_provenance(hiit_proc)

# Retrieve specific step
get_provenance(hiit_proc, step = 2)

## Retrieve the full provenance log
log <- get_provenance(hiit_proc)
# print(log) # very detailed

## ----preproc_provenance-indi--------------------------------------------------
# Retrieve parameter from a named step
get_provenance(hiit_proc, step = "calibrate", param = "target")

## ----provenance_note----------------------------------------------------------
# collect env varaibles
params <- list(
  agent    = paste0("snakemake/", Sys.getenv("SNAKEMAKE_VERSION", "v?")),
  runtime  = "docker",
  workflow = Sys.getenv("M8_WORKFLOW", "std_prof-urine"),
  run_id   = Sys.getenv("M8_RUN_ID", "m8-2605-001")
)
 
# create note with env parameters
hiit_proc <- add_note(hiit_proc, 
         'Excluding outlier sample XX - reason is ...', 
         params)

# check provenance out
get_provenance(hiit_proc , step = "user note")

## ----mva-data-----------------------------------------------------------------
Y <- covid$an$type
X <- covid$X

## ----model_setup--------------------------------------------------------------
# Define the scaling and centering strategy
uv = uv_scaling(center = TRUE)

## ----pca----------------------------------------------------------------------
pca_model <- pca(X, scaling = uv, ncomp=2)

# show / summary methods 
show(pca_model)
summary(pca_model)

## ----resampling---------------------------------------------------------------
mc_cv <- balanced_mc(k=15, split=2/3, type="DA")

## ----pls----------------------------------------------------------------------
pls_model <- pls(X, Y, validation_strategy=mc_cv, scaling = uv)

# model information
show(pls_model)

# model performance summary
summary(pls_model)

# model scores & loadings
Tp <- scores(pls_model)
Pp <- loadings(pls_model)

## ----opls---------------------------------------------------------------------
opls_model <- opls(X, Y, validation_strategy=mc_cv, scaling = uv)

# model information & performance
show(opls_model)

## ----opls-summary-------------------------------------------------------------
summary(opls_model)

## ----opls-t-p-----------------------------------------------------------------
# model scores, loadings & vip
Tp <- scores(opls_model)
To <- scores(opls_model, orth=TRUE)

Pp <- loadings(opls_model)
Po <- loadings(opls_model, orth=TRUE)

vip <- vip(opls_model)

## ----opls-dx, fig.width=7.2, fig.height=4-------------------------------------
# distance to the model in X space
dex <- dmodx(opls_model)
head(dex)

## ----citation, echo=FALSE, results='asis'-------------------------------------
citation("metabom8")

## ----session-info, echo=FALSE-------------------------------------------------
sessionInfo()

#' @title k-fold cross-validation index generator
#'
#' @description
#' Creates a list of training set indices for k-fold cross-validation (CV),
#' without considering class balance in Y. Each element of the list represents
#' the row indices used for training in one CV fold.
#'
#' @param k Integer. Number of CV folds. If \code{k} is not valid or too high relative to number of rows in Y,
#' it defaults to leave-one-out CV (LOO-CV).
#' @param Y Matrix. Outcome matrix (observations x variables).
#'
#' @return
#' A list of length \code{k}, where each element contains the training indices (integers)
#' for one CV fold.
#'
#' @keywords internal
.kFold <- function(k, Y){
  n <- nrow(Y)

  if (!is.numeric(k) || is.na(k) || is.infinite(k) || k < 2 || k > n) {
    stop("Invalid k value.", call. = FALSE)
  }

  if (k != n && n / k < 2) {
    message("Fold size is small; consider using fewer folds.")
  }

  sets <- sample(rep_len(seq_len(k), n))
  sets_list <- lapply(seq_len(k), function(i) {
    which(sets != i)
  })

  sets_list
}


#' @title Stratified k-fold cross-validation index generator
#'
#' @description
#' Generates a list of training set indices for stratified k-fold cross-validation (CV).
#' Stratification is performed based on the first column of \code{Y}. For regression tasks,
#' \code{Y} is binned by quantiles to emulate class balance. Ensures class proportions are
#' preserved across folds when possible.
#'
#' @param k Integer. Number of folds.
#' @param stratified List with three elements:
#' \itemize{
#'   \item \code{type}: Character. Either \code{"R"} for regression or \code{"DA"} for discriminant analysis.
#'   \item \code{Y}: Matrix. Outcome matrix with a single column.
#'   \item \code{probs}: Numeric vector. Probabilities used for stratification of numeric \code{Y} (regression only).
#' }
#'
#' @return
#' A list of length \code{k}, each element containing the training set row indices (integers) for one CV fold.
#' Returns \code{NULL} and a warning if stratification is not feasible due to class imbalance.
#'
#' @importFrom stats quantile
#' @keywords internal
.kFoldStratified <- function(k, stratified){

  if (!is.list(stratified) || length(stratified) != 3) {
    stop("'stratified' must be a list of length 3: type, Y, probs.", call. = FALSE)
  }

  Y <- stratified[[2]]
  if (anyNA(Y)) {
    stop("Stratified CV not possible: Y contains NA values.", call. = FALSE)
  }

  if (grepl('R', stratified[[1]])) {
    Yori <- Y
    breaks <- unique(quantile(Yori[, 1], stratified[[3]]))
    if (length(breaks) < 3) {
      warning("Insufficient variability in Y for stratification - use unstratified k-fold or MC.")
      return(NULL)
    }
    Y <- cbind(cut(Yori[, 1], breaks = breaks, include.lowest = TRUE))
  } else {
    Y <- Y[, 1]
  }

  ct <- table(Y)
  if (min(ct) / max(ct) < 0.1) {
    warning("Skewed class distribution detected. Consider using non-stratified k-fold CV.")
    return(NULL)
  }

  levs <- names(ct)
  if (min(ct) <= k || min(ct) / k < 2 || is.na(k) || is.infinite(k) || k < 2) {
    message(sprintf("Reducing k to %d due to small group size (min n = %d).", min(ct), min(ct)))
    k <- min(ct)
  }

  nobs <- floor(min(ct) / k)
  if (nobs == 0) stop("Too few observations per class for stratified CV.", call. = FALSE)

  set_lev <- lapply(levs, function(x, ks = k, y = Y, nob = nobs) {
    idx <- sample(which(y == x))
    k1 <- rep(seq_len(ks), length.out = length(idx))
    names(k1) <- idx
    return(k1)
  })

  idc <- unlist(set_lev)
  sets_list <- lapply(seq_len(k), function(i) {
    idx <- which(idc != i)
    as.numeric(names(idc[idx]))
  })

  sets_list
}



#' @title Generate Monte Carlo Cross-Validation (MCCV) Training Indices
#' @description Generates a list of training set indices for Monte Carlo Cross-Validation.
#' @param k Integer. Number of MCCV iterations (i.e., training sets to generate).
#' @param Y A matrix (n x p), where rows are observations and columns are outcomes. Only the number of rows is used here.
#' @param split Numeric. Fraction of the data to include in each training set. Should be between 0 and 1.
#' @return A list of length \code{k}, each containing a vector of training set indices.
#' @details For each fold, a random sample (without replacement) of size \code{floor(nrow(Y) * split)} is drawn.
#' @keywords internal
.mc <- function(k, Y, split) {

  k <- ceiling(k)

  if (k <= 0 || k > 1e6 || is.na(k) || is.infinite(k)) {
    stop("Invalid value for 'k'. Must be a positive integer and <= 1e6.", call. = FALSE)
  }

  if (split <= 0 || split >= 1 || is.na(split) || is.infinite(split)) {
    stop("Invalid value for 'split'. Must be between 0 and 1 (exclusive).", call. = FALSE)
  }

  if (split > 0.9 || split < 0.3) {
    message("Unusual value for 'split'. Typical values are between 0.5 and 0.9.")
  }

  n <- nrow(Y)
  n_train <- floor(n * split)
  if (n_train < 1) {
    stop("Split results in zero training samples. Increase split or use a larger dataset.", call. = FALSE)
  }

  sets_list <- lapply(seq_len(k), function(i) {
    sample(seq_len(n), size = n_train, replace = FALSE)
  })

  sets_list
}


#' @title Class-balanced Monte Carlo Cross-Validation (MCCV) splits
#' @description Generates class/group-balanced training set indices for each MCCV round.
#' @param k Integer. Number of MCCV training sets to generate.
#' @param split Numeric. Fraction of observations per class to include in each training set. Must be between 0 and 1.
#' @param stratified List of three elements:
#'   \itemize{
#'     \item{1: } Character. Outcome type: 'R' (regression) or 'DA' (discriminant analysis), optionally with suffix '-mY' for multi-column Y.
#'     \item{2: } Matrix. Response matrix \code{Y}, with \code{ncol(Y) == 1}.
#'     \item{3: } Numeric vector of probabilities used to stratify numeric Y if regression.
#'   }
#' @return A list of length \code{k}, each element containing a vector of row indices for the training set.
#' @details Each round samples a class-balanced subset without replacement. Useful for high-variance, imbalanced-class modeling.
#'
#' @importFrom stats quantile
#' @keywords internal
.mcBalanced <- function(k, split, stratified) {

  k <- ceiling(k)

  if (k <= 1 || k > 1e6 || is.na(k) || is.infinite(k)) {
    stop("Invalid 'k' value. Must be between 2 and 1e6.", call. = FALSE)
  }

  if (split <= 0 || split >= 1 || is.na(split) || is.infinite(split)) {
    stop("Split must be between 0 and 1 (exclusive).", call. = FALSE)
  }
  if (split < 0.3 || split > 0.9) {
    message("Unusual value for split. Typical values are between 0.5 and 0.9.")
  }

  type <- stratified[[1]]
  Ymat <- stratified[[2]]
  probs <- stratified[[3]]

  if (grepl("R", type)) {
    breaks <- quantile(Ymat[, 1], probs = probs)
    if (length(unique(breaks)) <= 2) {
      stop("Y has insufficient variability to create stratified bins - switch to unstratified CV.", call. = FALSE)
    }
    Y <- cut(Ymat[, 1], breaks = breaks, include.lowest = TRUE)
  } else {
    Y <- Ymat[, 1]
  }

  tabY <- table(Y)

  if (min(tabY) / max(tabY) < 0.1) {
    stop("Skewed class frequencies. Results will be biased. Consider using non-stratified CV.", call. = FALSE)
  }

  n_obs <- floor(min(tabY) * split)
  if (n_obs == 0) {
    stop("Insufficient observations in minority class. Consider increasing split or aggregating classes.", call. = FALSE)
  }

  sets_list <- lapply(seq_len(k), function(i) {
    idx_k <- unlist(lapply(names(tabY), function(class) {
      sample(which(Y == class), size = n_obs, replace = FALSE)
    }))
    return(idx_k)
  })

  sets_list
}

#' @title Class-balanced resampling (with replacement)
#'
#' @description
#' Generates k training index sets. Total train size per set is floor(N * split),
#' where N is the total number of samples. Each stratum contributes (approximately)
#' equally, sampling WITH replacement within each stratum.
#'
#' @param k Integer. Number of resampling sets.
#' @param split Numeric in (0,1). Fraction of total N used as training size.
#' @param stratified List of length 3:
#'   1) type: "DA" for classification or "R" for regression
#'   2) Y: matrix with one column (n x 1)
#'   3) probs: numeric vector of quantile probs for regression binning (e.g. c(0, .33, .66, 1))
#' @param remainder Character. How to handle n_train %% G remainder:
#'   - "drop": keep perfect balance, total size becomes G * floor(n_train/G)
#'   - "distribute": distribute leftover 1-by-1 to random strata (still near-balanced)
#'
#' @return List of length k. Each element is an integer vector of training indices.
#' @keywords internal
.balanced_bootstrap_mc <- function(k, split, stratified, remainder = c("distribute", "drop")) {

  remainder <- match.arg(remainder)

  k <- ceiling(k)
  if (!is.numeric(k) || is.na(k) || is.infinite(k) || k <= 0 || k > 1e6) {
    stop("Invalid 'k'. Must be a positive integer <= 1e6.", call. = FALSE)
  }

  if (!is.numeric(split) || is.na(split) || is.infinite(split) || split <= 0 || split >= 1) {
    stop("Invalid 'split'. Must be between 0 and 1 (exclusive).", call. = FALSE)
  }

  if (!is.list(stratified) || length(stratified) != 3) {
    stop("'stratified' must be a list of length 3: type, Y, probs.", call. = FALSE)
  }

  type  <- stratified[[1]]
  Ymat  <- as.matrix(stratified[[2]])
  probs <- stratified[[3]]

  if (ncol(Ymat) != 1) stop("Y must be a single-column matrix.", call. = FALSE)
  if (anyNA(Ymat)) stop("Y contains NA values; cannot stratify.", call. = FALSE)

  if (grepl("R", type)) {
    if (!is.numeric(probs) || length(probs) < 2) stop("For regression, 'probs' must be a numeric vector of quantile probs.", call. = FALSE)
    breaks <- unique(stats::quantile(Ymat[, 1], probs = probs, na.rm = TRUE))
    if (length(breaks) <= 2) stop("Insufficient variability in Y to create stratified bins.", call. = FALSE)
    Ygrp <- cut(Ymat[, 1], breaks = breaks, include.lowest = TRUE)
  } else {
    Ygrp <- Ymat[, 1]
  }

  N <- length(Ygrp)
  n_train <- floor(N * split)
  if (n_train < 1) stop("Split results in zero training samples.", call. = FALSE)

  tab <- table(Ygrp)
  groups <- names(tab)
  G <- length(groups)
  if (G < 2) stop("Need at least 2 strata/classes to balance.", call. = FALSE)

  n_per <- floor(n_train / G)
  if (n_per < 1) stop("Training size too small to sample at least 1 per group.", call. = FALSE)

  extra <- n_train - (n_per * G)  # remainder

  sets_list <- lapply(seq_len(k), function(i) {
    alloc <- rep(n_per, G)
    names(alloc) <- groups

    if (extra > 0 && remainder == "distribute") {
      add_to <- sample(groups, size = extra, replace = TRUE)
      alloc[add_to] <- alloc[add_to] + 1
    }

    unlist(lapply(groups, function(g) {
      idx_g <- which(Ygrp == g)
      sample(idx_g, size = alloc[[g]], replace = TRUE)
    }), use.names = FALSE)
  })

  sets_list
}


#' @title Generate Cross-Validation (CV) Training Set Indices
#' @description Generates a list of row indices representing training sets for various CV strategies including stratified and Monte Carlo methods. Handles both regression (R) and discriminant analysis (DA), and multi-column Y.
#'
#' @param Y matrix or data.frame. Outcome matrix. If multi-column, only the first column will be used.
#' @param type character. Type of analysis: "R" for regression or "DA" for discriminant analysis. Use suffix "-mY" to indicate multi-column Y.
#' @param method character. One of: `"k-fold"`, `"k-fold_stratified"`, `"MC"`, `"MC_balanced"`.
#' @param k integer. Number of CV iterations/folds.
#' @param split numeric. For Monte Carlo methods, the fraction of samples to use in the training set (0 < split < 1).
#' @param probs Numeric vector. Probabilities used for stratification of numeric \code{Y} (regression only).
#' @return A list of integer vectors. Each vector contains row indices representing a training set.
#'
#' @details
#' - In `"k-fold_stratified"` and `"MC_balanced"` modes, stratification is applied to preserve class proportions or outcome distribution.
#' - If `type` is regression (`"R"`), quantile-based binning is applied to stratify continuous Y values.
#' - If `type` is discriminant analysis (`"DA"`), class labels are used directly.
#' - If `Y` has multiple columns, only the first column is used to create resampling sets.

#' @keywords internal
.cvSetsMethod <- function(Y, type, method = "k-fold_stratified", k = 7, split = 2/3, probs=NULL) {

  valid_methods <- c("k-fold", "k-fold_stratified", "MC", "MC_balanced", "Boot_balanced")
  if (!method %in% valid_methods) {
    stop("Check method argument. Valid methods are: ", paste(valid_methods, collapse = ", "), call. = FALSE)
  }

  Y <- as.matrix(Y)
  if (ncol(Y) > 1) {
    Y <- Y[, 1, drop = FALSE]
  }

  if (k %% 1 == 0) {
    k <- as.integer(k)
  }

  if (grepl("MC", method) && (split <= 0 || split >= 1)) {
    stop("Parameter 'split' must be between 0 and 1 (exclusive).", call. = FALSE)
  }

  if (k < 2) stop("Number of folds or iterations (k) must be at least 2.", call. = FALSE)


  sets_list <- switch(method,
                      `k-fold_stratified` = .kFoldStratified(k, stratified = list(type, Y, probs=probs)),
                      `k-fold` = .kFold(k, Y),
                      MC = .mc(k, Y, split),
                      MC_balanced = .mcBalanced(k, split, stratified = list(type, Y, probs = probs)),
                      Boot_balanced = .balanced_bootstrap_mc(k, split, stratified = list(type, Y, probs = probs), remainder = "distribute")
  )

  if (is.null(sets_list)) {
    stop("Cross-validation set generation failed. Check input arguments.", call. = FALSE)
  }

  sets_list
}


#' K-fold cross-validation strategy
#' @param k Integer number of folds.
#' @details
#' Partitions the data into \code{k} folds. Each fold is used once as a test set,
#' with the remaining folds used for training.
#' No stratification is applied; folds are created by random partitioning.
#' @return A named \code{list} with elements:
#' \describe{
#'   \item{train}{List of integer vectors containing training set indices
#'   for each resampling iteration.}
#'   \item{strategy}{Character string indicating the resampling strategy.}
#'   \item{n}{Integer. Number of samples in the dataset.}
#'   \item{seed}{Integer. Random seed used to generate the resampling splits,
#'   ensuring reproducibility.}
#' }
#' @family resampling strategies
#' @examples
#' n <- 100
#' thr <- 1.5
#' Y <- c(rnorm(80, thr - 3, 0.3), rnorm(20, thr + 3, 0.3))  # unbalanced outcome
#' mean(Y > thr)
#' cv_k  <- kfold(k = 10)
#' k_inst <- metabom8:::.arg_check_cv(cv_pars=cv_k, model_type='R', n=n, Y_prepped=cbind(Y))
#' sapply(k_inst$train, function(i) length(i))
#' @export
kfold <- function(k){

  k <- as.integer(k)
  if (length(k) != 1L || is.na(k) || k < 2L)
    stop("k must be an integer >= 2.", call. = FALSE)

  list(method = 'kfold', k = k)
}


#' Y-stratified k-fold cross-validation strategy
#'
#' @param k Integer. Number of folds.
#' @param type Character. Either \code{"DA"} (classification)
#'   or \code{"R"} (regression).
#' @param probs Numeric vector of quantile probabilities used
#'   to stratify continuous \code{Y} when \code{type = "R"}.
#'
#' @details
#' For classification (\code{type = "DA"}), folds are generated such that
#' class proportions are approximately preserved in each fold.
#'
#' For regression (\code{type = "R"}), \code{Y} is discretised into bins
#' defined by \code{probs}, and folds are generated to approximately
#' preserve the bin distribution.
#'
#' @return A named \code{list} with elements:
#' \describe{
#'   \item{train}{List of integer vectors containing training set indices
#'   for each resampling iteration.}
#'   \item{strategy}{Character string indicating the resampling strategy.}
#'   \item{n}{Integer. Number of samples in the dataset.}
#'   \item{seed}{Integer. Random seed used to generate the resampling splits,
#'   ensuring reproducibility.}
#' }
#' @family resampling strategies
#' @examples
#' set.seed(1)
#' n <- 100
#' thr <- 1.5
#' Y <- c(rnorm(80, thr - 3, 0.3), rnorm(20, thr + 3, 0.3))  # unbalanced outcome
#' mean(Y > thr)
#'
#' q80 <- quantile(Y, 0.8)  # defines the rare "high" stratum (top 20%)
#'
#' cv_k  <- kfold(k = 10)
#' cv_sk  <- stratified_kfold(k = 10, type = "R", probs = c(0, 0.8, 1))
#'
#' k_inst <-  metabom8:::.arg_check_cv(cv_pars=cv_k, model_type='R', n=n, Y_prepped=cbind(Y))
#' sk_inst <-  metabom8:::.arg_check_cv(cv_pars=cv_sk, model_type='R', n=n, Y_prepped=cbind(Y))
#'
#' round(sapply(k_inst$train,  function(i) mean(Y[i] > q80)), 2)  # reflects imbalance
#' round(sapply(sk_inst$train, function(i) mean(Y[i] > q80)), 2)  # balanced across strata
#' @export
stratified_kfold <- function(k,
                             type = c("DA", "R"),
                             probs = NULL
                             ) {

  type <- match.arg(type)

  if (type == "R") {
    if (is.null(probs)) {
      stop("'probs' must be provided when type = 'R'", call. = FALSE)
    }
  } else {
    if (!is.null(probs)) {
      warning("'probs' ignored when type = 'DA'", call. = FALSE)
    }
  }

  k <- as.integer(k)
  if (is.na(k) || k < 2L)
    stop("k must be an integer >= 2.", call. = FALSE)

  list(method = 'StratifiedKFold', k = k, type  = type, probs = probs)

}

#' Monte-Carlo cross-validation strategy
#'
#' @param k Integer. Number of repeated random splits.
#' @param split Numeric. Fraction of samples assigned to the
#'   training set (e.g. \code{2/3}).
#'
#' @details
#' Monte-Carlo cross-validation generates \code{k} random train/test splits
#' without replacement. No stratification is applied; samples are drawn
#' uniformly at random.
#'
#' @return A named \code{list} with elements:
#' \describe{
#'   \item{train}{List of integer vectors containing training set indices
#'   for each resampling iteration.}
#'   \item{strategy}{Character string indicating the resampling strategy.}
#'   \item{n}{Integer. Number of samples in the dataset.}
#'   \item{seed}{Integer. Random seed used to generate the resampling splits,
#'   ensuring reproducibility.}
#' }
#' @family resampling strategies
#' @examples
#' n <- 100
#' # bivariate outcome
#' thr <- 1.5
#' Y <- c(rnorm(80, thr-3, 0.3), rnorm(20, thr+3, 0.3))  # unbalanced low/high outcome
#' mean(Y>thr)
#'
#' cv_mc <- mc(k = 10, split = 2/3)
#' mc_inst <- metabom8:::.arg_check_cv(cv_pars=cv_mc, model_type='R', n=n, Y_prepped=cbind(Y))
#' sapply(mc_inst$train, function(i) length(i))
#' @export
mc <- function(k, split) {

  k <- as.integer(k)
  if (is.na(k) || k < 1L)
    stop("k must be an integer >= 1.", call. = FALSE)

  if (!is.numeric(split) || length(split) != 1L)
    stop("split must be a single numeric value.", call. = FALSE)

  if (split <= 0 || split >= 1)
    stop("split must be strictly between 0 and 1.", call. = FALSE)

  list(method = "MonteCarlo", k = k, split = split)
}

#' Balanced Monte-Carlo resampling strategy
#'
#' @param k Integer. Number of repeated random splits.
#' @param split Numeric. Fraction of samples assigned to the training set
#'   (e.g. \code{2/3}).
#' @param type Character. Either \code{"DA"} (classification) or \code{"R"} (regression).
#' @param probs Numeric vector of quantile probabilities used to stratify
#'   continuous \code{Y} when \code{type = "R"}.
#'
#' @details
#' Generates \code{k} Monte-Carlo resampling splits by randomly partitioning
#' the data into training and test sets without replacement.
#'
#' Balancing ensures equal representation of strata in the training data:
#' \describe{
#'   \item{\code{type = "DA"}}{
#'     Class labels define the strata, and sampling is balanced across classes.
#'   }
#'   \item{\code{type = "R"}}{
#'     The response is discretised into bins using quantiles defined by
#'     \code{probs}, and each bin contributes equally to the training set.
#'   }
#' }
#'
#' This strategy can improve robustness of model evaluation in settings with
#' limited samples size and imbalanced or unevenly distributed outcome variables.
#'
#' @return A named \code{list} with elements:
#' \describe{
#'   \item{train}{List of integer vectors containing training set indices
#'   for each resampling iteration.}
#'   \item{strategy}{Character string indicating the resampling strategy.}
#'   \item{n}{Integer. Number of samples in the dataset.}
#'   \item{seed}{Integer. Random seed used to generate the resampling splits,
#'   ensuring reproducibility.}
#' }
#' @family resampling strategies
#' @examples
#' n <- 100
#' # bivariate outcome
#' thr <- 1.5
#' Y <- c(rnorm(80, thr-3, 0.3), rnorm(20, thr+3, 0.3))  # unbalanced low/high outcome
#' mean(Y>thr)
#'
#' cv_k <- kfold(k = 10)
#' cv_mc <- balanced_mc(k = 10, split = 2/3, type = "R", probs = c(0, 0.8, 1))
#'
#' k_inst <- metabom8:::.arg_check_cv(cv_pars=cv_k, model_type='R', n=n, Y_prepped=cbind(Y))
#' mc_inst <- metabom8:::.arg_check_cv(cv_pars=cv_mc, model_type='R', n=n, Y_prepped=cbind(Y))
#'
#' # balanced splits: proportion above global median stays ~0.5
#' q80 <- quantile(Y, 0.8)
#' round(sapply(k_inst$train, function(i) mean(Y[i] > q80)), 2) # resembles original Y distr.
#' round(sapply(mc_inst$train, function(i) mean(Y[i] > q80)), 2) # balanced strata (low/high)
#' @export
balanced_mc <- function(k,
                        split,
                        type = c("DA", "R"),
                        probs = NULL
                        ) {

  type <- match.arg(type)

  if (type == "R") {
    if (is.null(probs)) {
      stop("'probs' must be provided when type = 'R'", call. = FALSE)
    }
  } else {
    if (!is.null(probs)) {
      warning("'probs' ignored when type = 'DA'", call. = FALSE)
    }
  }

  k <- as.integer(k)
  if (is.na(k) || k < 1L)
    stop("k must be an integer >= 1.", call. = FALSE)

  if (!is.numeric(split) || length(split) != 1L)
    stop("split must be a single numeric value.", call. = FALSE)

  if (split <= 0 || split >= 1)
    stop("split must be strictly between 0 and 1.", call. = FALSE)

  list(method = "BalancedMonteCarlo", k = k, split = split,
       type  = type,  probs = probs)

}


#' Balanced bootstrap resampling strategy
#'
#' @param k Integer. Number of bootstrap resamples.
#' @param split Numeric. Fraction of samples drawn for the training set
#'   (e.g. \code{2/3}). Sampling is performed with replacement.
#' @param type Character. Either \code{"DA"} (classification) or \code{"R"} (regression).
#' @param probs Numeric vector of quantile probabilities used to stratify continuous
#'   \code{Y} when \code{type = "R"}.
#'
#' @details
#' Generates \code{k} bootstrap samples (training sets sampled with replacement).
#' The remaining samples (the out-of-bag set) can be used as a test set.
#'
#' Balancing ensures equal representation of strata in the training data:
#' \describe{
#'   \item{\code{type = "DA"}}{
#'     Class labels define the strata, and sampling is balanced across classes.
#'   }
#'   \item{\code{type = "R"}}{
#'     The response is discretised into bins using quantiles defined by
#'     \code{probs}, and each bin contributes equally to the training set.
#'   }
#' }
#'
#' @return A named \code{list} with elements:
#' \describe{
#'   \item{train}{List of integer vectors containing training set indices
#'   for each resampling iteration.}
#'   \item{strategy}{Character string indicating the resampling strategy.}
#'   \item{n}{Integer. Number of samples in the dataset.}
#'   \item{seed}{Integer. Random seed used to generate the resampling splits,
#'   ensuring reproducibility.}
#' }
#' @family resampling strategies
#' @examples
#' n <- 100
#' # bivariate outcome
#' thr <- 1.5
#' Y <- c(rnorm(80, thr-3, 0.3), rnorm(20, thr+3, 0.3))  # unbalanced low/high outcome
#' mean(Y>thr)
#'
#' cv_k <- kfold(k = 10)
#' cv_boot <- balanced_boot(k = 10, split = 2/3, type = "R", probs = c(0, 0.8, 1))
#'
#' k_inst <- metabom8:::.arg_check_cv(cv_pars=cv_k, model_type='R', n=n, Y_prepped=cbind(Y))
#' b_inst <- metabom8:::.arg_check_cv(cv_pars=cv_boot, model_type='R', n=n, Y_prepped=cbind(Y))
#'
#' # balanced splits: proportion above global median stays ~0.5
#' q80 <- quantile(Y, 0.8)
#' round(sapply(k_inst$train, function(i) mean(Y[i] > q80)), 2) # resembles original Y distr.
#' round(sapply(b_inst$train, function(i) mean(Y[i] > q80)), 2) # balanced strata (low/high)
#' @export
balanced_boot <- function(k,
                          split,
                          type = c("DA", "R"),
                          probs = NULL
                          ) {
  type <- match.arg(type)

  if (type == "R") {
    if (is.null(probs)) {
      stop("'probs' must be provided when type = 'R'", call. = FALSE)
    }
  } else {
    if (!is.null(probs)) {
      warning("'probs' ignored when type = 'DA'", call. = FALSE)
    }
  }

  k <- as.integer(k)
  if (is.na(k) || k < 1L)
    stop("k must be an integer >= 1.", call. = FALSE)

  if (!is.numeric(split) || length(split) != 1L)
    stop("split must be a single numeric value.", call. = FALSE)

  if (split <= 0 || split >= 1)
    stop("split must be strictly between 0 and 1.", call. = FALSE)

  list(method = "BalancedBoot", k = k, split = split, type=type, probs=probs)
}

set.seed(15)

###constructors

test_that("kfold constructor returns valid object", {

  cv <- kfold(k=5)
  expect_type(cv, "list")
  expect_equal(cv$k, 5)

})

test_that("kfold rejects invalid k", {

  expect_error(kfold(1))
  expect_error(kfold(NA))
})

test_that("stratified_kfold constructor works", {

  cv <- stratified_kfold(k = 5, type = "R", probs = c(0,0.5,1))

  expect_type(cv, "list")

  expect_equal(cv$k, 5)
  expect_equal(cv$type, "R")
  expect_equal(cv$probs, c(0,0.5,1))
})

test_that("mc constructor works", {

  cv <- mc(k = 10, split = 0.7)

  expect_type(cv, "list")

  expect_equal(cv$k, 10)
  expect_equal(cv$split, 0.7)

})

test_that("mc validates split", {

  expect_error(mc(10, -0.2))
  expect_error(mc(10, 1))
  expect_error(mc(10, 0))

})

test_that("balanced_mc constructor works", {

  cv <- balanced_mc(k = 5, split = 0.7, type = "DA")

  expect_type(cv, "list")

  expect_equal(cv$k, 5)
  expect_equal(cv$split, 0.7)
  expect_equal(cv$type, "DA")

})


### instantiation

test_that("kfold instantiates correct folds", {

  n = 100
  Y <- matrix(rnorm(n), ncol = 1)

  cv <- kfold(5)
  inst <- .arg_check_cv(cv, model_type='R', n=n, Y_prepped = Y)

  folds <- inst$train

  expect_length(folds, 5)
  expect_true(all(sapply(folds, is.integer)))
  expect_true(all(sapply(folds, function(i) all(i <= nrow(Y)))))
  expect_true(all(sapply(folds, length) > 0))
  expect_true(all(unlist(folds) >= 1))
  expect_true(all(unlist(folds) <= nrow(Y)))
  expect_true(length(unique(unlist(folds))) == nrow(Y))
})


test_that("stratified_kfold preserves class balance", {

  n = 100
  Y <- matrix(sample(c(0,1), n, replace = TRUE, prob = c(0.8,0.2)), ncol=1)

  cv <- stratified_kfold(5, type="DA")

  inst <- .arg_check_cv(cv, model_type='DA', n=n, Y_prepped = Y)
  folds <- inst$train

  props <- sapply(folds, function(i) mean(Y[i] == 1))

  expect_true(sd(props) < 0.1)
  expect_true(all(unlist(folds) >= 1))
  expect_true(all(unlist(folds) <= nrow(Y)))
  expect_true(all(props > 0))

})


test_that("mc generates correct training sizes", {

  n = 100
  Y <- matrix(rnorm(n), ncol = 1)
  cv <- mc(10, split = 0.7)

  inst <- .arg_check_cv(cv, model_type='R', n=n, Y_prepped = Y)
  folds <- inst$train

  sizes <- sapply(folds, length)

  expect_true(all(abs(sizes - 70) <= 2))
  expect_true(all(unlist(folds) >= 1))
  expect_true(all(sizes == floor(n * 0.7)))
})


test_that("balanced_mc returns correct number of splits", {

  Y <- c(rep(0,80), rep(1,20))
  Y <- matrix(Y, ncol = 1)

  strat <- list("DA", Y, NULL)

  sets <- .mcBalanced(k = 10, split = 0.7, stratified = strat)

  expect_length(sets, 10)
  expect_true(all(unlist(sets) >= 1))
  expect_true(all(unlist(sets) <= nrow(Y)))

})

test_that("balanced_mc produces balanced class samples", {

  Y <- c(rep(0,80), rep(1,20))
  Y <- matrix(Y, ncol = 1)

  strat <- list("DA", Y, NULL)

  sets <- .mcBalanced(10, 0.7, strat)

  props <- sapply(sets, function(i) mean(Y[i] == 1))

  expect_true(sd(props) < 0.05)

})

test_that("balanced_mc samples without replacement", {

  Y <- c(rep(0,80), rep(1,20))
  Y <- matrix(Y, ncol = 1)

  strat <- list("DA", Y, NULL)

  sets <- .mcBalanced(5, 0.7, strat)

  expect_false(any(sapply(sets, function(i) any(duplicated(i)))))
})

test_that("balanced_mc detects extreme imbalance", {

  Y <- c(rep(0,95), rep(1,5))
  Y <- matrix(Y, ncol = 1)

  strat <- list("DA", Y, NULL)

  expect_error(.mcBalanced(10, 0.7, strat))

})

test_that("bootstrap sampling contains duplicates", {

  n = 100
  Y <- matrix(rnorm(n), ncol = 1)

  cv <- balanced_boot(5, split = 0.7, type = 'R', probs=c(0, 0.5, 1))

  inst <- .arg_check_cv(cv, model_type='R', n=n, Y_prepped = Y)
  folds <- inst$train

  expect_true(any(sapply(folds, function(i) any(duplicated(i)))))
  expect_true(all(unlist(folds) >= 1))
  expect_true(all(unlist(folds) <= nrow(Y)))
})

# --- vips -----------------------------------------------------------

#' @rdname vip
setMethod("vip", "m8_model",
          function(object){

            if (!object@engine %in% c("opls", "pls"))
              stop("Method not available for this model", call. = FALSE)

            if (object@engine == "opls") {

              if(object@ctrl$ncomp_selected == 0)
                stop('No model components found', call. = FALSE)

              W <- t(as.matrix(object@fit$pred_comp$w_x))
              Tx <- as.matrix(object@fit$pred_comp$t_x)
              C <- object@fit$pred_comp$p_y

            } else if (object@engine == "pls") {

              W <- do.call(rbind, lapply(object@fit$comp, `[[`, "w_x"))
              Tx <- do.call(cbind, lapply(object@fit$comp, `[[`, "t_x"))
              C <- do.call(cbind, lapply(object@fit$comp, `[[`, "p_y"))

            }

            vip <- .ensure_matrix(.vip(Tx, W, C), ncol=ncol(W))

            return(vip)
          }
)

# --- yfitted -----------------------------------------------------------

#' @rdname fitted
setMethod('fitted', "m8_model",
          function(object){
            if(object@engine == "opls"){

              if(object@ctrl$ncomp_selected == 0)
                stop('No model components found', call. = FALSE)

              return(object@fit$pred_comp$y_pred)
            }
          }
          )


# --- weights -----------------------------------------------------------

#' @rdname weights
#' @param orth Logical indicating whether orthogonal scores should be returned
#'   (only applicable for OPLS models).
setMethod("weights","m8_model",
          function(object, orth = FALSE){


            if (!object@engine %in% c('opls', 'pls'))
              stop("Method not available for this model", call. = FALSE)


            if(isTRUE(orth) && object@engine != "opls")
              stop("Orthogonal weights not available for this model", call. = FALSE)


            key <- paste(if(isTRUE(orth) || is.numeric(orth)) "orth" else "pred",
                         "fit",
                         sep="_")

            if(object@engine == "opls"){

              if(object@ctrl$ncomp_selected == 0)
                stop('No model components found', call. = FALSE)

              map <- list(
                pred_fit = .ensure_matrix(object@fit$pred_comp$w_x, ncol=object@dims$p),
                orth_fit = object@fit$w_orth
              )} else if(object@engine == "pls"){
                map <- list(
                  pred_fit = .ensure_matrix(do.call(rbind, lapply(object@fit$comp, `[[`, "w_x")), ncol=object@dims$p)
                )
              }

            if(!key %in% names(map))
              stop("Requested weight type not available.", call. = FALSE)

            out <- map[[key]]

            if(is.numeric(orth)){

              if(orth > nrow(out))
                stop("Orthogonal component index exceeds available components", call. = FALSE)

              out <- out[orth, ,drop = FALSE]
            }


            return(out)
          }
    )


# --- loadings -----------------------------------------------------------

#' Model loadings
#' @param x An object of class \code{m8_model}.
#' @param orth Logical indicating whether orthogonal scores should be returned
#'   (only applicable for OPLS models).
#' @param ... Additional arguments (currently ignored).
#' @return Numeric vector or matrix containing loadings.
#' @examples
#' data(covid)
#' cv <- balanced_mc(k=5, split=2/3)
#' scaling <- uv_scaling(center=TRUE)
#' model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
#' show(model)
#' P <- loadings(model)
#' Po <- loadings(model, orth = TRUE)
#' dim(P)
#' dim(Po) == dim(P)
#' @export
setMethod("loadings","m8_model",
          function(x, orth = FALSE, ...){

            object <- x

            if (!object@engine %in% c('opls', 'pls', 'pca'))
              stop("Object needs to be of class m8_model.", call. = FALSE)

            if(isTRUE(orth) && object@engine != "opls")
              stop("Orthogonal loadings not available for this model", call. = FALSE)

            key <- paste(if(isTRUE(orth) || is.numeric(orth)) "orth" else "pred",
                         "fit",
                         sep="_")

            if(object@engine == "opls"){

              if(object@ctrl$ncomp_selected == 0)
                stop('No model components found', call. = FALSE)

              map <- list(
                pred_fit = .ensure_matrix(object@fit$pred_comp$p_x, ncol=length(object@fit$pred_comp$p_x)),
                orth_fit = object@fit$p_orth
              )} else if(object@engine == "pls"){
                map <- list(
                  pred_fit = do.call(rbind, lapply(object@fit$comp, `[[`, "p_x"))
                )
              }else if(object@engine == "pca"){
                map <- list(
                  pred_fit = object@fit$p
                )
              }

              out <- map[[key]]

              if(is.numeric(orth)){

                if(orth > nrow(out))
                  stop("Orthogonal component index exceeds available components", call. = FALSE)

                out <- out[orth, ,drop = FALSE]
              }


            return(out)


          })


# --- scores -----------------------------------------------------------

#' @rdname scores
setMethod("scores","m8_model",
          function(object, orth = FALSE, cv=FALSE){

            if (!object@engine %in% c('opls', 'pls', 'pca'))
              stop("Method not available for this model", call. = FALSE)


            needs_orth <- isTRUE(orth) || is.numeric(orth)

            if(needs_orth && object@engine != "opls")
              stop("Orthogonal scores not available for this model", call. = FALSE)

            if(needs_orth && isTRUE(cv) && object@engine == "opls")
              stop("Cross-validated orthogonal scores not available for this model", call. = FALSE)

            if(isTRUE(cv) && object@engine %in% c("pls", "pca"))
              stop("Cross-validated scores not available for this model", call. = FALSE)

            key <- paste(if(isTRUE(orth) || is.numeric(orth)) "orth" else "pred",
                         if(cv)"cv"   else "fit",
                         sep="_")

            if(object@engine == "opls"){

              if(object@ctrl$ncomp_selected == 0)
                stop('No model components found', call. = FALSE)

              map <- list(
                pred_fit = .ensure_matrix(object@fit$pred_comp$t_x, ncol=1),
                orth_fit = object@fit$t_orth,
                pred_cv  = object@fit$pred_comp$t_cv,
                orth_cv  = object@fit$pred_comp$t_o_cv
              )} else if(object@engine == "pls"){
                map <- list(
                  pred_fit = do.call(cbind, lapply(object@fit$comp, `[[`, "t_x"))
                  # pred_cv  = object@fit$pred_comp$t_cv
                )
              }else if(object@engine == "pca"){
                map <- list(
                  pred_fit = object@fit$t
                )
              }


              out <- map[[key]]

              if(is.numeric(orth)){

                if(object@engine != "opls")
                  stop("Orthogonal components only available for OPLS", call. = FALSE)

                if(orth > ncol(out))
                  stop("Orthogonal component index exceeds available components", call. = FALSE)

                out <- out[, orth, drop = FALSE]
              }


           return(out)
          })


# --- xres -----------------------------------------------------------

#' @rdname xres
setMethod("xres", "m8_model",
          function(object) {

            if (!identical(object@engine, "opls")) {
              stop("xres() is only available for OPLS models (engine = 'opls').", call. = FALSE)
            }

            if(object@ctrl$ncomp_selected == 0)
              stop('No model components found', call. = FALSE)

            Xcs <- object@fit$X_prepped
            if (is.null(Xcs)) stop("Missing fit$X_prepped.", call. = FALSE)

            To <- object@fit$t_orth
            Po <- object@fit$p_orth
            Tp <- .ensure_matrix(object@fit$pred_comp$t_x, ncol=1)
            Pp <- .ensure_matrix(object@fit$pred_comp$p_x, ncol=ncol(Po))

            if (is.null(To) || is.null(Po)) stop("Missing orthogonal components fit$t_o and/or fit$p_o.", call. = FALSE)
            if (is.null(Tp) || is.null(Pp)) stop("Missing predictive components fit$t_p and/or fit$p_p.", call. = FALSE)

            Xhat_o <- To %*% Po
            Xhat_p <- Tp %*% Pp

            Xcs - Xhat_p - Xhat_o
          }
)


# --- summary -----------------------------------------------------------

#' @describeIn m8_model-class Summarise model performance and component selection.
#' @param object An object of class \code{m8_model}.
#' @export
setMethod("summary", "m8_model",
          function(object) {

            eng  <- object@engine
            ctrl <- object@ctrl

            if (eng == "pls") {

              n_sel <- as.integer(ctrl$ncomp_selected)
              n_tst <- as.integer(ctrl$ncomp_tested)

              if (is.na(n_sel) || length(n_sel) == 0L) n_sel <- as.integer(ctrl$nc)
              if (is.na(n_tst) || length(n_tst) == 0L) n_tst <- length(ctrl$q2)

              n_sel <- max(0L, n_sel)
              n_tst <- max(n_sel, n_tst, 0L)

              df <- data.frame(comp = seq_len(n_tst))

              if (identical(ctrl$type, "R")) {
                df$R2 <- ctrl$r2[seq_len(min(length(ctrl$r2), n_tst))]
                df$Q2 <- ctrl$q2[seq_len(min(length(ctrl$q2), n_tst))]
              } else {
                df$AUCt <- ctrl$aucs_tr[seq_len(min(length(ctrl$aucs_tr), n_tst))]
                df$AUCv <- ctrl$aucs_te[seq_len(min(length(ctrl$aucs_te), n_tst))]
              }

              df$R2X <- NA_real_
              if (!is.null(ctrl$r2x_comp) && length(ctrl$r2x_comp) > 0L) {
                idx <- seq_len(min(n_tst, length(ctrl$r2x_comp)))
                df$R2X[idx] <- ctrl$r2x_comp[seq_len(length(idx))]
              }

              df$selected <- df$comp <= n_sel

              df$fit <- ifelse(df$selected, "fitted", "not fitted")

              df$stop_reason <- NA_character_
              df$stop_reason[n_tst] <- ctrl$stop_reason
              df$stop_metric <- NA_real_
              df$stop_metric[n_tst] <- ctrl$stop_metric
              df$stop_delta <- NA_real_
              df$stop_delta[n_tst] <- ctrl$stop_delta


              return(list(
                         perf   = df,
                         engine = eng,
                         y_type = ctrl$type
              ))
            }

            if (eng == "opls") {

              n_sel <- as.integer(ctrl$ncomp_selected)-1
              n_tst <- as.integer(ctrl$ncomp_tested)-1

              if (is.na(n_tst) || length(n_tst) == 0L) n_tst <- length(ctrl$q2)
              if (is.na(n_sel) || length(n_sel) == 0L) n_sel <- n_tst

              n_sel <- max(0L, n_sel)
              n_tst <- max(n_sel, n_tst, 0L)

              df <- data.frame(
                comp_total = seq_len(n_tst)+1,
                comp_pred  = 1,
                comp_orth  = seq_len(n_tst)
              )

              if (identical(ctrl$type, "R")) {
                df$R2 <- ctrl$r2[seq_len(min(length(ctrl$r2), n_tst))]
                df$Q2 <- ctrl$q2[seq_len(min(length(ctrl$q2), n_tst))]
              } else {
                df$AUCt <- ctrl$aucs_tr[seq_len(min(length(ctrl$aucs_tr), n_tst))]
                df$AUCv <- ctrl$aucs_te[seq_len(min(length(ctrl$aucs_te), n_tst))]
              }

              df$R2X <- NA_real_
              if (!is.null(ctrl$r2x_comp) && length(ctrl$r2x_comp) > 0L) {
                idx <- seq_len(min(n_tst, length(ctrl$r2x_comp)))
                df$R2X[idx] <- ctrl$r2x_comp[seq_len(length(idx))]
              }

              df$selected <- df$comp_total <= n_sel+1
              df$fit <- ifelse(df$selected, "fitted", "not fitted")

              df$stop_reason <- NA_character_
              df$stop_metric <- NA_real_
              df$stop_delta  <- NA_real_

              if (!is.null(ctrl$stop_reason) && n_tst >= 1L) {
                df$stop_reason[n_tst] <- ctrl$stop_reason
                df$stop_metric[n_tst] <- ctrl$stop_metric
                df$stop_delta[n_tst]  <- ctrl$stop_delta
              }

              return(list(
                         perf   = df,
                         engine = eng,
                         y_type = ctrl$type
              ))
            }

            if (eng == "pca") {

              n_sel <- as.integer(ctrl$ncomp_selected)
              df <- data.frame(comp = seq_len(n_sel))

              expl <- ctrl$r2x_comp
              if (is.null(expl)) {
                stop("PCA summary requires ctrl$r2x_comp (proportion variance explained per component).", call. = FALSE)
              }
              df$R2X <- expl
              df$R2X_cum <- cumsum(as.numeric(df$R2X))
              df$selected <- df$comp <= n_sel
              df$fit <- ifelse(df$selected, "fitted", "not fitted")

              alg <- object@ctrl$method

              return(list(
                         perf = df,
                         engine = paste0(eng, " (", alg, ")"),
                         y_type = "unsupervised"
              ))
            }

            list(
                perf   = data.frame(),
                engine = eng,
                y_type = if (!is.null(ctrl$type)) ctrl$type else NA_character_
            )
          }
)


# --- show -----------------------------------------------------------

#' @describeIn m8_model-class Show a compact model header.
#' @param object An object of class \code{m8_model}.
#' @export
setMethod("show", "m8_model",
          function(object) {

            cat("\n")
            cat("m8_model <", object@engine, ">\n", sep = "")
            cat(strrep("-", 40), "\n", sep = "")

            if (!is.null(object@dims)) {
              cat("Dimensions  :", object@dims$n, "samples x",
                  object@dims$p, "variables\n")
            }

            y_type <- tryCatch(object@ctrl$type, error = function(e) NULL)
            if (!is.null(y_type)) {
              if (object@engine == "pca") {
                cat("Mode        : unsupervised\n")
              } else {
                cat("Mode        :", if (y_type == "R") "regression" else "classification", "\n")
              }
            }

            prep <- tryCatch(object@prep, error = function(e) NULL)
            if (!is.null(prep)) {
              cat("Preprocess  :",
                  ifelse(prep$center, "center", "no-center"),
                  "|",
                  prep$scale, "\n")
            }

            ctrl <- object@ctrl

            n_selected <- tryCatch(ctrl$ncomp_selected, error = function(e) NULL)
            n_tested   <- tryCatch(ctrl$ncomp_tested,   error = function(e) NULL)
            if (is.null(n_selected)) {
              n_selected <- tryCatch(ctrl$nc, error = function(e) NULL)
            }

            if (!is.null(n_selected)) {
              cat("Components  :", n_selected)
              if (!is.null(n_tested) && n_tested > n_selected) {
                cat(" (", n_tested, " tested)", sep = "")
              }
              cat("\n")
            }

            if (!is.null(object@cv)) {
              cv_method <- tryCatch(object@cv$strategy, error = function(e) NULL)
              cv_k <- tryCatch(length(object@cv$train), error = function(e) NULL)
              if (!is.null(cv_method)) {
                line <- paste0("Validation  : ", cv_method)

                if (!is.null(cv_k)) {
                  line <- paste0(line, " (k = ", cv_k, ")")
                }

                cat(line, "\n")
              }
            }

            stop_reason <- tryCatch(ctrl$stop_reason, error = function(e) NULL)
            if (!is.null(stop_reason)) {
              cat("Stop rule   :", stop_reason, "\n")
            }

            cat(strrep("-", 40), "\n", sep = "")
            cat("Use summary() for performance metrics.\n\n")
          }
)

#' Statistical Total Correlation Spectroscopy (STOCSY)
#'
#' Performs STOCSY analysis on 1D NMR spectra. Computes correlation and covariance
#' of all variables in \code{X} against a driver signal (internal ppm position or
#' an external numeric vector).
#'
#' @param X Numeric matrix or data frame. Rows are spectra, columns are variables.
#' @param ppm Optional numeric vector of chemical shift values (length = ncol(X)).
#'   If missing or NULL, will try to read from \code{colnames(X)}.
#' @param driver Numeric scalar (ppm value, internal driver) or numeric vector
#'   (external driver, length = nrow(X)).
#' @param plotting Logical. If TRUE, plot the STOCSY spectrum.
#' @param title Optional character plot title.
#'
#' @return An object of class \code{m8_stocsy1d} (S3), a named list with entries:
#'   \code{version, X, ppm, driver, r, cov, extD}.
#'
#' @examples
#' # st <- stocsy(X, ppm, driver = 5.233, plotting = FALSE)
#' # plotStocsy(st, shift = c(5.15, 5.30))
#'
#' @family structural_annotation
#' @importFrom ggplot2 ggplot geom_line geom_vline scale_x_reverse
#'   scale_colour_gradientn labs theme_bw theme element_text
#' @importFrom colorRamps matlab.like2
#' @importFrom scales breaks_pretty
#' @examples
#' data(covid)
#' cs = 5.233
#' stocsy(covid$X, driver=cs, plotting = TRUE)
#' @export
stocsy <- function(X, ppm = NULL, driver, plotting = TRUE, title = NULL) {

  if (!is.matrix(X) && !is.data.frame(X)) {
    stop("X must be a numeric matrix or data frame.", call. = FALSE)
  }
  X <- as.matrix(X)
  if (!is.numeric(X)) stop("X must be numeric.", call. = FALSE)

  if (missing(ppm) || is.null(ppm)) {
    ppm_try <- as.numeric(colnames(X))
    if (is.null(ppm_try) || all(is.na(ppm_try))) {
      stop("Provide 'ppm' or ensure colnames(X) are numeric ppm values.", call. = FALSE)
    }
    ppm <- ppm_try
  }
  if (!is.numeric(ppm) || length(ppm) != ncol(X)) {
    stop("'ppm' must be numeric and have length ncol(X).", call. = FALSE)
  }

  if (!is.numeric(driver)) stop("'driver' must be numeric.", call. = FALSE)
  extD <- length(driver) > 1L

  if (extD) {
    if (length(driver) != nrow(X)) stop("External driver must have length nrow(X).", call. = FALSE)
    idx <- which(!is.na(driver))
    if (length(idx) < 2L) stop("External driver has <2 non-NA entries.", call. = FALSE)
    driver_use <- driver[idx]
    X_use <- X[idx, , drop = FALSE]
  } else {
    if (driver < min(ppm) || driver > max(ppm)) {
      stop("Internal STOCSY driver must lie within the ppm range.", call. = FALSE)
    }
    driver_use <- driver
    X_use <- X
  }

  l <- if (extD) {
    as.numeric(driver_use)
  } else {
    X_use[, which.min(abs(ppm - driver_use))]
  }

  cc <- apply(X_use, 2, function(x) cor(x, l, use = "pairwise.complete.obs"))
  cv <- apply(X_use, 2, function(x) cov(x, l, use = "pairwise.complete.obs"))

  stoc_mod <- list(
    # version = "0.9",
    X       = X_use,
    ppm     = as.numeric(ppm),
    driver  = driver_use,
    r       = as.numeric(cc),
    cov     = as.numeric(cv),
    extD    = extD
  )

  if (isTRUE(plotting)) {
    print(plotStocsy(stoc_mod, shift = range(ppm, na.rm = TRUE), title = title))
  }

  stoc_mod
}


#' Plot STOCSY result
#'
#' @description
#' Generates a STOCSY plot (covariance trace coloured by absolute correlation).
#'
#' @param stoc_mod An object of class \code{m8_stocsy1d} returned by \code{stocsy()}.
#' @param shift Numeric vector of length 2 specifying the chemical shift range (ppm).
#' @param title Optional character plot title.
#'
#' @return A \code{ggplot2} object.
#'
#' @examples
#' # st <- stocsy(X, ppm, driver = 5.233, plotting = FALSE)
#' # plotStocsy(st, shift = c(5.15, 5.30), title = "Glucose")
#'
#' @importFrom ggplot2 ggplot geom_line geom_vline scale_x_reverse
#'   scale_colour_gradientn labs theme_bw theme element_text ggtitle
#' @importFrom colorRamps matlab.like2
#' @importFrom scales breaks_pretty
#' @importFrom rlang .data
#' @examples
#' data(covid)
#' cs = 5.233 # anomeric H of gluc
#' s1 <- stocsy(covid$X, driver=cs, plotting = FALSE)
#' plotStocsy(s1)
#'
#' @export
plotStocsy <- function(stoc_mod, shift = c(0, 10), title = NULL) {

  if (!is.list(stoc_mod)) {
    stop("Input must be a list instance returned by `stocsy()`.")
  }
  if (!is.numeric(shift) || length(shift) != 2L) {
    stop("'shift' must be a numeric vector of length 2.")
  }

  ds <- data.frame(
    r   = stoc_mod$r,
    cov = stoc_mod$cov,
    ppm = stoc_mod$ppm
  )

  # Clamp shift range to available ppm range
  shift[1] <- max(shift[1], min(ds$ppm, na.rm = TRUE))
  shift[2] <- min(shift[2], max(ds$ppm, na.rm = TRUE))
  shift <- sort(shift)

  idx <- which(ds$ppm >= shift[1] & ds$ppm <= shift[2])
  if (length(idx) == 0L) stop("No ppm values in specified shift range.")
  ds <- ds[idx, , drop = FALSE]

  extD <- isTRUE(stoc_mod$extD)
  csc_lab <- if (extD) {
    "r(ext, X)"
  } else {
    paste0("r(d=", signif(stoc_mod$driver, 6), ", X)")
  }

  caption <- if (extD) {
    extD_stats <- round(summary(stoc_mod$driver), 2)
    paste0("Sample size: n=", length(stoc_mod$driver),
           "\nExternal driver: Median=", extD_stats[3],
           " Range=", extD_stats[1], "-", extD_stats[6])
  } else {
    paste0("Sample size: n=", nrow(stoc_mod$X))
  }

  ds$r_abs <- abs(ds$r)

  g1 <- ggplot(ds, aes(x = .data$ppm, y = .data$cov, colour = .data$r_abs)) +
    geom_line() +
    scale_x_reverse(breaks = breaks_pretty()) +
    scale_colour_gradientn(colours = matlab.like2(10), limits = c(0, 1), name = csc_lab) +
    labs(x = expression(delta^1 * H ~ "(ppm)"),
         y = gsub("^r", "cov", csc_lab),
         title = title, caption = caption) +
    theme_bw() +
    theme(axis.text = element_text(colour = "black")) +
    ggtitle(title)

  # Add internal driver vertical line if it lies within the shown window
  if (!extD) {
    d <- as.numeric(stoc_mod$driver)
    if (shift[1] <= d && d <= shift[2]) {
      g1 <- g1 + geom_vline(xintercept = d, linetype = 2, col = "grey")
    }
  }

  g1
}


#' @export
plot.m8_stocsy1d <- function(x, shift = c(0, 10), title = NULL, ...) {
  plotStocsy(x, shift = shift, title = title)
}

#' @export
print.m8_stocsy1d <- function(x, ...) {
  cat("m8_stocsy1d <", if (isTRUE(x$extD)) "external" else "internal", " driver>\n", sep = "")
  cat("Samples   :", nrow(x$X), "\n")
  cat("Variables :", ncol(x$X), "\n")
  invisible(x)
}

#' @title Cliff's Delta Effect Size
#' @description
#' Calculates Cliff's delta (\code{Cd}) as a non-parametric effect size for
#' comparing two numeric vectors. \code{Cd} quantifies the directional difference
#' between a reference (\code{ref}) and comparison (\code{comp}) distribution
#' based on pairwise comparisons of their values.
#'
#' \code{Cd} ranges from -1 to 1, where:
#' \itemize{
#'   \item{-1: values in \code{comp} are consistently larger than those in \code{ref}}
#'   \item{0: both distributions are similar, with no systematic difference}
#'   \item{1: values in \code{ref} are consistently larger than those in \code{comp}}
#' }
#'
#' @param ref Numeric vector representing the reference group.
#' @param comp Numeric vector representing the comparator group.
#' @return A single numeric value: Cliff's delta effect size.
#' @details
#' The effect size is calculated as the scaled difference in dominance between groups.
#' Missing or infinite values are removed with a message. Synonyms are
#' @references
#' Cliff, N. (1993). Dominance statistics: Ordinal analyses to answer ordinal questions.
#' \emph{Psychological Bulletin}, 114(3), 494–509. \doi{10.1037/0033-2909.114.3.494}
#' @examples
#' ref <- rnorm(100, mean = 0)
#' comp <- rnorm(100, mean = 1)
#'
#' hist(ref,  col=rgb(0,0,1,0.3), breaks=50, xlim=c(-4,4))
#' hist(comp, col=rgb(1,0,0,0.3), breaks=50, add=TRUE)
#' legend("topright", legend=c("ref","comp"), fill=c("blue","red"))
#'
#' cliffs_d(ref, comp)
#' cliffs_d(comp, ref)
#'
#' comp1 <- rnorm(100, mean = 10)
#' cliffs_d(ref, comp1)
#'
#' cliffs_d(ref, ref)
#' @export
cliffs_d <- function(ref, comp) {
  if (!is.numeric(ref) || !is.numeric(comp))
    stop("'ref' and 'comp' must be numeric.", call. = FALSE)

  ref0 <- length(ref); comp0 <- length(comp)

  ref <- ref[is.finite(ref)]
  comp <- comp[is.finite(comp)]

  if (length(ref) < ref0) message("Removed non-finite values from 'ref'")
  if (length(comp) < comp0) message("Removed non-finite values from 'comp'")

  n <- length(ref); m <- length(comp)
  if (n < 5 || m < 5)
    stop("Each group must contain at least 5 finite values.", call. = FALSE)

  r <- rank(c(ref, comp), ties.method = "average")
  R1 <- sum(r[seq_len(n)])
  U  <- R1 - n*(n+1)/2

  delta <- (2*U)/(n*m) - 1
  delta
}


#' @rdname cliffs_d
#' @export
es_cdelta <- cliffs_d

#' @title Full Width at Half Maximum (FWHM) Estimation
#'
#' @description
#' Estimates the full width at half maximum (FWHM; line width) of a singlet-like peak
#' within a specified chemical-shift range for each spectrum.
#'
#' @param X Numeric matrix (spectra in rows) \emph{or} a named list as returned by
#'   \code{\link{read1d}}/\code{read1d_proc} containing \code{X}, \code{ppm}, and \code{meta}.
#' @param ppm Numeric vector of chemical shift values (ppm) corresponding to columns of \code{X}.
#'   If \code{NULL}, \code{ppm} is inferred in the following order:
#'   \enumerate{
#'     \item \code{attr(X, "m8_axis")$ppm} (if present),
#'     \item numeric \code{colnames(X)} (if present).
#'   }
#' @param shift Numeric vector of length 2. Chemical shift range containing the peak
#'   (e.g., \code{c(-0.1, 0.1)} for TSP).
#' @param sf Spectrometer frequency in MHz. Either a single numeric (recycled across spectra)
#'   or a numeric vector of length \code{nrow(X)} (one value per spectrum).
#'   If \code{X} is a list input and \code{sf} is missing, \code{sf} is taken from
#'   \code{X$meta$a_SFO1} when available. If \code{sf} is missing for a matrix input,
#'   the function attempts to use \code{attr(X, "m8_meta")$a_SFO1} when present.
#'
#' @details
#' For each spectrum, the function:
#' \enumerate{
#'   \item extracts the region defined by \code{shift},
#'   \item finds the peak apex within that region,
#'   \item computes the half-height level relative to the local baseline (minimum in the window),
#'   \item estimates the left and right half-height crossing points by linear interpolation,
#'   \item converts the width from ppm to Hz using \code{sf} (MHz), i.e. \code{Hz = ppm * sf}.
#' }
#'
#' If no valid half-height crossings can be found (e.g., very low SNR or truncated peak),
#' \code{NA} is returned for that spectrum.
#'
#' @return Numeric vector of FWHM values in Hz (length \code{nrow(X)}).
#'
#' @note
#' The ppm axis may be increasing or decreasing; FWHM is computed as an absolute width
#' and is therefore independent of axis direction.
#'
#' @examples
#' # Simulated NMR peaks with different linewidths
#' ppm <- seq(-0.2, 0.2, length.out = 1000)
#'
#' # generate peaks with increasing width
#' sds <- seq(0.01, 0.03, length.out = 10)
#'
#' X <- t(sapply(sds, function(s)
#'   dnorm(ppm, mean = 0, sd = s)
#' ))
#'
#' sf <- 600  # spectrometer frequency in MHz
#'
#' fwhm_vals <- lw(X, ppm = ppm, shift = c(-0.1, 0.1), sf = sf)
#'
#' plot(sds, fwhm_vals,
#'   xlab = "Gaussian sd",
#'   ylab = "Estimated FWHM (Hz)",
#'   pch = 16)
#' @export
lw <- function(X, ppm = NULL, shift = c(-0.1, 0.1), sf) {

  if (is.list(X) && !is.null(X$X)) {
    dat <- X
    X <- dat$X
    if (is.null(ppm)) ppm <- dat$ppm
    if (missing(sf) && !is.null(dat$meta) && "a_SFO1" %in% names(dat$meta))
      sf <- dat$meta$a_SFO1
  }

  X <- .dimX(X)

  if (is.null(ppm)) {
    ax <- attr(X, "m8_axis", exact = TRUE)
    if (is.list(ax) && !is.null(ax$ppm)) {
      ppm <- ax$ppm
    } else if (!is.null(colnames(X))) {
      ppm <- as.numeric(colnames(X))
      if (anyNA(ppm)) ppm <- NULL
    }
  }
  if (is.null(ppm)) stop("`ppm` not provided and not found in attr(X,'m8_axis') or numeric colnames(X).", call. = FALSE)
  ppm <- as.numeric(ppm)
  if (!.check_X_ppm(X, ppm)) stop("X and ppm mismatch or ppm contains NA/Inf.", call. = FALSE)

  if (missing(sf)) {
    meta <- attr(X, "m8_meta", exact = TRUE)
    if (is.data.frame(meta) && "a_SFO1" %in% names(meta)) {
      sf <- meta$a_SFO1
    } else {
      stop("`sf` missing. Provide sf (MHz) or supply meta$a_SFO1 via list input or attr(X,'m8_meta').", call. = FALSE)
    }
  }

  sf <- as.numeric(sf)
  if (any(!is.finite(sf))) stop("`sf` contains non-finite values.", call. = FALSE)
  if (length(sf) == 1L) sf <- rep(sf, nrow(X))
  if (length(sf) != nrow(X)) stop("`sf` must be length 1 or length nrow(X).", call. = FALSE)

  idx <- get_idx(shift, ppm)
  if (length(idx) < 3L) stop("Selected `shift` region too small for FWHM estimation.", call. = FALSE)

  vapply(seq_len(nrow(X)), function(i) {
    x <- X[i, idx]
    x[is.na(x)] <- 0
    pp <- ppm[idx]

    baseline <- min(x, na.rm = TRUE)
    apex_i <- which.max(x)
    apex <- x[apex_i]
    height <- apex - baseline
    if (!is.finite(height) || height <= 0) return(NA_real_)

    hw <- baseline + 0.5 * height

    left_le <- which(x[seq_len(apex_i)] <= hw)
    if (length(left_le) == 0L) return(NA_real_)
    iL2 <- max(left_le)
    if (iL2 >= apex_i) return(NA_real_)
    iL1 <- iL2 + 1L

    right_le <- which(x[apex_i:length(x)] <= hw)
    if (length(right_le) == 0L) return(NA_real_)
    iR1 <- min(right_le) + apex_i - 1L
    if (iR1 <= apex_i) return(NA_real_)
    iR2 <- iR1 - 1L

    xL <- pp[iL2] + (pp[iL1] - pp[iL2]) * (hw - x[iL2]) / (x[iL1] - x[iL2])
    xR <- pp[iR2] + (pp[iR1] - pp[iR2]) * (hw - x[iR2]) / (x[iR1] - x[iR2])

    abs(xR - xL) * sf[i]
  }, numeric(1))
}



#' @title Estimate Noise Standard Deviation in 1D NMR Spectra
#'
#' @description
#' Estimates the noise standard deviation (\eqn{\sigma}) for each spectrum from a signal-free ppm region.
#' The default estimator is robust (MAD-based) and suitable for signal-to-noise calculations.
#'
#' @param X Numeric matrix (spectra in rows), numeric vector (single spectrum), or a named list
#'   as returned by \code{\link{read1d}}/\code{read1d_proc} containing \code{X}, \code{ppm}, and \code{meta}.
#' @param ppm Numeric vector of chemical shift values (ppm) corresponding to columns of \code{X}.
#'   If \code{NULL}, \code{ppm} is inferred in the following order:
#'   \enumerate{
#'     \item \code{X$ppm} if \code{X} is a list input,
#'     \item \code{attr(X, "m8_axis")$ppm} (if present),
#'     \item numeric \code{colnames(X)} (if present).
#'   }
#' @param where Numeric vector of length 2. ppm range used for noise estimation.
#'   Should be free of metabolite signals (default \code{c(14.6, 14.7)}).
#' @param method Character. Noise estimator: \code{"mad"} (default), \code{"sd"}, or \code{"p95"} (legacy amplitude).
#' @param baseline_correct Logical. If \code{TRUE}, subtract a smooth baseline in the noise window
#'   using asymmetric least squares (\code{\link[ptw]{asysm}}). Default \code{FALSE}.
#' @param lambda Numeric. Smoothing parameter for \code{\link[ptw]{asysm}} when \code{baseline_correct = TRUE}.
#' @param min_points Integer. Minimum number of points required in the noise window.
#'   May be a single value (recycled across spectra) or a numeric vector of
#'   length \code{nrow(X)}. If \code{X} is a list input as returned by
#'   \code{\link{read1d}}/\code{read1d_proc} and \code{ns} is \code{NULL},
#'   the function attempts to extract the number of scans from
#'   \code{X$meta$a_NS} when available.
#' @param ns Number of scans / transients (required if normalise_scans=TRUE)
#' @param normalise_scans Logical. If \code{TRUE}, noise estimates are
#'   multiplied by \eqn{\sqrt{NS}} to account for the theoretical scaling
#'   of noise with the number of scans (\eqn{\sigma \propto 1/\sqrt{NS}}).
#'   This is useful when comparing noise levels across spectra acquired
#'   with different numbers of scans. Default is \code{FALSE}.
#' @details
#' In NMR spectroscopy, noise scales predictably with the number of scans (\eqn{NS}).
#' For otherwise identical acquisition settings:
#'
#' \itemize{
#'   \item Signal increases approximately proportional to \eqn{NS}.
#'   \item Noise increases approximately proportional to \eqn{\sqrt{NS}}.
#'   \item Consequently, signal-to-noise ratio (SNR) increases proportional to \eqn{\sqrt{NS}}.
#' }
#'
#' Equivalently, the noise standard deviation scales as:
#'
#' \deqn{\sigma(NS) \propto \frac{1}{\sqrt{NS}},}
#'
#' assuming a fixed underlying signal scale and comparable acquisition conditions.
#'
#' To compare noise levels across datasets acquired with different numbers of scans,
#' a scan-normalised noise estimate may be used:
#'
#' \deqn{\sigma_{\mathrm{norm}} = \sigma \cdot \sqrt{NS}.}
#'
#' Under stable receiver gain and processing conditions, this normalised noise
#' should be approximately constant across runs.
#'
#' @return Numeric vector of noise estimates (length \code{nrow(X)}).
#'
#' @examples
#' data(hiit_raw)
#' X <- hiit_raw$X
#' ppm <- hiit_raw$ppm
#' sigma <- noise_sd(X, ppm, where = c(10,11))
#' plot(hiit_raw$meta$a_NS, sigma,
#'   xlab = "Number of scans (NS)",
#'   ylab = expression(sigma~"(noise estimate)"),
#'    pch = 16)
#'  lines(lowess(hiit_raw$meta$a_NS, sigma), col = "red", lwd = 2)
#' @importFrom stats sd quantile mad
#' @importFrom ptw asysm
#' @export
noise_sd <- function(X, ppm = NULL, where = c(14.6, 14.7),
                     method = c("mad","sd","p95"),
                     baseline_correct = FALSE, lambda = 1e4, min_points = 50L,
                     ns = NULL, normalise_scans = FALSE) {

  if (is.list(X) && !is.null(X$X)) {
    dat <- X
    X <- dat$X
    if (is.null(ppm)) ppm <- dat$ppm
    if (is.null(ns) && !is.null(dat$meta) && "a_NS" %in% names(dat$meta))
      ns <- dat$meta$a_NS
  }

  X <- .dimX(X)
  if (is.null(ppm)) {
    ax <- attr(X, "m8_axis", exact = TRUE)
    if (is.list(ax) && !is.null(ax$ppm)) {
      ppm <- ax$ppm
    } else if (!is.null(colnames(X))) {
      ppm <- as.numeric(colnames(X))
      if (anyNA(ppm)) ppm <- NULL
    }
  }
  if (is.null(ppm)) stop("`ppm` not provided and not found in list input, attr(X,'m8_axis'), or numeric colnames(X).", call. = FALSE)
  ppm <- as.numeric(ppm)
  if (!.check_X_ppm(X, ppm)) stop("X and ppm mismatch or ppm contains NA/Inf.", call. = FALSE)

  method <- match.arg(method)

  idx <- get_idx(where, ppm)
  if (length(idx) < as.integer(min_points)) {
    stop(sprintf("Insufficient number of points in selected region (need >= %d).", as.integer(min_points)), call. = FALSE)
  }

  sigma <- apply(X[, idx, drop = FALSE], 1, function(x) {
    x[is.na(x)] <- 0
    if (baseline_correct) x <- x - ptw::asysm(x, lambda = lambda)
    switch(method,
           mad = stats::mad(x, constant = 1.4826, na.rm = TRUE),
           sd  = stats::sd(x, na.rm = TRUE),
           p95 = as.numeric(stats::quantile(x, probs = 0.95, na.rm = TRUE, names = FALSE))
    )
  })

  if (isTRUE(normalise_scans)) {
    if (is.null(ns)) stop("normalise_scans=TRUE requires `ns` or meta$a_NS.", call. = FALSE)
    ns <- as.numeric(ns)
    if (length(ns) == 1L) ns <- rep(ns, length(sigma))
    if (length(ns) != length(sigma)) stop("`ns` must be length 1 or nrow(X).", call. = FALSE)
    if (any(!is.finite(ns)) || any(ns <= 0)) stop("`ns` must be positive finite values.", call. = FALSE)
    sigma <- sigma / sqrt(ns)
  }

  sigma
}

#' @noRd
noise.est <- function(...) {
  warning(msg = "noise.est() is deprecated; use noise_sd() instead.", call.=FALSE)
  noise_sd(...)
}

#' @title Min-Max Scaling to Arbitrary Range
#' @description
#' Rescales a numeric vector to a specified range using min-max scaling.
#' This is a generalized form of min-max normalization allowing any output range.
#'
#' @param x Numeric vector. Input values to be scaled.
#' @param ra Numeric vector of length 2. Desired output range (e.g., \code{c(5, 10)}).
#'
#' @return A numeric vector of the same length as \code{x}, scaled to the range \code{ra}.
#'
#' @details
#' The scaled values are computed as:
#' \deqn{x_{scaled} = \frac{x - \min(x)}{\max(x) - \min(x)} \cdot (r_{max} - r_{min}) + r_{min}}
#'
#' @examples
#' x <- rnorm(20)
#' plot(x, type = 'l'); abline(h = range(x), lty = 2)
#' points(scRange(x, ra = c(5, 10)), type = 'l', col = 'red');
#' abline(h = c(5, 10), col = 'red', lty = 2)
#'
#' @seealso [minmax()]
#' @export
scRange <- function(x, ra) {
  ra <- sort(ra)
  (ra[2] - ra[1]) * (x - min(x)) / (max(x) - min(x)) + ra[1]
}

#' @title Min-Max Scaling to \eqn{[0,1]}
#' @description
#' Scales a numeric vector to the range \eqn{[0,1]} using min-max normalization.
#' This is a special case of \code{\link{scRange}}.
#'
#' @param x Numeric vector. Input values to be scaled.
#' @param na.rm Logical; if `TRUE`, ignore `NA`s when computing the range.

#' @return A numeric vector of the same length as \code{x}, scaled to the range \eqn{[0,1]}.
#'
#' @details
#' The scaled values are computed as:
#' \deqn{x_{scaled} = \frac{x - \min(x)}{\max(x) - \min(x)}}
#'
#' Equivalent to \code{scRange(x, ra = c(0, 1))}.
#'
#' @examples
#' x <- rnorm(20)
#' plot(x, type = 'l'); abline(h = range(x), lty = 2)
#' points(minmax(x), type = 'l', col = 'blue')
#' abline(h = c(0, 1), col = 'blue', lty = 2)
#'
#' @seealso [scRange()] for flexible output ranges.
#' @export
minmax <- function(x, na.rm = FALSE) {
  stopifnot(is.numeric(x))
  r <- range(x, na.rm = na.rm)
  d <- r[2] - r[1]
  if (is.na(d) || d == 0) {
    return(ifelse(is.na(x), NA_real_, 0))
  }
  (x - r[1]) / d
}



#' @title Applies a preprocessing strategy to a numeric matrix.
#' @param preproc_strategy list with elements 'center' (logical) and 'scale' (character).
#' @param X numeric matrix - processing is done column-wise.
#' @return A list containing the processed matrix and parameters.
#' @examples
#' X <- matrix(c(10,10, 0,0, 0, 10), ncol=3)
#' autoscale <- uv_scaling(center=TRUE)
#' X_scaled <- prep_X(autoscale, X)
#' str(X_scaled)
#' @export
prep_X <- function(preproc_strategy, X){

            msd <- .sdRcpp(X)

            Xcs <- .scaleMatRcpp(
              X,
              0:(nrow(X)-1),
              center = preproc_strategy$center,
              scale_type =
                switch(preproc_strategy$scale,
                       "None"=0,
                       "UV"=1,
                       "Pareto"=2)
            )[[1]]

            preproc_strategy$X_mean <- msd[[2]]
            preproc_strategy$X_sd   <- msd[[1]]

            list(X=Xcs, prep=preproc_strategy)
          }


#' @title No Scaling
#' This function defines a preprocessing strategy that is applied via
#' \code{\link{prep_X}}.
#' @param center Logical. If TRUE, variables are mean-centered before scaling.
#' @return A \code{list} with elements:
#' \describe{
#'   \item{X}{Numeric matrix containing the scaled data.}
#'   \item{prep}{List describing the preprocessing, including centering and scaling
#'   parameters (\code{center}, \code{scale}, \code{X_mean}, \code{X_sd}).}
#' }#' @details Leaves variables unscaled. Optional centering.
#' @examples
#' just_centering <- unscaled(center=TRUE)
#' X <- matrix(c(10,10, 0,0, 0, 10), ncol=3)
#' X_centered <- prep_X(just_centering, X)
#' str(X_centered)
#' X_centered$X
#' @family scaling_strategies
#' @export
unscaled <- function(center = FALSE)
  list(center=center, scale="None")

#' @title Unit Variance Scaling
#' This function defines a preprocessing strategy that is applied via
#' \code{\link{prep_X}}.
#' @param center Logical. If TRUE, variables are mean-centered before scaling.
#' @return A \code{list} with elements:
#' \describe{
#'   \item{X}{Numeric matrix containing the scaled data.}
#'   \item{prep}{List describing the preprocessing, including centering and scaling
#'   parameters (\code{center}, \code{scale}, \code{X_mean}, \code{X_sd}).}
#' }
#' @details
#' Centers variables and scales each feature to unit variance.
#' UV scaling divides each variable by its standard deviation.
#' This is the default scaling in many multivariate methods such as PCA and PLS.
#' @family scaling_strategies
#' @examples
#' autoscale <- uv_scaling(center=TRUE)
#' X <- matrix(c(10,10, 0,0, 0, 10), ncol=3)
#' X_scaled <- prep_X(autoscale, X)
#' str(X_scaled)
#' X_scaled$X
#' @export
uv_scaling <- function(center = TRUE)
  list(center=center, scale="UV")


#' @title Pareto Scaling
#' Leaves variables unscaled. Optional centering.
#' @param center Logical. If TRUE, variables are mean-centered before scaling.
#' @return A \code{list} with elements:
#' \describe{
#'   \item{X}{Numeric matrix containing the scaled data.}
#'   \item{prep}{List describing the preprocessing, including centering and scaling
#'   parameters (\code{center}, \code{scale}, \code{X_mean}, \code{X_sd}).}
#' }
#' @details
#' Scales variables by the square root of their standard deviation.
#' @family scaling_strategies
#' @examples
#' paritalUV <- pareto_scaling(center=TRUE)
#' X <- matrix(c(10,10, 0,0, 0, 10, 0, 1000), ncol=4)
#' X_scaled <- prep_X(paritalUV, X)
#' str(X_scaled)
#' X_scaled$X
#' @export
pareto_scaling <- function(center = FALSE)
  list(center=center, scale="Pareto")

#' @title Normalise Spectra Using ERETIC Signal
#'
#' @description
#' Normalises 1D NMR spectra using an ERETIC reference signal. By default
#' the ERETIC position is discovered in a search window and the spectra are
#' normalised by the integral in a narrow window around the detected position.
#' Power users can supply \code{pos} to enforce a fixed ERETIC position.
#'
#' @param X Numeric matrix or vector. NMR spectra with spectra in rows.
#'   If \code{ppm} is not provided, it is inferred from \code{colnames(X)}.
#' @param integr Logical. If \code{TRUE}, returns the ERETIC integral per spectrum.
#'   If \code{FALSE}, returns the normalised spectra.
#' @param ppm Numeric vector of chemical shift positions. If \code{NULL}, inferred from \code{colnames(X)}.
#' @param pos Numeric scalar or \code{NULL}. If \code{NULL} (default), ERETIC position is discovered
#'   within \code{search}. If provided, the ERETIC window is centered on \code{pos}.
#' @param search Numeric vector of length 2. Ppm window used to discover ERETIC when \code{pos = NULL}.
#' @param width Numeric scalar. Full integration window width in ppm (default 0.2 gives \code{pos ± 0.1}).
#' @param warn_absent Logical. If \code{TRUE}, warn when ERETIC integral is zero/invalid for individual spectra.
#' @param noise_win Numeric vector of length 2. Ppm window for estimating noise intensity.
#' @param snr_factor Numeric scaler. Minimum Signal-to-Noise Ratio required for detecting the ERETIC peak.
#' @details
#' Binning/normalisation history is recorded in \code{attr(X, "m8_prep")} if present.
#' Ppm axis values are stored in \code{attr(X, "m8_axis")$ppm}.
#'
#' @return
#' If \code{integr = TRUE}, numeric vector of ERETIC integrals.
#' If \code{integr = FALSE}, numeric matrix of normalised spectra.
#'
#' @examples
#' set.seed(123)
#'
#' n <- 1000
#' ppm <- seq(0, 14, length.out = n)
#'
#' gauss <- function(x, c, h, w = 0.05) {
#'   h * exp(-((x - c)^2) / (2 * w^2))
#' }
#'
#' heights <- c(100, 80, 60, 40, 20)
#'
#' spectra <- sapply(heights, function(h)
#'   rnorm(n, 0, 0.01) + gauss(ppm, 12, h)
#' )
#'
#' spectra <- t(spectra)   # 5 spectra × 1000 variables
#'
#' plot_spec(spectra, ppm, shift = c(11, 13))
#'
#' X_norm <- norm_eretic(spectra, ppm=ppm)
#' plot_spec(X_norm, ppm, shift=c(11, 13))
#' @export
norm_eretic <- function(X, integr = FALSE,
                        ppm = NULL,
                        pos = NULL,
                        search = c(10, 16),
                        width = 0.2,
                        warn_absent = TRUE,
                        noise_win = c(9.8, 10.2),
                        snr_factor = 10) {

  .return_list <- is.list(X) && !is.null(X$X)
  meta <- NULL

  if (.return_list) {
    u <- .m8_unpack_dat(X, ppm = ppm)
    X   <- u$X
    ppm <- u$ppm
    meta <- u$meta
  }

  X0 <- X
  X  <- .dimX(X)

  if (is.null(ppm) && !is.null(colnames(X))) {
    ppm <- as.numeric(colnames(X))
    if (anyNA(ppm)) stop("ppm could not be inferred from colnames(X): non-numeric column names.", call. = FALSE)
  }
  if (!.check_X_ppm(X, ppm))
    stop("Non-matching dimensions X matrix and ppm vector or invalid ppm values.", call. = FALSE)
  ppm <- as.numeric(ppm)

  if (!is.null(pos)) {
    if (!is.numeric(pos) || length(pos) != 1L || !is.finite(pos))
      stop("`pos` must be a single finite numeric value or NULL.", call. = FALSE)
  }
  if (!is.numeric(search) || length(search) != 2L)
    stop("`search` must be a numeric vector of length 2.", call. = FALSE)
  if (!is.numeric(width) || length(width) != 1L || !is.finite(width) || width <= 0)
    stop("`width` must be a single positive numeric value.", call. = FALSE)
  if (!is.numeric(noise_win) || length(noise_win) != 2L)
    stop("`noise_win` must be a numeric vector of length 2.", call. = FALSE)
  if (!is.numeric(snr_factor) || length(snr_factor) != 1L || !is.finite(snr_factor) || snr_factor <= 0)
    stop("`snr_factor` must be a single positive numeric value.", call. = FALSE)

  discovered <- FALSE
  if (is.null(pos)) {
    idx_scan <- get_idx(search, ppm)
    if (length(idx_scan) < 3L)
      stop("Search window not present in ppm axis (too few points).", call. = FALSE)
    mean_spec <- colMeans(X, na.rm = TRUE)
    pos <- ppm[idx_scan][which.max(mean_spec[idx_scan])]
    discovered <- TRUE
  }

  half <- width / 2
  win <- c(pos - half, pos + half)
  idx_eretic <- get_idx(win, ppm)
  if (length(idx_eretic) < 2L)
    stop("ERETIC integration window not present in ppm axis.", call. = FALSE)

  idx_noise <- get_idx(noise_win, ppm)
  if (length(idx_noise) < 2L)
    stop("Noise window not present in ppm axis (adjust `noise_win`).", call. = FALSE)

  integrals <- rowSums(X[, idx_eretic, drop = FALSE], na.rm = TRUE)

  eretic_peak <- apply(X[, idx_eretic, drop = FALSE], 1, max, na.rm = TRUE)
  noise_level <- apply(abs(X[, idx_noise, drop = FALSE]), 1, max, na.rm = TRUE)

  present <- is.finite(eretic_peak) & is.finite(noise_level) & (eretic_peak > snr_factor * noise_level)

  if (warn_absent && any(!present)) {
    show <- paste(head(which(!present), 10L), collapse = ",")
    more <- if (sum(!present) > 10L) sprintf(" (+%d more)", sum(!present) - 10L) else ""
    warning(sprintf(
      "ERETIC likely absent (peak <= noise*%g) for %d spectra: %s%s",
      snr_factor, sum(!present), show, more
    ), call. = FALSE)
  }

  if (integr) {
    return(integrals)
  }

  denom <- integrals
  bad <- !is.finite(denom) | denom == 0 | !present
  denom[bad] <- 1
  Xn <- X / denom

  rownames(Xn) <- rownames(X)
  colnames(Xn) <- colnames(X)

  Xn <- .m8_copy_attrs(X0, Xn)

  ax <- attr(Xn, "m8_axis", exact = TRUE)
  if (!is.list(ax)) ax <- list()
  ax$ppm <- ppm
  attr(Xn, "m8_axis") <- ax

  Xn <- .m8_stamp(
    Xn,
    step = "normErectic",
    params = list(
      pos = pos,
      discovered = discovered,
      search = if (discovered) search else NULL,
      width = width,
      window = win,
      noise_win = noise_win,
      snr_factor = snr_factor,
      mode = "divide_by_integral"
    ),
    notes = "Normalised by ERETIC integral; spectra failing peak>noise*SNR left unchanged."
  )

  if (.return_list) {
    list(X = Xn, ppm = ppm, meta = meta)
  } else {
    Xn
  }
}

#' COVID-19 blood plasma proton NMR spectra (processed)
#'
#' 1D proton NMR spectra from SARS-CoV-2–positive patients (n = 10) and healthy controls (n = 13),
#' collected in Perth, Western Australia. Spectra were pre-processed (residual water and signal-free
#' regions excised, baseline corrected, and normalized to account for line-width differences).
#'
#' FIDs were acquired with a standard 90\eqn{^{\circ}} RF pulse sequence on a 600 MHz Bruker Avance II
#' spectrometer using IVDr methods for blood plasma (300 K, 32 scans). The spectrometer was equipped
#' with a double-resonance broadband (BBI) probe and a refrigerated autosampler.
#'
#' @format A numeric matrix/data frame with 23 rows (samples) and 27,819 columns
#'   (chemical shift variables in parts per million, ppm).
#' @source Australian National Phenome Centre (ANPC), Murdoch University.
#' @references Kimhofer et al. (2020) \doi{10.1021/acs.jproteome.0c00280}
#'
#' @examples
#' data(covid)
#' @docType data
#' @keywords datasets
#' @name covid
NULL

#' COVID-19 blood plasma proton NMR spectra (raw)
#'
#' 1D proton NMR spectra from SARS-CoV-2–positive patients (n = 10) and healthy controls (n = 13),
#' collected in a research study in Perth, Western Australia. Spectra are raw and require processing
#' before statistical analysis (see \code{?covid} for processed spectra).
#'
#' FIDs were acquired using a Carr–Purcell–Meiboom–Gill (CPMG) pulse sequence on a 600 MHz Bruker
#' Avance II spectrometer using IVDr methods for blood plasma (300 K, 32 scans). The spectrometer was
#' equipped with a double-resonance broadband (BBI) probe and a refrigerated autosampler at
#' 4\eqn{^{\circ}}C.
#'
#' @format A numeric matrix/data frame with 23 rows and 29,782 columns:
#' \describe{
#'   \item{rows}{Spectra (samples)}
#'   \item{columns}{Chemical shift variables in parts per million (ppm)}
#' }
#'
#' @source Australian National Phenome Centre (ANPC), Murdoch University.
#' @references Kimhofer et al. (2020) \doi{10.1021/acs.jproteome.0c00280}
#'
#' @examples
#' data(covid_raw)
#' @docType data
#' @keywords datasets
#' @name covid_raw
NULL

#' High-intensity interval training (HIIT) 1H NMR urine dataset
#'
#' Urine samples collected from a single individual performing a \eqn{VO_2max}-type
#' exercise protocol over a time period of 3h.
#'
#' @docType data
#'
#' @format A list with four elements:
#' \describe{
#'   \item{X}{Numeric matrix of spectral intensities (samples x variables).}
#'   \item{ppm}{Numeric vector of chemical shift values corresponding to columns of \code{X}.}
#'   \item{meta}{Data frame containing acquisition metadata.}
#'   \item{df}{Data frame containing sample annotations.}
#' }
#'
#' @details
#' The spectra were acquired on a 600 MHz Bruker NMR spectrometer.
#' The dataset is included for demonstration of preprocessing and
#' modelling workflows in \pkg{metabom8}.
#'
#' @source Example dataset bundled with the package.
#' @examples
#' data(hiit_raw)
#'
#' @keywords datasets
#' @name hiit_raw
NULL

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

.nipPcaCompRcpp <- function(X) {
    .Call(`_metabom8_nip_pca_comp_rcpp`, X)
}

multiY_Tw_rcpp <- function(X, Y, it_max, eps) {
    .Call(`_metabom8_multiY_Tw_rcpp`, X, Y, it_max, eps)
}

.nipPlsCompRcpp <- function(X, Y, it_max, eps) {
    .Call(`_metabom8_nip_PLS_comp_rcpp`, X, Y, it_max, eps)
}

.orthoGramSchmidtRcpp <- function(u, v) {
    .Call(`_metabom8_ortho_gram_schmidt_rcpp`, u, v)
}

.nipOplsRcpp <- function(X, Y) {
    .Call(`_metabom8_nip_opls_rcpp`, X, Y)
}

.tssRcpp <- function(X) {
    .Call(`_metabom8_tss_rcpp`, X)
}

.oplsPredRcpp <- function(opls_mod, pred_mod, Xnew) {
    .Call(`_metabom8_opls_pred_rcpp`, opls_mod, pred_mod, Xnew)
}

.plsPredRcpp <- function(pls_mod, Xnew) {
    .Call(`_metabom8_pls_pred_rcpp`, pls_mod, Xnew)
}

#' @keywords internal
.calcPhi <- function(ph0, ph1, le) {
    .Call(`_metabom8_calcPhi`, ph0, ph1, le)
}

#' @keywords internal
.phase1d <- function(sp_re, sp_im, ph0, ph1) {
    .Call(`_metabom8_phase1d`, sp_re, sp_im, ph0, ph1)
}

#' @keywords internal
.phaseTsp <- function(sp_re, sp_im, ppm, ph0, ph1, idx_tsp) {
    .Call(`_metabom8_phaseTsp`, sp_re, sp_im, ppm, ph0, ph1, idx_tsp)
}

#' @keywords internal
.zerofill <- function(fid, zf, le_ori) {
    .Call(`_metabom8_zerofil`, fid, zf, le_ori)
}

#' @keywords internal
.cplxFft <- function(fid) {
    .Call(`_metabom8_cplxFft`, fid)
}

#' @keywords internal
.defineChemShiftPpm <- function(sf_mhz, sw_hz, n_sp_re, dref, ref) {
    .Call(`_metabom8_defineChemShiftPpm`, sf_mhz, sw_hz, n_sp_re, dref, ref)
}

#' @title Calibrate
#' @return Shift-adjusted ppm vector
#' @keywords internal
.calibTsp <- function(spec, ppm) {
    .Call(`_metabom8_calibTsp`, spec, ppm)
}

#' @title Column-wise standard deviation and mean for a matrix
#'  Welford algorithm (single loop for mean/sd)
#' @param X num matrix
#' @keywords internal
#' @return list: 1. sd (num vec), 2. mean (sd vec)
.sdRcpp <- function(X) {
    .Call(`_metabom8_sd_rcpp`, X)
}

#' @title Column-wise matrix scaling  (C++ backend)
#' @param X num matrix
#' @param idc int row indices of X
#' @param center logical mean centering
#' @param scale_type int 0: no scaling, 1: SD scaling, 2: Pareto scaling
#' @return list: 1. scale X matrix, 2. mean (sd vec), 3: sd (num vec)
#' @keywords internal
.scaleMatRcpp <- function(X, idc, center, scale_type) {
    .Call(`_metabom8_scale_rcpp`, X, idc, center, scale_type)
}

### .scaleMatRcpp

test_that("mean is computed on training rows only", {

  X <- matrix(1:20, nrow = 5)
  idc <- 0:2   # rows 1,2,3 in R

  res <- .scaleMatRcpp(X, idc, center = TRUE, scale_type = 0)

  m_train <- colMeans(X[idc+1,,drop=FALSE])

  expect_equal(res$mean, m_train)

})

test_that("UV scaling matches base R", {

  set.seed(1)
  X <- matrix(rnorm(100), nrow = 10)
  idc <- 0:6

  res <- .scaleMatRcpp(X, idc, center = TRUE, scale_type = 1)

  m <- colMeans(X[idc+1,,drop=FALSE])
  s <- apply(X[idc+1,,drop=FALSE],2,sd)
  s[s < 1e-12] <- 1

  X_ref <- sweep(X,2,m,"-")
  X_ref <- sweep(X_ref,2,s,"/")

  expect_equal(res$X_prep, X_ref, tolerance = 1e-10)

})

test_that("Pareto scaling matches definition", {

  set.seed(1)
  X <- matrix(rnorm(100), nrow = 10)
  idc <- 0:6

  res <- .scaleMatRcpp(X, idc, center = TRUE, scale_type = 2)

  m <- colMeans(X[idc+1,,drop=FALSE])
  s <- apply(X[idc+1,,drop=FALSE],2,sd)

  denom <- sqrt(s)
  denom[denom < 1e-6] <- 1

  X_ref <- sweep(X,2,m,"-")
  X_ref <- sweep(X_ref,2,denom,"/")

  expect_equal(res$X_prep, X_ref, tolerance = 1e-10)

})

test_that("constant column becomes zero after UV scaling", {

  X <- cbind(rnorm(10), rep(5,10))
  idc <- 0:7

  res <- .scaleMatRcpp(X, idc, center = TRUE, scale_type = 1)

  expect_true(all(abs(res$X_prep[,2]) < 1e-12))

})


test_that("n_train = 1 is safe", {

  X <- matrix(rnorm(50), nrow = 5)
  idc <- 0L

  res <- .scaleMatRcpp(X, idc, center = TRUE, scale_type = 1)

  expect_false(any(is.nan(res$X_prep)))
  expect_false(any(is.infinite(res$X_prep)))

})

test_that("scaled matrix has same dimensions as input", {

  X <- matrix(rnorm(100), nrow = 10)
  idc <- 0:5

  res <- .scaleMatRcpp(X, idc, TRUE, 1)

  expect_equal(dim(res$X_prep), dim(X))

})

test_that("mean and sd have correct length", {

  X <- matrix(rnorm(100), nrow = 10)
  idc <- 0:5

  res <- .scaleMatRcpp(X, idc, TRUE, 1)

  expect_equal(length(res$mean), ncol(X))
  expect_equal(length(res$sd),   ncol(X))

})

test_that("single column remains matrix", {

  X <- matrix(rnorm(10), ncol = 1)
  idc <- 0:5

  res <- .scaleMatRcpp(X, idc, TRUE, 1)

  expect_true(is.matrix(res$X_prep))
  expect_equal(dim(res$X_prep), dim(X))

})

test_that("constant column does not drop dimensions", {

  X <- cbind(rnorm(10), rep(5,10))
  idc <- 0:5

  res <- .scaleMatRcpp(X, idc, TRUE, 1)

  expect_true(is.matrix(res$X_prep))
  expect_equal(ncol(res$X_prep), 2)

})

test_that("single row remains matrix", {

  X <- matrix(rnorm(5), nrow = 1)
  idc <- 0L

  res <- .scaleMatRcpp(X, idc, TRUE, 1)

  expect_true(is.matrix(res$X_prep))
  expect_equal(dim(res$X_prep), dim(X))

})

test_that("test rows use training estimates", {

  X <- matrix(rnorm(50), nrow = 5)
  idc <- 0:2

  res <- .scaleMatRcpp(X, idc, center = TRUE, scale_type = 1)

  m <- colMeans(X[idc+1,,drop=FALSE])
  s <- apply(X[idc+1,,drop=FALSE],2,sd)
  s[s < 1e-12] <- 1

  i_test <- 5

  expect_equal(
    res$X_prep[i_test,],
    (X[i_test,] - m)/s,
    tolerance = 1e-10
  )

})


### .sdRcpp

test_that("sd_rcpp agrees with base R mean/sd", {

  set.seed(1)
  X <- matrix(rnorm(1000), nrow = 100)

  res <- .sdRcpp(X)

  expect_equal(res$mean, colMeans(X), tolerance = 1e-12)
  expect_equal(res$sd,   apply(X, 2, sd), tolerance = 1e-12)

})

test_that("sd_rcpp handles constant columns", {

  X <- matrix(5, nrow = 10, ncol = 3)

  res <- .sdRcpp(X)

  expect_equal(res$mean, rep(5, 3))
  expect_equal(res$sd,   rep(0, 3))

})

test_that("sd_rcpp handles n = 1 correctly", {

  X <- matrix(c(1,2,3), nrow = 1)

  res <- .sdRcpp(X)

  expect_equal(res$mean, c(1,2,3))
  expect_true(all(is.na(res$sd)))

})

test_that("sd_rcpp is numerically stable for large values", {

  set.seed(1)
  X <- matrix(1e6 + rnorm(1000, sd = 1e-3), nrow = 100)

  res <- .sdRcpp(X)

  expect_equal(res$mean, colMeans(X), tolerance = 1e-8)
  expect_equal(res$sd,   apply(X, 2, sd), tolerance = 1e-8)

})


test_that("sd_rcpp returns finite output when possible", {

  X <- matrix(rnorm(1000), nrow = 100)

  res <- .sdRcpp(X)

  expect_false(any(is.nan(res$mean)))
  expect_false(any(is.infinite(res$mean)))
  expect_false(any(is.nan(res$sd)))
  expect_false(any(is.infinite(res$sd)))

})

test_that("scaling without centering works", {

  set.seed(1)
  X <- matrix(rnorm(100), nrow = 10)
  idc <- 0:6

  res <- .scaleMatRcpp(X, idc, center = FALSE, scale_type = 1)

  s <- apply(X[idc+1,,drop=FALSE],2,sd)
  s[s < 1e-12] <- 1

  X_ref <- sweep(X,2,s,"/")

  expect_equal(res$X_prep, X_ref, tolerance = 1e-10)

})

test_that("center only scaling works", {

  set.seed(1)
  X <- matrix(rnorm(100), nrow = 10)
  idc <- 0:6

  res <- .scaleMatRcpp(X, idc, center = TRUE, scale_type = 0)

  m <- colMeans(X[idc+1,,drop=FALSE])

  X_ref <- sweep(X,2,m,"-")

  expect_equal(res$X_prep, X_ref, tolerance = 1e-10)

})

#' @title Probabilistic Quotient Normalisation (PQN)
#'
#' @description
#' Applies probabilistic quotient normalisation (PQN) to spectra. PQN estimates a
#' sample-specific dilution factor from the median of quotients relative to a
#' reference spectrum and scales each spectrum accordingly.
#'
#' @param X Numeric matrix or data.frame. Each row is a sample spectrum and each
#'   column is a variable (e.g. chemical shift point or bin).
#' @param ref_index Integer vector of row indices used to compute the reference
#'   spectrum. If \code{NULL}, all rows are used.
#' @param total_area Logical. If \code{TRUE}, total area normalisation is applied
#'   to the working copy before estimating the PQN dilution factors. See Notes.
#' @param bin Optional named list controlling binning for reference estimation,
#'   e.g. \code{list(ppm = ppm, width = 0.05)} or \code{list(ppm = ppm, npoints = 400)}.
#'   If \code{NULL}, no binning is applied.
#' @param iref Deprecated. Use \code{ref_index}.
#' @param TArea Deprecated. Use \code{total_area}.
#'
#' @details
#' \strong{Mechanics.} Let \eqn{x_i} be spectrum \eqn{i} and \eqn{r} a reference spectrum.
#' PQN computes quotients \eqn{q_{ij} = x_{ij} / r_j} and defines the dilution factor as
#' \eqn{d_i = 1 / \mathrm{median}_j(q_{ij})}. The PQN-normalised spectrum is
#' \eqn{x_i^{(PQN)} = d_i \, x_i}.
#'
#' The reference spectrum \eqn{r} is typically the median spectrum across all samples
#' or across QC samples (\code{ref_index}).
#'
#' If \code{bin} is provided, \eqn{r} and dilution factors are computed on binned spectra,
#' but applied to the original spectra.
#'
#' Dilution factors are stored in \code{attr(X, "m8_pqn")$dilution_factor}.
#'
#' @section Notes on total area normalisation:
#' Total area normalisation prior to PQN is \emph{usually not recommended}. Total area
#' scaling removes global intensity differences by enforcing equal total signal per sample.
#' PQN is itself a global scaling method intended to estimate dilution. Applying both can
#' substantially change results because PQN no longer estimates dilution alone, but also
#' compensates compositional distortions introduced by total area scaling.
#'
#' Situations where \code{total_area = TRUE} can be defensible include:
#' \itemize{
#'   \item when spectra have large, non-dilution-related amplitude differences caused by
#'         acquisition artefacts (receiver gain / baseline offset) and you explicitly want
#'         to stabilise the reference estimation step;
#'   \item when the measured total signal is expected to be constant by design (e.g. strictly
#'         controlled sample mass/volume and stable overall metabolite pool), and the main
#'         goal is to reduce technical scaling variation before PQN.
#' }
#' In most metabolomics settings, prefer PQN without total area scaling.
#'
#'#' @section On spectral alignment and binning:
#' PQN assumes that corresponding variables represent the same chemical signal
#' across spectra. If spectra are not well aligned, small peak shifts can inflate
#' the variability of pointwise quotients \eqn{x_{ij} / r_j}, leading to unstable
#' dilution factor estimates.
#'
#' In such cases, slight binning (e.g. narrow fixed-width bins) prior to reference
#' estimation is recommended. Binning reduces sensitivity to minor misalignments
#' by aggregating neighbouring variables. However, excessive binning may obscure
#' narrow signals and should be avoided.
#'
#' Alternatively, prior spectral alignment is preferable when available.
#'
#' @return Numeric matrix of PQN-normalised spectra.
#'
#' @references
#' Dieterle F, Ross A, Schlotterbeck G, Senn H (2006).
#' Probabilistic Quotient Normalization as Robust Method to Account for Dilution of
#' Complex Biological Mixtures. \emph{Analytical Chemistry}, 78(13), 4281–4290.
#'
#'
#' @examples
#' set.seed(1)
#' ppm <- seq(0, 10, length.out = 1000)
#' ref <- dnorm(ppm, 3, 0.15) + dnorm(ppm, 6, 0.20) + dnorm(ppm, 7.5, 0.18)
#' dil <- c(1, 0.8, 0.6, 0.4, 0.2)            # true dilution factors
#' X <- t(sapply(dil, function(d) d * ref + rnorm(length(ref), 0, 0.005)))
#' plot_spec(X, ppm)
#'
#' Xn <- pqn(X, ref_index=1)
#' dil_est <- attr(Xn, "m8_pqn")$dilution_factor
#'
#' cbind(true = dil, estimated = dil_est)
#'
#' @family preprocessing
#' @export
pqn <- function(X,
                ref_index = NULL,
                total_area = FALSE,
                bin = NULL,
                iref = NULL,
                TArea = NULL) {

  if (!is.null(iref)) {
    warning("`iref` is deprecated; use `ref_index` instead.", call. = FALSE)
    ref_index <- iref
  }
  if (!is.null(TArea)) {
    warning("`TArea` is deprecated; use `total_area` instead.", call. = FALSE)
    total_area <- TArea
  }

  inp <- X
  u   <- .m8_unpack_dat(X)

  X0  <- u$X
  X   <- .dimX(X0)
  ppm <- u$ppm
  meta <- u$meta
  nams <- rownames(X)

  if (!is.null(bin)) {

    if (!is.list(bin) || is.null(names(bin)))
      stop("Argument 'bin' must be a named list.", call. = FALSE)

    if (!("ppm" %in% names(bin))) {

      if (!is.null(ppm)) {
        bin$ppm <- ppm
      } else if (!is.null(colnames(X))) {
        ppm_infer <- as.numeric(colnames(X))
        if (anyNA(ppm_infer))
          stop("ppm could not be inferred from colnames(X).", call. = FALSE)
        bin$ppm <- ppm_infer

      } else {
        stop("Binning requested but no ppm supplied.", call. = FALSE)
      }
    }

    if (!.check_X_ppm(X, bin$ppm))
      stop("Mismatch between X and ppm in binning specification.", call. = FALSE)

    has_width <- "width" %in% names(bin) && !is.null(bin$width)
    has_npts  <- "npoints" %in% names(bin) && !is.null(bin$npoints)

    if (!has_width && !has_npts)
      stop("Argument 'bin' must include either 'width' or 'npoints'.", call. = FALSE)

    X1 <- binning(
      X = X,
      ppm = bin$ppm,
      width = if (has_width) bin$width else NULL,
      npoints = if (has_npts) bin$npoints else NULL
    )
    ppm_est <- attr(X1, "m8_axis")$ppm
  } else {
    X1 <- X
    ppm_est <- ppm
  }

  if (isTRUE(total_area)) {
    rs <- rowSums(X1, na.rm = TRUE)
    rs[rs == 0 | !is.finite(rs)] <- NA_real_
    X1 <- X1 / rs
  }

  if (is.null(ref_index)) {
    ref <- apply(X1, 2, median, na.rm = TRUE)
  } else {
    ref_index <- as.integer(ref_index)
    if (any(!is.finite(ref_index)))
      stop("ref_index contains non-finite values.", call. = FALSE)
    if (any(ref_index < 1L | ref_index > nrow(X1)))
      stop("ref_index contains out-of-range indices.", call. = FALSE)

    if (length(ref_index) == 1L) {
      ref <- X1[ref_index, ]
    } else {
      ref <- apply(X1[ref_index, , drop = FALSE], 2, median, na.rm = TRUE)
    }
  }

  ref[!is.finite(ref) | ref == 0] <- 1e-4

  ref_safe <- ref
  ref_safe[!is.finite(ref_safe) | ref_safe == 0] <- 1e-12

  Q <- sweep(X1, 2, ref_safe, "/")
  Q[!is.finite(Q)] <- NA_real_
  quot_med <- apply(Q, 1, stats::median, na.rm = TRUE)

  dil_F <- 1 / quot_med
  dil_F[!is.finite(dil_F)] <- 1

  X_pqn <- sweep(X, 1, dil_F, "*")
  rownames(X_pqn) <- nams
  colnames(X_pqn) <- colnames(X)

  X_pqn <- .m8_copy_attrs(X0, X_pqn)
  attr(X_pqn, "m8_axis") <- list(ppm = ppm)

  attr(X_pqn, "m8_pqn") <- list(
    dilution_factor = quot_med,
    ref_index = ref_index,
    total_area = isTRUE(total_area),
    bin = bin,
    ppm_est = ppm_est
  )

  X_pqn <- .m8_stamp(
    X_pqn,
    step = "pqn",
    params = list(
      ref_index = ref_index,
      total_area = isTRUE(total_area),
      bin = bin
    ),
    notes = "Probabilistic quotient normalisation applied."
  )

  if (is.list(inp) && !is.null(inp$X)) {
    inp$X <- X_pqn
    inp$ppm <- ppm
    inp$meta <- meta
    return(inp)
  }

  X_pqn
}

.m8_unpack_dat <- function(x, ppm = NULL, meta = NULL) {
  if (is.list(x) && !is.null(x$X)) {
    if (is.null(ppm))  ppm  <- x$ppm
    if (is.null(meta)) meta <- x$meta
    x <- x$X
    if (!.check_X_ppm(x, ppm))
      stop("Dimensions of X and ppm do not match.", call. = FALSE)
  }


  list(X = x, ppm = ppm, meta = meta)
}

#' @title Find Local Extrema in NMR Spectra (Peak Picking)
#'
#' @description
#' Identifies local maxima, minima, or both from smoothed NMR spectra using Savitzky–Golay filtering.
#'
#' @param X Numeric matrix. NMR data with spectra in rows and chemical shifts in columns.
#' @param ppm Numeric vector. Chemical shift values corresponding to columns in \code{X}.
#' @param fil_p Integer. Polynomial order of the Savitzky–Golay filter.
#' @param fil_n Integer. Filter length (must be odd) of the Savitzky–Golay filter.
#' @param type Character. Type of extrema to return: \code{"max"}, \code{"min"}, or \code{"both"}.
#'
#' @details
#' The spectra are smoothed using a Savitzky–Golay filter to reduce noise. Extrema are then detected by identifying sign changes in the first derivative of the smoothed signal.
#'
#' @return
#' A list of data frames, one per spectrum. Each data frame contains:
#' \itemize{
#'   \item \code{idc}: Index of the detected peak.
#'   \item \code{ppm}: Chemical shift at the peak.
#'   \item \code{Int}: Intensity at the peak.
#'   \item \code{Etype}: Extrema type: \code{1} for minima, \code{-1} for maxima.
#' }
#'
#' @importFrom signal sgolayfilt
#' @seealso \code{\link{ppick2}}
#' @examples
#' data(covid)
#' X <- covid$X
#' ppm <- covid$ppm
#'
#' peaklist <- ppick(X, ppm)
#' plot_spec(X[1, ], ppm, shift = c(1.2, 1.4), backend='base')
#' points(peaklist[[1]]$ppm, peaklist[[1]]$Int, col = 'cyan')
#'
#' @importFrom utils head tail
#' @export
#' @export
ppick <- function(X, ppm, fil_p = 3, fil_n = 5, type = "max") {

  X <- .dimX(X)

  if (is.null(ppm)) stop("`ppm` must be provided.", call. = FALSE)
  ppm <- as.numeric(ppm)
  if (anyNA(ppm)) stop("`ppm` must be numeric and must not contain NA.", call. = FALSE)

  if (!.check_X_ppm(X, ppm))
    stop("Dimensions of X and ppm must match and ppm must not contain NA/Inf.", call. = FALSE)

  type <- match.arg(type, c("max", "min", "both"))

  if (!is.numeric(fil_p) || length(fil_p) != 1L || fil_p < 1)
    stop("`fil_p` must be a positive integer.", call. = FALSE)
  if (!is.numeric(fil_n) || length(fil_n) != 1L || fil_n < 3)
    stop("`fil_n` must be an integer >= 3.", call. = FALSE)
  fil_p <- as.integer(fil_p)
  fil_n <- as.integer(fil_n)

  if (fil_n %% 2L == 0L) stop("`fil_n` must be odd.", call. = FALSE)
  if (fil_n <= fil_p) stop("`fil_n` must be greater than `fil_p`.", call. = FALSE)

  lapply(seq_len(nrow(X)), function(i) {
    x <- X[i, ]
    x[is.na(x)] <- 0

    x_smooth <- signal::sgolayfilt(x, p = fil_p, n = fil_n)

    dsgn <- sign(diff(x_smooth))

    if (any(dsgn == 0)) {
      nz <- which(dsgn != 0)
      if (length(nz) > 0) {
        for (k in seq_along(dsgn)) {
          if (dsgn[k] == 0) dsgn[k] <- if (k == 1) dsgn[nz[1]] else dsgn[k - 1]
        }
      }
    }

    extrema_idx <- which(utils::head(dsgn, -1) != utils::tail(dsgn, -1)) + 1L
    if (length(extrema_idx) == 0L) return(NULL)

    extrema_type <- -dsgn[extrema_idx - 1L]

    res <- data.frame(
      idc = extrema_idx,
      ppm = ppm[extrema_idx],
      Int = x[extrema_idx],
      Etype = extrema_type
    )

    switch(type,
           "max"  = res[res$Etype < 0, , drop = FALSE],
           "min"  = res[res$Etype > 0, , drop = FALSE],
           "both" = res
    )
  })
}


#' @title Peak picking using Savitzky–Golay derivatives
#'
#' @description
#' Finds local extrema in 1D spectra using Savitzky–Golay first and second derivatives.
#' Candidate peaks are identified at zero-crossings of the first derivative and classified
#' by the sign of the second derivative. Optional filters control peak height, prominence,
#' SNR, curvature and minimum separation.
#'
#' @param X Numeric matrix or vector. Spectra in rows.
#' @param ppm Numeric vector or NULL. If NULL, inferred from colnames(X).
#' @param type Character. "max", "min", or "both".
#' @param fil_p Integer. Polynomial order for SG filter.
#' @param fil_n Integer. Window length for SG filter (odd).
#' @param noise_win Numeric length-2 vector or NULL. Ppm window used to estimate noise per spectrum.
#'   If NULL, a robust noise estimate is computed from the full spectrum (MAD of first differences).
#' @param min_snr Numeric. Minimum SNR (peak height / noise). Set NULL to disable.
#' @param min_height Numeric. Minimum absolute peak height (in original intensity units). NULL disables.
#' @param min_prominence Numeric. Minimum local prominence. NULL disables.
#' @param prom_half_window_ppm Numeric. Half-window (ppm) for prominence estimation around each peak.
#' @param min_distance_ppm Numeric. Minimum separation between peaks (ppm). NULL disables.
#' @param min_curvature Numeric. Minimum absolute curvature at peak (|d2|). NULL disables.
#' @param keep_cols Character. Extra columns to keep (default keeps all computed).
#'
#' @return List of data.frames (one per spectrum). Each contains:
#' idc, ppm, Int, Etype (+1 max, -1 min), height, snr, curvature, prominence.
#'
#' @importFrom signal sgolayfilt
#' @examples
#' data(covid)
#' X <- covid$X
#' ppm <- covid$ppm
#'
#' peaklist <- ppick2(X[1,], ppm, min_snr=50)
#'
#' plot_spec(X[1, ], ppm, shift = c(3, 4.5), backend='base')
#' points(peaklist[[1]]$ppm, peaklist[[1]]$Int, col = "cyan")
#' head(peaklist[[1]])
#' @export
ppick2 <- function(X,
                   ppm = NULL,
                   type = c("max", "min", "both"),
                   fil_p = 3L,
                   fil_n = 11L,
                   noise_win = NULL,
                   min_snr = 10,
                   min_height = NULL,
                   min_prominence = NULL,
                   prom_half_window_ppm = 0.02,
                   min_distance_ppm = 0.005,
                   min_curvature = NULL,
                   keep_cols = c("height", "snr", "curvature", "prominence")) {


  u <- .m8_unpack_dat(X, ppm = ppm)
  X <- .dimX(u$X)
  ppm <- u$ppm

  X <- .dimX(X)

  if (is.null(ppm)) {
    if (is.null(colnames(X))) stop("ppm is NULL and colnames(X) are missing.", call. = FALSE)
    ppm <-as.numeric(colnames(X))
    if (anyNA(ppm)) stop("ppm could not be inferred from colnames(X): non-numeric column names.", call. = FALSE)
  }
  ppm <- as.numeric(ppm)
  if (!.check_X_ppm(X, ppm)) stop("X and ppm dimensions mismatch or ppm contains NA/Inf.", call. = FALSE)

  type <- match.arg(type)

  fil_p <- as.integer(fil_p)
  fil_n <- as.integer(fil_n)
  if (fil_n < 5L) stop("fil_n should be >= 5 for stable derivative estimation.", call. = FALSE)
  if (fil_n %% 2L == 0L) stop("fil_n must be odd.", call. = FALSE)
  if (fil_n <= fil_p) stop("fil_n must be greater than fil_p.", call. = FALSE)

  if (!is.null(noise_win)) {
    if (!is.numeric(noise_win) || length(noise_win) != 2L) stop("noise_win must be numeric length 2 or NULL.", call. = FALSE)
    idx_noise <- get_idx(noise_win, ppm)
    if (length(idx_noise) < 5L) stop("noise_win too small or not present in ppm axis.", call. = FALSE)
  } else {
    idx_noise <- NULL
  }

  .noise_mad_diff <- function(x) {
    d <- diff(x)
    s <- stats::mad(d, constant = 1.4826, na.rm = TRUE)
    if (!is.finite(s) || s == 0)
      s <- sqrt(.Machine$double.eps)
    s
  }

  .prominence_local <- function(x, ppm, i0, half_ppm) {

    win <- c(ppm[i0] - half_ppm, ppm[i0] + half_ppm)
    idx <- get_idx(win, ppm)
    idx <- idx[idx != i0]
    if (length(idx) < 2L) return(NA_real_)

    left <- idx[idx < i0]
    right <- idx[idx > i0]
    if (length(left) < 1L || length(right) < 1L) return(NA_real_)
    base_left <- min(x[left], na.rm = TRUE)
    base_right <- min(x[right], na.rm = TRUE)
    x[i0] - max(base_left, base_right)
  }


  .enforce_min_distance <- function(tab, min_dist_ppm) {

    if (is.null(min_dist_ppm) || nrow(tab) <= 1L) return(tab)
    tab <- tab[order(-tab$height), , drop = FALSE]
    keep <- rep(TRUE, nrow(tab))
    for (i in seq_len(nrow(tab))) {
      if (!keep[i]) next
      too_close <- which(
        is.finite(tab$ppm) &
          abs(tab$ppm - tab$ppm[i]) < min_dist_ppm
      )
      too_close <- setdiff(too_close, i)
      keep[too_close] <- FALSE
    }
    tab2 <- tab[keep, , drop = FALSE]
    tab2[order(tab2$idc), , drop = FALSE]
  }

  ref <- if (nrow(X) > 1) apply(X, 2, median) else X[1,]
  d  <- diff(ref)
  sigma <- .noise_mad_diff(d)
  mask <- abs(ref) > (3 * sigma)
  k <- 9L
  mask <- stats::filter(mask,
                        rep(1, k),
                        sides = 2) > 0

  lapply(seq_len(nrow(X)), function(r) {
    x <- X[r, ]
    x[!mask] <- 0

    d1 <- signal::sgolayfilt(x, p = fil_p, n = fil_n, m = 1)
    d2 <- signal::sgolayfilt(x, p = fil_p, n = fil_n, m = 2)

    s1 <- sign(d1)
    if (any(s1 == 0)) {
      nz <- which(s1 != 0)
      if (length(nz) > 0) {
        for (k in seq_along(s1)) {
          if (s1[k] == 0) s1[k] <- if (k == 1) s1[nz[1]] else s1[k - 1]
        }
      }
    }

    zc <- which(diff(s1) != 0) + 1L

    if (length(zc) == 0L) return(NULL)

    etype <- ifelse(d2[zc] < 0, 1L, -1L)

    if (type == "max") {
      keep0 <- etype == 1L
    } else if (type == "min") {
      keep0 <- etype == -1L
    } else {
      keep0 <- rep(TRUE, length(zc))
    }
    zc <- zc[keep0]
    etype <- etype[keep0]
    if (length(zc) == 0L) return(NULL)

    height <- x[zc]
    curvature <- d2[zc]

    noise <- if (!is.null(idx_noise)) {
      apply(abs(x[idx_noise, drop = TRUE]), 1, max, na.rm = TRUE)
      max(abs(x[idx_noise]), na.rm = TRUE)
    } else {
      .noise_mad_diff(x)
    }
    snr <- if (is.finite(noise) && noise > 0) abs(height) / noise else NA_real_

    prominence <- vapply(zc, function(i0) .prominence_local(x, ppm, i0, prom_half_window_ppm), numeric(1))

    res <- data.frame(
      idc = zc,
      ppm = ppm[zc],
      Int = x[zc],
      Etype = etype,
      height = height,
      snr = snr,
      curvature = curvature,
      prominence = prominence
    )

    # filters
    if (!is.null(min_height)) {
      res <- res[res$height >= min_height, , drop = FALSE]
    }
    if (!is.null(min_snr)) {
      res <- res[is.finite(res$snr) & res$snr >= min_snr, , drop = FALSE]
    }
    if (!is.null(min_curvature)) {
      res <- res[abs(res$curvature) >= min_curvature, , drop = FALSE]
    }
    if (!is.null(min_prominence)) {
      res <- res[is.finite(res$prominence) & res$prominence >= min_prominence, , drop = FALSE]
    }
    if (nrow(res) == 0L) return(NULL)

    res <- .enforce_min_distance(res, min_distance_ppm)
    if (nrow(res) == 0L) return(NULL)

    if (!is.null(keep_cols)) {
      keep_cols <- intersect(keep_cols, names(res))
      base_cols <- c("idc", "ppm", "Int", "Etype")
      res <- res[, unique(c(base_cols, keep_cols)), drop = FALSE]
    }

    res
  })
}

set.seed(1)

## ppick

test_that("ppick returns expected output structure", {

  X <- matrix(rnorm(5 * 100), nrow = 5)
  ppm <- seq(0.5, 9.5, length.out = 100)

  peaks_max  <- ppick(X, ppm, type = "max")
  peaks_min  <- ppick(X, ppm, type = "min")
  peaks_both <- ppick(X, ppm, type = "both")

  expect_length(peaks_max, 5)
  expect_true(all(vapply(peaks_max, is.data.frame, logical(1))))

  expect_named(peaks_both[[1]], c("idc","ppm","Int","Etype"))

  expect_true(all(peaks_max[[1]]$Etype == -1))
  expect_true(all(peaks_min[[1]]$Etype == 1))
  expect_true(all(peaks_both[[1]]$Etype %in% c(-1,1)))

  expect_equal(peaks_both[[1]]$ppm,
               ppm[peaks_both[[1]]$idc])

})

## ppick2

test_that("ppick2 validates input parameters", {

  X <- matrix(rnorm(100), nrow = 1)
  ppm <- seq_len(100)

  expect_error(ppick2(X, ppm, fil_n = 4))
  expect_error(ppick2(X, ppm, fil_n = 3))
  expect_error(ppick2(X, ppm, fil_p = 5, fil_n = 5))

})

test_that("ppick2 returns correct structure", {

  ppm <- seq(0, 10, length.out = 200)

  peak1 <- exp(-(ppm - 3)^2 / 0.01)
  peak2 <- exp(-(ppm - 7)^2 / 0.02)

  X <- matrix(peak1 + peak2, nrow = 1)

  peaks <- ppick2(X, ppm, type = "max", min_snr = 1)

  expect_type(peaks, "list")
  expect_length(peaks, 1)

  expect_true(is.data.frame(peaks[[1]]))

  expect_named(peaks[[1]],
               c("idc","ppm","Int","Etype",
                 "height","snr","curvature","prominence"),
               ignore.order = TRUE)

})

test_that("ppick2 detects known peaks", {

  ppm <- seq(0, 10, length.out = 500)

  peak1 <- exp(-(ppm - 3)^2 / 0.02)
  peak2 <- exp(-(ppm - 7)^2 / 0.02)

  X <- matrix(peak1 + peak2, nrow = 1)

  peaks <- ppick2(X, ppm, type = "max", min_snr = 1)

  detected <- peaks[[1]]$ppm

  expect_true(any(abs(detected - 3) < 0.05))
  expect_true(any(abs(detected - 7) < 0.05))

})

test_that("ppick2 infers ppm from colnames", {

  ppm <- seq(0, 10, length.out = 200)

  X <- matrix(exp(-(ppm - 4)^2 / 0.02), nrow = 1)

  colnames(X) <- as.character(ppm)

  peaks <- ppick2(X, ppm = NULL, min_snr = 1)

  expect_true(is.list(peaks))
})

#' @title OPLS Component Estimation via Cross-Validation
#' @description
#' Performs orthogonal projections to latent structures (OPLS) modeling on
#' training folds of a cross-validation (CV) scheme.
#' The function fits one orthogonal and one predictive component per fold,
#' tracking scores, predictions, and residuals.
#' For the first orthogonal component (\code{nc = 1}), the full model structure
#' is initialized.
#' For later components (\code{nc > 1}), modeling proceeds on residual matrices
#' carried over from previous iterations.
#'
#' @param X Numeric matrix. Input data matrix (samples x features); only
#' required for \code{nc = 1}.
#' @param Ycs_fold List containing the response values (`Y`) for each
#' cross-validation split.
#' @param cv.set List of integer vectors. Each element contains training indices
#' for one CV round.
#'   These indices are adjusted internally by -1 due to 0-based indexing in the
#'   underlying C++ (Rcpp) routines.
#' @param nc Integer. Component number currently being estimated (\code{nc = 1}
#' for first orthogonal component).
#' @param mod.cv List. Model state to be updated with results from each CV iteration.
#'
#' @return A list of the same length as \code{cv.set}. Each list element contains:
#' \itemize{
#'   \item \code{t_xo}: Orthogonal component scores (matrix, samples x components).
#'   \item \code{t_xp}: Predictive component scores (matrix, samples x 1).
#'   \item \code{y_pred_train}: Predicted responses for training samples.
#'   \item \code{y_pred_test}: Predicted responses for held-out test samples.
#'   \item \code{x_res}: Residual matrix after filtering out orthogonal structure.
#' }
#'
#' @note Training indices in \code{cv.set} are internally adjusted by -1 before
#' being passed to Rcpp routines.
#' This is required because C++ uses zero-based indexing.
#'
#' @keywords internal
.oplsComponentCv <- function(X, Ycs_fold, cv.set, nc, mod.cv, acc) {

  n <- nrow(Ycs_fold[[1]]);

  if (nc == 1) {
    q <- ncol(Ycs_fold[[1]]);
    p <- ncol(X)

    mod.cv <- vector("list", length(cv.set))
    for(i in seq_along(cv.set)) {
      mod.cv[[i]] <- list(
        t_xo = matrix(0, n, 1),
        t_xp = matrix(0, n, 1),
        y_pred_train = matrix(0, n, q),
        y_pred_test = matrix(0, n, q),
        x_res = matrix(0, n, p)
      )
    }

  } else {
    for(i in seq_along(mod.cv)) {
      mod.cv[[i]]$t_xo <- cbind(mod.cv[[i]]$t_xo, 0)
      mod.cv[[i]]$t_xp <- cbind(mod.cv[[i]]$t_xp, 0)
    }
  }

  for (k in seq_along(cv.set)) {
    idc <- cv.set[[k]]
    not_idc <- setdiff(seq_len(n), idc)

    Xcs <- if (nc == 1) .scaleMatRcpp(X, idc - 1, center = TRUE, scale_type = 1)[[1]] else mod.cv[[k]]$x_res
    Ycs <- Ycs_fold[[k]]

    opls_filt  <- .nipOplsRcpp(X = Xcs[idc, , drop = FALSE], Y = Ycs[idc, , drop = FALSE])
    opls_filt$w_o <- .ensure_matrix(opls_filt$w_o)
    opls_filt$p_o <- .ensure_matrix(opls_filt$p_o)

    pred_comp  <- .nipPlsCompRcpp(opls_filt$X_res, Ycs[idc, , drop = FALSE], it_max = 800, eps = 1e-8)

    pred_comp$w_p <- .ensure_matrix(pred_comp$w_x)
    pred_comp$p_y <- .ensure_matrix(pred_comp$p_y)

    pred_comp$b   <- as.numeric(pred_comp$b)

    Xte <- as.matrix(Xcs[not_idc, , drop = FALSE])
    storage.mode(Xte) <- "double"

    opls_pred <- .oplsPredRcpp(opls_mod = opls_filt, pred_mod = pred_comp, Xnew = Xte)

    mod.cv[[k]]$t_xo[not_idc, nc]      <- opls_pred$t_xo_new
    mod.cv[[k]]$t_xp[not_idc, nc]      <- opls_pred$t_pred
    mod.cv[[k]]$y_pred_train[idc, ]    <- pred_comp$y_pred
    mod.cv[[k]]$y_pred_test[not_idc, ] <- opls_pred$y_pred
    mod.cv[[k]]$x_res[idc, ]           <- opls_filt$X_res
    mod.cv[[k]]$x_res[not_idc, ]       <- opls_pred$Xres

    acc$sum_test[not_idc, ] <- acc$sum_test[not_idc, ] + opls_pred$y_pred
    acc$n_test[not_idc]     <- acc$n_test[not_idc] + 1

    acc$sum_train[idc, ] <- acc$sum_train[idc, ] + pred_comp$y_pred
    acc$n_train[idc]     <- acc$n_train[idc] + 1
  }

  return(list(mod.cv = mod.cv, acc = acc))
}


.ensure_matrix <- function(x, ncol = 1L) {
  if (is.null(dim(x)))
    matrix(x, ncol = ncol)
  else
    x
}

#' @title Extract Cross-Validated OPLS Features
#'
#' @description
#' Extracts a specific data structure (e.g., scores or predictions) from cross-validation output returned by `.oplsComponentCv()`.
#' Depending on the cross-validation type and model structure, the output is aggregated across folds by computing the mean,
#' standard deviation (SD), and coverage.
#'
#' Here, **Coverage** refers to the proportion of cross-validation folds in which a given sample's value was observed
#' (i.e., not missing), serving as an estimate of how frequently that value was evaluated across folds.
#'
#' @param cv_obj Named list. Output of `.oplsComponentCv()` for a specific OPLS component.
#' @param feat Character. Feature name to extract from each CV fold object (e.g., `"t_xp"` or `"y_pred_test"`).
#'
#' @return A numeric matrix or array of the requested feature:
#' \itemize{
#'   \item For Monte Carlo CV and single-column Y: matrix with rows `"mean"`, `"sd"`, and `"coverage"` x samples.
#'   \item For Monte Carlo CV and multi-column Y: 3D array with shape \code{[3, samples, outcomes]}.
#'   \item For k-fold CV: matrix or array containing values for each sample (no aggregation, just fold values).
#' }
#'
#' @importFrom abind abind
#' @importFrom stats sd
#' @keywords internal
.extMeanCvFeat <- function(cv_obj, feat = "t_xp") {

  if (!feat %in% names(cv_obj[[1]]))
    stop("Feature name not in feature list.", call. = FALSE)

  inter <- lapply(cv_obj, `[[`, feat)

  f_mat <- abind::abind(inter, along = 3)

  apply(f_mat, c(1,2), function(x) {
    nz <- !is.na(x)
    if (any(nz)) sum(x[nz]) / sum(nz) else NA_real_
  })
}

# --- scores -----------------------------------------------------------

#' PLS/OPLS model scores
#' @param cv Logical indicating whether cross-validated scores should be returned
#' @param orth Logical indicating whether orthogonal scores should be returned
#'   (only applicable for OPLS models).
#' @param object An object of class \code{m8_model}.
#' @param ... Additional arguments (currently ignored).
#' @return Numeric vector or matrix containing scores.
#' @examples
#' data(covid)
#' cv <- balanced_mc(k=5, split=2/3)
#' scaling <- uv_scaling(center=TRUE)
#' model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
#' show(model)
#' scores(model, orth=FALSE)
#' scores(model, orth=TRUE)
#' scores(model, cv=TRUE)
#' @export
setGeneric("scores",
           function(object, ...)
             standardGeneric("scores"))


# --- vips -----------------------------------------------------------

#' Variable Importance in Projection (VIP)
#' @param object An object of class \code{m8_model}.
#' @return Numeric vector or matrix containing the variable importance in
#' projection (VIP) scores.
#' @examples
#' data(covid)
#' cv <- balanced_mc(k=5, split=2/3)
#' scaling <- uv_scaling(center=TRUE)
#' model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
#' show(model)
#' vip_scores <- vip(model)
#' dim(vip_scores)
#' @export
setGeneric("vip", function(object)
  standardGeneric("vip"))


# --- weights -----------------------------------------------------------

#' Extract model weights
#' @rdname weights
#' @param object An object of class \code{m8_model}.
#' @param ... Additional arguments (currently ignored).
#' @return Numeric vector or matrix containing model weights.
#' @examples
#' data(covid)
#' cv <- balanced_mc(k=5, split=2/3)
#' scaling <- uv_scaling(center=TRUE)
#' model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
#' show(model)
#' W <- weights(model)
#' Wo <- weights(model, orth = TRUE)
#' dim(W)
#' dim(Wo) == dim(W)
#' @export
setGeneric("weights", function(object, ...)
  standardGeneric("weights"))


# --- yfitted -----------------------------------------------------------

#' Extract fitted Y values
#' @param object An object of class \code{m8_model}.
#' @param ... Additional arguments (currently ignored).
#' @return Numeric vector or matrix containing the fitted response values.
#' @examples
#' data(covid)
#' cv <- balanced_mc(k=5, split=2/3)
#' scaling <- uv_scaling(center=TRUE)
#' model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
#' show(model)
#' Y_hat_dummy <- fitted(model)
#' @export
setGeneric("fitted", function(object, ...)
  standardGeneric("fitted"))


# --- xres -----------------------------------------------------------

#' Compute X residual matrix
#' Returns the residual matrix (E) of an OPLS model.
#' @param object An object.
#' @return Numeric matrix containing the X residuals.
#' @examples
#' data(covid)
#' cv <- balanced_mc(k = 5, split = 2/3)
#' scaling <- uv_scaling(center = TRUE)
#' model <- opls(X = covid$X, Y = covid$an$type, scaling, cv)
#' X_res <- xres(model)
#' dim(X_res) == dim(covid$X)
#' @export
setGeneric("xres", function(object) standardGeneric("xres"))

#' Prepare Spectral Data for Plotting
#'
#' Internal helper that validates input, enforces ppm ordering,
#' subsets the selected shift region, and returns matrix-form data.
#'
#' @param x Numeric vector or matrix of spectra.
#' @param ppm Numeric chemical shift vector.
#' @param shift Length-2 numeric vector specifying ppm window.
#'
#' @return A list with elements:
#'   \describe{
#'     \item{ppm}{Subsetted chemical shift axis}
#'     \item{X}{Matrix of subsetted spectra}
#'   }
#'
#' @keywords internal
.prep_spec <- function(x, ppm, shift) {

  if (is.vector(x)) {
    x <- matrix(x, nrow = 1)
  } else if (!is.matrix(x)) {
    stop("x must be vector or matrix", call. = FALSE)
  }

  if (ncol(x) != length(ppm))
    stop("ppm mismatch", call. = FALSE)

  if (ppm[1] < ppm[length(ppm)]) {
    ord <- order(ppm, decreasing = TRUE)
    ppm <- ppm[ord]
    x   <- x[, ord, drop = FALSE]
  }

  idx <- get_idx(shift, ppm)

  list(
    ppm = ppm[idx],
    X   = x[, idx, drop = FALSE]
  )
}


#' Draw NMR Spectra Using Base Graphics
#'
#' Internal backend renderer for base R plotting.
#'
#' @param dat List returned by `.prep_spec()`.
#' @param add Logical. Add to existing plot.
#' @param ... Additional graphical parameters.
#'
#' @keywords internal
.draw_base <- function(dat, add, ...) {

  if (add) {
    graphics::matlines(dat$ppm, t(dat$X), ...)
  } else {
    graphics::matplot(dat$ppm, t(dat$X),
                      type="l",
                      xlim = rev(range(dat$ppm)),
                      xlab="ppm", ylab="Intensity", ...)
  }

  invisible(NULL)
}


#' Draw NMR Spectra Using Plotly
#'
#' Internal backend renderer for interactive plotting.
#'
#' @param dat List returned by `.prep_spec()`.
#' @param ... Additional arguments passed to `plotly`.
#'
#' @keywords internal
.draw_plotly <- function(dat, col = NULL, ...) {

  X   <- dat$X
  ppm <- dat$ppm
  n   <- nrow(X)

  labels <- rownames(dat$meta)
  if (is.null(labels))
    labels <- paste0("spec_", seq_len(n))
    labels <- factor(labels, levels = labels)

  df <- data.frame(
    ppm = rep(ppm, times = n),
    intensity = as.vector(t(X)),
    spec = rep(seq_len(n), each = length(ppm)),
    label     = rep(labels, each = length(ppm))
  )

  p <- plotly::plot_ly(type = "scattergl", mode = "lines")

  # -------- numeric colouring --------
  if (!is.null(col) && is.numeric(col)) {

    df$z <- rep(col, each = length(ppm))

    p <- plotly::add_trace(
      p,
      data  = df,
      x     = ~ppm,
      y     = ~intensity,
      split = ~label,
      text  = ~label,
      hovertemplate = "%{text}<extra></extra>",
      line  = list(
        color = ~z,
        colorscale = "Viridis",
        showscale  = TRUE,
        width = 1
      ),
      hoverinfo = "none"
    )

    # -------- categorical colouring --------
  } else if (!is.null(col)) {

    df$grp <- rep(as.factor(col), each = length(ppm))

    p <- plotly::add_trace(
      p,
      data  = df,
      x     = ~ppm,
      y     = ~intensity,
      split = ~label,
      color = ~grp,
      text  = ~label,
      hovertemplate = "%{text}<extra></extra>",
      line  = list(width = 1),
      hoverinfo = "none"
    )

    # -------- default --------
  } else {

    p <- plotly::add_trace(
      p,
      data  = df,
      x     = ~ppm,
      y     = ~intensity,
      split = ~label,
      text  = ~label,
      hovertemplate = "%{text}<extra></extra>",
      line  = list(width = 1),
      hoverinfo = "none"
    )
  }

  plotly::layout(
    p,
    xaxis = list(autorange = "reversed"),
    yaxis = list(title = "Intensity")
  )
}

#' Draw NMR Spectra Using ggplot2
#'
#' Internal backend renderer using `ggplot2`. Converts spectra
#' to long format prior to plotting.
#'
#' @param dat List returned by `.prep_spec()`.
#' @param ... Additional arguments for `ggplot2`.
#' @importFrom rlang .data
#' @keywords internal
.draw_ggplot2 <- function(dat, ...) {

  df <- reshape2::melt(dat$X)
  df$ppm <- rep(dat$ppm, each = nrow(dat$X))

  g <- ggplot2::ggplot(df,
                  ggplot2::aes(.data$ppm, .data$value, group = .data$Var1)) +
    ggplot2::geom_line() +
    ggplot2::scale_x_reverse()+
    ggplot2::ylab("Intensity") +
    ggplot2::theme_bw()

  return(g)
}

#' Plot 1D NMR Spectra
#'
#' Plot one or multiple 1D ¹H NMR spectra using different rendering backends.
#'
#' @param x Numeric vector (single spectrum) or numeric matrix (spectra in rows).
#' @param ppm Numeric vector of chemical shift values. Must match `ncol(x)`.
#' @param shift Numeric vector of length 2. Chemical shift window to display
#'   (e.g., `c(0, 10)`).
#' @param backend Character. Rendering backend. One of `"plotly"`, `"base"`,
#'   or `"ggplot"`. Defaults to `"plotly"`.
#' @param add Logical. If `TRUE` and `backend = "base"`, add spectra to an
#'   existing plot.
#' @param ... Additional arguments passed to the selected backend.
#'
#' @details
#' The function accepts both single spectra and matrices of spectra. Input is
#' internally normalized to matrix form, subset to the selected ppm region,
#' and then rendered using the chosen backend.
#'
#' For large NMR datasets (e.g., >500 spectra × >10k ppm), the `"base"` and
#' `"plotly"` backends are substantially more memory-efficient than `"ggplot"`,
#' which requires reshaping to long format.
#'
#' Chemical shift axes are automatically displayed in decreasing order
#' (NMR convention).
#'
#' @return
#' - `"plotly"`: a `plotly` object.
#' - `"ggplot"`: a `ggplot2` object.
#' - `"base"`: `NULL` (invisibly).
#' @examples
#' data(hiit_raw)
#' plot_spec(hiit_raw)
#' plot_spec(hiit_raw, shift=c(-0.05,0.05), backend='base')
#' plot_spec(hiit_raw, shift=c(-0.05,0.05), backend='ggplot2')
#' @export
plot_spec <- function(x,
                      ppm=NULL,
                      shift = c(0,10),
                      backend = c("plotly","base","ggplot2"),
                      add = FALSE,
                      ...) {

  u   <- .m8_unpack_dat(x, ppm = ppm)
  X   <- .dimX(u$X)
  ppm <- u$ppm
  meta <- u$meta

  backend <- match.arg(backend)

  dat <- .prep_spec(X, ppm, shift)
  dat$meta <- meta

  switch(
    backend,
    base   = .draw_base(dat, add, ...),
    plotly = .draw_plotly(dat, ...),
    ggplot2 = .draw_ggplot2(dat, ...)
  )
}


#' @rdname plot_spec
#' @export
spec <- function(...) {
  warning("spec() is deprecated; use plot_spec() instead.", call. = FALSE)
  plot_spec(..., backend = "plotly")
}

#' @rdname plot_spec
#' @export
matspec <- function(...) {
  warning("matspec() is deprecated; use plot_spec() instead.", call.=FALSE)
  plot_spec(..., backend = "plotly")
}

#' metabom8: A High-Performance R Package for Metabolomics Modeling and Analysis
#'
#' @description
#' metabom8 (pronounced \emph{metabo-mate}) provides pipelines for 1D NMR data import,
#' preprocessing, multivariate modeling (PCA, OPLS), metabolite identification,
#' and visualisation. Core functions are accelerated in C++ via \pkg{Rcpp},
#' \pkg{RcppArmadillo}, and \pkg{RcppEigen} for improved computational performance.
#'
#' @section Features:
#' \itemize{
#'   \item Import, preprocessing, and analysis of 1D NMR spectra.
#'   \item \strong{Principal Components Analysis (PCA)} with \emph{back-scaled} loadings
#'         (projected back to the spectral domain and visualized as spectra).
#'   \item \strong{Orthogonal Partial Least Squares (OPLS)} fitted iteratively via the NIPALS algorithm,
#'         with an \emph{automatic stopping criterion} to determine the optimal number of components.
#'   \item Automatic model selection using cross-validated performance (\eqn{R^2}, \eqn{Q^2},
#'         and cross-validated AUC for classification), with safeguards against overfitting.
#'   \item Robust statistical validation for small to large sample sizes:
#'         stratified Monte Carlo cross-validation and k-fold CV.
#'   \item Model diagnostics and validation (DModX, permutation testing).
#'   \item Metabolite identification via STOCSY and STORM.
#'   \item Native C++ acceleration via \pkg{RcppArmadillo} and \pkg{RcppEigen}.
#' }
#'
#' @section Vignettes:
#' \itemize{
#'   \item Getting Started: \code{vignette("Getting Started")}
#' }
#'
#' @name metabom8
"_PACKAGE"

---
title: "Getting Started"
author: "Torben Kimhofer"
date: "`r Sys.Date()`"
output: 
  BiocStyle::html_document:
    toc: true
    number_sections: true
vignette: >
  %\VignetteIndexEntry{Getting Started}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  markdown: 
    wrap: 72
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  error = FALSE,
  collapse = TRUE,
  comment = "#>"
)
```

# Introduction

This vignette demonstrates data import and preprocessing workflow for NMR-based 
metabolomics with **metabom8**. 

```{r get-data}
library(metabom8)
```

# Importing spectral data

In routine high-throughput NMR workflows, raw time-domain data (the
free-induction decay) are typically processed directly on the
spectrometer using the vendor software TopSpin. Downstream metabolomics
analysis therefore commonly starts from frequency-domain spectra. Consequently, 
this vignette focuses on importing spectra generated by TopSpin.

For completeness and teaching purposes, the section
([Starting point: raw FID](#import-fid)) briefly illustrates importing and 
transforming time domain data to obtain spectra.


## TopSpin-processed spectra {#import-spectra}

The function `read1d_proc()` imports TopSpin-processed 1D NMR
spectra from the `pdata` subdirectory of Bruker experiment folders
(typically `pdata/1`). It reads the processed absorption-mode spectrum
(e.g. `1r`) together with associated acquisition (`acqus`) and processing
(`procs`) parameters.

At minimum two arguments are required:

-   `path`: path to the directory that encloses NMR experiments
-   `exp_type`: list of spectrometer parameters to filter desired
    experiment types based on acquisition and processing metadata.
    
Further details on supported filtering options are provided in `?read1d_proc`.

```{r import, fig.width=7.2, fig.height=4}
## Example: import 1D spectra
exp_dir <- system.file("extdata", package = "metabom8")
exp_type <- list(pulprog="noesygppr1d")

nmr_data <- read1d_proc(exp_dir, exp_type, n_max = 100)

names(nmr_data)
```

Both import functions (`read1d_proc()` and `read1d_raw()`) return a named
list containing three core components used throughout this vignette:
the spectral matrix (`X`), the chemical-shift axis (`ppm`), and associated
metadata (`meta`). 

For interactive exploration, setting `to_global = TRUE` assigns the
components `X`, `ppm`, and `meta` directly to the global environment
(`.GlobalEnv`). Existing objects with these names will be overwritten.

The row names of `X` and `meta` correspond to experiment directories and can be
used to join external sample annotations. The `meta` data frame
contains full TopSpin acquisition and processing parameters for each
spectrum.

Example acquisition and processing parameters as they appear in `meta`:

```{r pars_table, echo=FALSE}
knitr::kable(
  data.frame(
    Prefix = c("a_", "p_"),
    Meaning = c("Acquisition parameters",
                "Processing parameters"),
    Examples = c("a_PULPROG, a_SFO1, a_RG, a_SW_h, a_TE",
                 "p_WDW, p_LB, p_SI, p_PHC*, p_BC*, p_NC_proc")
  ),
  align = "lll"
)
```

See `?read1d_proc()` for further details.

## Raw FID data {#import-fid}

For completeness, `metabom8` also provides functionality to import and
process raw FIDs. This allows users to reproduce basic spectral processing
steps within R or to explore how different processing parameters influence
the resulting spectra.

Processing an FID converts the time-domain signal into a frequency-domain
spectrum. Typical processing steps include digital-filter correction 
(group-delay), apodisation (windowing), zero-filling, Fourier 
transformation, phase correction.

The function `read1d_raw()` performs these operations:

```{r read-in-raw, fig.show='hold'}
# import 1D NMR data
nmr_raw <- read1d_raw(
  exp_dir,
  exp_type,
  apodisation = list(fun = "exponential", lb = 0.2),
  zerofil = 1L,
  mode = "absorption"
  )

names(nmr_raw)
```

The following examples illustrate several commonly used apodisation
profiles implemented internally in `metabom8`.

```{r apod, eval=TRUE, fig.width=8}
# selected FID apodisation functions
f_apod <- c("sine","cosine","exponential","sem")

# create toy fid
n <- 200; t <- seq(0,1,len=n)
fid <- exp(-5*t)*cos(20*pi*t)

# apply apodisation
A  <- sapply(f_apod, \(f) metabom8:::.fidApodisationFct(n, list(fun=f, lb=-0.2)))
Fp <- sweep(A, 1, fid, `*`)
S  <- apply(Fp, 2, \(x) Mod(fft(x))[1:(n/2)])

# compare graphically
cols <- 1:ncol(A)

par(mfrow=c(2,2), mar=c(3,3,2,1))
plot(t, fid, type="l", lwd=2, main="FID")
matplot(t, A, type="l", lwd=2, col=cols, main="Apodisation windows")
matplot(t, Fp, type="l", lwd=2, col=cols, main="Windowed FID")
matplot(S, type="l", lwd=2, col=cols, main="Spectra")

legend("topright", f_apod, col=cols, lty=1, bty="n", cex=.8)
```

## Visual inspection

Spectra can be visualised with the function `plot_spec()`. The minimum
set of arguments include the spectral data `X` (array or matrix) and the
chemical shift vector `ppm`.

Three plotting backends are available: `"plotly"` (default), `"base"`, and
`"ggplot2"`. The `"plotly"` backend is set up to use WebGL, enabling
smooth interactive graphics even when the number of data spectra is
large. The `"ggplot2"` backend is not iteractive and renders much slower. 
However, the returned object integrates with the extensive `"ggplot2"` ecosystem 
and can be further customised to generate publication-quality figures.

See `?plot_spec` for additional details.

```{r viz_covid, fig.width=7.2, fig.height=4}
data("covid", package = "metabom8")

## plot directly from the dataset object
plot_spec(covid, shift = c(0.5, 2))

## alternatively extract data components: X, ppm
X <- covid$X
ppm <- covid$ppm

dim(X)
head(ppm)

plot_spec(X[1:3, ], ppm, shift = c(0.5, 2)) # backend='plotly' by default
plot_spec(X[1:3, ], ppm, shift = c(0.5, 2), backend='base')
plot_spec(X[1:3, ], ppm, shift = c(0.5, 2), backend='ggplot2')
```

# Preprocessing

`metabom8` provides a modular preprocessing API for 1D NMR spectra. Each
preprocessing function is composable and designed to support both
interactive exploration and reproducible pipelines.

For a list of preprocessing functions call `list_preprocessing()` from
the R console.

## Pipeline-style usage

When operating on a `metabom8`-style dataset object, preprocessing functions return
the same structured object: a named list containing 
the elements `X`, `ppm`, and `meta`. This allows steps to be chained using the
base R pipe operator (`|>`):

```{r preproc_pipe}
data("hiit_raw", package = "metabom8") # urine NMR (HIIT exercise experiment)

names(hiit_raw) # X, ppm, meta

# piped preprocessing 
hiit_proc <- hiit_raw |> 
  calibrate(type = "tsp") |>
  excise() |>
  correct_baseline(method='asls') |>
  align_spectra() |>
  pqn()

dim(hiit_proc$X)
```

This API design enables clear, declarative workflows suitable for production 
analyses.



```{r viz-raw-hiit, fig.width=7.2, fig.height=4}
plot_spec(hiit_raw, shift = c(4,4.1))
```

```{r viz-proc-hiit, fig.width=7.2, fig.height=4}
plot_spec(hiit_proc, shift = c(4,4.1))
```

## Stepwise execution

Each operation can also be applied independently to a spectral matrix
and ppm vector. This explicit alternative is useful for pipeline development,
parameter tuning and visual inspection.

```{r preproc_stepwise, fig.width=7.2, fig.height=4}
X <- hiit_raw$X
ppm <- hiit_raw$ppm

# perform TSP calibration
X_cal <- calibrate(X, ppm, type='tsp')

# excise chemical shift regions
regions <- list(
      upfield_noise   = c(min(ppm), 0.25),
      residual_water  = c(4.5, 5.2),
      downfield_noise = c(9.7, max(ppm))
    )
X_exi <- excise(X_cal, ppm, regions)
ppm_exi <- attr(X_exi, 'm8_axis')$ppm

# baseline correction
X_bli <- correct_baseline(X_exi)

# plot spectrum
plot_spec(X_bli[1,], ppm_exi, shift = c(0,10))
```

## Provenance

Data provenance refers to the documented record of how data were generated,
processed, and transformed. Recording this information ensures that analyses
remain reproducible and that individual preprocessing steps can be inspected,
verified, or repeated at a later stage. 

In `metabom8`, particular emphasis is placed on transparent preprocessing, as 
these decisions can substantially influence downstream statistical analyses and 
biological interpretation. Each preprocessing operation automatically records 
its parameters and execution context as structured metadata attached to the 
spectral matrix `X`.

The resulting dataset therefore remains self-describing: the complete processing 
history is linked to the data matrix and can be inspected programmatically at 
any time.

```{r preproc_provenance}

## Inspect preprocessing steps
print_provenance(hiit_proc)

# Retrieve specific step
get_provenance(hiit_proc, step = 2)

## Retrieve the full provenance log
log <- get_provenance(hiit_proc)
# print(log) # very detailed
```

Each entry stores: 

* sequential execution index
* preprocessing step name 
* parameter settings
* optional notes
* timestamp and versioning info

Individual parameter values can be retrieved as well, allowing the full transformation 
history of a dataset to be inspected programmatically.

```{r preproc_provenance-indi}
# Retrieve parameter from a named step
get_provenance(hiit_proc, step = "calibrate", param = "target")
```

Custom notes can also be included in the provenance log. This is particularly userful
when data are processed in automated fashion, for example when using workflow managers:

```{r provenance_note}
# collect env varaibles
params <- list(
  agent    = paste0("snakemake/", Sys.getenv("SNAKEMAKE_VERSION", "v?")),
  runtime  = "docker",
  workflow = Sys.getenv("M8_WORKFLOW", "std_prof-urine"),
  run_id   = Sys.getenv("M8_RUN_ID", "m8-2605-001")
)
 
# create note with env parameters
hiit_proc <- add_note(hiit_proc, 
         'Excluding outlier sample XX - reason is ...', 
         params)

# check provenance out
get_provenance(hiit_proc , step = "user note")
```


> **Note on provenance**  
`metabom8` records preprocessing history as structured attributes on
standard R matrices or named lists. This design is lightweigth and prioritises
interoperability and flexibility over strict container classes
(e.g., `SummarizedExperiment`). Provenance is preserved when objects are saved/loaded
based on standard R serialisation (`save()`/`saveRDS()`, `load()`), but may be lost 
if the object is modified by functions outside the `metabom8`
API that drop or overwrite attributes (for example when objects are
coerced, reconstructed, or exported to external formats).


# Modelling

The section demonstrates data modelling using PCA (unsupervised) and PLS/OPLS (supervised).

All modelling calls return an instance of class `m8_model`, which can be used 
for further processing and diagnostics.


```{r mva-data}
Y <- covid$an$type
X <- covid$X
```

To prevent the natural dynamic range of metabolite concentrations from 
dominating statistical modelling, data are scaled prior to analysis. In 
**metabom8**, scaling and centering are passed explicitly to the model function:

```{r model_setup}
# Define the scaling and centering strategy
uv = uv_scaling(center = TRUE)
```

See `?uv_scaling` for further details and additional scaling strategies.



## Principal Components Analysis (PCA)

Unsupervised modelling with PCA is done by an iterative procedure
(NIPALS) that calculates components iteratively, one at a time. Therefore, the
number of desired components is provided in the function call.

Here we compute two principal components using unit-variance scaling:

```{r pca}
pca_model <- pca(X, scaling = uv, ncomp=2)

# show / summary methods 
show(pca_model)
summary(pca_model)
```

## Partial-Least Squares (PLS)
Partial-Least-Squares (PLS) as supervised modelling technique utilises 
a statistical re-sampling techniques to determine the optimal number of components 
and to safeguard against overfitting.

In this vignette, we employ balanced Monte Carlo cross-validation (`balanced_mc`). 
This process repeatedly splits the data into training and testing sets to ensure the 
model predictive power generalizes well to unseen data.

```{r resampling}
mc_cv <- balanced_mc(k=15, split=2/3, type="DA")
```

See `?balanced_mc` for further details and additional statistical validation strategies.

```{r pls}
pls_model <- pls(X, Y, validation_strategy=mc_cv, scaling = uv)

# model information
show(pls_model)

# model performance summary
summary(pls_model)

# model scores & loadings
Tp <- scores(pls_model)
Pp <- loadings(pls_model)
```

## Orthogonal-Partial-Least Squares (OPLS)

OPLS can be though of as PLS with an integrated orthogonality filter, which has 
been reported to improve model interpretability. While standard 
PLS mixes predictive and non-predictive variation, OPLS partitions the variance 
into two parts: variation correlated to the response ($Y$) and variation "orthogonal" 
(unrelated) to it.

Similar to PLS, the OPLS algorithm evaluates model performance 
using statistical re-sampling techniques to ensure the results are robust and 
not driven by noise.

```{r opls}
opls_model <- opls(X, Y, validation_strategy=mc_cv, scaling = uv)

# model information & performance
show(opls_model)
```
The OPLS algorithm fitted one predictive and one orthogonal component.
A second orthogonal component was evaluated but the improvement in
cross-validated class prediction (`AUCv`)based on the left-out sets was
negligible (`stop reason`). Therefore, a third component was not fitted to the full
spectral matrix.

```{r opls-summary}
summary(opls_model)
```

For detailed information on component stopping criteria used in `opls()` and`pls()` 
see `?.evalFit()`.

Methods implemented for both, `pls()` and `opls()`, models include `scores()`, 
`loadings()` and `fitted()` (Y-fitted values from statistical resampling).
For OPLS models, these former two functions accept argument `orth = TRUE`
to extract orthogonal component information.

Further OPLS model interpretation can be performed using Variable Importance in 
Projection (`vip()`), which identifies the most influential variables in the model.

```{r opls-t-p}
# model scores, loadings & vip
Tp <- scores(opls_model)
To <- scores(opls_model, orth=TRUE)

Pp <- loadings(opls_model)
Po <- loadings(opls_model, orth=TRUE)

vip <- vip(opls_model)
```


To ensure the model is statistically robust, `metabom8` provides several 
diagnostic tools:

* `dmodx()`: Calculates the "Distance to the Model X" to identify spectral outliers
* `opls_perm()`:  Executes Y-permutations to validate that the observed $Q^2$ is 
significantly higher than what would be expected by chance.
* `cv_anova()`: Performs a Cross-Validated ANOVA to test the significance of the 
model's predictive ability (only for regression)


```{r opls-dx, fig.width=7.2, fig.height=4}
# distance to the model in X space
dex <- dmodx(opls_model)
head(dex)
```

This vignette demonstrated the core data import, preprocessing, and modelling workflow 
implemented in `metabom8`. For further details on individual functions, see the 
corresponding help pages.

# Citation

```{r citation, echo=FALSE, results='asis'}
citation("metabom8")
```

# Session Info

```{r session-info, echo=FALSE} 
sessionInfo()
```

# Contributor Code of Conduct

The contributors and maintainers of this project, pledge to respect all people who 
contribute through reporting issues, posting feature requests, updating documentation,
submitting pull requests or patches, and other activities.

We are committed to making participation in this project a harassment-free experience, 
regardless of level of experience, gender, gender identity and expression,
sexual orientation, disability, personal appearance, body size, race, ethnicity, age, or religion.

Project maintainers have the right and responsibility to remove, edit, or reject comments,
commits, code, wiki edits, issues, and other contributions that are not aligned to this 
Code of Conduct. Project maintainers who do not follow the Code of Conduct may be removed 
from the project team.

Instances of abusive, harassing, or otherwise unacceptable behavior may be reported by 
opening an issue or contacting the project maintainers.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_vip.R
\name{.vip}
\alias{.vip}
\title{Compute Variable Importance in Projection (VIP)}
\usage{
.vip(Tx, W, C)
}
\arguments{
\item{Tx}{Numeric matrix (n x A). Score matrix.}

\item{W}{Numeric matrix (A x p). Weight matrix.
Rows correspond to components, columns to variables.}

\item{C}{Numeric vector or matrix of length A. Y-loadings.}
}
\value{
Numeric vector of length p containing VIP scores.
}
\description{
Calculates VIP scores for a PLS/OPLS model using predictive
components only.
}
\details{
VIP is computed as:

\deqn{
VIP_j = \sqrt{ p \frac{\sum_a SSY_a w_{aj}^2}{\sum_a SSY_a} }
}

where
SSY_a = (t_a^T t_a) * c_a^2

By construction, mean(VIP^2) = 1.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_extract.R
\name{.oplsComponentCv}
\alias{.oplsComponentCv}
\title{OPLS Component Estimation via Cross-Validation}
\usage{
.oplsComponentCv(X, Ycs_fold, cv.set, nc, mod.cv, acc)
}
\arguments{
\item{X}{Numeric matrix. Input data matrix (samples x features); only
required for \code{nc = 1}.}

\item{Ycs_fold}{List containing the response values (\code{Y}) for each
cross-validation split.}

\item{cv.set}{List of integer vectors. Each element contains training indices
for one CV round.
These indices are adjusted internally by -1 due to 0-based indexing in the
underlying C++ (Rcpp) routines.}

\item{nc}{Integer. Component number currently being estimated (\code{nc = 1}
for first orthogonal component).}

\item{mod.cv}{List. Model state to be updated with results from each CV iteration.}
}
\value{
A list of the same length as \code{cv.set}. Each list element contains:
\itemize{
\item \code{t_xo}: Orthogonal component scores (matrix, samples x components).
\item \code{t_xp}: Predictive component scores (matrix, samples x 1).
\item \code{y_pred_train}: Predicted responses for training samples.
\item \code{y_pred_test}: Predicted responses for held-out test samples.
\item \code{x_res}: Residual matrix after filtering out orthogonal structure.
}
}
\description{
Performs orthogonal projections to latent structures (OPLS) modeling on
training folds of a cross-validation (CV) scheme.
The function fits one orthogonal and one predictive component per fold,
tracking scores, predictions, and residuals.
For the first orthogonal component (\code{nc = 1}), the full model structure
is initialized.
For later components (\code{nc > 1}), modeling proceeds on residual matrices
carried over from previous iterations.
}
\note{
Training indices in \code{cv.set} are internally adjusted by -1 before
being passed to Rcpp routines.
This is required because C++ uses zero-based indexing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_cvanova.R
\name{cv_anova}
\alias{cv_anova}
\title{Cross-validated ANOVA for O-PLS models}
\usage{
cv_anova(smod)
}
\arguments{
\item{smod}{An object of class \code{m8_model}, generated by function \pkg{opls} in the \pkg{metabom8} package.}
}
\value{
A \code{data.frame} containing:
\itemize{
\item \code{SS} - Sum of Squares
\item \code{DF} - Degrees of Freedom
\item \code{MS} - Mean Squares
\item \code{F_value} - F statistic
\item \code{p_value} - P-value from the F-test
}
}
\description{
Performs a cross-validated ANOVA (CV-ANOVA) test for OPLS models. The function compares residuals from a null model and a model using cross-validated predictive scores to assess the significance of the OPLS model.
}
\details{
\strong{Interpretation of the CV-ANOVA table}

The CV-ANOVA compares two nested linear models:
\itemize{
\item \emph{Null model}: \eqn{Y = \beta_0 + \epsilon}
\item \emph{Full model}: \eqn{Y = \beta_0 + \beta_1 t_{\text{pred,cv}} + \epsilon}
}

where \eqn{t_{\text{pred,cv}}} represents the cross-validated predictive component
score(s) obtained from the OPLS model.

The ANOVA table contains:
\itemize{
\item \strong{SS} – Sum of Squares
\item \strong{DF} – Degrees of Freedom
\item \strong{MS} – Mean Squares (SS / DF)
\item \strong{F_value} – F statistic comparing model vs. null
\item \strong{p_value} – P-value from the F-test
}

The \emph{Regression} row quantifies the reduction in residual variance
achieved by including the cross-validated predictive score(s).
The \emph{Residual} row represents the unexplained variance of the full model.

The F-statistic is computed as:
\deqn{
F = \frac{(RSS_0 - RSS_1) / df_{reg}}{RSS_1 / df_{res}}
}
where \eqn{RSS_0} and \eqn{RSS_1} are the residual sums of squares of the
null and full models, respectively.

\strong{Predictive interpretation}

A small p-value (typically < 0.05) indicates that the cross-validated
predictive score(s) significantly reduce residual variance compared to an
intercept-only model. This suggests that the OPLS model captures
statistically meaningful predictive structure in \eqn{Y}.

A large p-value indicates that the predictive component does not explain
\eqn{Y} significantly better than chance, implying weak or unstable
predictive performance.

Importantly, CV-ANOVA evaluates the \emph{linear explanatory power of the
cross-validated predictive scores}, not the descriptive separation of the
latent space. A model may show visual class separation or moderate R\eqn{^2}
yet fail CV-ANOVA if predictive performance is weak.

Sample size strongly affects statistical power. With small \eqn{n}, even
models with moderate predictive strength may not reach statistical
significance.

CV-ANOVA is intended for continuous response (regression) models.
}
\examples{
data("covid")

X <- covid$X
Y <- as.numeric(factor(covid$an$type)) - 1

scaling <- uv_scaling(center=TRUE)
cv <- balanced_mc(10, split=2/3, type='R', probs = c(0, 0.5, 1))
mod <- opls(X, Y, scaling, cv)

cv_anova(mod)

}
\references{
Eriksson, L., et al. (2008). CV-ANOVA for significance testing of PLS and OPLS models. \emph{Journal of Chemometrics}, 22(11-12), 594–600.
}
\seealso{
Other model_validation: 
\code{\link{dmodx}()},
\code{\link{opls_perm}()}
}
\concept{model_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_allviz.R
\name{.draw_base}
\alias{.draw_base}
\title{Draw NMR Spectra Using Base Graphics}
\usage{
.draw_base(dat, add, ...)
}
\arguments{
\item{dat}{List returned by \code{.prep_spec()}.}

\item{add}{Logical. Add to existing plot.}

\item{...}{Additional graphical parameters.}
}
\description{
Internal backend renderer for base R plotting.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_alignSegment.R
\name{.alignSegment}
\alias{.alignSegment}
\title{Align NMR Spectra via Cross-Correlation}
\usage{
.alignSegment(
  seg,
  idx_ref = 1,
  clim = 0.7,
  med = TRUE,
  norm = FALSE,
  lag.max = 20
)
}
\arguments{
\item{seg}{Numeric matrix. Each row is a 1D NMR spectrum segment.}

\item{idx_ref}{Integer. Row index to use as reference spectrum. Ignored if
\code{med = TRUE}.}

\item{clim}{Numeric. Minimum cross-correlation threshold. Segments with lower
similarity are not shifted.}

\item{med}{Logical. If \code{TRUE}, the row-wise median spectrum is used as
the reference.}

\item{norm}{Logical. If \code{TRUE}, rows are z-scaled prior to alignment.}
}
\value{
A numeric matrix of aligned spectra with the same dimensions as
\code{seg}. If \code{med = TRUE}, the output excludes the median row.
}
\description{
Aligns rows of an NMR spectral matrix to a reference spectrum by maximizing
cross-correlation. Optionally uses the row-wise median spectrum as reference.
Useful for minor spectral misalignments.
}
\details{
Each spectrum is aligned to the reference by computing the cross-correlation
and applying a lag-based linear interpolated shift. The shift is applied only
if the cross-correlation exceeds \code{clim}. This function is intended for
fine-tuning alignment across short ppm windows (i.e., segments).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_helpers.R
\name{.checkXclassNas}
\alias{.checkXclassNas}
\title{Check X matrix (PLS context)}
\usage{
.checkXclassNas(X)
}
\arguments{
\item{X}{Input data (univariate or multivariate), formatted as a matrix.}
}
\value{
\code{NULL} if all checks pass; otherwise throws an informative error.
}
\description{
Validates that input \code{X} is a numeric matrix and does not contain missing, NaN, or infinite values.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_helpers.R
\name{.dimX}
\alias{.dimX}
\title{Ensure Input is a Row Matrix}
\usage{
.dimX(X)
}
\arguments{
\item{X}{Numeric vector, matrix, or array. Typically NMR data where rows represent spectra.}
}
\value{
A numeric matrix with one row if input was a vector; otherwise, returns \code{X} unchanged.
}
\description{
Ensures that the input object \code{X} is in row-matrix form. If \code{X} is a numeric
vector  (i.e., has no dimensions), it is converted to a matrix with 1 row
and \code{length(X)} columns. This is helpful for standardizing inputs before
applying matrix operations.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_helpers.R
\name{.checkYclassNas}
\alias{.checkYclassNas}
\title{Check Y for missing values (PLS context)}
\usage{
.checkYclassNas(Y)
}
\arguments{
\item{Y}{A vector, matrix, or data.frame. Target variable.}
}
\value{
A list: cleaned Y matrix, levels (if applicable), and type ('R' or 'DA')
}
\description{
Checks input Y for NA/NaN/Inf values, determines analysis type
(Regression or Discriminant Analysis), and returns cleaned Y matrix,
associated levels, and type.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_sets_methods.R
\name{.cvSetsMethod}
\alias{.cvSetsMethod}
\title{Generate Cross-Validation (CV) Training Set Indices}
\usage{
.cvSetsMethod(
  Y,
  type,
  method = "k-fold_stratified",
  k = 7,
  split = 2/3,
  probs = NULL
)
}
\arguments{
\item{Y}{matrix or data.frame. Outcome matrix. If multi-column, only the first column will be used.}

\item{type}{character. Type of analysis: "R" for regression or "DA" for discriminant analysis. Use suffix "-mY" to indicate multi-column Y.}

\item{method}{character. One of: \code{"k-fold"}, \code{"k-fold_stratified"}, \code{"MC"}, \code{"MC_balanced"}.}

\item{k}{integer. Number of CV iterations/folds.}

\item{split}{numeric. For Monte Carlo methods, the fraction of samples to use in the training set (0 < split < 1).}

\item{probs}{Numeric vector. Probabilities used for stratification of numeric \code{Y} (regression only).}
}
\value{
A list of integer vectors. Each vector contains row indices representing a training set.
}
\description{
Generates a list of row indices representing training sets for various CV strategies including stratified and Monte Carlo methods. Handles both regression (R) and discriminant analysis (DA), and multi-column Y.
}
\details{
\itemize{
\item In \code{"k-fold_stratified"} and \code{"MC_balanced"} modes, stratification is applied to preserve class proportions or outcome distribution.
\item If \code{type} is regression (\code{"R"}), quantile-based binning is applied to stratify continuous Y values.
\item If \code{type} is discriminant analysis (\code{"DA"}), class labels are used directly.
\item If \code{Y} has multiple columns, only the first column is used to create resampling sets.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{covid_raw}
\alias{covid_raw}
\title{COVID-19 blood plasma proton NMR spectra (raw)}
\format{
A numeric matrix/data frame with 23 rows and 29,782 columns:
\describe{
\item{rows}{Spectra (samples)}
\item{columns}{Chemical shift variables in parts per million (ppm)}
}
}
\source{
Australian National Phenome Centre (ANPC), Murdoch University.
}
\description{
1D proton NMR spectra from SARS-CoV-2–positive patients (n = 10) and healthy controls (n = 13),
collected in a research study in Perth, Western Australia. Spectra are raw and require processing
before statistical analysis (see \code{?covid} for processed spectra).
}
\details{
FIDs were acquired using a Carr–Purcell–Meiboom–Gill (CPMG) pulse sequence on a 600 MHz Bruker
Avance II spectrometer using IVDr methods for blood plasma (300 K, 32 scans). The spectrometer was
equipped with a double-resonance broadband (BBI) probe and a refrigerated autosampler at
4\eqn{^{\circ}}C.
}
\examples{
data(covid_raw)
}
\references{
Kimhofer et al. (2020) \doi{10.1021/acs.jproteome.0c00280}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features_ppick.R
\name{ppick}
\alias{ppick}
\title{Find Local Extrema in NMR Spectra (Peak Picking)}
\usage{
ppick(X, ppm, fil_p = 3, fil_n = 5, type = "max")
}
\arguments{
\item{X}{Numeric matrix. NMR data with spectra in rows and chemical shifts in columns.}

\item{ppm}{Numeric vector. Chemical shift values corresponding to columns in \code{X}.}

\item{fil_p}{Integer. Polynomial order of the Savitzky–Golay filter.}

\item{fil_n}{Integer. Filter length (must be odd) of the Savitzky–Golay filter.}

\item{type}{Character. Type of extrema to return: \code{"max"}, \code{"min"}, or \code{"both"}.}
}
\value{
A list of data frames, one per spectrum. Each data frame contains:
\itemize{
\item \code{idc}: Index of the detected peak.
\item \code{ppm}: Chemical shift at the peak.
\item \code{Int}: Intensity at the peak.
\item \code{Etype}: Extrema type: \code{1} for minima, \code{-1} for maxima.
}
}
\description{
Identifies local maxima, minima, or both from smoothed NMR spectra using Savitzky–Golay filtering.
}
\details{
The spectra are smoothed using a Savitzky–Golay filter to reduce noise. Extrema are then detected by identifying sign changes in the first derivative of the smoothed signal.
}
\examples{
data(covid)
X <- covid$X
ppm <- covid$ppm

peaklist <- ppick(X, ppm)
plot_spec(X[1, ], ppm, shift = c(1.2, 1.4), backend='base')
points(peaklist[[1]]$ppm, peaklist[[1]]$Int, col = 'cyan')

}
\seealso{
\code{\link{ppick2}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_sets_methods.R
\name{.mc}
\alias{.mc}
\title{Generate Monte Carlo Cross-Validation (MCCV) Training Indices}
\usage{
.mc(k, Y, split)
}
\arguments{
\item{k}{Integer. Number of MCCV iterations (i.e., training sets to generate).}

\item{Y}{A matrix (n x p), where rows are observations and columns are outcomes. Only the number of rows is used here.}

\item{split}{Numeric. Fraction of the data to include in each training set. Should be between 0 and 1.}
}
\value{
A list of length \code{k}, each containing a vector of training set indices.
}
\description{
Generates a list of training set indices for Monte Carlo Cross-Validation.
}
\details{
For each fold, a random sample (without replacement) of size \code{floor(nrow(Y) * split)} is drawn.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_read1d.R
\name{.extract_pars1d}
\alias{.extract_pars1d}
\title{Read Bruker NMR Parameter Files}
\usage{
.extract_pars1d(f_list)
}
\arguments{
\item{f_list}{List of file paths. Output from \code{.detect1d_procs()}.}
}
\value{
A data frame of extracted acquisition and processing metadata. Row names correspond to spectrum filenames.
}
\description{
Helper function used by \code{read1d()} to extract acquisition (\code{acqus}) and processing (\code{procs}) parameters
from Bruker-formatted 1D NMR experiments.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_alignSegment.R
\name{align_segment}
\alias{align_segment}
\title{Align NMR Spectra in a Selected Shift Region}
\usage{
align_segment(
  dat,
  shift,
  idx_ref = 1,
  med = TRUE,
  clim = 0.7,
  norm = FALSE,
  lag.max = 20
)
}
\arguments{
\item{dat}{Named list with elements:
\describe{
\item{X}{Numeric matrix (spectra in rows)}
\item{ppm}{Numeric vector of chemical shift values}
\item{meta}{Optional metadata}
}}

\item{shift}{Numeric vector of length 2 specifying ppm region to align.}

\item{idx_ref}{Integer. Row index to use as reference spectrum.}

\item{med}{Logical. Use row-wise median spectrum as reference?}

\item{clim}{Numeric. Minimum correlation threshold.}

\item{norm}{Logical. Z-scale before alignment.}

\item{lag.max}{Integer. Maximum lag allowed.}
}
\value{
Updated \code{dat} list with aligned spectra in selected region.
}
\description{
Aligns spectra within a specified ppm window using cross-correlation.
Only the selected region is aligned; the remainder of the spectra
is unchanged.
}
\examples{
data(hiit_raw)
plot_spec(hiit_raw, shift=c(1.3,1.4))
hiit_aligned <- align_segment(hiit_raw, c(1.3, 1.35))
plot_spec(hiit_aligned, shift=c(1.3,1.4)) # aligned segment
plot_spec(hiit_aligned, shift=c(3, 3.1)) # segment not aligned
}
\seealso{
Other preprocessing: 
\code{\link{align_spectra}()},
\code{\link{binning}()},
\code{\link{calibrate}()},
\code{\link{correct_baseline}()},
\code{\link{correct_lw}()},
\code{\link{pqn}()},
\code{\link{print_preprocessing}()}
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_read1d.R
\name{read1d}
\alias{read1d}
\alias{read1d_proc}
\title{Import 1D NMR spectra (TopSpin processed)}
\usage{
read1d(
  path,
  exp_type = list(pulprog = "noesygppr1d"),
  n_max = 1000,
  filter = TRUE,
  recursive = TRUE,
  verbose = 1,
  to_global = FALSE
)

read1d_proc(
  path,
  exp_type = list(pulprog = "noesygppr1d"),
  n_max = 1000,
  filter = TRUE,
  recursive = TRUE,
  verbose = 1,
  to_global = FALSE
)
}
\arguments{
\item{path}{Character. Directory path containing Bruker NMR experiments.}

\item{exp_type}{Named list. Optional filtering specification based on
acquisition or processing metadata. Each list element must correspond to a
metadata field (e.g. \code{pulprog}, \code{ns}, \code{rg}). Filtering supports:
\itemize{
\item Exact match: \code{list(pulprog = "noesygppr1d")}
\item Membership: \code{list(pulprog = c("zg30", "noesygppr1d"))}
\item Numeric range: \code{list(ns = list(range = c(16, 128)))}
\item Generic comparison: \code{list(ns = list(op = ">=", value = 32))}
}
Multiple fields are combined using logical AND.}

\item{n_max}{Integer. Maximum number of spectra to import. Default: 1000.}

\item{filter}{Logical. Filter out experiments with incomplete file systems.}

\item{recursive}{Logical. Search \code{path} recursively. Default: TRUE.}

\item{verbose}{Logical or numeric. Verbosity level.}

\item{to_global}{Logical. If \code{TRUE}, the returned objects are additionally
assigned to the global environment.}
}
\value{
A named list with three elements:
\describe{
\item{X}{A numeric matrix of spectra (rows = samples, columns = ppm values).}
\item{ppm}{A numeric vector of chemical shift values (ppm).}
\item{meta}{A data frame of acquisition and processing metadata,
row-aligned with \code{X}.}
}

If \code{to_global = TRUE}, objects with the same names in the global
environment will be overwritten.
}
\description{
Imports TopSpin-processed 1D NMR spectra together with spectrometer acquisition
and TopSpin processing parameters (\code{acqus} and \code{procs}, respectively).
}
\examples{
path <- system.file("extdata", package = "metabom8")

read1d_proc(path, exp_type = list(pulprog = "noesygppr1d"), n_max = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{covid}
\alias{covid}
\title{COVID-19 blood plasma proton NMR spectra (processed)}
\format{
A numeric matrix/data frame with 23 rows (samples) and 27,819 columns
(chemical shift variables in parts per million, ppm).
}
\source{
Australian National Phenome Centre (ANPC), Murdoch University.
}
\description{
1D proton NMR spectra from SARS-CoV-2–positive patients (n = 10) and healthy controls (n = 13),
collected in Perth, Western Australia. Spectra were pre-processed (residual water and signal-free
regions excised, baseline corrected, and normalized to account for line-width differences).
}
\details{
FIDs were acquired with a standard 90\eqn{^{\circ}} RF pulse sequence on a 600 MHz Bruker Avance II
spectrometer using IVDr methods for blood plasma (300 K, 32 scans). The spectrometer was equipped
with a double-resonance broadband (BBI) probe and a refrigerated autosampler.
}
\examples{
data(covid)
}
\references{
Kimhofer et al. (2020) \doi{10.1021/acs.jproteome.0c00280}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_scalings.R
\name{uv_scaling}
\alias{uv_scaling}
\title{Unit Variance Scaling
This function defines a preprocessing strategy that is applied via
\code{\link{prep_X}}.}
\usage{
uv_scaling(center = TRUE)
}
\arguments{
\item{center}{Logical. If TRUE, variables are mean-centered before scaling.}
}
\value{
A \code{list} with elements:
\describe{
\item{X}{Numeric matrix containing the scaled data.}
\item{prep}{List describing the preprocessing, including centering and scaling
parameters (\code{center}, \code{scale}, \code{X_mean}, \code{X_sd}).}
}
}
\description{
Unit Variance Scaling
This function defines a preprocessing strategy that is applied via
\code{\link{prep_X}}.
}
\details{
Centers variables and scales each feature to unit variance.
UV scaling divides each variable by its standard deviation.
This is the default scaling in many multivariate methods such as PCA and PLS.
}
\examples{
autoscale <- uv_scaling(center=TRUE)
X <- matrix(c(10,10, 0,0, 0, 10), ncol=3)
X_scaled <- prep_X(autoscale, X)
str(X_scaled)
X_scaled$X
}
\seealso{
Other scaling_strategies: 
\code{\link{pareto_scaling}()},
\code{\link{unscaled}()}
}
\concept{scaling_strategies}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/00-s4_classes.R, R/03-s4_methods.R
\docType{class}
\name{m8_model-class}
\alias{m8_model-class}
\alias{m8_model}
\alias{summary,m8_model-method}
\alias{show,m8_model-method}
\title{m8_model class
Model object returned by \code{pca()}, \code{pls()}, and \code{opls()}.}
\usage{
\S4method{summary}{m8_model}(object)

\S4method{show}{m8_model}(object)
}
\arguments{
\item{object}{An object of class \code{m8_model}.}
}
\value{
An object of class \code{m8_model}.
}
\description{
m8_model class
Model object returned by \code{pca()}, \code{pls()}, and \code{opls()}.
}
\section{Functions}{
\itemize{
\item \code{summary(m8_model)}: Summarise model performance and component selection.

\item \code{show(m8_model)}: Show a compact model header.

}}
\section{Slots}{

\describe{
\item{\code{engine}}{Character. Model engine ("pca", "pls", "opls").}

\item{\code{ctrl}}{List. Engine-specific control and performance information.}

\item{\code{fit}}{List. Fitted data (engine specific).}

\item{\code{cv}}{Resampling instance (may be NULL if not used).}

\item{\code{prep}}{Scaling and centering information}

\item{\code{provenance}}{Preprocessing attributes of spectral matrix X}

\item{\code{session}}{R-session information}

\item{\code{call}}{Function call}

\item{\code{dims}}{List with \code{n} and \code{p}.}
}}

\examples{
data(covid)
cv <- balanced_mc(k=5, split=2/3)
scaling <- uv_scaling(center=TRUE)
model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
class(model)
show(model)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_extract.R
\name{.extMeanCvFeat}
\alias{.extMeanCvFeat}
\title{Extract Cross-Validated OPLS Features}
\usage{
.extMeanCvFeat(cv_obj, feat = "t_xp")
}
\arguments{
\item{cv_obj}{Named list. Output of \code{.oplsComponentCv()} for a specific OPLS component.}

\item{feat}{Character. Feature name to extract from each CV fold object (e.g., \code{"t_xp"} or \code{"y_pred_test"}).}
}
\value{
A numeric matrix or array of the requested feature:
\itemize{
\item For Monte Carlo CV and single-column Y: matrix with rows \code{"mean"}, \code{"sd"}, and \code{"coverage"} x samples.
\item For Monte Carlo CV and multi-column Y: 3D array with shape \code{[3, samples, outcomes]}.
\item For k-fold CV: matrix or array containing values for each sample (no aggregation, just fold values).
}
}
\description{
Extracts a specific data structure (e.g., scores or predictions) from cross-validation output returned by \code{.oplsComponentCv()}.
Depending on the cross-validation type and model structure, the output is aggregated across folds by computing the mean,
standard deviation (SD), and coverage.

Here, \strong{Coverage} refers to the proportion of cross-validation folds in which a given sample's value was observed
(i.e., not missing), serving as an estimate of how frequently that value was evaluated across folds.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_sets_methods.R
\name{balanced_mc}
\alias{balanced_mc}
\title{Balanced Monte-Carlo resampling strategy}
\usage{
balanced_mc(k, split, type = c("DA", "R"), probs = NULL)
}
\arguments{
\item{k}{Integer. Number of repeated random splits.}

\item{split}{Numeric. Fraction of samples assigned to the training set
(e.g. \code{2/3}).}

\item{type}{Character. Either \code{"DA"} (classification) or \code{"R"} (regression).}

\item{probs}{Numeric vector of quantile probabilities used to stratify
continuous \code{Y} when \code{type = "R"}.}
}
\value{
A named \code{list} with elements:
\describe{
\item{train}{List of integer vectors containing training set indices
for each resampling iteration.}
\item{strategy}{Character string indicating the resampling strategy.}
\item{n}{Integer. Number of samples in the dataset.}
\item{seed}{Integer. Random seed used to generate the resampling splits,
ensuring reproducibility.}
}
}
\description{
Balanced Monte-Carlo resampling strategy
}
\details{
Generates \code{k} Monte-Carlo resampling splits by randomly partitioning
the data into training and test sets without replacement.

Balancing ensures equal representation of strata in the training data:
\describe{
\item{\code{type = "DA"}}{
Class labels define the strata, and sampling is balanced across classes.
}
\item{\code{type = "R"}}{
The response is discretised into bins using quantiles defined by
\code{probs}, and each bin contributes equally to the training set.
}
}

This strategy can improve robustness of model evaluation in settings with
limited samples size and imbalanced or unevenly distributed outcome variables.
}
\examples{
n <- 100
# bivariate outcome
thr <- 1.5
Y <- c(rnorm(80, thr-3, 0.3), rnorm(20, thr+3, 0.3))  # unbalanced low/high outcome
mean(Y>thr)

cv_k <- kfold(k = 10)
cv_mc <- balanced_mc(k = 10, split = 2/3, type = "R", probs = c(0, 0.8, 1))

k_inst <- metabom8:::.arg_check_cv(cv_pars=cv_k, model_type='R', n=n, Y_prepped=cbind(Y))
mc_inst <- metabom8:::.arg_check_cv(cv_pars=cv_mc, model_type='R', n=n, Y_prepped=cbind(Y))

# balanced splits: proportion above global median stays ~0.5
q80 <- quantile(Y, 0.8)
round(sapply(k_inst$train, function(i) mean(Y[i] > q80)), 2) # resembles original Y distr.
round(sapply(mc_inst$train, function(i) mean(Y[i] > q80)), 2) # balanced strata (low/high)
}
\seealso{
Other resampling strategies: 
\code{\link{balanced_boot}()},
\code{\link{kfold}()},
\code{\link{mc}()},
\code{\link{stratified_kfold}()}
}
\concept{resampling strategies}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_indexing.R
\name{get_idx}
\alias{get_idx}
\alias{get.idx}
\title{Select Indices for a Chemical Shift Region}
\usage{
get_idx(range = c(1, 5), ppm)

get.idx(range = c(1, 5), ppm)
}
\arguments{
\item{range}{Numeric vector of length 2 specifying lower and upper bounds
(order does not matter).}

\item{ppm}{Numeric vector. The full chemical shift axis (in ppm).}
}
\value{
Integer vector of indices corresponding to \code{ppm} values within the given range.
}
\description{
Returns the indices of the chemical shift vector (\code{ppm}) that fall
within the specified range.
}
\examples{
data(covid_raw)
X <- covid_raw$X
ppm <- covid_raw$ppm
idx_tsp <- get_idx(c(-0.1, 0.1), ppm)
ppm[range(idx_tsp)]
plot(ppm[idx_tsp], X[1, idx_tsp], type = 'l')

}
\seealso{
Other NMR: 
\code{\link{.dynamicIntervalsMedian}()}
}
\concept{NMR}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_norm_eretic.R
\name{norm_eretic}
\alias{norm_eretic}
\title{Normalise Spectra Using ERETIC Signal}
\usage{
norm_eretic(
  X,
  integr = FALSE,
  ppm = NULL,
  pos = NULL,
  search = c(10, 16),
  width = 0.2,
  warn_absent = TRUE,
  noise_win = c(9.8, 10.2),
  snr_factor = 10
)
}
\arguments{
\item{X}{Numeric matrix or vector. NMR spectra with spectra in rows.
If \code{ppm} is not provided, it is inferred from \code{colnames(X)}.}

\item{integr}{Logical. If \code{TRUE}, returns the ERETIC integral per spectrum.
If \code{FALSE}, returns the normalised spectra.}

\item{ppm}{Numeric vector of chemical shift positions. If \code{NULL}, inferred from \code{colnames(X)}.}

\item{pos}{Numeric scalar or \code{NULL}. If \code{NULL} (default), ERETIC position is discovered
within \code{search}. If provided, the ERETIC window is centered on \code{pos}.}

\item{search}{Numeric vector of length 2. Ppm window used to discover ERETIC when \code{pos = NULL}.}

\item{width}{Numeric scalar. Full integration window width in ppm (default 0.2 gives \code{pos ± 0.1}).}

\item{warn_absent}{Logical. If \code{TRUE}, warn when ERETIC integral is zero/invalid for individual spectra.}

\item{noise_win}{Numeric vector of length 2. Ppm window for estimating noise intensity.}

\item{snr_factor}{Numeric scaler. Minimum Signal-to-Noise Ratio required for detecting the ERETIC peak.}
}
\value{
If \code{integr = TRUE}, numeric vector of ERETIC integrals.
If \code{integr = FALSE}, numeric matrix of normalised spectra.
}
\description{
Normalises 1D NMR spectra using an ERETIC reference signal. By default
the ERETIC position is discovered in a search window and the spectra are
normalised by the integral in a narrow window around the detected position.
Power users can supply \code{pos} to enforce a fixed ERETIC position.
}
\details{
Binning/normalisation history is recorded in \code{attr(X, "m8_prep")} if present.
Ppm axis values are stored in \code{attr(X, "m8_axis")$ppm}.
}
\examples{
set.seed(123)

n <- 1000
ppm <- seq(0, 14, length.out = n)

gauss <- function(x, c, h, w = 0.05) {
  h * exp(-((x - c)^2) / (2 * w^2))
}

heights <- c(100, 80, 60, 40, 20)

spectra <- sapply(heights, function(h)
  rnorm(n, 0, 0.01) + gauss(ppm, 12, h)
)

spectra <- t(spectra)   # 5 spectra × 1000 variables

plot_spec(spectra, ppm, shift = c(11, 13))

X_norm <- norm_eretic(spectra, ppm=ppm)
plot_spec(X_norm, ppm, shift=c(11, 13))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01-s4_generics.R, R/03-s4_methods.R
\name{weights}
\alias{weights}
\alias{weights,m8_model-method}
\title{Extract model weights}
\usage{
weights(object, ...)

\S4method{weights}{m8_model}(object, orth = FALSE)
}
\arguments{
\item{object}{An object of class \code{m8_model}.}

\item{...}{Additional arguments (currently ignored).}

\item{orth}{Logical indicating whether orthogonal scores should be returned
(only applicable for OPLS models).}
}
\value{
Numeric vector or matrix containing model weights.
}
\description{
Extract model weights
}
\examples{
data(covid)
cv <- balanced_mc(k=5, split=2/3)
scaling <- uv_scaling(center=TRUE)
model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
show(model)
W <- weights(model)
Wo <- weights(model, orth = TRUE)
dim(W)
dim(Wo) == dim(W)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_allviz.R
\name{.draw_plotly}
\alias{.draw_plotly}
\title{Draw NMR Spectra Using Plotly}
\usage{
.draw_plotly(dat, col = NULL, ...)
}
\arguments{
\item{dat}{List returned by \code{.prep_spec()}.}

\item{...}{Additional arguments passed to \code{plotly}.}
}
\description{
Internal backend renderer for interactive plotting.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_binning.R
\name{binning}
\alias{binning}
\title{Spectral data binning}
\usage{
binning(X, ppm = NULL, width = NULL, npoints = NULL, fun = sum)
}
\arguments{
\item{X}{Numeric matrix or data frame with spectra in rows, or a named list as
returned by \code{\link{read1d}}/\code{read1d_proc} containing \code{X} and \code{ppm}.}

\item{ppm}{Numeric vector of chemical shift positions (length must match \code{ncol(X)}).
If \code{NULL}, \code{ppm} is inferred in the following order:
\enumerate{
\item \code{X$ppm} if \code{X} is a list input,
\item \code{attr(X, "m8_axis")$ppm} (if present),
\item numeric \code{colnames(X)} (if present).
}}

\item{width}{Numeric. Bin size in ppm, or \code{NULL} if \code{npoints} is specified.}

\item{npoints}{Integer. Desired number of bins per spectrum, or \code{NULL} if \code{width} is specified.
If both are provided, \code{npoints} is used.}

\item{fun}{Function. Summary function applied to intensities within each bin.
Must return a single numeric value (e.g. \code{sum}, \code{mean}, \code{max}).}
}
\value{
Numeric matrix with spectra in rows and binned ppm variables in columns.
}
\description{
Equidistant binning of spectra by summarising intensities within ppm bins.
}
\details{
If present, preprocessing provenance is appended to \code{attr(X, "m8_prep")} using
\code{.m8_stamp()}. The ppm axis is updated in \code{attr(X, "m8_axis")$ppm} and
column names are set to the bin centres.

When \code{width} is specified, spectra are interpolated onto a regular ppm grid and
then aggregated within bins (\code{interp = TRUE} in provenance). When \code{npoints}
is specified, aggregation is performed by index bins on the original grid
(\code{interp = FALSE} in provenance).
}
\examples{
set.seed(1)
X <- matrix(rnorm(2 * 100), nrow = 2)
ppm <- round(seq(10, 0.5, length.out = 100), 3)
colnames(X) <- ppm

Xb <- binning(X, ppm, width = 0.5)
Xb_mean <- binning(X, ppm, width = 0.5, fun = mean)
dim(Xb)

}
\seealso{
Other preprocessing: 
\code{\link{align_segment}()},
\code{\link{align_spectra}()},
\code{\link{calibrate}()},
\code{\link{correct_baseline}()},
\code{\link{correct_lw}()},
\code{\link{pqn}()},
\code{\link{print_preprocessing}()}
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_sets_methods.R
\name{.kFold}
\alias{.kFold}
\title{k-fold cross-validation index generator}
\usage{
.kFold(k, Y)
}
\arguments{
\item{k}{Integer. Number of CV folds. If \code{k} is not valid or too high relative to number of rows in Y,
it defaults to leave-one-out CV (LOO-CV).}

\item{Y}{Matrix. Outcome matrix (observations x variables).}
}
\value{
A list of length \code{k}, where each element contains the training indices (integers)
for one CV fold.
}
\description{
Creates a list of training set indices for k-fold cross-validation (CV),
without considering class balance in Y. Each element of the list represents
the row indices used for training in one CV fold.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_sets_methods.R
\name{stratified_kfold}
\alias{stratified_kfold}
\title{Y-stratified k-fold cross-validation strategy}
\usage{
stratified_kfold(k, type = c("DA", "R"), probs = NULL)
}
\arguments{
\item{k}{Integer. Number of folds.}

\item{type}{Character. Either \code{"DA"} (classification)
or \code{"R"} (regression).}

\item{probs}{Numeric vector of quantile probabilities used
to stratify continuous \code{Y} when \code{type = "R"}.}
}
\value{
A named \code{list} with elements:
\describe{
\item{train}{List of integer vectors containing training set indices
for each resampling iteration.}
\item{strategy}{Character string indicating the resampling strategy.}
\item{n}{Integer. Number of samples in the dataset.}
\item{seed}{Integer. Random seed used to generate the resampling splits,
ensuring reproducibility.}
}
}
\description{
Y-stratified k-fold cross-validation strategy
}
\details{
For classification (\code{type = "DA"}), folds are generated such that
class proportions are approximately preserved in each fold.

For regression (\code{type = "R"}), \code{Y} is discretised into bins
defined by \code{probs}, and folds are generated to approximately
preserve the bin distribution.
}
\examples{
set.seed(1)
n <- 100
thr <- 1.5
Y <- c(rnorm(80, thr - 3, 0.3), rnorm(20, thr + 3, 0.3))  # unbalanced outcome
mean(Y > thr)

q80 <- quantile(Y, 0.8)  # defines the rare "high" stratum (top 20\%)

cv_k  <- kfold(k = 10)
cv_sk  <- stratified_kfold(k = 10, type = "R", probs = c(0, 0.8, 1))

k_inst <-  metabom8:::.arg_check_cv(cv_pars=cv_k, model_type='R', n=n, Y_prepped=cbind(Y))
sk_inst <-  metabom8:::.arg_check_cv(cv_pars=cv_sk, model_type='R', n=n, Y_prepped=cbind(Y))

round(sapply(k_inst$train,  function(i) mean(Y[i] > q80)), 2)  # reflects imbalance
round(sapply(sk_inst$train, function(i) mean(Y[i] > q80)), 2)  # balanced across strata
}
\seealso{
Other resampling strategies: 
\code{\link{balanced_boot}()},
\code{\link{balanced_mc}()},
\code{\link{kfold}()},
\code{\link{mc}()}
}
\concept{resampling strategies}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_provenance.R
\name{list_preprocessing}
\alias{list_preprocessing}
\title{List available preprocessing steps}
\usage{
list_preprocessing()
}
\value{
A named character vector with preprocessing function
identifiers as names and short descriptions as values.
}
\description{
Returns a named character vector describing the preprocessing
operations implemented in \pkg{metabom8}. For more information
on individual functionalities please refer to the function help
pages.
}
\examples{
list_preprocessing()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_opls_model_perm.R
\name{opls_perm}
\alias{opls_perm}
\title{OPLS Model Validation via Y-Permutation}
\usage{
opls_perm(smod, n = 10, plot = TRUE, mc = FALSE)
}
\arguments{
\item{smod}{An OPLS model object of class \code{OPLS_metabom8}.}

\item{n}{Integer. Number of permutations to perform.}

\item{plot}{Logical. If \code{TRUE}, generates a visual summary of permutation statistics.}

\item{mc}{Logical. If \code{TRUE}, enables multicore processing (currently not implemented).}
}
\value{
A \code{data.frame} with model metrics (e.g., R2, Q2, AUROC) from both permuted and original models.
}
\description{
Performs Y-permutation tests to assess the robustness of OPLS models by comparing model performance statistics on real vs. permuted response labels.
}
\details{
Each permutation shuffles the response labels and fits a new OPLS model. The function captures model statistics (R2, Q2, AUC) to compare against the non-permuted model. This helps determine whether the original model performance is better than expected by chance.
}
\references{
Wiklund, S. et al. (2008). A Tool for Improved Validation of OPLS Models. \emph{Journal of Chemometrics}, 22(11–12), 594–600.
}
\seealso{
Other model_validation: 
\code{\link{cv_anova}()},
\code{\link{dmodx}()}
}
\concept{model_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_fit_pca.R
\name{pca}
\alias{pca}
\title{Principal Component Analysis (PCA)}
\usage{
pca(X, scaling, ncomp, method = "nipals")
}
\arguments{
\item{X}{Numeric matrix (rows = samples, columns = variables).}

\item{scaling}{A \code{ScalingStrategy} object defining model-internal
centering and scaling (e.g. \code{uv_scaling(center = TRUE)}).}

\item{ncomp}{Integer. Number of principal components to compute.}

\item{method}{Character. PCA backend. Use \code{"nipals"} for the
internal implementation or any method supported by
\code{pcaMethods::pca()}.}
}
\value{
An object of class \code{m8_model} with \code{engine = "pca"}.
The object contains:
\itemize{
\item \code{fit$t}: Score matrix (samples × components)
\item \code{fit$p}: Loading matrix (components × variables)
\item \code{ctrl}: Model control information (variance explained, scaling settings)
\item \code{provenance}: Attributes inherited from the input matrix
}
}
\description{
Fits an unsupervised Principal Component Analysis (PCA) model
to a spectral data matrix. Model-internal centering and scaling
are controlled via a \code{ScalingStrategy} object and do not
modify the input matrix.

The default backend uses an internal NIPALS implementation.
Alternatively, PCA algorithms from the \pkg{pcaMethods} package
(e.g. \code{"ppca"}, \code{"svd"}, \code{"robustPca"}) can be used.
}
\details{
PCA decomposes \eqn{X} into orthogonal score and loading matrices:
\deqn{X \approx T P}
where:
\itemize{
\item \eqn{T} contains the principal component scores
\item \eqn{P} contains the loadings
}

The number of components is fixed by \code{ncomp}.
Unlike supervised models, PCA does not use cross-validation
or stopping rules.

Scaling and centering are applied internally during model fitting.
The original input matrix is not modified.
}
\examples{
data(covid)

uv <- uv_scaling(center=TRUE)
model <- pca(X=covid$X, scaling=uv, ncomp=2)

show(model)
summary(model)

Tx <- scores(model)
Px <- loadings(model)

t2 <- hotellingsT2(Tx)
ell <-ellipse2d(t2)

# scores plot
plot(Tx, asp = 1,
  col = as.factor(covid$an$type),
  xlim = range(c(Tx[1,], ell$x)),
  ylim = range(c(Tx[2,], ell$y))
 )
lines(ell$x, ell$y, col = "grey", lty=2)
}
\seealso{
\code{\link{uv_scaling}}

Other modelling: 
\code{\link{opls}()},
\code{\link{pls}()}
}
\concept{modelling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_nmr_qc.R
\name{noise_sd}
\alias{noise_sd}
\title{Estimate Noise Standard Deviation in 1D NMR Spectra}
\usage{
noise_sd(
  X,
  ppm = NULL,
  where = c(14.6, 14.7),
  method = c("mad", "sd", "p95"),
  baseline_correct = FALSE,
  lambda = 10000,
  min_points = 50L,
  ns = NULL,
  normalise_scans = FALSE
)
}
\arguments{
\item{X}{Numeric matrix (spectra in rows), numeric vector (single spectrum), or a named list
as returned by \code{\link{read1d}}/\code{read1d_proc} containing \code{X}, \code{ppm}, and \code{meta}.}

\item{ppm}{Numeric vector of chemical shift values (ppm) corresponding to columns of \code{X}.
If \code{NULL}, \code{ppm} is inferred in the following order:
\enumerate{
\item \code{X$ppm} if \code{X} is a list input,
\item \code{attr(X, "m8_axis")$ppm} (if present),
\item numeric \code{colnames(X)} (if present).
}}

\item{where}{Numeric vector of length 2. ppm range used for noise estimation.
Should be free of metabolite signals (default \code{c(14.6, 14.7)}).}

\item{method}{Character. Noise estimator: \code{"mad"} (default), \code{"sd"}, or \code{"p95"} (legacy amplitude).}

\item{baseline_correct}{Logical. If \code{TRUE}, subtract a smooth baseline in the noise window
using asymmetric least squares (\code{\link[ptw]{asysm}}). Default \code{FALSE}.}

\item{lambda}{Numeric. Smoothing parameter for \code{\link[ptw]{asysm}} when \code{baseline_correct = TRUE}.}

\item{min_points}{Integer. Minimum number of points required in the noise window.
May be a single value (recycled across spectra) or a numeric vector of
length \code{nrow(X)}. If \code{X} is a list input as returned by
\code{\link{read1d}}/\code{read1d_proc} and \code{ns} is \code{NULL},
the function attempts to extract the number of scans from
\code{X$meta$a_NS} when available.}

\item{ns}{Number of scans / transients (required if normalise_scans=TRUE)}

\item{normalise_scans}{Logical. If \code{TRUE}, noise estimates are
multiplied by \eqn{\sqrt{NS}} to account for the theoretical scaling
of noise with the number of scans (\eqn{\sigma \propto 1/\sqrt{NS}}).
This is useful when comparing noise levels across spectra acquired
with different numbers of scans. Default is \code{FALSE}.}
}
\value{
Numeric vector of noise estimates (length \code{nrow(X)}).
}
\description{
Estimates the noise standard deviation (\eqn{\sigma}) for each spectrum from a signal-free ppm region.
The default estimator is robust (MAD-based) and suitable for signal-to-noise calculations.
}
\details{
In NMR spectroscopy, noise scales predictably with the number of scans (\eqn{NS}).
For otherwise identical acquisition settings:

\itemize{
\item Signal increases approximately proportional to \eqn{NS}.
\item Noise increases approximately proportional to \eqn{\sqrt{NS}}.
\item Consequently, signal-to-noise ratio (SNR) increases proportional to \eqn{\sqrt{NS}}.
}

Equivalently, the noise standard deviation scales as:

\deqn{\sigma(NS) \propto \frac{1}{\sqrt{NS}},}

assuming a fixed underlying signal scale and comparable acquisition conditions.

To compare noise levels across datasets acquired with different numbers of scans,
a scan-normalised noise estimate may be used:

\deqn{\sigma_{\mathrm{norm}} = \sigma \cdot \sqrt{NS}.}

Under stable receiver gain and processing conditions, this normalised noise
should be approximately constant across runs.
}
\examples{
data(hiit_raw)
X <- hiit_raw$X
ppm <- hiit_raw$ppm
sigma <- noise_sd(X, ppm, where = c(10,11))
plot(hiit_raw$meta$a_NS, sigma,
  xlab = "Number of scans (NS)",
  ylab = expression(sigma~"(noise estimate)"),
   pch = 16)
 lines(lowess(hiit_raw$meta$a_NS, sigma), col = "red", lwd = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_scalings.R
\name{scRange}
\alias{scRange}
\title{Min-Max Scaling to Arbitrary Range}
\usage{
scRange(x, ra)
}
\arguments{
\item{x}{Numeric vector. Input values to be scaled.}

\item{ra}{Numeric vector of length 2. Desired output range (e.g., \code{c(5, 10)}).}
}
\value{
A numeric vector of the same length as \code{x}, scaled to the range \code{ra}.
}
\description{
Rescales a numeric vector to a specified range using min-max scaling.
This is a generalized form of min-max normalization allowing any output range.
}
\details{
The scaled values are computed as:
\deqn{x_{scaled} = \frac{x - \min(x)}{\max(x) - \min(x)} \cdot (r_{max} - r_{min}) + r_{min}}
}
\examples{
x <- rnorm(20)
plot(x, type = 'l'); abline(h = range(x), lty = 2)
points(scRange(x, ra = c(5, 10)), type = 'l', col = 'red');
abline(h = c(5, 10), col = 'red', lty = 2)

}
\seealso{
\code{\link[=minmax]{minmax()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/03-s4_methods.R
\name{loadings,m8_model-method}
\alias{loadings,m8_model-method}
\title{Model loadings}
\usage{
\S4method{loadings}{m8_model}(x, orth = FALSE, ...)
}
\arguments{
\item{x}{An object of class \code{m8_model}.}

\item{orth}{Logical indicating whether orthogonal scores should be returned
(only applicable for OPLS models).}

\item{...}{Additional arguments (currently ignored).}
}
\value{
Numeric vector or matrix containing loadings.
}
\description{
Model loadings
}
\examples{
data(covid)
cv <- balanced_mc(k=5, split=2/3)
scaling <- uv_scaling(center=TRUE)
model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
show(model)
P <- loadings(model)
Po <- loadings(model, orth = TRUE)
dim(P)
dim(Po) == dim(P)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_excise.R
\name{excise}
\alias{excise}
\title{Excise Chemical Shift Regions from 1D NMR Spectra}
\usage{
excise(x, ppm = NULL, regions = NULL)
}
\arguments{
\item{x}{Numeric matrix or vector. Spectra in rows and variables in columns.}

\item{ppm}{Numeric vector of chemical shift positions (ppm).
If omitted, \code{ppm} is inferred from \code{colnames(X)}.}

\item{regions}{Named list of numeric vectors (length 2), specifying ppm
regions to remove. Each element must define lower and upper bounds.}
}
\value{
Numeric matrix with specified chemical shift regions removed.
}
\description{
Removes specified chemical shift regions from 1D \eqn{^1}H NMR spectra.
By default, commonly excluded metabolomics regions are removed
(upfield/downfield noise, water, urea).
}
\details{
Default regions removed (ppm):
\itemize{
\item Upfield noise: \code{[min(ppm), 0.25]}
\item Residual water: \code{[4.5, 5.2]}
\item Urea region: \code{[5.5, 6.0]}
\item Downfield noise: \code{[9.7, max(ppm)]}
}

Removed regions are recorded in \code{attr(X, "m8_prep")} if present.
Updated ppm values are stored in column names and in
\code{attr(X, "m8_axis")$ppm}.
}
\examples{
set.seed(1)
ppm <- seq(0, 10, length.out = 1000)
X <- matrix(rnorm(100 * length(ppm)), nrow = 100)

Xe <- excise(X, ppm)

dim(Xe)
names(attributes(Xe))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_helpers.R
\name{.filterExp_files}
\alias{.filterExp_files}
\title{Filter Bruker NMR Experiments}
\usage{
.filterExp_files(pars, exp_type, f_list, n_max)
}
\arguments{
\item{pars}{Data frame. Parsed acquisition and processing parameters.}

\item{exp_type}{Named list. Filtering conditions for parameters. Elements may
be character vectors, numeric values/vectors, or operator/range lists for
numeric fields.}

\item{f_list}{List. File paths of spectra and associated metadata (e.g., f_fid, f_1r).}

\item{n_max}{Integer. Maximum number of experiments to retain.}
}
\value{
A list containing filtered and ordered \code{f_list} and \code{pars}.
}
\description{
Filters Bruker 1D experiments based on acquisition metadata and
optionally limits the number of returned entries. Character parameters are
matched by value, while numeric parameters support equality, set inclusion,
range filtering (list(range = c(min, max))), and generic operators
(list(op = ">", value = x)).
}
\seealso{
\code{\link{.extract_acq_pars1d}}, \code{\link{.check1d_files_fid}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_read1d_raw.R
\name{read1d_raw}
\alias{read1d_raw}
\title{Read raw FIDs and process to spectra}
\usage{
read1d_raw(
  path,
  exp_type = list(exp = "PROF_PLASMA_CPMG128_3mm", pulprog = "noesygppr1d"),
  apodisation = list(fun = "exponential", lb = 0.2),
  zerofill = 1L,
  mode = c("absorption", "dispersion", "magnitude"),
  verbose = 1,
  recursive = TRUE,
  n_max = 1000,
  filter = TRUE,
  to_global = FALSE
)
}
\arguments{
\item{path}{Character. Path to the root directory containing Bruker experiment folders.}

\item{exp_type}{Named list. Acquisition-parameter filter to select experiments
(e.g., \code{list(PULPROG = "noesygppr1d")}).}

\item{apodisation}{Named list. Apodisation function and parameters. \code{fun}
must be one of \code{"exponential"}, \code{"cosine"}, \code{"sine"}, \code{"sem"}.}

\item{zerofill}{Integer. Zero-filling exponent (\code{1} doubles points, \code{2} quadruples, …).}

\item{mode}{Character. Spectrum type to return: \code{"absorption"}, \code{"dispersion"}, or \code{"magnitude"}.}

\item{verbose}{Integer. Verbosity level: \code{0} = silent, \code{1} = summary (default),
\code{2} = detailed, \code{3} = debug.}

\item{recursive}{Logical. Recursively search subdirectories for FIDs.}

\item{n_max}{Integer. Maximum number of experiments to process.}

\item{filter}{Logical. Remove experiments with incomplete file structures.}

\item{to_global}{Logical. If TRUE, objects are also assigned to the global
environment; otherwise, only an invisible list is returned.}
}
\value{
A named list with three elements:
\describe{
\item{X}{A numeric matrix of spectra (rows = samples, columns = ppm values).}
\item{ppm}{A numeric vector of chemical-shift axis (ppm).}
\item{meta}{A data frame of acquisition metadata, row-aligned with \code{X}.}
}

If \code{to_global = TRUE}, these objects are also assigned to the global environment.
In that case, any existing objects with the same names will be overwritten.
}
\description{
Reads Bruker 1D NMR FIDs, corrects the digital filter (group delay),
applies apodisation (windowing), optional zero-filling, FFT, phasing and
ppm calibration.

Returns either absorption-, dispersion-, or magnitude-mode
spectra.

If \code{to_global = TRUE}, objects are assigned to the global environment.
}
\details{
FIDs are read from Bruker acquisition folders and processed by the following pipeline:
\enumerate{
\item Digital-filter (group-delay) correction: the initial \emph{n} complex points
are invalid due to the causal DSP decimation filter and are discarded; \emph{n}
equals \code{GRPDLY} when present, or is looked up from Bruker tables indexed by
\code{DECIM} and \code{DSPFVS} on older systems.
\item Apodisation (windowing).
\item Zero-filling (optional).
\item FFT to the frequency domain.
\item Phase correction.
\item PPM calibration (e.g., to TSP).
}
A common ppm scale is then interpolated across spectra.

\strong{Digital filter note:} On newer systems \code{GRPDLY} is written in \code{acqus}/\code{acqu2s}
and should be used directly. For older data sets (\code{GRPDLY < 0} or missing), the group delay
is derived from \code{DECIM} and \code{DSPFVS} via an internal look-up table.

\strong{Apodisation functions}:
\itemize{
\item \code{"uniform"}
\item \code{"cosine"},
\item \code{"sine"},
\item \code{"exponential"} (parameter: \code{lb})
\item \code{"sem"} (sine * exponential, parameter: \code{lb})
\item \code{"gauss"} (parameter: \code{lb}, \code{'gb'}, and \code{'para'})
\item \code{"expGaus_resyG"} (parameter: \code{lb}, \code{'gb'}, and \code{'aq_t'})
}
}
\examples{
path <- system.file("extdata", package = "metabom8")
read1d_raw(
  path,
  exp_type    = list(PULPROG = "noesygppr1d"),
  apodisation = list(fun = "exponential", lb = 0.2),
  zerofill    = 1,
  n_max       = 3
)

}
\seealso{
\code{\link{read1d_proc}} for importing TopSpin-processed spectra
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_storm.R
\name{storm}
\alias{storm}
\title{Subset Optimisation by Reference Matching (STORM)}
\usage{
storm(X, ppm, b = 30, q = 0.05, idx.refSpec, shift)
}
\arguments{
\item{X}{Numeric matrix (or data.frame) of NMR spectra with samples in rows
and spectral variables in columns.}

\item{ppm}{Numeric vector of chemical shift values corresponding to the columns of \code{X}.}

\item{b}{Integer. Half-window size expressed as number of spectral variables
(data points). The effective window width therefore depends on the ppm spacing.}

\item{q}{Numeric. P-value threshold for including spectral variables when
updating the reference region.}

\item{idx.refSpec}{Integer. Row index of \code{X} defining the initial reference spectrum.}

\item{shift}{Numeric vector of length 2 giving the ppm range (minimum, maximum)
defining the initial target signal region.}
}
\value{
Integer vector of row indices of \code{X} defining the selected spectral subset.
}
\description{
Selects an optimal subset of spectra that best match a specified target signal
region, improving downstream correlation-based structural analysis such as STOCSY.
}
\details{
STORM iteratively refines a subset of spectra exhibiting consistent signal
position and multiplicity within a specified ppm region.

Starting from an initial reference spectrum and ppm window:
\enumerate{
\item Spectra positively correlated with the current reference signal are retained.
\item A driver peak (maximum intensity within the reference window) is identified.
\item Correlation and covariance are evaluated within a local window of size \code{b}
around the driver peak.
\item The reference region is updated using variables that satisfy the p-value
threshold (\code{q}) and show positive correlation.
}

The procedure continues until the selected subset stabilises. The resulting
row indices define spectra that most consistently represent the structural
pattern of the target signal.

STORM does not perform metabolite identification directly. Instead, it refines
the dataset to enhance structural coherence prior to correlation-based
interpretation methods such as STOCSY.
}
\examples{
## Simulated example with three Gaussian signals
set.seed(123)

n <- 100          # number of spectra
S <- 1000         # number of spectral variables
ppm <- seq(10, 0, length.out = S)

gauss <- function(x, centre, width, height) {
  height * exp(-((x - centre)^2) / (2 * width^2))
}

## Generate base signals
sig1 <- gauss(ppm, 7,   0.05, 10)
sig2 <- gauss(ppm, 3.5, 0.10,  8)
sig3 <- gauss(ppm, 1,   0.07,  5)

spectra <- matrix(0, n, S)

for (i in seq_len(n)) {
  spectra[i, ] <- sig1 + sig2 + rnorm(S, 0, 0.1)
  if (i <= 25) {
    spectra[i, ] <- spectra[i, ] + sig3
  }
}

## Apply STORM to refine spectra containing the 1 ppm signal
idx <- storm(
  X = spectra,
  ppm = ppm,
  b = 30,
  q = 0.05,
  idx.refSpec = 1,
  shift = c(0.75, 1.25)
)

length(idx)   # number of spectra retained
}
\references{
Posma, J. M., et al. (2012). Subset optimisation by reference matching (STORM):
an optimised statistical approach for recovery of metabolic biomarker structural
information from \eqn{^1H} NMR spectra of biofluids.
\emph{Analytical Chemistry}, 84(24), 10694–10701.
}
\seealso{
Other structural_annotation: 
\code{\link{stocsy}()}
}
\concept{structural_annotation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01-s4_generics.R, R/03-s4_methods.R
\name{vip}
\alias{vip}
\alias{vip,m8_model-method}
\title{Variable Importance in Projection (VIP)}
\usage{
vip(object)

\S4method{vip}{m8_model}(object)
}
\arguments{
\item{object}{An object of class \code{m8_model}.}
}
\value{
Numeric vector or matrix containing the variable importance in
projection (VIP) scores.
}
\description{
Variable Importance in Projection (VIP)
}
\examples{
data(covid)
cv <- balanced_mc(k=5, split=2/3)
scaling <- uv_scaling(center=TRUE)
model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
show(model)
vip_scores <- vip(model)
dim(vip_scores)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_pqn.R
\name{pqn}
\alias{pqn}
\title{Probabilistic Quotient Normalisation (PQN)}
\usage{
pqn(
  X,
  ref_index = NULL,
  total_area = FALSE,
  bin = NULL,
  iref = NULL,
  TArea = NULL
)
}
\arguments{
\item{X}{Numeric matrix or data.frame. Each row is a sample spectrum and each
column is a variable (e.g. chemical shift point or bin).}

\item{ref_index}{Integer vector of row indices used to compute the reference
spectrum. If \code{NULL}, all rows are used.}

\item{total_area}{Logical. If \code{TRUE}, total area normalisation is applied
to the working copy before estimating the PQN dilution factors. See Notes.}

\item{bin}{Optional named list controlling binning for reference estimation,
e.g. \code{list(ppm = ppm, width = 0.05)} or \code{list(ppm = ppm, npoints = 400)}.
If \code{NULL}, no binning is applied.}

\item{iref}{Deprecated. Use \code{ref_index}.}

\item{TArea}{Deprecated. Use \code{total_area}.}
}
\value{
Numeric matrix of PQN-normalised spectra.
}
\description{
Applies probabilistic quotient normalisation (PQN) to spectra. PQN estimates a
sample-specific dilution factor from the median of quotients relative to a
reference spectrum and scales each spectrum accordingly.
}
\details{
\strong{Mechanics.} Let \eqn{x_i} be spectrum \eqn{i} and \eqn{r} a reference spectrum.
PQN computes quotients \eqn{q_{ij} = x_{ij} / r_j} and defines the dilution factor as
\eqn{d_i = 1 / \mathrm{median}_j(q_{ij})}. The PQN-normalised spectrum is
\eqn{x_i^{(PQN)} = d_i \, x_i}.

The reference spectrum \eqn{r} is typically the median spectrum across all samples
or across QC samples (\code{ref_index}).

If \code{bin} is provided, \eqn{r} and dilution factors are computed on binned spectra,
but applied to the original spectra.

Dilution factors are stored in \code{attr(X, "m8_pqn")$dilution_factor}.
}
\section{Notes on total area normalisation}{

Total area normalisation prior to PQN is \emph{usually not recommended}. Total area
scaling removes global intensity differences by enforcing equal total signal per sample.
PQN is itself a global scaling method intended to estimate dilution. Applying both can
substantially change results because PQN no longer estimates dilution alone, but also
compensates compositional distortions introduced by total area scaling.

Situations where \code{total_area = TRUE} can be defensible include:
\itemize{
\item when spectra have large, non-dilution-related amplitude differences caused by
acquisition artefacts (receiver gain / baseline offset) and you explicitly want
to stabilise the reference estimation step;
\item when the measured total signal is expected to be constant by design (e.g. strictly
controlled sample mass/volume and stable overall metabolite pool), and the main
goal is to reduce technical scaling variation before PQN.
}
In most metabolomics settings, prefer PQN without total area scaling.

#' @section On spectral alignment and binning:
PQN assumes that corresponding variables represent the same chemical signal
across spectra. If spectra are not well aligned, small peak shifts can inflate
the variability of pointwise quotients \eqn{x_{ij} / r_j}, leading to unstable
dilution factor estimates.

In such cases, slight binning (e.g. narrow fixed-width bins) prior to reference
estimation is recommended. Binning reduces sensitivity to minor misalignments
by aggregating neighbouring variables. However, excessive binning may obscure
narrow signals and should be avoided.

Alternatively, prior spectral alignment is preferable when available.
}

\examples{
set.seed(1)
ppm <- seq(0, 10, length.out = 1000)
ref <- dnorm(ppm, 3, 0.15) + dnorm(ppm, 6, 0.20) + dnorm(ppm, 7.5, 0.18)
dil <- c(1, 0.8, 0.6, 0.4, 0.2)            # true dilution factors
X <- t(sapply(dil, function(d) d * ref + rnorm(length(ref), 0, 0.005)))
plot_spec(X, ppm)

Xn <- pqn(X, ref_index=1)
dil_est <- attr(Xn, "m8_pqn")$dilution_factor

cbind(true = dil, estimated = dil_est)

}
\references{
Dieterle F, Ross A, Schlotterbeck G, Senn H (2006).
Probabilistic Quotient Normalization as Robust Method to Account for Dilution of
Complex Biological Mixtures. \emph{Analytical Chemistry}, 78(13), 4281–4290.
}
\seealso{
Other preprocessing: 
\code{\link{align_segment}()},
\code{\link{align_spectra}()},
\code{\link{binning}()},
\code{\link{calibrate}()},
\code{\link{correct_baseline}()},
\code{\link{correct_lw}()},
\code{\link{print_preprocessing}()}
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_sets_methods.R
\name{mc}
\alias{mc}
\title{Monte-Carlo cross-validation strategy}
\usage{
mc(k, split)
}
\arguments{
\item{k}{Integer. Number of repeated random splits.}

\item{split}{Numeric. Fraction of samples assigned to the
training set (e.g. \code{2/3}).}
}
\value{
A named \code{list} with elements:
\describe{
\item{train}{List of integer vectors containing training set indices
for each resampling iteration.}
\item{strategy}{Character string indicating the resampling strategy.}
\item{n}{Integer. Number of samples in the dataset.}
\item{seed}{Integer. Random seed used to generate the resampling splits,
ensuring reproducibility.}
}
}
\description{
Monte-Carlo cross-validation strategy
}
\details{
Monte-Carlo cross-validation generates \code{k} random train/test splits
without replacement. No stratification is applied; samples are drawn
uniformly at random.
}
\examples{
n <- 100
# bivariate outcome
thr <- 1.5
Y <- c(rnorm(80, thr-3, 0.3), rnorm(20, thr+3, 0.3))  # unbalanced low/high outcome
mean(Y>thr)

cv_mc <- mc(k = 10, split = 2/3)
mc_inst <- metabom8:::.arg_check_cv(cv_pars=cv_mc, model_type='R', n=n, Y_prepped=cbind(Y))
sapply(mc_inst$train, function(i) length(i))
}
\seealso{
Other resampling strategies: 
\code{\link{balanced_boot}()},
\code{\link{balanced_mc}()},
\code{\link{kfold}()},
\code{\link{stratified_kfold}()}
}
\concept{resampling strategies}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_allviz.R
\name{.draw_ggplot2}
\alias{.draw_ggplot2}
\title{Draw NMR Spectra Using ggplot2}
\usage{
.draw_ggplot2(dat, ...)
}
\arguments{
\item{dat}{List returned by \code{.prep_spec()}.}

\item{...}{Additional arguments for \code{ggplot2}.}
}
\description{
Internal backend renderer using \code{ggplot2}. Converts spectra
to long format prior to plotting.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_alignSegment.R
\name{.dynamicIntervalsMedian}
\alias{.dynamicIntervalsMedian}
\title{Dynamic Interval Segmentation (RSPA-style)}
\usage{
.dynamicIntervalsMedian(X, ppm, half_win_ppm = 0.02, min_snr = 8)
}
\arguments{
\item{X}{Numeric matrix (spectra in rows).}

\item{ppm}{Numeric ppm vector.}

\item{half_win_ppm}{Numeric. Half window size around detected peaks.}
}
\value{
List of index vectors defining alignment intervals.
}
\description{
Defines peak-system intervals from the median spectrum for
recursive segment-wise alignment (RSPA).
}
\seealso{
Other NMR: 
\code{\link{get_idx}()}
}
\concept{NMR}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_stocsy.R
\name{stocsy}
\alias{stocsy}
\title{Statistical Total Correlation Spectroscopy (STOCSY)}
\usage{
stocsy(X, ppm = NULL, driver, plotting = TRUE, title = NULL)
}
\arguments{
\item{X}{Numeric matrix or data frame. Rows are spectra, columns are variables.}

\item{ppm}{Optional numeric vector of chemical shift values (length = ncol(X)).
If missing or NULL, will try to read from \code{colnames(X)}.}

\item{driver}{Numeric scalar (ppm value, internal driver) or numeric vector
(external driver, length = nrow(X)).}

\item{plotting}{Logical. If TRUE, plot the STOCSY spectrum.}

\item{title}{Optional character plot title.}
}
\value{
An object of class \code{m8_stocsy1d} (S3), a named list with entries:
\code{version, X, ppm, driver, r, cov, extD}.
}
\description{
Performs STOCSY analysis on 1D NMR spectra. Computes correlation and covariance
of all variables in \code{X} against a driver signal (internal ppm position or
an external numeric vector).
}
\examples{
# st <- stocsy(X, ppm, driver = 5.233, plotting = FALSE)
# plotStocsy(st, shift = c(5.15, 5.30))

data(covid)
cs = 5.233
stocsy(covid$X, driver=cs, plotting = TRUE)
}
\seealso{
Other structural_annotation: 
\code{\link{storm}()}
}
\concept{structural_annotation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_provenance.R
\name{add_note}
\alias{add_note}
\title{Add user note to metabom8 provenance
Appends a user annotation to the \code{"m8_prep"} attribute.
The step title is formatted as \code{"note {username}"}.
The timestamp is stored in \code{params}, and the user message
is stored in \code{notes}.}
\usage{
add_note(x, note, params = NULL)
}
\arguments{
\item{x}{A metabom8 object or matrix with \code{"m8_prep"} metadata.}

\item{note}{Character string describing the annotation.}

\item{params}{Named list providing paramter key-value pairs.}
}
\value{
The input object with updated provenance metadata.
}
\description{
Add user note to metabom8 provenance
Appends a user annotation to the \code{"m8_prep"} attribute.
The step title is formatted as \code{"note {username}"}.
The timestamp is stored in \code{params}, and the user message
is stored in \code{notes}.
}
\examples{
params <- list(
  runtime  = "docker",
  image    = Sys.getenv("IMAGE", "docker-image-dummy"),
  workflow = Sys.getenv("M8_WORKFLOW", "std_prof-urine"),
  agent    = paste0("snakemake/", Sys.getenv("SNAKEMAKE_VERSION", "v?")),
  run_id   = Sys.getenv("M8_RUN_ID", "m8-2605-001")
)

data(hiit_raw)
print_provenance(hiit_raw)

hiit_proc <- hiit_raw |>
  calibrate(type = "tsp") |>
  excise() |>
  add_note('dilution-adaptive acquisition mode -> verify snr after normalisation',
    params)

print_provenance(hiit_proc)

}
\seealso{
Other provenance: 
\code{\link{get_provenance}()},
\code{\link{print_provenance}()}
}
\concept{provenance}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01-s4_generics.R, R/03-s4_methods.R
\name{fitted}
\alias{fitted}
\alias{fitted,m8_model-method}
\title{Extract fitted Y values}
\usage{
fitted(object, ...)

\S4method{fitted}{m8_model}(object)
}
\arguments{
\item{object}{An object of class \code{m8_model}.}

\item{...}{Additional arguments (currently ignored).}
}
\value{
Numeric vector or matrix containing the fitted response values.
}
\description{
Extract fitted Y values
}
\examples{
data(covid)
cv <- balanced_mc(k=5, split=2/3)
scaling <- uv_scaling(center=TRUE)
model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
show(model)
Y_hat_dummy <- fitted(model)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_sets_methods.R
\name{balanced_boot}
\alias{balanced_boot}
\title{Balanced bootstrap resampling strategy}
\usage{
balanced_boot(k, split, type = c("DA", "R"), probs = NULL)
}
\arguments{
\item{k}{Integer. Number of bootstrap resamples.}

\item{split}{Numeric. Fraction of samples drawn for the training set
(e.g. \code{2/3}). Sampling is performed with replacement.}

\item{type}{Character. Either \code{"DA"} (classification) or \code{"R"} (regression).}

\item{probs}{Numeric vector of quantile probabilities used to stratify continuous
\code{Y} when \code{type = "R"}.}
}
\value{
A named \code{list} with elements:
\describe{
\item{train}{List of integer vectors containing training set indices
for each resampling iteration.}
\item{strategy}{Character string indicating the resampling strategy.}
\item{n}{Integer. Number of samples in the dataset.}
\item{seed}{Integer. Random seed used to generate the resampling splits,
ensuring reproducibility.}
}
}
\description{
Balanced bootstrap resampling strategy
}
\details{
Generates \code{k} bootstrap samples (training sets sampled with replacement).
The remaining samples (the out-of-bag set) can be used as a test set.

Balancing ensures equal representation of strata in the training data:
\describe{
\item{\code{type = "DA"}}{
Class labels define the strata, and sampling is balanced across classes.
}
\item{\code{type = "R"}}{
The response is discretised into bins using quantiles defined by
\code{probs}, and each bin contributes equally to the training set.
}
}
}
\examples{
n <- 100
# bivariate outcome
thr <- 1.5
Y <- c(rnorm(80, thr-3, 0.3), rnorm(20, thr+3, 0.3))  # unbalanced low/high outcome
mean(Y>thr)

cv_k <- kfold(k = 10)
cv_boot <- balanced_boot(k = 10, split = 2/3, type = "R", probs = c(0, 0.8, 1))

k_inst <- metabom8:::.arg_check_cv(cv_pars=cv_k, model_type='R', n=n, Y_prepped=cbind(Y))
b_inst <- metabom8:::.arg_check_cv(cv_pars=cv_boot, model_type='R', n=n, Y_prepped=cbind(Y))

# balanced splits: proportion above global median stays ~0.5
q80 <- quantile(Y, 0.8)
round(sapply(k_inst$train, function(i) mean(Y[i] > q80)), 2) # resembles original Y distr.
round(sapply(b_inst$train, function(i) mean(Y[i] > q80)), 2) # balanced strata (low/high)
}
\seealso{
Other resampling strategies: 
\code{\link{balanced_mc}()},
\code{\link{kfold}()},
\code{\link{mc}()},
\code{\link{stratified_kfold}()}
}
\concept{resampling strategies}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features_ppick.R
\name{ppick2}
\alias{ppick2}
\title{Peak picking using Savitzky–Golay derivatives}
\usage{
ppick2(
  X,
  ppm = NULL,
  type = c("max", "min", "both"),
  fil_p = 3L,
  fil_n = 11L,
  noise_win = NULL,
  min_snr = 10,
  min_height = NULL,
  min_prominence = NULL,
  prom_half_window_ppm = 0.02,
  min_distance_ppm = 0.005,
  min_curvature = NULL,
  keep_cols = c("height", "snr", "curvature", "prominence")
)
}
\arguments{
\item{X}{Numeric matrix or vector. Spectra in rows.}

\item{ppm}{Numeric vector or NULL. If NULL, inferred from colnames(X).}

\item{type}{Character. "max", "min", or "both".}

\item{fil_p}{Integer. Polynomial order for SG filter.}

\item{fil_n}{Integer. Window length for SG filter (odd).}

\item{noise_win}{Numeric length-2 vector or NULL. Ppm window used to estimate noise per spectrum.
If NULL, a robust noise estimate is computed from the full spectrum (MAD of first differences).}

\item{min_snr}{Numeric. Minimum SNR (peak height / noise). Set NULL to disable.}

\item{min_height}{Numeric. Minimum absolute peak height (in original intensity units). NULL disables.}

\item{min_prominence}{Numeric. Minimum local prominence. NULL disables.}

\item{prom_half_window_ppm}{Numeric. Half-window (ppm) for prominence estimation around each peak.}

\item{min_distance_ppm}{Numeric. Minimum separation between peaks (ppm). NULL disables.}

\item{min_curvature}{Numeric. Minimum absolute curvature at peak (|d2|). NULL disables.}

\item{keep_cols}{Character. Extra columns to keep (default keeps all computed).}
}
\value{
List of data.frames (one per spectrum). Each contains:
idc, ppm, Int, Etype (+1 max, -1 min), height, snr, curvature, prominence.
}
\description{
Finds local extrema in 1D spectra using Savitzky–Golay first and second derivatives.
Candidate peaks are identified at zero-crossings of the first derivative and classified
by the sign of the second derivative. Optional filters control peak height, prominence,
SNR, curvature and minimum separation.
}
\examples{
data(covid)
X <- covid$X
ppm <- covid$ppm

peaklist <- ppick2(X[1,], ppm, min_snr=50)

plot_spec(X[1, ], ppm, shift = c(3, 4.5), backend='base')
points(peaklist[[1]]$ppm, peaklist[[1]]$Int, col = "cyan")
head(peaklist[[1]])
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_provenance.R
\name{get_provenance}
\alias{get_provenance}
\title{Retrieve metabom8 provenance metadata}
\usage{
get_provenance(x, step = NULL, param = NULL)
}
\arguments{
\item{x}{A metabom8 object (named list with element \code{X}) or a numeric
matrix containing metabom8 provenance metadata.}

\item{step}{Optional. Either:
\itemize{
\item Numeric index of the preprocessing step
\item Character string matching the recorded step name
}
If \code{NULL}, the full provenance log is returned.}

\item{param}{Optional character string specifying a parameter name within
the selected step. If provided, only this parameter value is returned.}
}
\value{
Depending on the arguments:
\itemize{
\item Full provenance list (if \code{step = NULL})
\item A single preprocessing step (if \code{step} specified)
\item A single parameter value (if \code{param} specified)
}
}
\description{
Extracts preprocessing provenance stored in the \code{"m8_prep"} attribute.
Allows access to the full processing log, a specific step, or a specific
parameter within a step.
}
\details{
Provenance metadata are recorded automatically by metabom8 preprocessing
functions and stored as structured attributes on the spectral matrix.
This function provides programmatic access to these records.
}
\examples{
data(hiit_raw)

hiit_proc <- hiit_raw |>
  calibrate(type = "tsp") |>
  excise()

# Retrieve full log
log <- get_provenance(hiit_proc)

# Retrieve specific step
get_provenance(hiit_proc, step = 2)

# Retrieve parameter from a named step
get_provenance(hiit_proc, step = "calibrate", param = "target")

}
\seealso{
Other provenance: 
\code{\link{add_note}()},
\code{\link{print_provenance}()}
}
\concept{provenance}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01-s4_generics.R, R/03-s4_methods.R
\name{scores}
\alias{scores}
\alias{scores,m8_model-method}
\title{PLS/OPLS model scores}
\usage{
scores(object, ...)

\S4method{scores}{m8_model}(object, orth = FALSE, cv = FALSE)
}
\arguments{
\item{object}{An object of class \code{m8_model}.}

\item{...}{Additional arguments (currently ignored).}

\item{orth}{Logical indicating whether orthogonal scores should be returned
(only applicable for OPLS models).}

\item{cv}{Logical indicating whether cross-validated scores should be returned}
}
\value{
Numeric vector or matrix containing scores.
}
\description{
PLS/OPLS model scores
}
\examples{
data(covid)
cv <- balanced_mc(k=5, split=2/3)
scaling <- uv_scaling(center=TRUE)
model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
show(model)
scores(model, orth=FALSE)
scores(model, orth=TRUE)
scores(model, cv=TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_alignSegment.R
\name{.shift_interp}
\alias{.shift_interp}
\title{Linear Interpolated Shift (Internal)}
\usage{
.shift_interp(x, lag, ppm = NULL)
}
\arguments{
\item{x}{Numeric vector.}

\item{lag}{Numeric scalar. Shift in index units. Positive values shift the
signal to the right, negative values shift to the left.}
}
\value{
Numeric vector of the same length as \code{x}.
}
\description{
Applies a linear shift to a numeric vector using interpolation.
Values outside the original domain are extended
using the boundary values (no zero padding).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_scalings.R
\name{prep_X}
\alias{prep_X}
\title{Applies a preprocessing strategy to a numeric matrix.}
\usage{
prep_X(preproc_strategy, X)
}
\arguments{
\item{preproc_strategy}{list with elements 'center' (logical) and 'scale' (character).}

\item{X}{numeric matrix - processing is done column-wise.}
}
\value{
A list containing the processed matrix and parameters.
}
\description{
Applies a preprocessing strategy to a numeric matrix.
}
\examples{
X <- matrix(c(10,10, 0,0, 0, 10), ncol=3)
autoscale <- uv_scaling(center=TRUE)
X_scaled <- prep_X(autoscale, X)
str(X_scaled)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_hotellingsT2.R
\name{ellipse2d}
\alias{ellipse2d}
\title{Calculate 2D Hotelling T^2 Ellipse}
\usage{
ellipse2d(obj, dims = c(1, 2), npoints = 100)
}
\arguments{
\item{obj}{A \code{list} as returned by \code{hotellingsT2()}.}

\item{dims}{Integer vector of length 2 specifying which dimensions to project
onto. Default is \code{c(1, 2)}.}

\item{npoints}{Integer. Number of points used to approximate the ellipse.
Default is 100.}
}
\value{
A \code{data.frame} with columns \code{x} and \code{y} containing
ellipse coordinates.
}
\description{
Generates coordinates of a two-dimensional ellipse corresponding
to a Hotelling T^2 region projected onto selected dimensions.
}
\details{
The ellipse is obtained by projecting the multivariate T^2 ellipsoid onto
the specified dimensions. The scaling is derived from the Hotelling T^2
statistic and accounts for sample size and dimensionality.
}
\examples{
set.seed(1)
X <- cbind(rnorm(100), rnorm(100) + 0.5)

t2 <- hotellingsT2(X)
ell <- ellipse2d(t2)

plot(X[,1], X[,2], asp = 1, pch = 16,
     xlab = "Score 1", ylab = "Score 2")
lines(ell$x, ell$y, col = "red", lwd = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_allviz.R
\name{plot_spec}
\alias{plot_spec}
\alias{spec}
\alias{matspec}
\title{Plot 1D NMR Spectra}
\usage{
plot_spec(
  x,
  ppm = NULL,
  shift = c(0, 10),
  backend = c("plotly", "base", "ggplot2"),
  add = FALSE,
  ...
)

spec(...)

matspec(...)
}
\arguments{
\item{x}{Numeric vector (single spectrum) or numeric matrix (spectra in rows).}

\item{ppm}{Numeric vector of chemical shift values. Must match \code{ncol(x)}.}

\item{shift}{Numeric vector of length 2. Chemical shift window to display
(e.g., \code{c(0, 10)}).}

\item{backend}{Character. Rendering backend. One of \code{"plotly"}, \code{"base"},
or \code{"ggplot"}. Defaults to \code{"plotly"}.}

\item{add}{Logical. If \code{TRUE} and \code{backend = "base"}, add spectra to an
existing plot.}

\item{...}{Additional arguments passed to the selected backend.}
}
\value{
\itemize{
\item \code{"plotly"}: a \code{plotly} object.
\item \code{"ggplot"}: a \code{ggplot2} object.
\item \code{"base"}: \code{NULL} (invisibly).
}
}
\description{
Plot one or multiple 1D ¹H NMR spectra using different rendering backends.
}
\details{
The function accepts both single spectra and matrices of spectra. Input is
internally normalized to matrix form, subset to the selected ppm region,
and then rendered using the chosen backend.

For large NMR datasets (e.g., >500 spectra × >10k ppm), the \code{"base"} and
\code{"plotly"} backends are substantially more memory-efficient than \code{"ggplot"},
which requires reshaping to long format.

Chemical shift axes are automatically displayed in decreasing order
(NMR convention).
}
\examples{
data(hiit_raw)
plot_spec(hiit_raw)
plot_spec(hiit_raw, shift=c(-0.05,0.05), backend='base')
plot_spec(hiit_raw, shift=c(-0.05,0.05), backend='ggplot2')
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_bline.R
\name{correct_baseline}
\alias{correct_baseline}
\alias{bline}
\title{Baseline Correction for Spectral Data}
\usage{
correct_baseline(X, method = c("asls", "linear"), ...)

bline(X, ...)
}
\arguments{
\item{X}{Numeric matrix or metabom8 \code{dat} object containing spectra
in rows and spectral variables in columns.}

\item{method}{Character specifying the baseline correction algorithm.
One of:
\describe{
\item{\code{"asls"}}{Asymmetric least squares baseline estimation.}
\item{\code{"linear"}}{Linear baseline estimated from edge regions.}
}}

\item{...}{Additional parameters passed to the selected method.

\strong{Arguments for \code{method = "asls"}}
\describe{
\item{lambda}{Numeric smoothing parameter controlling baseline stiffness.
Larger values produce smoother baselines. Default \code{1e7}.}
\item{iter_max}{Maximum number of iterations used in the baseline
estimation procedure. Default \code{30}.}
}

\strong{Arguments for \code{method = "linear"}}
\describe{
\item{ppm}{Numeric vector describing the spectral axis (e.g. chemical shift).
Must have length equal to \code{ncol(X)}.}
\item{edge_frac}{Fraction of points at each spectrum edge used to
estimate the baseline. Default \code{0.1}.}
}}
}
\value{
Numeric matrix containing baseline-corrected spectra with the
same dimensions as \code{X}. If \code{X} is a metabom8 \code{dat} object, the corrected
matrix replaces the \code{X} component while preserving associated metadata.
}
\description{
Applies baseline correction to each spectrum (row) of a spectral matrix.
Multiple correction algorithms are available and selected via the
\code{method} argument.
}
\details{
Baseline correction is performed independently for each spectrum.

The \code{"asls"} method estimates a smooth baseline using asymmetric least
squares smoothing, which is well suited for spectra with positive peaks
such as NMR metabolomics data.

The \code{"linear"} method estimates a straight baseline from the edges of
each spectrum and subtracts the fitted trend.

The returned object includes a \code{.stamp} attribute recording the
baseline correction method and parameters used.
}
\examples{
data(hiit_raw)

plot_spec(hiit_raw$X[1,], hiit_raw$ppm, shift=c(3.1,4), backend='base')
hiit_proc <-
  hiit_raw |>
  excise() |>
  correct_baseline()

plot_spec(hiit_proc$X[1,], hiit_proc$ppm, shift=c(3.1,4), backend='base', add=TRUE, col='red')

}
\references{
Eilers PHC, Boelens HFM (2005).
Baseline correction with asymmetric least squares smoothing.
}
\seealso{
Other preprocessing: 
\code{\link{align_segment}()},
\code{\link{align_spectra}()},
\code{\link{binning}()},
\code{\link{calibrate}()},
\code{\link{correct_lw}()},
\code{\link{pqn}()},
\code{\link{print_preprocessing}()}
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_helpers.R
\name{.r2}
\alias{.r2}
\title{\eqn{R^2} and \eqn{Q^2} Calculation for OPLS Models}
\usage{
.r2(Y, Yhat, ytss = NULL)
}
\arguments{
\item{Y}{Numeric matrix. True response values. For classification, this should be a dummy matrix.}

\item{Yhat}{Numeric matrix. Predicted response values from the model.}

\item{ytss}{Numeric (optional). Total sum of squares of \code{Y}. If not provided, it will be computed internally.}
}
\value{
A single numeric value representing \eqn{R^2} or \eqn{Q^2}.
}
\description{
Computes the coefficient of determination (\eqn{R^2} or \eqn{Q^2}) for OPLS regression models using the formula:
\deqn{R^2 = 1 - \frac{PRESS}{TSS}}, where PRESS is the prediction error sum of squares,
and TSS is the total sum of squares. If \code{ytss} is not provided, it is computed directly from \code{Y}.

This function supports matrix inputs (e.g., for multi-class outcomes) and averages across columns
to return a single summary \eqn{R^2}/\eqn{Q^2} value.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_sets_methods.R
\name{.mcBalanced}
\alias{.mcBalanced}
\title{Class-balanced Monte Carlo Cross-Validation (MCCV) splits}
\usage{
.mcBalanced(k, split, stratified)
}
\arguments{
\item{k}{Integer. Number of MCCV training sets to generate.}

\item{split}{Numeric. Fraction of observations per class to include in each training set. Must be between 0 and 1.}

\item{stratified}{List of three elements:
\itemize{
\item{1: } Character. Outcome type: 'R' (regression) or 'DA' (discriminant analysis), optionally with suffix '-mY' for multi-column Y.
\item{2: } Matrix. Response matrix \code{Y}, with \code{ncol(Y) == 1}.
\item{3: } Numeric vector of probabilities used to stratify numeric Y if regression.
}}
}
\value{
A list of length \code{k}, each element containing a vector of row indices for the training set.
}
\description{
Generates class/group-balanced training set indices for each MCCV round.
}
\details{
Each round samples a class-balanced subset without replacement. Useful for high-variance, imbalanced-class modeling.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_effectsize.R
\name{cliffs_d}
\alias{cliffs_d}
\alias{es_cdelta}
\title{Cliff's Delta Effect Size}
\usage{
cliffs_d(ref, comp)

es_cdelta(ref, comp)
}
\arguments{
\item{ref}{Numeric vector representing the reference group.}

\item{comp}{Numeric vector representing the comparator group.}
}
\value{
A single numeric value: Cliff's delta effect size.
}
\description{
Calculates Cliff's delta (\code{Cd}) as a non-parametric effect size for
comparing two numeric vectors. \code{Cd} quantifies the directional difference
between a reference (\code{ref}) and comparison (\code{comp}) distribution
based on pairwise comparisons of their values.

\code{Cd} ranges from -1 to 1, where:
\itemize{
\item{-1: values in \code{comp} are consistently larger than those in \code{ref}}
\item{0: both distributions are similar, with no systematic difference}
\item{1: values in \code{ref} are consistently larger than those in \code{comp}}
}
}
\details{
The effect size is calculated as the scaled difference in dominance between groups.
Missing or infinite values are removed with a message. Synonyms are
}
\examples{
ref <- rnorm(100, mean = 0)
comp <- rnorm(100, mean = 1)

hist(ref,  col=rgb(0,0,1,0.3), breaks=50, xlim=c(-4,4))
hist(comp, col=rgb(1,0,0,0.3), breaks=50, add=TRUE)
legend("topright", legend=c("ref","comp"), fill=c("blue","red"))

cliffs_d(ref, comp)
cliffs_d(comp, ref)

comp1 <- rnorm(100, mean = 10)
cliffs_d(ref, comp1)

cliffs_d(ref, ref)
}
\references{
Cliff, N. (1993). Dominance statistics: Ordinal analyses to answer ordinal questions.
\emph{Psychological Bulletin}, 114(3), 494–509. \doi{10.1037/0033-2909.114.3.494}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_helpers.R
\name{.is_numeric_trycatch}
\alias{.is_numeric_trycatch}
\title{Check if input is numeric-like using tryCatch}
\usage{
.is_numeric_trycatch(x)
}
\arguments{
\item{x}{A vector to test for numeric coercion.}
}
\value{
Logical \code{TRUE} if \code{x} is numeric-like, \code{FALSE} otherwise.
}
\description{
Tests whether a vector \code{x} can be safely coerced to numeric without
producing \code{NA}s (excluding original \code{NA}s) or warnings.

This function attempts to coerce the input to numeric and returns \code{TRUE}
if all non-missing elements convert without coercion failure; otherwise \code{FALSE}.
It catches warnings and errors during coercion and treats those as non-numeric.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_helpers.R
\name{.check_X_ppm}
\alias{.check_X_ppm}
\title{Check Consistency of Spectral Data and ppm Vector}
\usage{
.check_X_ppm(X, ppm)
}
\arguments{
\item{X}{Numeric matrix. NMR spectral matrix with samples in rows and variables in columns (e.g., intensities).}

\item{ppm}{Numeric vector. Chemical shift reference values corresponding to columns of \code{X}.}
}
\value{
Logical. Returns \code{TRUE} if inputs are valid, \code{FALSE} otherwise.
}
\description{
Validates that the NMR data matrix and the chemical shift reference vector
(\code{ppm}) are consistent.
Specifically, it checks that
\itemize{
\item \code{ppm} contains no NA or infinite values.
\item The number of columns in \code{X} matches the length of \code{ppm}.
}

This function is typically used as a preliminary data validation step before downstream spectral analysis.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_sets_methods.R
\name{kfold}
\alias{kfold}
\title{K-fold cross-validation strategy}
\usage{
kfold(k)
}
\arguments{
\item{k}{Integer number of folds.}
}
\value{
A named \code{list} with elements:
\describe{
\item{train}{List of integer vectors containing training set indices
for each resampling iteration.}
\item{strategy}{Character string indicating the resampling strategy.}
\item{n}{Integer. Number of samples in the dataset.}
\item{seed}{Integer. Random seed used to generate the resampling splits,
ensuring reproducibility.}
}
}
\description{
K-fold cross-validation strategy
}
\details{
Partitions the data into \code{k} folds. Each fold is used once as a test set,
with the remaining folds used for training.
No stratification is applied; folds are created by random partitioning.
}
\examples{
n <- 100
thr <- 1.5
Y <- c(rnorm(80, thr - 3, 0.3), rnorm(20, thr + 3, 0.3))  # unbalanced outcome
mean(Y > thr)
cv_k  <- kfold(k = 10)
k_inst <- metabom8:::.arg_check_cv(cv_pars=cv_k, model_type='R', n=n, Y_prepped=cbind(Y))
sapply(k_inst$train, function(i) length(i))
}
\seealso{
Other resampling strategies: 
\code{\link{balanced_boot}()},
\code{\link{balanced_mc}()},
\code{\link{mc}()},
\code{\link{stratified_kfold}()}
}
\concept{resampling strategies}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.sdRcpp}
\alias{.sdRcpp}
\title{Column-wise standard deviation and mean for a matrix
Welford algorithm (single loop for mean/sd)}
\usage{
.sdRcpp(X)
}
\arguments{
\item{X}{num matrix}
}
\value{
list: 1. sd (num vec), 2. mean (sd vec)
}
\description{
Column-wise standard deviation and mean for a matrix
Welford algorithm (single loop for mean/sd)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_dmodx.R
\name{dmodx}
\alias{dmodx}
\title{Distance to the Model in X-Space (DModX)}
\usage{
dmodx(mod, plot = TRUE)
}
\arguments{
\item{mod}{An OPLS model object of class \code{m8_model} (engine = "opls").}

\item{plot}{Logical. If TRUE, a plot of DModX values with an approximate cutoff is shown.}
}
\value{
A data frame with columns \code{ID}, \code{DmodX}, and \code{passed}.
}
\description{
Calculates the orthogonal distance of each observation to an OPLS model in X-space.
DModX can be used for identifying outliers.
}
\details{
DModX is computed from the X-residual matrix as a scaled RMSE of residuals.
The empirical cutoff (dashed line in plot) uses a t-based confidence interval
and assumes approximate normality of DModX values. This assumption may not be
satisfied in all datasets, so the resulting threshold should be regarded
as a pragmatic heuristic for outlier detection.
}
\examples{
data(covid)
cv <- balanced_mc(k=5, split=2/3)
scaling <- uv_scaling(center=TRUE)
model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)
dX <- dmodx(model)
print(dX[[1]])
df <- dX[[2]]; head(df)

}
\seealso{
Other model_validation: 
\code{\link{cv_anova}()},
\code{\link{opls_perm}()}
}
\concept{model_validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{hiit_raw}
\alias{hiit_raw}
\title{High-intensity interval training (HIIT) 1H NMR urine dataset}
\format{
A list with four elements:
\describe{
\item{X}{Numeric matrix of spectral intensities (samples x variables).}
\item{ppm}{Numeric vector of chemical shift values corresponding to columns of \code{X}.}
\item{meta}{Data frame containing acquisition metadata.}
\item{df}{Data frame containing sample annotations.}
}
}
\source{
Example dataset bundled with the package.
}
\description{
Urine samples collected from a single individual performing a \eqn{VO_2max}-type
exercise protocol over a time period of 3h.
}
\details{
The spectra were acquired on a 600 MHz Bruker NMR spectrometer.
The dataset is included for demonstration of preprocessing and
modelling workflows in \pkg{metabom8}.
}
\examples{
data(hiit_raw)

}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_helpers.R
\name{.fidApodisationFct}
\alias{.fidApodisationFct}
\title{Apodisation function dispatcher}
\usage{
.fidApodisationFct(n, pars)
}
\arguments{
\item{n}{Integer. Number of data points.}

\item{pars}{List of parameters defining the apodisation.
Must contain element \code{fun} specifying the window type.}
}
\value{
Numeric vector of length \code{n} with apodisation weights.
}
\description{
Generates apodisation (window) functions for FID processing.
The function applies one of several predefined window shapes,
selected via \code{pars$fun}.
}
\details{
Supported apodisation functions and parameters:
\describe{
\item{\code{"uniform"}}{No apodisation (all weights = 1).}

\item{\code{"exponential"}}{
Exponential decay window.
\describe{
\item{\code{lb}}{Line broadening parameter.}
}
}

\item{\code{"cosine"}}{
Cosine window (no additional parameters).
}

\item{\code{"sine"}}{
Sine window (no additional parameters).
}

\item{\code{"sem"}}{
Sine-modulated exponential window.
\describe{
\item{\code{lb}}{Exponential decay parameter.}
}
}

\item{\code{"expGaus_resyG"}}{
Exponential-Gaussian hybrid window.
\describe{
\item{\code{lb}}{Exponential decay parameter.}
\item{\code{gb}}{Gaussian broadening parameter.}
\item{\code{aq_t}}{Acquisition time.}
}
}

\item{\code{"gauss"}}{
Gaussian window (Bruker-style).
\describe{
\item{\code{lb}}{Line broadening parameter.}
\item{\code{gb}}{Gaussian broadening factor.}
\item{\code{para}}{List containing acquisition parameters
(e.g. \code{a_SW_h}, \code{a_TD}).}
}
}
}
}
\examples{
# Internal FID apidisation functions (illustrative only)
f_apod <- c("sine","cosine","exponential","sem")

# create toy fid
n <- 200; t <- seq(0,1,len=n)
fid <- exp(-5*t)*cos(20*pi*t)

# generate & apply apodisation windows
A  <- sapply(f_apod, \(f) metabom8:::.fidApodisationFct(n, list(fun=f, lb=-0.2)))
Fp <- sweep(A, 1, fid, `*`)

# generate spectra
S  <- apply(Fp, 2, \(x) Mod(fft(x))[1:(n/2)])

# graphhics
cols <- 1:ncol(A)

par(mfrow=c(2,2), mar=c(3,3,2,1))
plot(t, fid, type="l", lwd=2, main="FID")
matplot(t, A, type="l", lwd=2, col=cols, main="Windows")
matplot(t, Fp, type="l", lwd=2, col=cols, main="Windowed FID")
matplot(S, type="l", lwd=2, col=cols, main="Spectrum")

legend("topright", f_apod, col=cols, lty=1, bty="n", cex=.8)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_calibrate.R
\name{calibrate}
\alias{calibrate}
\title{Chemical Shift Calibration}
\usage{
calibrate(X, ppm = NULL, type = "tsp")
}
\arguments{
\item{X}{Numeric matrix/vector of spectra or a metabom8 data list.}

\item{ppm}{Numeric chemical shift vector. If \code{X} is a metabom8 list,
this is taken from \code{X$ppm}.}

\item{type}{Character (\code{"tsp"}, \code{"glucose"}, \code{"alanine"}),
or list.}
}
\value{
Calibrated spectra in the same structure as input.
}
\description{
Aligns 1D \eqn{^1}H NMR spectra to a reference signal on the existing ppm grid.
Supports singlet (e.g. TSP or custom range) and predefined doublet references
(glucose, alanine).
}
\details{
In addition to predefined references, custom calibration targets can be supplied
as a list with elements:
\itemize{
\item \code{mode}: \code{"singlet"} or \code{"doublet"}
\item \code{window}: numeric vector of length 2 defining the ppm search region
\item \code{centre}: target ppm position (optional; defaults to mean(window))
\item \code{j}: numeric vector of length 2 specifying expected J-coupling
(required for doublet calibration)
}
Custom doublet calibration requires the expected J-coupling range (\code{j})
in ppm to distinguish the two peaks of the multiplet.

For example:
\preformatted{
calibrate(
  X, ppm,
  list(
    mode   = "doublet",
    window = c(1.2, 1.35),
    j      = c(0.007, 0.009)
  )
)
}
}
\examples{
data("covid_raw")
X=covid_raw$X
ppm=covid_raw$ppm
X_tsp <- calibrate(X, ppm, type = "tsp")
X_glu <- calibrate(X, ppm, type = "glucose")
X_custom <- calibrate(X, ppm, type = c(1.9, 2.1))
}
\seealso{
Other preprocessing: 
\code{\link{align_segment}()},
\code{\link{align_spectra}()},
\code{\link{binning}()},
\code{\link{correct_baseline}()},
\code{\link{correct_lw}()},
\code{\link{pqn}()},
\code{\link{print_preprocessing}()}
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_scalings.R
\name{unscaled}
\alias{unscaled}
\title{No Scaling
This function defines a preprocessing strategy that is applied via
\code{\link{prep_X}}.}
\usage{
unscaled(center = FALSE)
}
\arguments{
\item{center}{Logical. If TRUE, variables are mean-centered before scaling.}
}
\value{
A \code{list} with elements:
\describe{
\item{X}{Numeric matrix containing the scaled data.}
\item{prep}{List describing the preprocessing, including centering and scaling
parameters (\code{center}, \code{scale}, \code{X_mean}, \code{X_sd}).}
}#' @details Leaves variables unscaled. Optional centering.
}
\description{
No Scaling
This function defines a preprocessing strategy that is applied via
\code{\link{prep_X}}.
}
\examples{
just_centering <- unscaled(center=TRUE)
X <- matrix(c(10,10, 0,0, 0, 10), ncol=3)
X_centered <- prep_X(just_centering, X)
str(X_centered)
X_centered$X
}
\seealso{
Other scaling_strategies: 
\code{\link{pareto_scaling}()},
\code{\link{uv_scaling}()}
}
\concept{scaling_strategies}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metabom8-package.R
\docType{package}
\name{metabom8}
\alias{metabom8-package}
\alias{metabom8}
\title{metabom8: A High-Performance R Package for Metabolomics Modeling and Analysis}
\description{
metabom8 (pronounced \emph{metabo-mate}) provides pipelines for 1D NMR data import,
preprocessing, multivariate modeling (PCA, OPLS), metabolite identification,
and visualisation. Core functions are accelerated in C++ via \pkg{Rcpp},
\pkg{RcppArmadillo}, and \pkg{RcppEigen} for improved computational performance.
}
\section{Features}{

\itemize{
\item Import, preprocessing, and analysis of 1D NMR spectra.
\item \strong{Principal Components Analysis (PCA)} with \emph{back-scaled} loadings
(projected back to the spectral domain and visualized as spectra).
\item \strong{Orthogonal Partial Least Squares (OPLS)} fitted iteratively via the NIPALS algorithm,
with an \emph{automatic stopping criterion} to determine the optimal number of components.
\item Automatic model selection using cross-validated performance (\eqn{R^2}, \eqn{Q^2},
and cross-validated AUC for classification), with safeguards against overfitting.
\item Robust statistical validation for small to large sample sizes:
stratified Monte Carlo cross-validation and k-fold CV.
\item Model diagnostics and validation (DModX, permutation testing).
\item Metabolite identification via STOCSY and STORM.
\item Native C++ acceleration via \pkg{RcppArmadillo} and \pkg{RcppEigen}.
}
}

\section{Vignettes}{

\itemize{
\item Getting Started: \code{vignette("Getting Started")}
}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://tkimhofer.github.io/metabom8/}
  \item Report bugs at \url{https://github.com/tkimhofer/metabom8/issues}
}

}
\author{
\strong{Maintainer}: Torben Kimhofer \email{tkimhofer@gmail.com} (\href{https://orcid.org/0000-0001-7158-9930}{ORCID})

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_provenance.R
\name{print_preprocessing}
\alias{print_preprocessing}
\title{List available preprocessing functions
Returns the preprocessing utilities provided by \pkg{metabom8}.}
\usage{
print_preprocessing()
}
\value{
A named character vector describing preprocessing functions.
}
\description{
List available preprocessing functions
Returns the preprocessing utilities provided by \pkg{metabom8}.
}
\examples{
list_preprocessing()
}
\seealso{
Other preprocessing: 
\code{\link{align_segment}()},
\code{\link{align_spectra}()},
\code{\link{binning}()},
\code{\link{calibrate}()},
\code{\link{correct_baseline}()},
\code{\link{correct_lw}()},
\code{\link{pqn}()}
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_scalings.R
\name{pareto_scaling}
\alias{pareto_scaling}
\title{Pareto Scaling
Leaves variables unscaled. Optional centering.}
\usage{
pareto_scaling(center = FALSE)
}
\arguments{
\item{center}{Logical. If TRUE, variables are mean-centered before scaling.}
}
\value{
A \code{list} with elements:
\describe{
\item{X}{Numeric matrix containing the scaled data.}
\item{prep}{List describing the preprocessing, including centering and scaling
parameters (\code{center}, \code{scale}, \code{X_mean}, \code{X_sd}).}
}
}
\description{
Pareto Scaling
Leaves variables unscaled. Optional centering.
}
\details{
Scales variables by the square root of their standard deviation.
}
\examples{
paritalUV <- pareto_scaling(center=TRUE)
X <- matrix(c(10,10, 0,0, 0, 10, 0, 1000), ncol=4)
X_scaled <- prep_X(paritalUV, X)
str(X_scaled)
X_scaled$X
}
\seealso{
Other scaling_strategies: 
\code{\link{unscaled}()},
\code{\link{uv_scaling}()}
}
\concept{scaling_strategies}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_nmr_qc.R
\name{lw}
\alias{lw}
\title{Full Width at Half Maximum (FWHM) Estimation}
\usage{
lw(X, ppm = NULL, shift = c(-0.1, 0.1), sf)
}
\arguments{
\item{X}{Numeric matrix (spectra in rows) \emph{or} a named list as returned by
\code{\link{read1d}}/\code{read1d_proc} containing \code{X}, \code{ppm}, and \code{meta}.}

\item{ppm}{Numeric vector of chemical shift values (ppm) corresponding to columns of \code{X}.
If \code{NULL}, \code{ppm} is inferred in the following order:
\enumerate{
\item \code{attr(X, "m8_axis")$ppm} (if present),
\item numeric \code{colnames(X)} (if present).
}}

\item{shift}{Numeric vector of length 2. Chemical shift range containing the peak
(e.g., \code{c(-0.1, 0.1)} for TSP).}

\item{sf}{Spectrometer frequency in MHz. Either a single numeric (recycled across spectra)
or a numeric vector of length \code{nrow(X)} (one value per spectrum).
If \code{X} is a list input and \code{sf} is missing, \code{sf} is taken from
\code{X$meta$a_SFO1} when available. If \code{sf} is missing for a matrix input,
the function attempts to use \code{attr(X, "m8_meta")$a_SFO1} when present.}
}
\value{
Numeric vector of FWHM values in Hz (length \code{nrow(X)}).
}
\description{
Estimates the full width at half maximum (FWHM; line width) of a singlet-like peak
within a specified chemical-shift range for each spectrum.
}
\details{
For each spectrum, the function:
\enumerate{
\item extracts the region defined by \code{shift},
\item finds the peak apex within that region,
\item computes the half-height level relative to the local baseline (minimum in the window),
\item estimates the left and right half-height crossing points by linear interpolation,
\item converts the width from ppm to Hz using \code{sf} (MHz), i.e. \code{Hz = ppm * sf}.
}

If no valid half-height crossings can be found (e.g., very low SNR or truncated peak),
\code{NA} is returned for that spectrum.
}
\note{
The ppm axis may be increasing or decreasing; FWHM is computed as an absolute width
and is therefore independent of axis direction.
}
\examples{
# Simulated NMR peaks with different linewidths
ppm <- seq(-0.2, 0.2, length.out = 1000)

# generate peaks with increasing width
sds <- seq(0.01, 0.03, length.out = 10)

X <- t(sapply(sds, function(s)
  dnorm(ppm, mean = 0, sd = s)
))

sf <- 600  # spectrometer frequency in MHz

fwhm_vals <- lw(X, ppm = ppm, shift = c(-0.1, 0.1), sf = sf)

plot(sds, fwhm_vals,
  xlab = "Gaussian sd",
  ylab = "Estimated FWHM (Hz)",
  pch = 16)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_scalings.R
\name{minmax}
\alias{minmax}
\title{Min-Max Scaling to \eqn{[0,1]}}
\usage{
minmax(x, na.rm = FALSE)
}
\arguments{
\item{x}{Numeric vector. Input values to be scaled.}

\item{na.rm}{Logical; if \code{TRUE}, ignore \code{NA}s when computing the range.}
}
\value{
A numeric vector of the same length as \code{x}, scaled to the range \eqn{[0,1]}.
}
\description{
Scales a numeric vector to the range \eqn{[0,1]} using min-max normalization.
This is a special case of \code{\link{scRange}}.
}
\details{
The scaled values are computed as:
\deqn{x_{scaled} = \frac{x - \min(x)}{\max(x) - \min(x)}}

Equivalent to \code{scRange(x, ra = c(0, 1))}.
}
\examples{
x <- rnorm(20)
plot(x, type = 'l'); abline(h = range(x), lty = 2)
points(minmax(x), type = 'l', col = 'blue')
abline(h = c(0, 1), col = 'blue', lty = 2)

}
\seealso{
\code{\link[=scRange]{scRange()}} for flexible output ranges.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_helpers.R
\name{.list_to_string}
\alias{.list_to_string}
\title{Convert named list to compact string representation}
\usage{
.list_to_string(lst)
}
\arguments{
\item{lst}{A named list.}
}
\value{
A character string.
}
\description{
Internal utility that converts a named list into a single
character string of the form \code{name=value}. Nested
sublists are represented recursively using braces.
}
\details{
This function is primarily used for compact parameter logging
in provenance metadata and should not be called directly by users.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_hotellingsT2.R
\name{hotellingsT2}
\alias{hotellingsT2}
\title{Hotelling T^2 Statistic}
\usage{
hotellingsT2(X, alpha = 0.95)
}
\arguments{
\item{X}{Numeric matrix with observations in rows and variables (dimensions)
in columns.}

\item{alpha}{Numeric scalar. Confidence level for the T^2 region. Default is 0.95.}
}
\value{
A named \code{list} with elements:
\describe{
\item{center}{Numeric vector of column means.}
\item{cov}{Sample covariance matrix.}
\item{c2}{Squared radius of the Hotelling T^2 region.}
}
}
\description{
Computes the Hotelling T^2 statistic defining a multivariate
confidence region for a set of observations. The region corresponds to an
ellipsoid in p-dimensional space and is commonly visualised as an ellipse
in two-dimensional (OPLS) score plots.
}
\details{
The Hotelling T^2 region is defined as
\deqn{(x - \mu)^T S^{-1} (x - \mu) \le c^2}
where \eqn{c^2 = (p (n - 1) / (n - p)) F_{p, n-p}(\alpha)}.
}
\examples{
set.seed(1)
X <- matrix(rnorm(200), ncol = 2)
t2 <- hotellingsT2(X)
ell <- ellipse2d(t2)
plot(X, asp = 1)
lines(ell$x, ell$y, col = "red", lwd = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_opls_model_perm.R
\name{.perm_test_from_table}
\alias{.perm_test_from_table}
\title{Permutation-test summary from an opls_perm out_df table}
\usage{
.perm_test_from_table(
  out_df,
  observed_label = "non-permuted",
  alternative = c("greater", "less"),
  add_one = TRUE,
  na_rm = TRUE
)
}
\arguments{
\item{out_df}{data.frame with columns like q2_comp, r2_comp, aucs_te, aucs_tr,
model (with one row "non-permuted"), and optionally r/r_abs.}

\item{observed_label}{character. Label used in \code{model} for the observed row.}

\item{alternative}{character. "greater" (default) tests obs > perm; "less" tests obs < perm.}

\item{add_one}{logical. If TRUE, uses (1 + count)/(B + 1) p-value correction.}

\item{na_rm}{logical. Drop NA values before computing p-values.}
}
\value{
A list with:
\itemize{
\item \code{observed}: named numeric vector of observed metrics
\item \code{perm}: list of numeric vectors for each metric
\item \code{p_value}: named numeric vector of permutation p-values
\item \code{B}: number of permutations used
}
}
\description{
Takes the \code{out_df} you showed (perm rows + one "non-permuted" row) and returns
observed values, permutation distributions, and permutation p-values.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_helpers.R
\name{.prepareY}
\alias{.prepareY}
\title{Prepare Response Vector for OPLS/OPLS-DA}
\usage{
.prepareY(Y)
}
\arguments{
\item{Y}{A numeric, factor, or character vector.}
}
\value{
A list with:
\describe{
\item{\code{[[1]]}}{Numeric matrix of processed response.}
\item{\code{[[2]]}}{Data frame mapping class labels to encoding (empty for numeric input).}
}
}
\description{
Converts a response vector \code{Y} into a numeric matrix suitable for regression or classification models.
\itemize{
\item Numeric input is returned as a column matrix.
\item Factor or character input:
\itemize{
\item If binary (2 levels): returns a single numeric column (values 1 and 2).
\item If multiclass (>2 levels): returns a dummy matrix with 1 for presence and -1 for absence.
Also returns a mapping between original labels and numeric values (only for categorical input).
}
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_read1d.R
\name{.detect1d_procs}
\alias{.detect1d_procs}
\title{Check for Intact File Systems - Helper Function for read1d}
\usage{
.detect1d_procs(
  datapath,
  n_max = 10,
  filter = TRUE,
  recursive = TRUE,
  verbose = 1
)
}
\arguments{
\item{datapath}{character. Path to directory containing spectra.}

\item{n_max}{integer. Maximum number of spectra to read (default 10).}

\item{filter}{logical. Whether to filter out incomplete file sets (default TRUE).}

\item{recursive}{logical. Whether to search directories recursively.}

\item{verbose}{integer. Verbosity level for messaging.}
}
\value{
A list with elements:
\describe{
\item{path}{Character vector of experiment folder paths (intact).}
\item{exp_no}{Character vector of experiment folder IDs.}
\item{f_procs}{Character vector of paths to 'procs' files.}
\item{f_acqus}{Character vector of paths to 'acqus' files.}
\item{f_1r}{Character vector of paths to '1r' files.}
}
}
\description{
Scans the specified directory recursively to find intact sets of Bruker NMR files:
'procs', 'acqus', and '1r' files. Optionally filters incomplete experiments.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_sets_methods.R
\name{.kFoldStratified}
\alias{.kFoldStratified}
\title{Stratified k-fold cross-validation index generator}
\usage{
.kFoldStratified(k, stratified)
}
\arguments{
\item{k}{Integer. Number of folds.}

\item{stratified}{List with three elements:
\itemize{
\item \code{type}: Character. Either \code{"R"} for regression or \code{"DA"} for discriminant analysis.
\item \code{Y}: Matrix. Outcome matrix with a single column.
\item \code{probs}: Numeric vector. Probabilities used for stratification of numeric \code{Y} (regression only).
}}
}
\value{
A list of length \code{k}, each element containing the training set row indices (integers) for one CV fold.
Returns \code{NULL} and a warning if stratification is not feasible due to class imbalance.
}
\description{
Generates a list of training set indices for stratified k-fold cross-validation (CV).
Stratification is performed based on the first column of \code{Y}. For regression tasks,
\code{Y} is binned by quantiles to emulate class balance. Ensures class proportions are
preserved across folds when possible.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_helpers.R
\name{.extract_acq_pars1d}
\alias{.extract_acq_pars1d}
\title{Extract Bruker NMR Acquisition Parameters}
\usage{
.extract_acq_pars1d(f_list)
}
\arguments{
\item{f_list}{List. Contains paths to NMR experiment folders (e.g., as returned by \code{.check1d_files_fid()}).}
}
\value{
A data frame containing extracted acquisition parameters for each experiment.
}
\description{
Parses \code{acqus} files to extract acquisition parameters for each experiment.
}
\seealso{
\code{\link{.filterExp_files}}, \code{\link{.check1d_files_fid}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_helpers.R
\name{.check1d_files_fid}
\alias{.check1d_files_fid}
\title{Check for Intact Bruker NMR File Structures}
\usage{
.check1d_files_fid(datapath, n_max = 10, filter = TRUE, recursive, verbose)
}
\arguments{
\item{datapath}{Character. Directory containing NMR spectra.}

\item{n_max}{Integer. Maximum number of experiments to process.}

\item{filter}{Logical. Whether to filter out incomplete experiments.}

\item{recursive}{Logical. Whether to search directories recursively.}

\item{verbose}{Integer. Controls verbosity level (0 = silent, 1 = basic, 2 = detailed).}
}
\value{
A list with paths to intact experiments and matching metadata.
}
\description{
Verifies that each NMR experiment has a matching \code{acqus} and \code{fid} file.
}
\seealso{
\code{\link{.extract_acq_pars1d}}, \code{\link{.filterExp_files}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_cv_sets_methods.R
\name{.balanced_bootstrap_mc}
\alias{.balanced_bootstrap_mc}
\title{Class-balanced resampling (with replacement)}
\usage{
.balanced_bootstrap_mc(
  k,
  split,
  stratified,
  remainder = c("distribute", "drop")
)
}
\arguments{
\item{k}{Integer. Number of resampling sets.}

\item{split}{Numeric in (0,1). Fraction of total N used as training size.}

\item{stratified}{List of length 3:
\enumerate{
\item type: "DA" for classification or "R" for regression
\item Y: matrix with one column (n x 1)
\item probs: numeric vector of quantile probs for regression binning (e.g. c(0, .33, .66, 1))
}}

\item{remainder}{Character. How to handle n_train \%\% G remainder:
\itemize{
\item "drop": keep perfect balance, total size becomes G * floor(n_train/G)
\item "distribute": distribute leftover 1-by-1 to random strata (still near-balanced)
}}
}
\value{
List of length k. Each element is an integer vector of training indices.
}
\description{
Generates k training index sets. Total train size per set is floor(N * split),
where N is the total number of samples. Each stratum contributes (approximately)
equally, sampling WITH replacement within each stratum.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/01-s4_generics.R, R/03-s4_methods.R
\name{xres}
\alias{xres}
\alias{xres,m8_model-method}
\title{Compute X residual matrix
Returns the residual matrix (E) of an OPLS model.}
\usage{
xres(object)

\S4method{xres}{m8_model}(object)
}
\arguments{
\item{object}{An object.}
}
\value{
Numeric matrix containing the X residuals.
}
\description{
Compute X residual matrix
Returns the residual matrix (E) of an OPLS model.
}
\examples{
data(covid)
cv <- balanced_mc(k = 5, split = 2/3)
scaling <- uv_scaling(center = TRUE)
model <- opls(X = covid$X, Y = covid$an$type, scaling, cv)
X_res <- xres(model)
dim(X_res) == dim(covid$X)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.calibTsp}
\alias{.calibTsp}
\title{Calibrate}
\usage{
.calibTsp(spec, ppm)
}
\value{
Shift-adjusted ppm vector
}
\description{
Calibrate
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/io_read1d.R
\name{.chemShift}
\alias{.chemShift}
\title{Calculate Chemical Shift Axis}
\usage{
.chemShift(swidth, offset, si)
}
\arguments{
\item{swidth}{Numeric. Sweep width (Hz).}

\item{offset}{Numeric. Frequency offset (ppm).}

\item{si}{Integer. Number of data points.}
}
\value{
Numeric vector of chemical shift values (ppm).
}
\description{
Computes a 1D NMR chemical shift axis based on sweep width, offset, and size.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_provenance.R
\name{print_provenance}
\alias{print_provenance}
\title{Print metabom8 preprocessing pipeline}
\usage{
print_provenance(x, detail = FALSE, max_items = 8)
}
\arguments{
\item{x}{A metabom8 object or spectral matrix with attached
\code{"m8_prep"} provenance metadata.}

\item{detail}{Prints full information trail, incl. timestamp / versioning}

\item{max_items}{Limits individual list entries (e.g., parameter) to specified
number}
}
\value{
Invisibly returns \code{NULL}. This function is called for
its side effect of printing pipeline information.
}
\description{
Displays the recorded preprocessing history attached to a metabom8
spectral matrix. The function reads the \code{"m8_prep"} attribute
and prints each processing step in chronological order, including
parameters and notes.
}
\details{
The input can be either:
\itemize{
\item A metabom8-style list containing \code{$X}, or
\item A matrix with metabom8 provenance attributes attached.
}

If no preprocessing metadata is found, a message is printed.

metabom8 records preprocessing steps as an ordered list of
transformation descriptors stored in the \code{"m8_prep"} attribute.
Each step typically contains:
\itemize{
\item \code{step}: Name of the preprocessing operation
\item \code{params}: Parameters used
\item \code{notes}: Optional description
\item \code{time}: Timestamp
\item \code{pkg}: Package name and version
}

This function provides a compact audit trail of the processing
workflow, facilitating reproducibility and provenance inspection.
}
\examples{
params <- list(
  runtime  = "docker",
  image    = Sys.getenv("IMAGE", "docker-image-dummy"),
  workflow = Sys.getenv("M8_WORKFLOW", "std_prof-urine"),
  agent    = paste0("snakemake/", Sys.getenv("SNAKEMAKE_VERSION", "v?")),
  run_id   = Sys.getenv("M8_RUN_ID", "m8-2605-001")
)

data(hiit_raw)
print_provenance(hiit_raw)

hiit_proc <- hiit_raw |>
  calibrate(type = "tsp") |>
  excise() |>
  add_note('dilution-adaptive acquisition mode -> verify snr after normalisation',
    params)

print_provenance(hiit_proc)
}
\seealso{
Other provenance: 
\code{\link{add_note}()},
\code{\link{get_provenance}()}
}
\concept{provenance}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_fit_opls.R
\name{opls}
\alias{opls}
\title{Fit an Orthogonal Partial Least Squares (O-PLS) model}
\usage{
opls(X, Y, scaling, validation_strategy)
}
\arguments{
\item{X}{Numeric matrix of predictors (rows = samples, columns = variables).}

\item{Y}{Numeric matrix or factor vector of responses.}

\item{scaling}{A scaling strategy object (e.g., \code{uv_scaling(center = TRUE)}),
specifying model-internal centering and/or scaling applied during fitting.
This does not modify the original spectral matrix.}

\item{validation_strategy}{A cross-validation strategy object defining how
resampling is performed (e.g., k-fold, Monte Carlo).}
}
\value{
An object of class \code{m8_model} containing the fitted O-PLS model,
cross-validation results, and performance statistics.
}
\description{
Fits a supervised Orthogonal Partial Least Squares (O-PLS) model
using a NIPALS-based algorithm with optional cross-validation and
automatic component selection.
}
\details{
O-PLS decomposes the predictor matrix into:
\itemize{
\item One predictive component capturing variation correlated with \code{Y}
\item Orthogonal components capturing structured variation in \code{X}
unrelated to \code{Y}
}

Predictive and orthogonal components are estimated sequentially.
Cross-validated performance metrics (e.g., Q², R², classification AUC)
are computed for each model configuration according to the supplied
\code{validation_strategy}.

The model extracts a single predictive component and iteratively
adds orthogonal components until the \code{stopRule} indicates
overfitting or \code{maxPCo} is reached.

Scaling specified via \code{scaling} is applied internally during model
fitting and does not alter the input matrix \code{X}. Spectral preprocessing
(e.g., alignment or baseline correction) should be performed prior to
model fitting.

The returned model object stores:
\itemize{
\item Predictive and orthogonal component models
\item Cross-validation results
\item Performance metrics (R², Q², AUC)
\item Model control parameters
\item Input data provenance metadata
\item Session information for reproducibility
}
}
\examples{
data(covid)

cv <- balanced_mc(k=5, split=2/3)
scaling <- uv_scaling(center=TRUE)
model <-opls(X=covid$X, Y=covid$an$type, scaling, cv)

show(model)
summary(model)

# scores
Tp <- scores(model)
To <- scores(model, orth=TRUE)

t2 <- hotellingsT2(cbind(Tp, To))
ell <-ellipse2d(t2)

plot(Tp, To, asp = 1,
  col = as.factor(covid$an$type),
  xlim = range(c(Tp, ell$x)),
  ylim = range(c(To, ell$y))
 )
lines(ell$x, ell$y, col = "grey", lty=2)

# loadings & vip's
Pp <- loadings(model)
Po <- loadings(model, orth=TRUE)
vips <- vip(model)

x=covid$ppm
y = Pp * apply(covid$X, 2, sd)

palette <-  colorRampPalette(c("blue", "cyan", "yellow", "red"))(100)
idx <- cut(vips, breaks = 100, labels = FALSE)
plot(x, y, type = "n", xlim = rev(range(x)), xlab='ppm', ylab='t_pred_sc')

for (i in seq_len(length(x) - 1)) {
  segments(x[i], y[i], x[i+1], y[i+1], col = palette[idx[i]], lwd = 2)
}

}
\seealso{
\code{\link{pls}}, \code{\link{uv_scaling}}

Other modelling: 
\code{\link{pca}()},
\code{\link{pls}()}
}
\concept{modelling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_helpers.R
\name{.check_pc_model}
\alias{.check_pc_model}
\title{Check Validity of Selected Principal or Orthogonal Component}
\usage{
.check_pc_model(pc, mod, le = 1, type = "p")
}
\arguments{
\item{pc}{Character or numeric. Selected component, e.g., \code{1} or \code{"o1"}.}

\item{mod}{An object of class \code{PCA_metabom8} or \code{OPLS_metabom8}.}

\item{le}{Integer. Expected maximum length of \code{pc}; usually 1.}

\item{type}{Character. Type of component to check: \code{"p"} for projection or \code{"t"} for scores. Currently only used for logic branching.}
}
\value{
Nothing. Function is used for its side effect (i.e., error checking).
}
\description{
Internal consistency check for principal (or orthogonal) component selection used in plotting or extraction functions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_allviz.R
\name{.prep_spec}
\alias{.prep_spec}
\title{Prepare Spectral Data for Plotting}
\usage{
.prep_spec(x, ppm, shift)
}
\arguments{
\item{x}{Numeric vector or matrix of spectra.}

\item{ppm}{Numeric chemical shift vector.}

\item{shift}{Length-2 numeric vector specifying ppm window.}
}
\value{
A list with elements:
\describe{
\item{ppm}{Subsetted chemical shift axis}
\item{X}{Matrix of subsetted spectra}
}
}
\description{
Internal helper that validates input, enforces ppm ordering,
subsets the selected shift region, and returns matrix-form data.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_helpers.R
\name{.checkDimXY}
\alias{.checkDimXY}
\title{Check Dimensions of X and Y Matrices (PLS Context)}
\usage{
.checkDimXY(X, Y)
}
\arguments{
\item{X}{Numeric matrix. Predictor data matrix (observations x variables).}

\item{Y}{Numeric matrix. Response data matrix (observations x variables).}
}
\value{
Invisibly returns \code{NULL} if checks pass.
Throws an error if dimensions are incompatible.
}
\description{
Checks that the input matrices \code{X} and \code{Y} have compatible dimensions
for Partial Least Squares (PLS) analysis.
Specifically, the number of observations (rows) must match, and
the number of variables (columns) in \code{X} must be greater than in \code{Y}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R
\name{.scaleMatRcpp}
\alias{.scaleMatRcpp}
\title{Column-wise matrix scaling  (C++ backend)}
\usage{
.scaleMatRcpp(X, idc, center, scale_type)
}
\arguments{
\item{X}{num matrix}

\item{idc}{int row indices of X}

\item{center}{logical mean centering}

\item{scale_type}{int 0: no scaling, 1: SD scaling, 2: Pareto scaling}
}
\value{
list: 1. scale X matrix, 2. mean (sd vec), 3: sd (num vec)
}
\description{
Column-wise matrix scaling  (C++ backend)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_alignSegment.R
\name{align_spectra}
\alias{align_spectra}
\title{Cohort-Guided Interval Alignment for 1D NMR Spectra}
\usage{
align_spectra(x, ppm = NULL, half_win_ppm = 0.007, lag.max = 200)
}
\arguments{
\item{x}{A numeric matrix/vector of spectra or a named list with elements
\code{X}, \code{ppm}, and optionally \code{meta}.}

\item{ppm}{Numeric vector of chemical shift values (ppm). Ignored if \code{x} is
a \code{dat} list.}

\item{half_win_ppm}{Numeric scalar controlling alignment interval width.}

\item{lag.max}{Integer. Maximum allowed lag for cross-correlation alignment.}
}
\value{
If input is \code{dat}, returns updated \code{dat}. Otherwise returns aligned matrix.
}
\description{
Aligns 1D NMR spectra using signal-adaptive ppm intervals derived from the
cohort median spectrum. Intervals are constructed around median peak systems
and aligned locally via cross-correlation.
The function accepts either a metabom8-style \code{dat} list or plain matrix/vector with \code{ppm}.
}
\details{
Alignment intervals are derived from peaks in the cohort median spectrum.

The \code{half_win_ppm} parameter controls interval granularity:

\itemize{
\item Smaller values (e.g. 0.005–0.007 ppm) generate more, narrower
intervals (RSPA-like behaviour).
\item Larger values (e.g. 0.008–0.015 ppm) merge nearby multiplets into
broader regions (icoshift-like behaviour).
}

Typical 600–800 MHz \eqn{^1}H NMR data:
\itemize{
\item 0.006–0.008 ppm: stable multiplet-level alignment
\item <0.005 ppm: may split J-coupled systems
\item >0.015 ppm: may merge unrelated resonances
}

Alignment parameters and interval definitions are recorded in
\code{attr(X, "m8_prep")}.
}
\examples{
data(hiit_raw)
plot_spec(hiit_raw, shift=c(1.3,1.4))
hiit_aligns <- align_spectra(hiit_raw)
plot_spec(hiit_aligns, shift=c(1.3,1.4)) # aligned segment
plot_spec(hiit_aligns, shift=c(3, 3.1)) # aligned segment
}
\seealso{
\code{\link{align_segment}}

Other preprocessing: 
\code{\link{align_segment}()},
\code{\link{binning}()},
\code{\link{calibrate}()},
\code{\link{correct_baseline}()},
\code{\link{correct_lw}()},
\code{\link{pqn}()},
\code{\link{print_preprocessing}()}
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/preproc_correct_lw.R
\name{correct_lw}
\alias{correct_lw}
\title{Linewidth correction by scaling spectra to a reference linewidth}
\usage{
correct_lw(
  x,
  lw_ref = 0.8,
  beta = 0.6,
  shift = c(-0.1, 0.1),
  ppm = NULL,
  sf = NULL,
  estimate_beta = TRUE,
  beta_min_n = 6,
  only_if_improves = TRUE,
  notes = NULL
)
}
\arguments{
\item{x}{A numeric matrix (samples x variables) or a named list containing
\code{X}, \code{ppm}, and optionally \code{meta}.}

\item{lw_ref}{Numeric reference linewidth (FWHM) to which spectra are
normalized. If \code{NULL}, the median linewidth across samples is used.}

\item{beta}{Numeric exponent governing the linewidth-to-height relationship.
Ignored if \code{estimate_beta = TRUE} and sufficient samples are available.}

\item{shift}{Numeric vector of length 2 specifying the ppm region used for
linewidth estimation and peak height measurement (e.g. \code{c(-0.1, 0.1)}
for TSP).}

\item{ppm}{Optional numeric ppm axis. If \code{NULL}, inferred from
\code{attr(X, "m8_axis")$ppm} or numeric \code{colnames(X)}.}

\item{sf}{Optional spectrometer frequency (MHz). If \code{NULL}, obtained from
\code{meta$a_SFO1} in list input or \code{attr(X, "m8_meta")$a_SFO1}.}

\item{estimate_beta}{Logical; if \code{TRUE}, estimate \code{beta} from a
log-log regression of reference peak height versus FWHM.}

\item{beta_min_n}{Minimum number of valid samples required to estimate
\code{beta}.}

\item{only_if_improves}{Logical; if \code{TRUE}, apply correction only if the
CV of the reference peak height decreases after correction.}

\item{notes}{Optional character string appended to the preprocessing log.}
}
\value{
An object of the same type as input:
\itemize{
\item If input is a matrix, returns a corrected matrix with attributes
preserved and updated.
\item If input is a list, returns the same list with corrected \code{$X}.
}
}
\description{
Applies a multiplicative correction to each spectrum to compensate for
linewidth-induced peak height variation, using a reference peak region
(e.g. TSP). The correction assumes an empirical power-law relationship
between peak height and linewidth:
}
\details{
\deqn{I \propto \mathrm{FWHM}^{-\beta}}

Spectra are scaled such that all samples are adjusted to a common reference
linewidth \code{lw_ref}:

\deqn{X_{\mathrm{adj}} = X \cdot \left(\frac{\mathrm{FWHM}}{lw_{\mathrm{ref}}}\right)^{\beta}}

where FWHM is estimated per spectrum within the specified \code{shift}
region (typically containing an internal reference signal).

If \code{estimate_beta = TRUE}, the exponent \code{beta} is estimated from a
log-log regression between peak height (max within \code{shift}) and FWHM
across samples. Otherwise, the user-supplied \code{beta} is used.

The correction is only applied if it reduces the coefficient of variation (CV)
of the reference peak height across samples when \code{only_if_improves = TRUE}.

Input may be either:
\itemize{
\item a numeric matrix \code{X} (samples x variables), with ppm axis supplied
via \code{attr(X, "m8_axis")$ppm} or numeric \code{colnames(X)}, or
\item a named list with elements \code{X}, \code{ppm}, and optionally \code{meta}.
}

Attributes \code{"m8_prep"}, \code{"m8_axis"}, and \code{"m8_meta"} are
preserved and the applied correction is appended to the preprocessing log
via \code{.m8_stamp()}.

This correction removes systematic peak height variation due to spectral
broadening (e.g. shimming differences) under the assumption that the internal
reference signal has constant concentration across samples. The exponent
\code{beta} reflects the empirical sensitivity of peak height to linewidth
in the current acquisition and processing regime.

This adjustment improves comparability of signal intensities across spectra
by reducing linewidth-induced amplitude bias, thereby enhancing the
detectability of small effects in downstream multivariate analyses
(e.g. PLS-type models).
}
\examples{
data(covid_raw)
# matrix input with ppm in m8_axis
Xcor <- correct_lw(covid_raw, shift = c(-0.1, 0.1))

}
\seealso{
\code{\link{lw}}

Other preprocessing: 
\code{\link{align_segment}()},
\code{\link{align_spectra}()},
\code{\link{binning}()},
\code{\link{calibrate}()},
\code{\link{correct_baseline}()},
\code{\link{pqn}()},
\code{\link{print_preprocessing}()}
}
\concept{preprocessing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_model_helpers.R
\name{.evalFit}
\alias{.evalFit}
\title{Evaluate model fit progression based on cross-validated performance}
\usage{
.evalFit(
  type,
  q2s,
  cv_auc,
  pc_max = 5,
  min_delta = 0.05,
  sat_q2 = 0.98,
  sat_auc = 0.97,
  min_q2 = 0.05,
  min_auc = 0.6
)
}
\arguments{
\item{type}{Character. Model type: \code{"R"} for regression or \code{"DA"} for
discriminant analysis. The value may optionally include a suffix
such as \code{"-mY"} to indicate multi-response models.}

\item{q2s}{Numeric vector of \eqn{Q^2} values for the fitted
components (used for regression models).}

\item{cv_auc}{Numeric vector of cross-validated AUROC values for
the fitted components (used for classification models).}

\item{pc_max}{Integer. Maximum number of components allowed.}

\item{min_delta}{Numeric. Minimum improvement in the cross-validated
performance metric required to justify fitting an additional
component.}

\item{sat_q2}{Numeric. Saturation threshold for \eqn{Q^2} in
regression models.}

\item{sat_auc}{Numeric. Saturation threshold for cross-validated
AUROC in discriminant analysis models.}

\item{min_q2}{Numeric. Minimum acceptable \eqn{Q^2} value for the
first component.}

\item{min_auc}{Numeric. Minimum acceptable cross-validated AUROC for
the first component.}
}
\value{
A list containing:
\describe{
\item{stop}{Logical indicating whether model fitting should stop.}
\item{reason}{Character string describing the stopping condition
(e.g., \code{"pc_max"}, \code{"cv_decreased"}, \code{"cv_improvement_negligible"}).}
\item{metric}{The current cross-validated performance value.}
\item{delta}{Improvement relative to the previous component.}
}
}
\description{
Determines whether an additional component should be fitted based on
cross-validated generalization performance. For regression models
(\code{type = "R"}), the decision is based on the \eqn{Q^2} statistic.
For discriminant analysis models (\code{type = "DA"}), the decision is
based on the cross-validated AUROC (area under the ROC curve).
}
\details{
AUROC is computed on cross-validation test folds to ensure that it
reflects out-of-sample classification performance, analogous to the
use of \eqn{Q^2} in regression.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_opls_model_perm.R
\name{.permYmod}
\alias{.permYmod}
\title{OPLS Y-permutation Modeling}
\usage{
.permYmod(XcsTot, Y, cv, type, nc_o)
}
\arguments{
\item{XcsTot}{Numeric matrix. Input data matrix (samples x features).}

\item{Y}{Numeric matrix. Response variable (numeric or dummy-coded).}

\item{cv}{List. Cross-validation parameters, including \code{method} and \code{cv_sets}.}

\item{type}{Character. Model type: \code{"DA"}, \code{"R"}, possibly with \code{"-mY"} suffix for multi-column Y.}

\item{nc_o}{Integer. Number of orthogonal components to be fitted.}
}
\value{
List with elements:
\itemize{
\item \code{r2_comp}: Numeric, R2 statistic for test data.
\item \code{q2_comp}: Numeric, Q2 statistic from cross-validation.
\item \code{aucs_tr}: Numeric, AUROC on training data (for classification only).
\item \code{aucs_te}: Numeric, AUROC on test data (for classification only).
}
}
\description{
Performs OPLS modeling on permuted Y to estimate cross-validated model performance under the null hypothesis.
Extracts predictive performance (R2, Q2, AUROC) from cross-validation for either regression or classification.
}
\details{
When \code{nc_o > 1}, the function fits additional orthogonal components sequentially, updating the model object.
Classification performance is computed using AUROC (via \code{pROC}). Regression performance uses R2 and Q2.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_fit_pls.r
\name{pls}
\alias{pls}
\title{Fit a Partial Least Squares (PLS) model}
\usage{
pls(X, Y, scaling, validation_strategy, maxPCo = 5)
}
\arguments{
\item{X}{Numeric matrix of predictors (rows = samples, columns = variables).}

\item{Y}{Numeric matrix or factor vector of responses.}

\item{scaling}{A scaling strategy object (e.g., \code{uv_scaling(center = TRUE)}),
specifying model-internal centering and/or scaling applied during fitting.
This does not modify the original spectral matrix.}

\item{validation_strategy}{A cross-validation strategy object defining how
resampling is performed (e.g., k-fold, Monte Carlo).}

\item{maxPCo}{Integer. Maximum number of predictive components to evaluate.}
}
\value{
An object of class \code{m8_model} containing the fitted PLS model,
cross-validation results, and performance statistics.
}
\description{
Fits a supervised Partial Least Squares (PLS) model using a NIPALS-based
algorithm with optional cross-validation and automatic component selection.
}
\details{
Model components are estimated sequentially using a NIPALS-based algorithm.
For each component, cross-validated performance metrics (e.g., Q², R²,
classification AUC) are computed according to the supplied
\code{validation_strategy}. Component extraction stops when the
\code{stopRule} indicates overfitting or when \code{maxPCo} is reached.

Scaling specified via \code{scaling} is applied internally during model
fitting and does not alter the input matrix \code{X}. Spectral preprocessing
steps (e.g., alignment, baseline correction) should be performed prior to
model fitting.

The returned model object stores:
\itemize{
\item Fitted component models
\item Cross-validation results
\item Performance metrics (R², Q², AUC)
\item Model control parameters
\item Input data provenance metadata
\item Session information for reproducibility
}
}
\examples{
data(covid)

cv <- balanced_mc(k=5, split=2/3)
scaling <- uv_scaling(center=TRUE)
model <-pls(X=covid$X, Y=covid$an$type, scaling, cv)

show(model)
summary(model)

Tp <- scores(model)
Pp <- loadings(model)
}
\seealso{
\code{\link{opls}}, \code{\link{uv_scaling}}

Other modelling: 
\code{\link{opls}()},
\code{\link{pca}()}
}
\concept{modelling}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stats_stocsy.R
\name{plotStocsy}
\alias{plotStocsy}
\title{Plot STOCSY result}
\usage{
plotStocsy(stoc_mod, shift = c(0, 10), title = NULL)
}
\arguments{
\item{stoc_mod}{An object of class \code{m8_stocsy1d} returned by \code{stocsy()}.}

\item{shift}{Numeric vector of length 2 specifying the chemical shift range (ppm).}

\item{title}{Optional character plot title.}
}
\value{
A \code{ggplot2} object.
}
\description{
Generates a STOCSY plot (covariance trace coloured by absolute correlation).
}
\examples{
# st <- stocsy(X, ppm, driver = 5.233, plotting = FALSE)
# plotStocsy(st, shift = c(5.15, 5.30), title = "Glucose")

data(covid)
cs = 5.233 # anomeric H of gluc
s1 <- stocsy(covid$X, driver=cs, plotting = FALSE)
plotStocsy(s1)

}
